@@ -1,4 +1,4 @@
-/*  $Id: messages.pl,v 1.66 2004/09/30 10:41:35 jan Exp $
+/*  $Id: messages.pl,v 1.81 2006/08/14 11:28:11 jan Exp $
 
     Part of SWI-Prolog
 
@@ -213,10 +213,10 @@ swi_context(X) -->
 	{ var(X)
 	}, !,
 	[].
-swi_context(context(Name/Arity, _Msg)) -->
-	{ nonvar(Name)
+swi_context(context(Caller, _Msg)) -->
+	{ ground(Caller)
 	}, !,
-	[ '~q/~w: '-[Name, Arity] ].
+	caller(Caller).
 swi_context(file(Path, Line, LinePos, _CharNo)) -->
 	[ '~w:~d:~d: '-[Path, Line, LinePos] ].
 swi_context(stream(Stream, Line, LinePos, _CharNo)) -->
@@ -224,6 +224,17 @@ swi_context(stream(Stream, Line, LinePos, _CharNo)) -->
 swi_context(_) -->
 	[].
 
+caller(Module:Name/Arity) --> !,
+	(   { \+ hidden_module(Module) }
+	->  [ '~q:~q/~w: '-[Module, Name, Arity] ]
+	;   [ '~q/~w: '-[Name, Arity] ]
+	).
+caller(Name/Arity) -->
+	[ '~q/~w: '-[Name, Arity] ].
+caller(Caller) -->
+	[ '~p: '-[Caller] ].
+
+
 swi_extra(X) -->
 	{ var(X)
 	}, !,
@@ -263,10 +274,22 @@ prolog_message(missing_feature(Name)) -->
 	[ 'This version of SWI-Prolog does not support ~w'-[Name] ].
 prolog_message(singletons(List)) -->
 	[ 'Singleton variables: ~w'-[List] ].
+prolog_message(multitons(List)) -->
+	[ 'Singleton-marked variables appearing more than once: ~w'-[List] ].
 prolog_message(profile_no_cpu_time) -->
 	[ 'No CPU-time info.  Check the SWI-Prolog manual for details' ].
 prolog_message(non_ascii(Text, Type)) -->
 	[ 'Unquoted ~w with non-portable characters: ~w'-[Type, Text] ].
+prolog_message(io_warning(Stream, Message)) -->
+	{ stream_property(Stream, Position),
+	  stream_position_data(line_count, Position, LineNo),
+	  stream_position_data(line_position, Position, LinePos),
+	  (   stream_property(Stream, file_name(File))
+	  ->  Obj = File
+	  ;   Obj = Stream
+	  )
+	},
+	[ '~p:~d:~d: ~w'-[Obj, LineNo, LinePos, Message] ].
 
 
 		 /*******************************
@@ -323,6 +346,9 @@ prolog_message(minus_in_identifier) -->
 	[ 'The "-" character should not be used to seperate words in an', nl,
 	  'identifier.  Check the SWI-Prolog FAQ for details.'
 	].
+prolog_message(qlf(removed_after_error(File))) -->
+	[ 'Removed incomplete QLF file ~w'-[File] ].
+
 
 used_search([]) -->
 	[].
@@ -395,7 +421,7 @@ prolog_message(make(reload(Files))) -->
 	{ length(Files, N)
 	},
 	[ 'Make: reloading ~D files'-[N] ].
-prolog_message(make(done)) -->
+prolog_message(make(done(_Files))) -->
 	[ 'Make: finished' ].
 prolog_message(make(library_index(Dir))) -->
 	[ 'Updating index for library ~w'-[Dir] ].
@@ -423,7 +449,7 @@ prolog_message(threads) -->
 prolog_message(threads) -->
 	[].
 prolog_message(copyright) -->
-	[ 'Copyright (c) 1990-2003 University of Amsterdam.', nl,
+	[ 'Copyright (c) 1990-2006 University of Amsterdam.', nl,
 	  'SWI-Prolog comes with ABSOLUTELY NO WARRANTY. This is free software,', nl,
 	  'and you are welcome to redistribute it under certain conditions.', nl,
 	  'Please visit http://www.swi-prolog.org for details.'
@@ -475,11 +501,6 @@ prolog_message(query(eof)) -->
 
 bindings([]) -->
 	[].
-bindings([Name = _|T]) -->
-	{ sub_atom(Name, 0, _, _, '_'),
-	  current_prolog_flag(toplevel_print_anon, false)
-	}, !,
-	bindings(T).
 bindings([Name = Value|T]) -->
 	{ current_prolog_flag(toplevel_print_options, Options)
 	},
@@ -670,6 +691,16 @@ clean_goal(M:Goal, Goal) :-
 clean_goal(Goal, Goal).
 
 
+		 /*******************************
+		 *	  COMPATIBILITY		*
+		 *******************************/
+
+prolog_message(compatibility(renamed(Old, New))) -->
+	[ 'The predicate ~p has been renamed to ~p.'-[Old, New], nl,
+	  'Please update your sources for compatibility with future versions.'
+	].
+
+
 		 /*******************************
 		 *	      THREADS		*
 		 *******************************/
@@ -679,7 +710,8 @@ prolog_message(abnormal_thread_completion(Goal, exception(Ex))) --> !,
 	translate_message(Ex).
 prolog_message(abnormal_thread_completion(Goal, fail)) -->
 	[ 'Thread running "~p" died due to failure'-[Goal] ].
-
+prolog_message(threads_not_died(Count)) -->
+	[ '~D threads wouldn\'t die'-[Count] ].
 
 		 /*******************************
 		 *	PRINTING MESSAGES	*
@@ -696,11 +728,15 @@ prolog_message(abnormal_thread_completion(Goal, fail)) -->
 %	system.
 
 print_message(Level, Term) :-
-	translate_message(Term, Lines, []), !,
-	(   nonvar(Term),
-	    notrace(user:message_hook(Term, Level, Lines))
-	->  true
-	;   print_system_message(Term, Level, Lines)
+	(   must_print(Level, Term)
+	->  (   translate_message(Term, Lines, [])
+	    ->  (   nonvar(Term),
+		    notrace(user:message_hook(Term, Level, Lines))
+		->  true
+		;   print_system_message(Term, Level, Lines)
+		)
+	    )
+	;   true
 	).
 
 %	print_system_message(+Term, +Level, +Lines)
@@ -788,7 +824,7 @@ print_message_line(S, [Fmt|T0], T) :-
 message_to_string(Term, Str) :-
         translate_message(Term, Actions, []), !,
         actions_to_format(Actions, Fmt, Args),
-        sformat(Str, Fmt, Args).
+        format(string(Str), Fmt, Args).
 
 actions_to_format([], '', []) :- !.
 actions_to_format([nl], '', []) :- !.
@@ -810,4 +846,28 @@ actions_to_format([Term|Tail], Fmt, Args) :-
         atom_concat('~w', Fmt1, Fmt),
         $append([Term], Args1, Args).
 
+
+		 /*******************************
+		 *    MESSAGES TO PRINT ONCE	*
+		 *******************************/
+
+:- dynamic
+	printed/2.
+
+%	print_once(Message, Level)
+%	
+%	True for messages that must be printed only once.
+
+print_once(compatibility(_), _).
+
+%	must_print(+Level, +Message)
+%	
+%	True if the message must be printed.
+
+must_print(Level, Message) :-
+	nonvar(Message),
+	print_once(Message, Level), !,
+	\+ printed(Message, Level),
+	assert(printed(Message, Level)).
+must_print(_, _).
 	
