@@ -1,4 +1,4 @@
-/*  $Id: syspred.pl,v 1.66 2004/09/30 10:41:35 jan Exp $
+/*  $Id: syspred.pl,v 1.75 2006/07/27 19:25:05 jan Exp $
 
     Part of SWI-Prolog
 
@@ -39,12 +39,14 @@
 	, trace/2
 	, nospyall/0
 	, debugging/0
+	, rational/3
 	, concat_atom/2
 	, term_to_atom/2
-	, int_to_atom/2
+	, atom_prefix/2
 	, dwim_match/2
 	, source_file/1
 	, prolog_load_context/2
+	, stream_position_data/3
 	, current_predicate/2
 	, $defined_predicate/1
 	, predicate_property/2
@@ -65,8 +67,6 @@
 	, open_shared_object/2
 	, open_shared_object/3
 	, format/1
-	, sformat/2
-	, sformat/3
 	, garbage_collect/0
 	, arithmetic_function/1
         , default_module/2
@@ -92,7 +92,7 @@ $map_bits(Pred, +Name, Old, New) :- !, 		% set a bit
 $map_bits(Pred, -Name, Old, New) :- !, 		% clear a bit
 	call(Pred, Name, Bits), !,
 	New is Old /\ (\Bits).
-$map_bits(Pred, ?Name, Old, Old) :-		% ask a bit
+$map_bits(Pred, ?(Name), Old, Old) :-		% ask a bit
 	call(Pred, Name, Bits),
 	Old /\ Bits > 0.
 
@@ -131,7 +131,7 @@ style_check(+string) :- !,
 	set_prolog_flag(double_quotes, string).
 style_check(-string) :- !,
 	set_prolog_flag(double_quotes, codes).
-style_check(?string) :- !,
+style_check(?(string)) :- !,
 	current_prolog_flag(double_quotes, string).
 style_check(Spec) :-
 	$style_check(Old, Old),
@@ -286,6 +286,24 @@ trace_ports(Head, Ports) :-
 		Ports).
 
 
+		 /*******************************
+		 *	      RATIONAL		*
+		 *******************************/
+
+%	rational(+Rat, -M, -N)
+%	
+%	Get parts of a rational number.
+
+rational(Rat, M, N) :-
+	rational(Rat),
+	(   Rat = rdiv(M, N)
+	->  true
+	;   integer(Rat)
+	->  M = Rat,
+	    N = 1
+	).
+
+
 		/********************************
 		*             ATOMS             *
 		*********************************/
@@ -298,12 +316,12 @@ concat_atom(L, Atom) :-
 term_to_atom(Term, Atom) :-
 	atom_to_term(Atom, Term, 0).
 
-int_to_atom(Int, Atom) :-
-	int_to_atom(Int, 10, Atom).
-
 dwim_match(A1, A2) :-
 	dwim_match(A1, A2, _).
 
+atom_prefix(Atom, Prefix) :-
+	sub_atom(Atom, 0, _, _, Prefix).
+
 
 		/********************************
 		*             SOURCE            *
@@ -318,9 +336,9 @@ source_file(File) :-
 
 %	prolog_load_context(+Key, -Value)
 %
-%	Provides context information for term_expansion and directives.
-%	Note that only the line-number info is valid for the
-%	'$stream_position'.  Largely Quintus compatible.
+%	Provides context information for  term_expansion and directives.
+%	Note  that  only  the  line-number  info    is   valid  for  the
+%	'$stream_position'. Largely Quintus compatible.
 
 :- module_transparent
 	prolog_load_context/2.
@@ -329,6 +347,8 @@ prolog_load_context(module, Module) :-
 	$set_source_module(Module, Module).
 prolog_load_context(file, F) :-
 	source_location(F, _).
+prolog_load_context(source, F) :-	% SICStus compatibility
+	source_location(F, _).
 prolog_load_context(stream, S) :-
 	source_location(F, _),
 	(   system:$load_input(F, S0)
@@ -340,6 +360,21 @@ prolog_load_context(directory, D) :-
 prolog_load_context(term_position, '$stream_position'(0,L,0,0,0)) :-
 	source_location(_, L).
 
+		 /*******************************
+		 *	      STREAMS		*
+		 *******************************/
+
+%	stream_position_data(?Field, +Pos, ?Date)
+%	
+%	Extract values from stream position objects. $stream_position is
+%	of the format '$stream_position'(Byte, Char, Line, LinePos)
+
+stream_position_data(char_count,    '$stream_position'(Char,_,_,_), Char).
+stream_position_data(line_count,    '$stream_position'(_,Line,_,_), Line).
+stream_position_data(line_position, '$stream_position'(_,_,LPos,_), LPos).
+stream_position_data(byte_count,    '$stream_position'(_,_,_,Byte), Byte).
+
+
 		 /*******************************
 		 *	      CONTROL		*
 		 *******************************/
@@ -470,8 +505,7 @@ $predicate_property(references(N), Pred) :-
 	$get_predicate_attribute(Pred, references, N),
 	N \== 0.			% show negative for debugging!
 $predicate_property(number_of_clauses(N), Pred) :-
-	$get_predicate_attribute(Pred, number_of_clauses, N),
-	N \== 0.
+	$get_predicate_attribute(Pred, number_of_clauses, N).
 $predicate_property(noprofile, Pred) :-
 	$get_predicate_attribute(Pred, noprofile, 1).
 
@@ -629,8 +663,11 @@ shell(Command) :-
 	shell(Command, 0).
 
 shell :-
-	getenv('SHELL', Shell), !,
+	getenv('SHELL', Shell), !,	% Unix, also Cygwin
 	shell(Shell).
+shell :-
+	getenv(comspec, ComSpec), !,	% Windows
+	shell(ComSpec).
 shell :-
 	shell('/bin/sh').
 
@@ -683,18 +720,13 @@ open_shared_object(File, Handle) :-
 	open_shared_object(File, [], Handle). % use pl-load.c defaults
 
 
-		/********************************
-		*              I/O              *
-		*********************************/
+		 /*******************************
+		 *	       I/O		*
+		 *******************************/
 
 format(Fmt) :-
 	format(Fmt, []).
 
-sformat(String, Format, Arguments) :-
-	$write_on_string(format(Format, Arguments), String).
-sformat(String, Format) :-
-	$write_on_string(format(Format), String).
-
 		 /*******************************
 		 *	      FILES		*
 		 *******************************/
