@@ -1,4 +1,4 @@
-/*  $Id: bounds.pl,v 1.2 2004/09/21 14:30:24 jan Exp $
+/*  $Id: bounds.pl,v 1.14 2006/08/11 08:02:16 toms Exp $
 
     Part of SWI-Prolog
 
@@ -79,9 +79,17 @@
 		(#\)/1,
 		(in)/2,
 		label/1,
-		all_different/1
+		labeling/2,
+		all_different/1,
+		sum/3,
+		lex_chain/1,
+		indomain/1,
+		check/1,
+		tuples_in/2
 	]).
 
+:- use_module(library('clp/clp_events')).
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 % exported predicates
 X #>= Y :-
@@ -182,6 +190,33 @@ reify(#\ C, B) :-
 	reify(C,BC),
 	mynot(BC,B).
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+check(X #= Y) :-
+	parse_expression(X,XR),
+	parse_expression(Y,YR),
+	check_eq(XR,YR).
+
+check(X #=< Y) :-
+	parse_expression(X,XR),
+	parse_expression(Y,YR),
+	check_leq(XR,YR).
+
+check(X #>= Y) :-
+	parse_expression(X,XR),
+	parse_expression(Y,YR),
+	check_leq(YR,XR).
+
+check(X #< Y) :-
+	parse_expression(X,XR),
+	parse_expression(Y,YR),
+	check_lt(XR,YR).
+
+check(X #> Y) :-
+	parse_expression(X,XR),
+	parse_expression(Y,YR),
+	check_lt(YR,XR).
+
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 parse_expression(Expr,Result) :-
 	( var(Expr) ->
@@ -215,24 +250,114 @@ parse_expression(Expr,Result) :-
 	; Expr = L mod R ->
 		parse_expression(L,RL),
 		parse_expression(R,RR),
-		mymod(RL,RR,Result)
+		mymod(RL,RR,Result,yes)
 	; Expr = abs(E) ->
 		parse_expression(E,RE),
-		myabs(RE,Result)
+		myabs(RE,Result,yes)
+	; Expr = (L / R) ->
+		parse_expression(L,RL),
+		parse_expression(R,RR),
+		mydiv(RL,RR,Result,yes)
 	).
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-label([]).
-label([V|Vs]) :-
-	( get(V,L,U,_) -> 
-		between(L,U,W),
-		%format('\tlabelling ~w with ~w\n',[V,W]),
-		V = W
+indomain(Var) :-
+	( var(Var) ->
+		bounds(Var,L,U),
+		between(L,U,Var)
 	;
 		true
-	),
-	label(Vs).
-	
+	).
+
+label(Vs) :- labeling([],Vs).
+
+labeling(Options,Vars) :-
+	label(Options,leftmost,Vars).
+
+label([Option|Options],_Selection,Vars) :-
+	selection(Option),
+	label(Options,Option,Vars).
+label([],Selection,Vars) :-
+	label(Vars,Selection).
+
+label([],_) :- !.
+label(Vars,Selection) :-
+	select_var(Selection,Vars,Var,RVars),
+	indomain(Var),
+	label(RVars,Selection).
+
+selection(ff).
+selection(min).
+selection(max).
+selection(leftmost).
+
+
+select_var(leftmost,[Var|Vars],Var,Vars).
+select_var(min,[V|Vs],Var,RVars) :-
+		find_min(Vs,V,Var),
+		delete_eq([V|Vs],Var,RVars).
+select_var(max,[V|Vs],Var,RVars) :-
+		find_max(Vs,V,Var),
+		delete_eq([V|Vs],Var,RVars).
+select_var(ff,[V|Vs],Var,RVars) :-
+		find_ff(Vs,V,Var),
+		delete_eq([V|Vs],Var,RVars).
+
+find_min([],Var,Var).
+find_min([V|Vs],CM,Min) :-
+	( min_lt(V,CM) ->
+		find_min(Vs,V,Min)
+	;
+		find_min(Vs,CM,Min)
+	).
+
+find_max([],Var,Var).
+find_max([V|Vs],CM,Max) :-
+	( max_gt(V,CM) ->
+		find_min(Vs,V,Max)
+	;
+		find_min(Vs,CM,Max)
+	).
+
+find_ff([],Var,Var).
+find_ff([V|Vs],CM,FF) :-
+	( ff_lt(V,CM) ->
+		find_min(Vs,V,FF)
+	;
+		find_min(Vs,CM,FF)
+	).
+
+ff_lt(X,Y) :-
+	bounds(X,LX,UX),
+	bounds(Y,LY,UY),
+	UX - LX < UY - LY.	
+
+min_lt(X,Y) :-
+	bounds(X,LX,_),
+	bounds(Y,LY,_),
+	LX < LY.	
+
+max_gt(X,Y) :-
+	bounds(X,_,UX),
+	bounds(Y,_,UY),
+	UX > UY.	
+
+bounds(X,L,U) :-
+	( var(X) ->
+		get(X,L,U,_)
+	;
+		X = L,
+		X = U
+	).	
+
+delete_eq([],_,[]).
+delete_eq([X|Xs],Y,List) :-
+	( X == Y ->
+		List = Xs
+	;
+		List = [X|Tail],
+		delete_eq(Xs,Y,Tail)
+	).
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 all_different([]).
 all_different([X|Xs]) :-
@@ -244,6 +369,45 @@ different([Y|Ys],X) :-
 	neq(X,Y,yes),
 	different(Ys,X).
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+sum(Xs,Op,Value) :-
+	sum1(Xs,0,Op,Value).
+
+sum1([],Sum,Op,Value) :-
+	call(Op,Sum,Value).
+sum1([X|Xs],Acc,Op,Value) :-
+	NAcc #= Acc + X,
+	sum1(Xs,NAcc,Op,Value).
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+% contributed by Markus Triska
+
+lex_le([],[]).
+lex_le([V1|V1s], [V2|V2s]) :-
+      V1 #=< V2,
+      ( integer(V1) ->
+              ( integer(V2) ->
+                      (V1 == V2 ->
+                              lex_le(V1s,V2s)
+                      ;
+                              true
+                      )
+              ;
+                      freeze(V2,lex_le([V1|V1s],[V2|V2s]))
+              )
+      ;
+              freeze(V1,lex_le([V1|V1s],[V2|V2s]))
+      ).
+
+lex_chain([]).
+lex_chain([L|Ls]) :-
+	lex_chain_lag(Ls, L).
+
+lex_chain_lag([], _).
+lex_chain_lag([L|Ls], Prev) :-
+	lex_le(Prev, L),
+	lex_chain_lag(Ls, L).
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 domain(X,L,U) :-
 	( var(X) ->
 		get(X,XL,XU,Exp),
@@ -325,7 +489,11 @@ neq(X,Y,New) :-
 			; U == X ->
 				NU is U - 1,
 				put(Y,L,NU,Exp)
-			;
+		        ; L > X ->
+		                true
+		        ; U < X ->
+		                true
+		        ;
 				( New == yes ->
 					active_state(State),
 					put(Y,L,U,[neq(X,State)|Exp])
@@ -555,6 +723,7 @@ mytimes(X,Y,Z,New) :-
 			NZU is min(ZU2,TZU),
 			put(Z,NZL,NZU,ZExp2)
 		;
+
 			true
 		)	
 	).
@@ -576,6 +745,108 @@ min_divide(L1,U1,L2,U2,Min) :-
 		Min is min(min(div(L1,L2),div(L1,U2)),min(div(U1,L2),div(U1,U2)))
 	).
 
+
+mydiv(X,Y,Z,New) :-
+	( Y == 0 -> fail ; true	),
+	( nonvar(X) ->
+		( nonvar(Y) ->
+			Z is X // Y
+		;
+			get(Y,YL,YU,YExp),
+			( New == yes ->
+				put(Y,YL,YU,[mydiv3(X,Z)|YExp])
+			;
+				true
+			),
+			( nonvar(Z) ->
+				% TODO: cover this
+				true
+			;	
+				get(Z,ZL,ZU,ZExp),
+				( YL =< 0, YU >= 0 ->
+					NZL is max(-abs(X),ZL),
+					NZU is min(abs(X),ZU)
+				; X >= 0, YL > 0 ->
+					NZL is max(X // YU, ZL),
+					NZU is min(X // YL, ZU)
+				;	% TODO: cover more cases
+					NZL = ZL,
+					NZU = ZU
+				),
+				( New == yes ->
+					put(Z,NZL,NZU,[mydiv2(X,Y)|ZExp])
+				;
+					put(Z,NZL,NZU,ZExp)
+				)
+			)
+		)
+	; nonvar(Y) ->
+		get(X,XL,XU,XExp),
+		( nonvar(Z) ->
+			(Z >= 0, Y >= 0 ->
+				NXL is max(Z*Y,XL),
+				NXU is min((Z+1)*Y - 1,XU)
+			;
+				% TODO: cover more cases
+				NXL = XL,
+				NXU = XU
+			),
+			( New == yes ->
+				put(X,NXL,NXU,[mydiv(Y,Z)|XExp])
+			;
+				put(X,NXL,NXU,XExp)
+			)
+		;
+			get(Z,ZL,ZU,ZExp),
+			( XL >= 0, Y >= 0 ->
+				NZL is max(XL // Y,ZL),
+				NZU is min(XU // Y,ZU),
+				NXL is max(Y*NZL, XL),
+				NXU is min((NZU+1)*Y - 1, XU)
+			; % TODO: cover more cases
+				NZL is max(-max(abs(XL),abs(XU)),ZL),
+				NZU is min(max(abs(XL),abs(XU)),ZU),
+				NXL = XL,
+				NXU = XU
+			),
+			( New == yes ->
+				put(X,NXL,NXU,[mydiv(Y,Z)|XExp]),
+				put(Z,NZL,NZU,[mydiv2(X,Y)|ZExp])
+			;
+				put(X,NXL,NXU,XExp),
+				( nonvar(Z) -> true ; put(Z,NZL,NZU,ZExp) )
+			)
+		)
+	; nonvar(Z) ->
+		get(X,XL,XU,XExp),
+		get(Y,YL,YU,YExp),
+		(YL >= 0, XL >= 0 ->
+			NXL is max(YL*Z,XL),
+			NXU is min(YU*(Z + 1) - 1,XU)
+		; % TODO: cover more cases
+			NXL = XL,
+			NXU = XU
+		),
+		( New == yes ->
+			put(X,NXL,NXU,[mydiv(Y,Z)|XExp]),
+			put(Y,YL,YU,[mydiv3(X,Z)|YExp])
+		;
+			put(X,NXL,NXU,XExp)
+		)
+	;
+		get(X,XL,XU,XExp),
+		get(Y,YL,YU,YExp),
+		get(Z,ZL,ZU,ZExp),
+		( New == yes ->
+			put(Y,YL,YU,[mydiv3(X,Z)|YExp]),
+			put(Z,ZL,ZU,[mydiv2(X,Y)|ZExp]),
+			put(X,XL,XU,[mydiv(Y,Z)|XExp])
+		;
+			true
+		)
+	).
+
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 mymax(X,Y,Z,New) :-
 	( nonvar(X) ->
@@ -805,7 +1076,7 @@ mymin(X,Y,Z,New) :-
 		)
 	).
 
-myabs(X,Y) :-
+myabs(X,Y,New) :-
 	( nonvar(X) ->
 		Y is abs(X)
 	; nonvar(Y) ->
@@ -817,20 +1088,32 @@ myabs(X,Y) :-
 			X = Y
 		;
 			NLX is (-Y),
-			put(X,NLX,Y,[myabs(Y)|ExpX])
+			( New == yes ->
+				put(X,NLX,Y,[myabs(Y)|ExpX])
+			;
+				put(X,NLX,Y,ExpX)
+			)
 		)
 	;
 		get(X,LX,UX,ExpX),
 		get(Y,LY,UY,ExpY),
 		UY > 0,
-		put(X,LX,UX,[myabs(Y)|ExpX]),
+		( New == yes ->
+			put(X,LX,UX,[myabs(Y)|ExpX])
+		;
+			true
+		),
 		( LX =< 0, UX >= 0 ->
 			NLY is max(0,LY)
 		;
 			NLY is max(min(abs(LX),abs(UX)),max(LY,0))
 		),
 		NUY is min(UY,max(abs(LX),abs(UX))),
-		put(Y,NLY,NUY,[myabs2(X)|ExpY]),
+		( New == yes ->
+			put(Y,NLY,NUY,[myabs2(X)|ExpY])
+		;
+			put(Y,NLY,NUY,ExpY)
+		),
 		( var(X) ->
 			get(X,LX2,UX2,ExpX2),
 			( LX2 == LX, UX2 == UX ->
@@ -857,7 +1140,7 @@ myabs(X,Y) :-
 		)
 	).
 
-mymod(X,Y,Z) :- % Z is X mod Y
+mymod(X,Y,Z,New) :- % Z is X mod Y
 	( nonvar(X) ->
 		( nonvar(Y) ->
 			Z is X mod Y
@@ -866,36 +1149,64 @@ mymod(X,Y,Z) :- % Z is X mod Y
 			( Z > 0 ->
 				( X == Z ->
 					no_overlap(-X,X,YL,YU,NYL,NYU),
-					put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					( New == yes ->
+						put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					;
+						put(Y,NYL,NYU,YExp)
+					)
 				; % Z < X
 					no_overlap(-Z,Z,YL,YU,YL1,YU1),
 					in_between(-X,X,YL1,YU1,NYL,NYU),
-					put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					( New == yes ->
+						put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					;
+						put(Y,NYL,NYU,YExp)
+					)
 				)
 			; Z < 0 ->
 				( X == Z ->
 					no_overlap(X,-X,YL,YU,NYL,NYU),
-					put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					( New == yes ->
+						put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					;
+						put(Y,NYL,NYU,YExp)
+					)
 				; % Z < X
 					no_overlap(Z,-Z,YL,YU,YL1,YU1),
 					in_between(X,-X,YL1,YU1,NYL,NYU),
-					put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					( New == yes ->
+						put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					;
+						put(Y,NYL,NYU,YExp)
+					)
 				)
 			; % Z == 0
 				( X == 0 ->
 					no_overlap(0,0,YL,YU,NYL,NYU),% Y \== 0
-					put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					( New == yes ->
+						put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					;
+						put(Y,NYL,NYU,YExp)
+					)
 				; % X \== 0
 					PX is abs(X),
 					in_between(-PX,PX,YL,YU,NYL,NYU),% |Y| < |X|
-					put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					( New == yes ->
+						put(Y,NYL,NYU,[mymod2(X,Z)|YExp])
+					;
+						put(Y,NYL,NYU,YExp)
+					)
 				)
 			)
 		; % var(Y), var(Z)
-			get(Y,YL,YU,YExp),
-			get(Z,ZL,ZU,ZExp),
-			put(Y,YL,YU,[mymod2(X,Z)|YExp]),
-			put(Z,ZL,ZU,[mymod3(X,Y)|ZExp])
+			( New == yes ->
+				get(Y,YL,YU,YExp),
+				get(Z,ZL,ZU,ZExp),
+				put(Y,YL,YU,[mymod2(X,Z)|YExp]),
+				put(Z,ZL,ZU,[mymod3(X,Y)|ZExp])
+			;
+				true
+			)
 		)
 	; nonvar(Y) ->
 		Y \== 0,
@@ -903,31 +1214,55 @@ mymod(X,Y,Z) :- % Z is X mod Y
 			get(X,XL,XU,XExp),
 			( Z > 0 ->
 				NXL is max(XL,1),
-				put(X,NXL,XU,[mymod1(Y,Z)|XExp])
+				( New == yes ->
+					put(X,NXL,XU,[mymod1(Y,Z)|XExp])
+				;
+					put(X,NXL,XU,XExp)
+				)
 			; Z < 0 ->
 				NXU is min(XU,-1),
-				put(X,XL,NXU,[mymod1(Y,Z)|XExp])
+				( New == yes ->
+					put(X,XL,NXU,[mymod1(Y,Z)|XExp])
+				;
+					put(X,XL,NXU,XExp)
+				)
 			;
-				put(X,XL,XU,[mymod1(Y,Z)|XExp])	
+				( New == yes ->
+					put(X,XL,XU,[mymod1(Y,Z)|XExp])
+				;
+					put(X,XL,XU,XExp)
+				)
 			)
 		;
-			get(X,XL,XU,XExp),
-			get(Z,ZL,ZU,ZExp),
-			put(X,XL,XU,[mymod1(Y,Z)|XExp]),
-			put(Z,ZL,ZU,[mymod3(X,Y)|ZExp])
+			( New == yes ->
+				get(X,XL,XU,XExp),
+				get(Z,ZL,ZU,ZExp),
+				put(X,XL,XU,[mymod1(Y,Z)|XExp]),
+				put(Z,ZL,ZU,[mymod3(X,Y)|ZExp])
+			;
+				true
+			)
 		)
 	; nonvar(Z) ->
-		get(X,XL,XU,XExp),
-		get(Y,YL,YU,YExp),
-		put(X,XL,XU,[mymod1(Y,Z)|XExp]),
-		put(Y,YL,YU,[mymod2(X,Z)|YExp])
+		( New == yes ->
+			get(X,XL,XU,XExp),
+			get(Y,YL,YU,YExp),
+			put(X,XL,XU,[mymod1(Y,Z)|XExp]),
+			put(Y,YL,YU,[mymod2(X,Z)|YExp])
+		;
+			true
+		)
 	;
-		get(X,XL,XU,XExp),
-		get(Y,YL,YU,YExp),
-		get(Z,ZL,ZU,ZExp),
-		put(X,XL,XU,[mymod1(Y,Z)|XExp]),
-		put(Y,YL,YU,[mymod2(X,Z)|YExp]),
-		put(Z,ZL,ZU,[mymod3(X,Y)|ZExp])
+		( New == yes ->
+			get(X,XL,XU,XExp),
+			get(Y,YL,YU,YExp),
+			get(Z,ZL,ZU,ZExp),
+			put(X,XL,XU,[mymod1(Y,Z)|XExp]),
+			put(Y,YL,YU,[mymod2(X,Z)|YExp]),
+			put(Z,ZL,ZU,[mymod3(X,Y)|ZExp])
+		;
+			true
+		)
 	).
 			
 no_overlap(XL,XU,YL,YU,NYL,NYU) :-
@@ -1014,6 +1349,44 @@ reified_geq(X,Y,B) :-
 		X #< Y
 	).
 
+check_leq(X,Y) :-
+	( nonvar(X) ->
+		( nonvar(Y) ->
+			X =< Y
+		;
+			bounds(Y,L,_),
+			X =< L
+		)
+	;
+		( nonvar(Y) ->
+			bounds(X,_,U),
+			U =< Y
+		;
+			bounds(X,_,UX),
+			bounds(Y,LY,_),
+			UX =< LY
+		)
+	).
+
+check_lt(X,Y) :-
+	( nonvar(X) ->
+		( nonvar(Y) ->
+			X < Y
+		;
+			bounds(Y,L,_),
+			X < L
+		)
+	;
+		( nonvar(Y) ->
+			bounds(X,_,U),
+			U < Y
+		;
+			bounds(X,_,UX),
+			bounds(Y,LY,_),
+			UX < LY
+		)
+	).
+			
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 reified_eq(X,Y,B) :-
 	( var(B) ->
@@ -1065,6 +1438,9 @@ reified_eq(X,Y,B) :-
 		X #\= Y
 	).
 		
+
+check_eq(X,Y) :-
+	X == Y.
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 reified_neq(X,Y,B) :-
 	mynot(B,B1),
@@ -1226,21 +1602,31 @@ put(X,L,U,Exp) :-
 				true
 			;
 				%format('\t~w in ~w .. ~w\n',[X,L,U]),
-				trigger_exps(Exp,X)
+				trigger_exps(Exp,X),
+				notify(X,bounds)
 			)
 		; 
 			%format('\t~w in ~w .. ~w\n',[X,L,U]),
-			put_attr(X,bounds,bounds(L,U,Exp))
+			put_attr(X,bounds,bounds(L,U,Exp)),
+			notify(X,bounds)
 		)
 	).
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
+%	SWI-Prolog 5.7.x has unbounded arithmetic.  What to do?
+
 min_inf(Inf) :-
-	current_prolog_flag(min_integer,MInf),
-	Inf is MInf + 1.
+	(   current_prolog_flag(min_integer,MInf)
+	->  Inf is MInf + 1
+	;   Inf = -9223372036854775807
+	).
 
 max_inf(Inf) :-
-	current_prolog_flag(max_integer,Inf).
+	(   current_prolog_flag(max_integer,Inf)
+	->  true
+	;   Inf = 9223372036854775807
+	).
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 attr_unify_hook(bounds(L,U,Exp),Other) :-
@@ -1345,16 +1731,41 @@ trigger_exp(myxor2(X,Y),Z) :-
 	myxor(X,Y,Z).
 
 trigger_exp(myabs(Y),X) :-
-	myabs(X,Y).
+	myabs(X,Y,no).
 trigger_exp(myabs2(X),Y) :-
-	myabs(X,Y).
+	myabs(X,Y,no).
 
 trigger_exp(mymod1(Y,Z),X) :-
-	mymod(X,Y,Z).
+	mymod(X,Y,Z,no).
 trigger_exp(mymod2(X,Z),Y) :-
-	mymod(X,Y,Z).
+	mymod(X,Y,Z,no).
 trigger_exp(mymod3(X,Y),Z) :-
-	mymod(X,Y,Z).
+	mymod(X,Y,Z,no).
+
+trigger_exp(mydiv(Y,Z),X) :-
+	mydiv(X,Y,Z,no).
+trigger_exp(mydiv2(A,B),X) :-
+	mydiv(A,B,X,no).
+trigger_exp(mydiv3(A,B),X) :-
+	mydiv(A,X,B,no).
+
+trigger_exp(rel_tuple(Relation,Tuple), _) :-
+	( ground(Tuple) ->
+		memberchk(Tuple, Relation)
+	;
+		relation_unifiable(Relation, Tuple, Us),
+		Us \= [],
+		( Us = [Single] ->
+			Single = Tuple
+		;
+			( Us == Relation ->
+				true
+			;
+				tuple_domain(Tuple, Us)
+			)
+		)
+	).
+	
 
 memberchk_eq(X,[Y|Ys],Z) :-
    (   X == Y ->
@@ -1362,6 +1773,102 @@ memberchk_eq(X,[Y|Ys],Z) :-
    ;   memberchk_eq(X,Ys,Z)
    ).
 
+
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
+tuples_in(Tuples, Relation) :-
+	ground(Relation),
+	bind_unique(Tuples, Relation),
+	approx_tuples_domain(Tuples, Relation),
+	tuples_freeze(Tuples, Relation).
+
+bind_unique([], _).
+bind_unique([Tuple|Ts], Relation) :-
+	relation_unifiable(Relation, Tuple, Us),
+	Us \= [],
+	( Us = [Single] ->
+		Single = Tuple
+	;
+		true
+	),
+	bind_unique(Ts, Relation).
+
+
+   % The following predicates find component-wise extrema in the relation.
+
+relation_mins_maxs(Relation, Mins, Maxs) :-
+	Relation = [R|_],
+	relation_mins_maxs(Relation, R, Mins, R, Maxs).
+
+relation_mins_maxs([], Mins, Mins, Maxs, Maxs).
+relation_mins_maxs([R|Rs], Mins0, Mins, Maxs0, Maxs) :-
+	components_mins_maxs(R, Mins0, Mins1, Maxs0, Maxs1),
+	relation_mins_maxs(Rs, Mins1, Mins, Maxs1, Maxs).
+
+components_mins_maxs([], [], [], [], []).
+components_mins_maxs([C|Cs], [Min0|Mins0], [Min|Mins], [Max0|Maxs0], [Max|Maxs]) :-
+	Min is min(Min0, C),
+	Max is max(Max0, C),
+	components_mins_maxs(Cs, Mins0, Mins, Maxs0, Maxs).
+
+
+   % Approximate the domains of each tuple variable for all tuples.
+   % Heuristics: Take component-wise mins/maxs of the relation.
+
+approx_tuples_domain(Tuples, Relation) :-
+	relation_mins_maxs(Relation, Mins, Maxs),
+	constrain_domains(Tuples, Mins, Maxs).
+
+constrain_domains([], _, _).
+constrain_domains([Tuple|Tuples], Mins, Maxs) :-
+	constrain_domain(Tuple, Mins, Maxs),
+	constrain_domains(Tuples, Mins, Maxs).
+
+constrain_domain([], [], []).
+constrain_domain([T|Ts], [Min|Mins], [Max|Maxs]) :-
+	T in Min..Max,
+	constrain_domain(Ts, Mins, Maxs).
+
+tuple_domain(Tuple, Relation) :-
+	relation_mins_maxs(Relation, Mins, Maxs),
+	constrain_domain(Tuple, Mins, Maxs).
+
+   % Set up the attributes for each tuple variable.
+   % Attributes are rel_tuple(Rel, Tuple) terms:
+   %    Rel: the relation of which Tuple must be an element
+   %    Tuple: the tuple to which this variable belongs
+   % Note that the variable is itself part of the Tuple of its attribute.
+
+tuple_freeze([],  _, _).
+tuple_freeze([T|Ts], Tuple, Relation) :-
+	( var(T) ->
+		get(T, L, U, Exp),
+		put(T, L, U, [rel_tuple(Relation,Tuple)|Exp])
+	;
+		true
+	),
+	tuple_freeze(Ts, Tuple, Relation).
+
+tuples_freeze([], _).
+tuples_freeze([Tuple|Tuples], Relation) :-
+	tuple_freeze(Tuple, Tuple, Relation),
+	tuples_freeze(Tuples, Relation).
+
+   % Find all tuples in the relation that can unify with Tuple, not
+   % taking into account attributes (like constraints), as that would
+   % trigger attr_unify_hook recursively.
+
+relation_unifiable([], _, []).
+relation_unifiable([R|Rs], Tuple, Us) :-
+	( unifiable(R, Tuple, _) ->
+		Us = [R|Rest]
+	;
+		Rest = Us
+	),
+	relation_unifiable(Rs, Tuple, Rest).
+
+
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 active_state(state(active)).
 is_active(state(active)).
