@@ -1,9 +1,32 @@
-/*  progman.pl,v 1.5 1995/08/16 11:49:43 jan Exp
+/*  $Id: progman.pl,v 1.20 2006/08/31 13:27:59 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Simple program consistency check predicates
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License
+    as published by the Free Software Foundation; either version 2
+    of the License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+    As a special exception, if you link this library with other files,
+    compiled with a Free Software compiler, to produce an executable, this
+    library does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however
+    invalidate any other reasons why the executable file might be covered by
+    the GNU General Public License.
 */
 
 :- module(progman,
@@ -13,13 +36,14 @@
 	    progman_make_group/1,	% +Group
 	    progman_make_group/2,	% +Group, +GroupFile
 	    progman_make_item/4,	% +Group, +Title, +CmdLine, +Cwd
+	    progman_make_item/5,	% +Group, +Title, +CmdLine, +Cwd, +Icon
 
 	    progman_setup/0		% Installs icons
 	  ]).
 
-:- (   feature(dde, true)
+:- (   current_prolog_flag(dde, true)
    ->  true
-   ;   '$warning'('Module "library(progman)" requires DDE support')
+   ;   print_message(error, missing_feature('DDE'))
    ).
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -27,7 +51,7 @@ MS-Windows  PROGMAN  interface  and  installation   for  SWI-Prolog  and
 XPCE/SWI-Prolog.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-%	progman_groups(-Groups)
+%%	progman_groups(-Groups)
 %
 %	Return list of atoms containing the titles of the currently
 %	available program groups.
@@ -35,11 +59,12 @@ XPCE/SWI-Prolog.
 progman_groups(Groups) :-
 	open_dde_conversation(progman, progman, DDE),
 	dde_request(DDE, groups, Lines),
-	lines_to_atoms(Lines, Groups0),
 	close_dde_conversation(DDE),
+	check_error(Lines),
+	lines_to_atoms(Lines, Groups0),
 	Groups = Groups0.
 
-%	progman_group_info(+Group, -File, -Items)
+%%	progman_group_info(+Group, -File, -Items)
 %
 %	Extracts info on the given group: Filename used to store the
 %	group and a list of item(Title, CmdLine, Dir) terms.
@@ -48,45 +73,54 @@ progman_group_info(Group, File, Items) :-
 	open_dde_conversation(progman, progman, DDE),
 	dde_request(DDE, Group, Info),
 	close_dde_conversation(DDE),
+	check_error(Info),
 	lines_to_atoms(Info, [GrounInfo|ItemLines]),
 	line_to_args(GrounInfo, [_,File|_]),
 	maplist(map_item_info, ItemLines, Items).
 
-%	progman_make_group(+Name, [+File])
+%%	progman_make_group(+Name) is det.
+%%	progman_make_group(+Name, +File) is det.
 %
 %	Create a group in the program manager.  If file is given, this
 %	is the file used by Windows to store the group info.
 
 progman_make_group(Name) :-
 	open_dde_conversation(progman, progman, DDE),
-	dde_execute(DDE, '[CreateGroup("~w")]', [Name]),
+	dde_fmt_execute(DDE, '[CreateGroup("~w")]', [Name]),
 	close_dde_conversation(DDE).
 progman_make_group(Name, File) :-
 	open_dde_conversation(progman, progman, DDE),
-	dde_execute(DDE, '[CreateGroup("~w", "~w")]', [Name, File]),
+	dde_fmt_execute(DDE, '[CreateGroup("~w", "~w")]', [Name, File]),
 	close_dde_conversation(DDE).
 
-%	progman_make_itme(+Group, +Title, +CmdLine, +Dir)
+%%	progman_make_item(+Group, +Title, +CmdLine, +Dir)
 %	
 %	Make a new program item in the named group.  If the item already
-%	exists, ask to delete the item.
+%	exists, delete it.
 
 progman_make_item(Group, Title, CmdLine, Dir) :-
+	progman_make_item(Group, Title, CmdLine, Dir, -).
+progman_make_item(Group, Title, CmdLine, Dir, Icon) :-
+	(   nonvar(Icon),
+	    Icon = IconFile:IconNum
+	->  true
+	;   Icon = IconFile
+	->  IconNum = ''
+	;   IconFile = '',
+	    IconNum = ''
+	),
 	progman_group_info(Group, _File, Items),
 	open_dde_conversation(progman, progman, DDE),
+	dde_fmt_execute(DDE, '[ShowGroup("~w", 1)]', Group),
 	(   memberchk(item(Title, _, _), Items)
-	->  (   '$confirm'('Replace item ~w from group ~w', [Title, Group])
-	    ->	dde_execute(DDE, '[ReplaceItem("~w")]', [Title])
-	    ;	format('Adding duplicate item~n', [])
-	    )
+	->  dde_fmt_execute(DDE, '[ReplaceItem("~w")]', [Title])
 	;   true
 	),
-	dde_execute(DDE, '[ShowGroup("~w", 1)]', Group),
-	dde_execute(DDE, '[addItem("~w", "~w",,,,, "~w",,)]',
-		    [CmdLine, Title, Dir]),
+	dde_fmt_execute(DDE, '[addItem(~w,~w,~w,~w,,,~w,,)]',
+			[CmdLine, Title, IconFile, IconNum, Dir]),
 	close_dde_conversation(DDE).
 
-%	program_group(+Default, -Group)
+%%	program_group(+Default, -Group)
 %
 %	Given a default group name, ask for a new name if this group
 %	already exists.
@@ -94,7 +128,7 @@ progman_make_item(Group, Title, CmdLine, Dir) :-
 program_group(Default, Group) :-
 	progman_groups(Existing),
 	memberchk(Default, Existing), !,
-	(   '$confirm'('Put (replace) items in existing group ~w', [Default])
+	(   '$confirm'(progman_replace(Default))
 	->  Group = Default
 	;   format('Enter new group name: '),
 	    read_line(NewDef),
@@ -103,50 +137,39 @@ program_group(Default, Group) :-
 program_group(Default, Default).
 
 		 /*******************************
-		 *	       INSTALL		*
+		 *     ERRORS AND WARNINGS	*
 		 *******************************/
 
-progman_setup :-
-	feature(xpce, true), !,
+%%	check_error(+Return)
+%
+%	Map return-codes into exeptions.
 
-	explain(start),
-	program_group('XPCE/SWI-Prolog', Group),
+check_error(error(Error)) :-
+	throw(error(dde_error, Error)).
+check_error(_).
 
-	feature(symbol_file, EXE),
-	feature(home, PlHomeDir),
-	file_directory_name(EXE, Bindir),
-	file_directory_name(Bindir, PceHomeDir),
-	concat(PlHomeDir, '/bin/pl.exe', PlExe),
-	prolog_to_os_filename(PlExe, OsPlExe),
-	prolog_to_os_filename(EXE, OsPceExe),
-	prolog_to_os_filename(PlHomeDir, OsPlHomeDir),
-	prolog_to_os_filename(PceHomeDir, OsPceHomeDir),
 
-	progman_make_group(Group),
-	progman_make_item(Group, 'SWI-Prolog', OsPlExe, 'c:'),
-	progman_make_item(Group, 'SWI-Prolog (Boot)',
-			  'bin\pl.exe -b boot/init.pl',
-			  OsPlHomeDir),
-	progman_make_item(Group, 'XPCE/SWI-Prolog', OsPceExe, 'c:'),
-	progman_make_item(Group, 'XPCE/SWI-Prolog (Boot)',
-			  'bin\xpce.exe -g pce_host:pce_banner -b ../pl/boot/init.pl -c prolog/load.pl',
-			  OsPceHomeDir),
-	explain(end).
+:- multifile
+	prolog:message/3.
+
+prolog:message(progman_replace(Default)) -->
+	[ 'Put (replace) items in existing group ~w? '-[Default], flush ].
+prolog:message(error(dde_error, Msg)) -->
+	[ 'DDE error: ~w'-[Msg] ].
+
+		 /*******************************
+		 *	       INSTALL		*
+		 *******************************/
 
 progman_setup :-
 	explain(start),
 
 	program_group('SWI-Prolog', Group),
-	feature(symbol_file, EXE),
-	feature(home, HomeDir),
-	prolog_to_os_filename(EXE, OsEXE),
-	prolog_to_os_filename(HomeDir, OsHomeDir),
+	current_prolog_flag(executable, PlExe),
+	prolog_to_os_filename(PlExe, OsPlExe),
 
 	progman_make_group(Group),
-	progman_make_item(Group, 'SWI-Prolog', OsEXE, 'c:'),
-	progman_make_item(Group, 'SWI-Prolog (Boot)',
-			  'bin\pl.exe -b boot/init.pl',
-			  OsHomeDir),
+	progman_make_item(Group, 'SWI-Prolog', OsPlExe, 'c:'),
 	explain(end).
 
 
@@ -161,12 +184,7 @@ explanation(start, '*******************************************************').
 explanation(start, '').
 
 explanation(end, '').
-explanation(end, 'I''ve created a program group containing the icons to').
-explanation(end, 'boot and run the system.  First you must execute the').
-explanation(end, 'icons tagged (Boot) to create the startup file.').
-explanation(end, '').
-explanation(end, 'After successfully running the boot compilation you').
-explanation(end, 'may use the normal execution icon to start the application').
+explanation(end, 'Program manager setup completed').
 explanation(end, '').
 
 explain(Id) :-
@@ -174,7 +192,7 @@ explain(Id) :-
 	format('~w~n', [X]),
 	fail ; true.
 
-%	line_to_args(+Line, -Args)
+%%	line_to_args(+Line, -Args)
 %
 %	Translate a line (Atom) as returned by PROGMAN's request for
 %	the contents of a group into a list of atomic arguments.  Arguments
@@ -220,7 +238,7 @@ string_val([C|M]) -->
 char([C], [C|T], T).
 end_of_string([], []).
 
-%	lines_to_atoms(+Lines, -Atoms)
+%%	lines_to_atoms(+Lines, -Atoms)
 %       
 %	Break a multiline answer from PROGMAN in multiple atoms, each
 %	describing a single line of the answer without the \r\n.
@@ -237,16 +255,20 @@ string_to_atoms([13,10|Rest], S0, [A|T]) :- !,
 	reverse(S0, S),
 	name(A, S),
 	string_to_atoms(Rest, [], T).
+string_to_atoms([10|Rest], S0, [A|T]) :- !,
+	reverse(S0, S),
+	name(A, S),
+	string_to_atoms(Rest, [], T).
 string_to_atoms([C|T], M, A) :-
 	string_to_atoms(T, [C|M], A).
 
-%	read_line(-Line)
+%%	read_line(-Line)
 %
 %	Flush pending output and read input upto a newline.  Return the
 %	entered line as an atom.
 
 read_line(Line) :-
-	flush,
+	flush_output,
 	prompt(O, ''),
 	read_chars(Chars),
 	prompt(_, O),
@@ -258,10 +280,10 @@ read_chars([C|T]) :-
 	read_chars(T).
 read_chars([]).
 
-%	dde_execute(+DdeId, +Format, +Args)
+%%	dde_fmt_execute(+DdeId, +Format, +Args)
 %
 %	Utility predicate to create DDE commands from a formatted spec.
 
-dde_execute(DDE, Fmt, Args) :-
-	sformat(Cmd, Fmt, Args),
+dde_fmt_execute(DDE, Fmt, Args) :-
+	format(string(Cmd), Fmt, Args),
 	dde_execute(DDE, Cmd).
