@@ -1,11 +1,11 @@
-/*  $Id: quintus.pl,v 1.38 2004/09/30 10:41:35 jan Exp $
+/*  $Id: quintus.pl,v 1.44 2006/08/18 07:01:53 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -81,12 +81,29 @@
 	  raise_exception/1,		% +Exception
 	  on_exception/3		% +Ball, :Goal, :Recover
 	]).
+:- use_module(library(lists), [member/2]).
+
+/** <module> Quintus compatibility
+
+This  module  defines  several  predicates    from  the  Quintus  Prolog
+libraries. Note that our library structure is totally different. If this
+library were complete, Prolog  code  could   be  ported  by removing the
+use_module/1 declarations, relying on the SWI-Prolog autoloader.
+
+Bluffers guide to porting:
+
+	* Remove =|use_module(library(...))|=
+	* Run =|?- list_undefined.|=
+	* Fix problems
+
+Of course, this library is incomplete ...
+*/
 
 		/********************************
 		*      SYSTEM INTERACTION       *
 		*********************************/
 
-%	unix(+Action)
+%%	unix(+Action)
 %	interface to  Unix.
 
 unix(system(Command)) :-
@@ -113,19 +130,12 @@ to_prolog(S, A) :-
 	name(A, L).
 
 
-%	file_exists(+File)
-%	Succeeds if `File' exists as a file or directory in the Unix file
-%	system.
-
-file_exists(File) :-
-	exists_file(File).
-
-
 		/********************************
 		*        META PREDICATES        *
 		*********************************/
 
-%	otherwise/0
+%%	otherwise
+%
 %	For (A -> B ; otherwise -> C)
 
 otherwise.
@@ -135,13 +145,20 @@ otherwise.
 		*          ARITHMETIC           *
 		*********************************/
 
-%	abs(+Number, -Absolute)
+%%	abs(+Number, -Absolute)
 %	Unify `Absolute' with the absolute value of `Number'.
 
 abs(Number, Absolute) :-
 	Absolute is abs(Number).
 
-%	Math library predicates
+%%	sin(+Angle, -Sine) is det.
+%%	cos(+Angle, -Cosine) is det.
+%%	tan(+Angle, -Tangent) is det.
+%%	log(+X, -NatLog) is det.
+%%	log10(+X, -Log) is det.
+%
+%	Math library predicates. SWI-Prolog (and   ISO) support these as
+%	functions under is/2, etc.
 
 sin(A, V) :-	  V is sin(A).
 cos(A, V) :-	  V is cos(A).
@@ -159,7 +176,7 @@ atan(X, V) :-	  V is atan(X).
 atan2(Y, X, V) :- V is atan(Y, X).
 sign(X, V) :-	  V is sign(X).
 
-%	random(+Min, +Max, -Value)
+%%	random(+Min, +Max, -Value)
 %
 %	Value is a random integer with Min <= Value < Max.  Note that
 %	Quintus random/3 also deals with floats.  This one doesn't.
@@ -171,6 +188,10 @@ random(Min, Max, Value) :-
 		 *	TERM MANIPULATION	*
 		 *******************************/
 
+%%	genarg(?Index, +Term, ?Arg) is nondet.
+%	
+%	Generalised version of ISO arg/3.  SWI-Prolog's arg/3 is already
+%	genarg/3.
 
 genarg(N, T, A) :-			% SWI-Prolog arg/3 is generic
 	arg(N, T, A).
@@ -180,9 +201,11 @@ genarg(N, T, A) :-			% SWI-Prolog arg/3 is generic
 		 *	      FLAGS		*
 		 *******************************/
 
-%	prolog_flag(?Flag, ?Value)
+%%	prolog_flag(?Flag, ?Value) is nondet.
 %
-%	Should map relevant Quintus flag identifiers.
+%	Same as ISO current_prolog_flag/2.  Maps =version=.
+%	
+%	@bug	Should map relevant Quintus flag identifiers.
 
 prolog_flag(version, Version) :- !,
 	current_prolog_flag(version, N),
@@ -211,19 +234,27 @@ prolog_flag(Flag, Value) :-
 		 *	     DATE/TIME		*
 		 *******************************/
 
-%	date(-Date)
+%%	date(-Date) is det.
 %
-%	Get current date.
+%	Get current date as date(Y,M,D)
 
-date(date(Year,Month,Day)) :-
+date(Date) :-
 	get_time(T),
-	convert_time(T, Year, Month, Day, _, _, _, _).
+	stamp_date_time(T, DaTime, local),
+	date_time_value(date, DaTime, Date).
 
 
 		/********************************
 		*          STYLE CHECK          *
 		*********************************/
 
+%%	no_style_check(Style) is det.
+%
+%	Same as SWI-Prolog =|style_check(-Style)|=.   The Quintus option
+%	=single_var= is mapped to =singleton=.
+%	
+%	@see style_check/1.
+
 q_style_option(single_var, singleton) :- !.
 q_style_option(Option, Option).
 
@@ -246,9 +277,11 @@ public(_).
 		 *	TERM MANIPULATION	*
 		 *******************************/
 
-%	temporary hack for subsumes_chk/2 in ordinary Prolog world.
-%	This comes from the SWI-Prolog port of ALE.
-%	If all works fine we might move this to the kernel.
+%%	subsumes_chk(@X, @Y) is semidet.
+%
+%	Temporary hack for subsumes_chk/2 in ordinary Prolog world. This
+%	comes from the SWI-Prolog port of  ALE.   If  all  works fine we
+%	might move this to the kernel.
 
 subsumes_chk(X,Y) :-
   \+ \+ (copy_term(Y,Y2),
@@ -263,6 +296,10 @@ subsumes_chk(X,Y) :-
 		 *	      TYPES		*
 		 *******************************/
 
+%%	simple(@Term) is semidet.
+%
+%	Term is atomic or a variable.
+
 simple(X) :-
 	(   atomic(X)
 	->  true
@@ -301,7 +338,7 @@ module_expansion_argument(N) :- integer(N).
 		 *	      STREAMS		*
 		 *******************************/
 
-%	current_stream(?Object, ?Mode, ?Stream)
+%%	current_stream(?Object, ?Mode, ?Stream)
 %
 %	SICStus/Quintus and backward compatible predicate.  New code should
 %	be using the ISO compatible stream_property/2.
@@ -321,14 +358,18 @@ stream_mode(write,  write).
 stream_mode(append, write).
 stream_mode(update, write).
 
-%	stream_position(+Stream, -Old, +New)
+%%	stream_position(+Stream, -Old, +New)
 
 stream_position(Stream, Old, New) :-
 	stream_property(Stream, position(Old)),
 	set_stream_position(Stream, New).
 
 
-%	skip_line(Stream)
+%%	skip_line is det.
+%%	skip_line(Stream) is det.
+%
+%	Skip  the  rest  of  the  current  line  (on  Stream).  Same  as
+%	=|skip(0'\n)|=.
 
 skip_line :-
 	skip(10).
@@ -340,6 +381,13 @@ skip_line(Stream) :-
 		 *	   COMPILATION		*
 		 *******************************/
 
+%%	compile(+Files) is det.
+%
+%	Compile   files.   SWI-Prolog   doesn't    distinguish   between
+%	compilation and consult.
+%	
+%	@see load_files/2.
+
 :- meta_predicate
 	compile(:).
 
@@ -350,12 +398,18 @@ compile(Files) :-
 		 *	   ATOM-HANDLING	*
 		 *******************************/
 
-%	atom_char/2
+%%	atom_char(+Char, -Code) is det.
+%%	atom_char(-Char, +Code) is det.
+%
+%	Same as ISO char_code/2.
 
 atom_char(Char, Code) :-
 	char_code(Char, Code).
 
-%	midstring(?ABC, ?B, ?AC, LenA, LenB, LenC)
+%%	midstring(?ABC, ?B, ?AC) is nondet.
+%%	midstring(?ABC, ?B, ?AC, LenA) is nondet.
+%%	midstring(?ABC, ?B, ?AC, LenA, LenB) is nondet.
+%%	midstring(?ABC, ?B, ?AC, LenA, LenB, LenC) is nondet.
 %
 %	Too difficult to explain.  See the Quintus docs.  As far as I
 %	understand them the code below emulates this function just fine.
@@ -389,14 +443,14 @@ midstring(ABC, B, AC, LenA, LenB, LenC) :-
 		 *	     EXCEPTIONS		*
 		 *******************************/
 
-%	raise_exception(+Term)
+%%	raise_exception(+Term)
 %
 %	Quintus compatible exception handling
 
 raise_exception(Term) :-
 	throw(Term).
 
-%	on_exception(+Template, :Goal, :Recover)
+%%	on_exception(+Template, :Goal, :Recover)
 
 :- meta_predicate
 	on_exception(+, :, :).
