@@ -1,4 +1,4 @@
-/*  $Id: system.pl,v 1.6 2004/09/30 10:41:35 jan Exp $
+/*  $Id: system.pl,v 1.9 2006/08/17 10:05:18 jan Exp $
 
     Part of SWI-Prolog
 
@@ -38,7 +38,17 @@
 
 :- style_check(+dollar).
 
-%	system_mode(+OnOff)
+/** <module> System utilities
+
+This module provides some tools to deal with system predicates. System
+predicates cannot be traced or redefined.
+
+@depreciated	Use :- set_prolog_flag(generate_debug_info, false) to
+		hide predicate internals from the tracer.
+*/
+
+%%	system_mode(+OnOff)
+%
 %	Switch the system into system or user mode.  When in system mode,
 %	system predicates loose most of their special properties, so it
 %	becomes possible to trace and even redefine them.  Use the latter
@@ -48,7 +58,7 @@
 
 system_mode(X) :-
 	var(X), !,
-	(   style_check(?dollar)
+	(   style_check(?(dollar))
 	->  X = on
 	;   X = off
 	).
@@ -57,7 +67,8 @@ system_mode(on) :-
 system_mode(off) :-
 	style_check(-dollar).
 
-%	system_module
+%%	system_module
+%
 %	Any predicate defined after this declaraction uptill the end of
 %	the file will become a system predicate. Normally invoked by a
 %	directive immediately following the module declaration.
@@ -69,7 +80,8 @@ system_module :-
 	lock_predicate/2,
 	unlock_predicate/2.
 
-%	lock_predicate(+Name, Arity)
+%%	lock_predicate(+Name, Arity)
+%
 %	Transform a predicate into a system predicate. 
 
 lock_predicate(Spec, Arity) :-
@@ -77,7 +89,8 @@ lock_predicate(Spec, Arity) :-
 	functor(Head, Name, Arity ),
 	$set_predicate_attribute(Module:Head, system, 1).
 
-%	unlock_predicate(+Name, Arity)
+%%	unlock_predicate(+Name, Arity)
+%
 %	Transform a system predicate into a normal system predicate.
 
 unlock_predicate(Spec, Arity) :-
