@@ -1,4 +1,4 @@
-/*  $Id: nonblockio.c,v 1.3.2.2 2004/12/16 14:13:14 jan Exp $
+/*  $Id: nonblockio.c,v 1.10 2005/12/27 20:32:29 jan Exp $
 
     Part of SWI-Prolog
 
@@ -333,7 +333,7 @@ waitRequest(plsocket *s)
     { TranslateMessage(&msg);
       DispatchMessage(&msg);
     } else
-    { ExitProcess(0);			/* WM_QUIT received */
+    { ExitThread(0);			/* WM_QUIT received */
       return FALSE;			/* NOTREACHED */
     }
   }
@@ -374,7 +374,7 @@ nbio_wait(int socket, nbio_request request)
     { TranslateMessage(&msg);
       DispatchMessage(&msg);
     } else
-    { ExitProcess(0);			/* WM_QUIT received */
+    { ExitThread(0);			/* WM_QUIT received */
       return -1;			/* NOTREACHED */
     }
   }
@@ -457,7 +457,7 @@ nbio_select(int n,
 	{ TranslateMessage(&msg);
 	  DispatchMessage(&msg);
 	} else
-	{ ExitProcess(0);		/* WM_QUIT received */
+	{ ExitThread(0);		/* WM_QUIT received */
 	  return -1;			/* NOTREACHED */
 	}
       } else if ( rc == WAIT_TIMEOUT )
@@ -470,7 +470,7 @@ nbio_select(int n,
       { TranslateMessage(&msg);
 	DispatchMessage(&msg);
       } else
-      { ExitProcess(0);			/* WM_QUIT received */
+      { ExitThread(0);			/* WM_QUIT received */
 	return -1;			/* NOTREACHED */
       }
     }
@@ -734,7 +734,7 @@ SocketHiddenWindow()
 			   NULL, NULL, hinstance, NULL);
     assert(s->hwnd);
     DEBUG(1, Sdprintf("%d created hidden window %p\n",
-		   PL_thread_self(), s->hwnd));
+		      PL_thread_self(), s->hwnd));
   }
 
   return s->hwnd;
@@ -848,6 +848,10 @@ nbio_select(int n,
 
 static functor_t FUNCTOR_module2;
 static functor_t FUNCTOR_ip4;
+static functor_t FUNCTOR_ip1;
+static atom_t ATOM_any;
+static atom_t ATOM_broadcast;
+static atom_t ATOM_loopback;
 
 static plsocket *sockets;
 static int initialised = FALSE;		/* Windows only */
@@ -1140,6 +1144,10 @@ nbio_init(const char *module)
 
   FUNCTOR_module2 = PL_new_functor(PL_new_atom(":"), 2);
   FUNCTOR_ip4     = PL_new_functor(PL_new_atom("ip"), 4);
+  FUNCTOR_ip1     = PL_new_functor(PL_new_atom("ip"), 1);
+  ATOM_any	  = PL_new_atom("any");
+  ATOM_broadcast  = PL_new_atom("broadcast");
+  ATOM_loopback   = PL_new_atom("loopback");
 
 #ifdef WIN32
 { WSADATA WSAData;
@@ -1265,6 +1273,18 @@ nbio_setopt(int socket, nbio_option opt, ...)
 
       break;
     }
+    case UDP_BROADCAST:
+    { int val = va_arg(args, int);
+
+      if( setsockopt(socket, SOL_SOCKET, SO_BROADCAST,
+		     (const char *)&val, sizeof(val)) == -1)
+      { nbio_error(h_errno, TCP_HERRNO);
+	rc = -1;
+      } else
+	rc = 0;
+
+      break;
+    }
     case TCP_DISPATCH:
     { int val = va_arg(args, int);
       plsocket *s = lookupSocket(socket);
@@ -1354,18 +1374,27 @@ or the name of a registered port (e.g. 'smtp').
 
 int
 nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr)
-{ struct hostent *host;
-  char           *hostName = NULL;
-  int		  port;
+{ int		  port;
 	
   addr->sin_family = AF_INET;
+  addr->sin_addr.s_addr = INADDR_ANY;
 
   if ( PL_is_functor(Address, FUNCTOR_module2) )
-  { term_t arg = PL_new_term_ref();
+  { char *hostName;
+    term_t arg = PL_new_term_ref();
 
     PL_get_arg(1, Address, arg);
-    if ( !PL_get_atom_chars(arg, &hostName) )
-      return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1, arg, "atom");
+    if ( PL_get_atom_chars(arg, &hostName) )
+    { struct hostent *host;
+
+      if( !(host = gethostbyname(hostName)) )
+	return nbio_error(h_errno, TCP_HERRNO);
+      if ( (int)sizeof(addr->sin_addr) < host->h_length )
+	return PL_warning("Oops, host address too long!");
+      memcpy(&addr->sin_addr, host->h_addr, host->h_length);
+    } else if ( !nbio_get_ip(arg, &addr->sin_addr) )
+    { return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1, arg, "atom|ip/4");
+    }
 
     PL_get_arg(2, Address, arg);
     if ( !nbio_get_port(arg, &port) )
@@ -1375,15 +1404,6 @@ nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr)
   } else if ( !nbio_get_port(Address, &port) )
     return FALSE;
 
-  if ( hostName )
-  { if( !(host = gethostbyname(hostName)) )
-      return nbio_error(h_errno, TCP_HERRNO);
-    if ( (int)sizeof(addr->sin_addr) < host->h_length )
-      return PL_warning("Oops, host address too long!");
-    memcpy(&addr->sin_addr, host->h_addr, host->h_length);
-  } else
-    addr->sin_addr.s_addr = INADDR_ANY;
-	
   addr->sin_port = htons((short)port);
 
   return TRUE;
@@ -1409,6 +1429,23 @@ nbio_get_ip(term_t ip4, struct in_addr *ip)
     memcpy(ip, &hip, sizeof(hip));
 
     return TRUE;
+  } else if ( PL_is_functor(ip4, FUNCTOR_ip1) )
+  { term_t a = PL_new_term_ref();
+    atom_t id;
+
+    PL_get_arg(1, ip4, a);
+    if ( PL_get_atom(a, &id) )
+    { if ( id == ATOM_any )
+	ip->s_addr = INADDR_ANY;
+      else if ( id == ATOM_broadcast )
+	ip->s_addr = INADDR_BROADCAST;
+      else if ( id == ATOM_loopback )
+	ip->s_addr = INADDR_LOOPBACK;
+      else
+	return FALSE;
+
+      return TRUE;
+    }
   }
 
   return FALSE;
@@ -1485,7 +1522,7 @@ nbio_connect(int socket,
 
 
 int
-nbio_accept(int master, struct sockaddr *addr, size_t *addrlen)
+nbio_accept(int master, struct sockaddr *addr, socklen_t *addrlen)
 { int slave;
   plsocket *m;
 	
