@@ -1,4 +1,4 @@
-/*  $Id: http_header.pl,v 1.3 2004/09/21 14:22:24 jan Exp $
+/*  $Id: http_header.pl,v 1.22 2006/08/21 09:34:44 jan Exp $
 
     Part of SWI-Prolog
 
@@ -35,13 +35,19 @@
 	    http_reply/2,		% +What, +Stream
 	    http_reply/3,		% +What, +Stream, +HdrExtra
 
+	    http_timestamp/2,		% +Time, -HTTP string
+
 	    http_post_data/3,		% +Stream, +Data, +HdrExtra
 
 	    http_read_header/2,		% +Fd, -Header
-	    http_join_headers/3		% +Default, +InHdr, -OutHdr
+	    http_join_headers/3,	% +Default, +InHdr, -OutHdr
+	    http_update_encoding/3	% +HeaderIn, -Encoding, -HeaderOut
 	  ]).
 :- use_module(library(readutil)).
 :- use_module(library(debug)).
+:- use_module(library(lists)).
+:- use_module(library(url)).
+:- use_module(library(memfile)).
 :- use_module(dcg_basics).
 :- use_module(html_write).
 :- use_module(mimetype).
@@ -52,26 +58,35 @@
 		 *	    READ REQUEST	*
 		 *******************************/
 
-%	http_read_request(+FdIn, -Request)
+%%	http_read_request(+FdIn, -Request)
 %
 %	Read an HTTP request-header from FdIn and return the broken-down
 %	request fields as +Name(+Value) pairs in a list.
 
 http_read_request(In, [input(In)|Request]) :-
 	read_line_to_codes(In, Codes),
-	debug(header, 'First line: ~s~n', [Codes]),
-	phrase(request(In, Request), Codes).
+	(   Codes == end_of_file
+	->  debug(header, 'end-of-file', [])
+	;   debug(header, 'First line: ~s~n', [Codes]),
+	    phrase(request(In, Request), Codes)
+	).
 
-%	http_read_reply_header(+FdIn, -Reply)
+
+%%	http_read_reply_header(+FdIn, -Reply)
 %
-%	Read the HTTP reply header (if any).
+%	Read the HTTP reply header. Throws   an exception if the current
+%	input does not contain a valid reply header.
 
 http_read_reply_header(In, [input(In)|Reply]) :-
 	read_line_to_codes(In, Codes),
-	debug(header, 'First line: ~s~n', [Codes]),
-	(   phrase(reply(In, Reply), Codes)
-	->  true
-	;   throw(error(syntax(http_reply_header, Codes), _))
+	(   Codes == end_of_file
+	->  debug(header, 'end-of-file', []),
+	    throw(error(syntax(http_reply_header, end_of_file), _))
+	;   debug(header, 'First line: ~s~n', [Codes]),
+	    (   phrase(reply(In, Reply), Codes)
+	    ->  true
+	    ;   throw(error(syntax(http_reply_header, Codes), _))
+	    )
 	).
 
 
@@ -79,6 +94,30 @@ http_read_reply_header(In, [input(In)|Reply]) :-
 		 *	  FORMULATE REPLY	*
 		 *******************************/
 
+%%	http_reply(+Data, +Out:stream) is det.
+%%	http_reply(+Data, +Out:stream, +HdrExtra) is det.
+%
+%	Data is one of
+%	
+%		* html(HTML)
+%		HTML tokens as produced by html//1 from html_write.pl
+%		
+%		* file(+MimeType, +FileName)
+%		Reply content of FileName using MimeType
+%		
+%		* tmp_file(+MimeType, +FileName)
+%		Same as =file=, but do not include modification time
+%		
+%		* stream(+In, +Len)
+%		Reply content of stream.
+%		
+%		* cgi_stream(+In, +Len)
+%		Reply content of stream, which should start with an
+%		HTTP header, followed by a blank line.  This is the
+%		typical output from a CGI script.
+%		
+%	@tbd	Complete documentation	
+
 http_reply(What, Out) :-
 	http_reply(What, Out, [connection(close)]).
 
@@ -89,15 +128,15 @@ http_reply(html(HTML), Out, HrdExtra) :- !,
 http_reply(file(Type, File), Out, HrdExtra) :- !,
 	phrase(reply_header(file(Type, File), HrdExtra), Header),
 	format(Out, '~s', [Header]),
-	open(File, read, In),
-	copy_stream_data(In, Out),
-	close(In).
+	open(File, read, In, [type(binary)]),
+	call_cleanup(copy_stream_data(In, Out),
+		     close(In)).
 http_reply(tmp_file(Type, File), Out, HrdExtra) :- !,
 	phrase(reply_header(tmp_file(Type, File), HrdExtra), Header),
 	format(Out, '~s', [Header]),
-	open(File, read, In),
-	copy_stream_data(In, Out),
-	close(In).
+	open(File, read, In, [type(binary)]),
+	call_cleanup(copy_stream_data(In, Out),
+		     close(In)).
 http_reply(stream(In, Len), Out, HdrExtra) :- !,
 	phrase(reply_header(cgi_data(Len), HdrExtra), Header),
 	format(Out, '~s', [Header]),
@@ -178,11 +217,11 @@ http_reply(not_modified, Out, HrdExtra) :- !,
 	format(Out, '~s', [Header]),
 	print_html(Out, HTML).
 http_reply(server_error(ErrorTerm), Out, HrdExtra) :-
-	message_to_html(ErrorTerm, Tokens),
+	'$messages':translate_message(ErrorTerm, Lines, []),
 	phrase(page([ title('500 Internal server error')
 		    ],
 		    [ h1('Internal server error'),
-		      p(Tokens),
+		      p(\html_message_lines(Lines)),
 		      address(httpd)
 		    ]),
 	       HTML),
@@ -191,30 +230,26 @@ http_reply(server_error(ErrorTerm), Out, HrdExtra) :-
 	print_html(Out, HTML).
 
 
-
-message_to_html(Term, Tokens) :-
-	'$messages':translate_message(Term, Lines, []),
-	phrase(html_message_lines(Lines), Tokens).
-
 html_message_lines([]) -->
 	[].
 html_message_lines([nl|T]) --> !,
-	['<br>'],
+	html([br([])]),
 	html_message_lines(T).
 html_message_lines([flush]) -->
 	[].
 html_message_lines([Fmt-Args|T]) --> !,
 	{ sformat(S, Fmt, Args)
 	},
-	[S],
+	html([S]),
 	html_message_lines(T).
 html_message_lines([Fmt|T]) --> !,
 	{ sformat(S, Fmt, [])
 	},
-	[S],
+	html([S]),
 	html_message_lines(T).
 
-%	http_join_headers(+Default, +Header, -Out)
+%%	http_join_headers(+Default, +Header, -Out)
+%	
 %	Append headers from Default to Header if they are not
 %	already part of it.
 
@@ -227,10 +262,63 @@ http_join_headers([H|T], Hdr0, Hdr) :-
 http_join_headers([H|T], Hdr0, [H|Hdr]) :-
 	http_join_headers(T, Hdr0, Hdr).
 
+
+%%	http_update_encoding(+HeaderIn, -Encoding, -HeaderOut)
+%	
+%	Allow for rewrite of the  header,   adjusting  the  encoding. We
+%	distinguish three options. If  the   user  announces  `text', we
+%	always use UTF-8 encoding. If   the user announces charset=utf-8
+%	we  use  UTF-8  and  otherwise  we  use  octet  (raw)  encoding.
+%	Alternatively we could dynamically choose for ASCII, ISO-Latin-1
+%	or UTF-8.
+
+http_update_encoding(Header0, utf8, [content_type(Type)|Header]) :-
+	select(content_type(Type0), Header0, Header),
+	sub_atom(Type0, 0, _, _, 'text/'), !,
+	(   sub_atom(Type0, S, _, _, ';')
+	->  sub_atom(Type0, 0, B, _, S)
+	;   B = Type0
+	),
+	atom_concat(B, '; charset=UTF-8', Type).
+http_update_encoding(Header, utf8, Header) :-
+	memberchk(content_type(Type), Header),
+	(   sub_atom(Type, _, _, _, 'UTF-8')
+	;   sub_atom(Type, _, _, _, 'utf-8')
+	), !.
+http_update_encoding(Header, octet, Header).
+
+
+%%	content_length_in_encoding(+Encoding, +In, -Bytes)
+%	
+%	Determine hom much bytes are required to represent the data from
+%	stream In using the given encoding.  Fails if the data cannot be
+%	represented with the given encoding.
+
+content_length_in_encoding(Enc, Stream, Bytes) :-
+	open_null_stream(Out),
+	set_stream(Out, encoding(Enc)),
+	stream_property(Stream, position(Here)),
+	(   catch((copy_stream_data(Stream, Out),
+		   flush_output(Out)), _, fail)
+	->  byte_count(Out, Bytes0)
+	;   true
+	),
+	close(Out),
+	set_stream_position(Stream, Here),
+	(   var(Bytes0)
+	->  fail
+	;   Bytes = Bytes0
+	).
+
+
 		 /*******************************
 		 *	    POST SUPPORT	*
 		 *******************************/
 
+%%	http_post_data(+Data, +Out:stream, +HdrExtra) is det.
+%
+%	@tbd	Complete docs
+
 http_post_data(html(HTML), Out, HdrExtra) :-
 	phrase(post_header(html(HTML), HdrExtra), Header),
 	format(Out, '~s', [Header]),
@@ -244,17 +332,22 @@ http_post_data(file(File), Out, HdrExtra) :- !,
 http_post_data(file(Type, File), Out, HdrExtra) :- !,
 	phrase(post_header(file(Type, File), HdrExtra), Header),
 	format(Out, '~s', [Header]),
-	open(File, read, In),
-	copy_stream_data(In, Out),
-	close(In).
-http_post_data(cgi_stream(In, Len), Out, HdrExtra) :- !,
-	http_read_header(In, CgiHeader),
-	seek(In, 0, current, Pos),
-	Size is Len - Pos,
-	http_join_headers(HdrExtra, CgiHeader, Hdr2),
-	phrase(reply_header(cgi_data(Size), Hdr2), Header),
-	format(Out, '~s', [Header]),
-	copy_stream_data(In, Out, Size).
+	open(File, read, In, [type(binary)]),
+	call_cleanup(copy_stream_data(In, Out),
+		     close(In)).
+http_post_data(cgi_stream(In, _Len), Out, HdrExtra) :- !,
+	debug(obsolete, 'Obsolete 2nd argument in cgi_stream(In,Len)', []),
+	http_post_data(cgi_stream(In), Out, HdrExtra).
+http_post_data(cgi_stream(In), Out, HdrExtra) :- !,
+	http_read_header(In, Header0),
+	http_update_encoding(Header0, Encoding, Header),
+	content_length_in_encoding(Encoding, In, Size),
+	http_join_headers(HdrExtra, Header, Hdr2),
+	phrase(reply_header(cgi_data(Size), Hdr2), HeaderText),
+	format(Out, '~s', [HeaderText]),
+	set_stream(Out, encoding(Encoding)),
+	call_cleanup(copy_stream_data(In, Out),
+		     set_stream(Out, encoding(octet))).
 http_post_data(form(Fields), Out, HdrExtra) :- !,
 	parse_url_search(Codes, Fields),
 	length(Codes, Size),
@@ -322,9 +415,6 @@ post_header(cgi_data(Size), HdrExtra) -->
 		 *       OUTPUT HEADER DCG	*
 		 *******************************/
 
-reply_header(What) -->
-	reply_header(What, []).
-
 reply_header(string(String), HdrExtra) -->
 	reply_header(string(text/plain, String), HdrExtra).
 reply_header(string(Type, String), HdrExtra) -->
@@ -480,7 +570,7 @@ header_field(Name, Value) -->
 	}, !,
 	field_name(Name),
 	":",
-	blanks,
+	whites,
 	string(ValueChars),
 	blanks_to_nl, !,
 	{ field_to_prolog(Name, ValueChars, Value)
@@ -499,6 +589,13 @@ field_to_prolog(cookie, ValueChars, Cookies) :- !,
 field_to_prolog(set_cookie, ValueChars, SetCookie) :- !,
 	debug(cookie, 'SetCookie: ~s', [ValueChars]),
 	phrase(set_cookie(SetCookie), ValueChars).
+field_to_prolog(host, ValueChars, Host) :- !,
+	(   append(HostChars, [0':|PortChars], ValueChars),
+	    catch(number_codes(Port, PortChars), _, fail)
+	->  atom_codes(HostName, HostChars),
+	    Host = HostName:Port
+	;   atom_codes(Host, ValueChars)
+	).
 field_to_prolog(_, ValueChars, Atom) :-
 	atom_codes(Atom, ValueChars).
 
@@ -533,7 +630,7 @@ header_fields([H|T]) -->
 	"\r\n",
 	header_fields(T).
 
-%	field_name(?PrologName)
+%%	field_name(?PrologName)
 %
 %	Convert between prolog_name and HttpName
 
@@ -580,28 +677,30 @@ wr_field_chars2([]) -->
 	[].
 
 %	now
-%	rfc_date(+Time)
+%%	rfc_date(+Time)
 
 now -->
 	{ get_time(Time)
 	},
 	rfc_date(Time).
 
-rfc_date(Time) -->
-	{ convert_time(Time, CDate)
-	},
-	sub(CDate, 1-3),
-	", ",
-	sub(CDate, 5-7),
-	sub(CDate, 21-4),
-	sub(CDate, 11-9).
+%%	rfc_date(+Time)// is det.
+%	
+%	Write time according to RFC1123 specification as required by the
+%	RFC2616 HTTP protocol specs. 
 
-sub(String, From-Len) -->
-	{ substring(String, From, Len, S),
-	  string_to_list(S, Chars)
-	},
-	string(Chars).
+rfc_date(Time, String, Tail) :-
+	stamp_date_time(Time, Date, 'UTC'),
+	format_time(codes(String, Tail), '%a, %d %b %Y %H:%M:%S GMT', Date).
 	
+%%	http_timestamp(+Time, -Atom)
+%	
+%	Generate a description of a Time in HTTP format (RFC1123)
+
+http_timestamp(Time, Atom) :-
+	stamp_date_time(Time, Date, 'UTC'),
+	format_time(atom(Atom), '%a, %d %b %Y %H:%M:%S GMT', Date).
+
 
 		 /*******************************
 		 *	   REQUEST DCG		*
@@ -625,8 +724,10 @@ request(Fd, [unknown(What)|Header]) -->
 
 method(get) -->
 	"GET", !.
-method(post) -->
+method(put) -->
 	"PUT", !.
+method(head) -->
+	"HEAD", !.
 method(post) -->
 	"POST", !.
 
@@ -759,7 +860,7 @@ reply(Fd, [http_version(HttpVersion), status(Status, Comment)|Header]) -->
 		 *	      READ HEADER	*
 		 *******************************/
 
-%	http_read_header(+Fd, -Header)
+%%	http_read_header(+Fd, -Header)
 %
 %	Read Name: Value lines from FD until an empty line is encountered.
 %	Field-name are converted to Prolog conventions (all lower, _ instead
@@ -785,13 +886,16 @@ parse_header(Text, Header) :-
 	phrase(header(Header), Text),
 	debug(header, 'Fields: ~w~n', [Header]).
 
-
-header([Att|T]) -->
+header(List) -->
 	header_field(Name, Value), !,
-	{ Att =.. [Name, Value]
+	{ mkfield(Name, Value, List, Tail)
 	},
 	blanks,
-	header(T).
+	header(Tail).
 header([]) -->
 	blanks,
 	[].
+
+mkfield(host, Host:Port, [host(Host),port(Port)|Tail], Tail) :- !.
+mkfield(Name, Value, [Att|Tail], Tail) :-
+	Att =.. [Name, Value].
