@@ -1,11 +1,11 @@
-/*  $Id: latex2html.pl,v 1.44 2004/10/05 10:15:24 jan Exp $
+/*  $Id: latex2html.pl,v 1.66 2006/08/29 10:25:57 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -42,7 +42,6 @@
 	    translate_command/4,	% +In, +ModeIn, -ModeOut, -HTML
 	    translate_environment/4,	% +In, +ModeIn, -ModeOut, -HTML
 	    translate_reference/4,	% +Kind, +RefPrefix, +Label, -HTML
-	    translate_footnote/2,	% +Text, -HTML
 	    translate_table/3,		% +Format, +BodyTokens, -HTML
 	    translate_section/4,	% +Level, -/*, +Title, -HTML
 	    translate_section/5,	% +Level, -/*, +Title, -HTML, +FileBase
@@ -55,9 +54,10 @@
 	  ]).
 :- use_module(library(quintus)).
 
-version('0.96').			% for SWI-Prolog 5.0.6
-page_header('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2//EN">\n\n').
+version('0.98').			% for SWI-Prolog 5.6.18
 
+page_header('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" \
+               "http://www.w3.org/TR/html4/strict.dtd">\n\n').
 :- dynamic			
 	html_output_dir/1,		% output relative to this dir
 	tex_file_base/1,		% Basename of the main file
@@ -186,7 +186,6 @@ author('Anonymous').
 
 
 run_latex2html(TeXFile) :-
-	reset_footnotes,
 	reset_counters,
 	reset_sections,
 	reset_index,
@@ -202,11 +201,9 @@ run_latex2html(TeXFile) :-
 	),
 	once(html_file_base(Base)),
 	expand_macros([tell(Base), #header], Header),
-	footnote_html(FootNotes),
 	make_index(Index),
 	flatten([ Header,
 		  HTML,
-		  FootNotes,
 		  Index
 		],
 		HtmlDocument),
@@ -336,7 +333,7 @@ translate_2($(Expr), Mode, Mode, #var(HTML)) :-		% $...$
 	tex_atom_to_tokens(Expr, Tokens),
 	translate(Tokens, math, _, HTML).
 translate_2(verbatim(_Cmd, Text), Mode, Mode, 		% \begin{verbatim} ...
-	    #listing(#pre(pre(Text)))).
+	    #pre(code, pre(Text))).
 translate_2(verb(_, Text), pcecode, pcecode, pre(Text)).
 translate_2(Layout, pcecode, pcecode, []) :-
 	atomic(Layout), !.
@@ -350,6 +347,7 @@ translate_2(Group, Mode, Mode, HTML) :-			% {...}
 	Group = [_|_],
 	translate_group(Group, HTML).
 translate_2(~, Mode, Mode, html('&nbsp;')).		% ~
+translate_2('\\[]', Mode, Mode, html('[]')). 		% []
 translate_2(Atom0, Mode, Mode, Atom) :-			% Normal word
 	atomic(Atom0),
 	(   Mode = group(Atts),
@@ -423,8 +421,17 @@ language_map(table,	'Table').
 
 #(tell(_File),		[]) :- onefile(true).
 #(tell(File),		tell(File)).
-#(head(Head),		[html('<HEAD>'), HtmlHead, html('</HEAD>')]) :-
-	expand_macros(Head, HtmlHead).
+#(head(Head),		[html('<HEAD>'), HtmlHead, Style, html('</HEAD>')]) :-
+	expand_macros(Head, HtmlHead),
+	expand_macros(#style, Style).
+#(style,		[ html('<STYLE type="text/css">'),
+			  html(Style),
+			  html('</STYLE>')
+			]) :-
+	absolute_file_name(library('latex2html.css'), StyleFile,
+			   [ access(read) ]),
+	read_file_to_codes(StyleFile, Codes, []),
+	atom_codes(Style, Codes).
 #(beginbody,		html(Body)) :-
 	bodycolor(Colour), !,
 	sformat(Body, '<BODY BGCOLOR="~w">', [Colour]).
@@ -440,6 +447,8 @@ language_map(table,	'Table').
 #(var(Text),		[html('<VAR>'),    Text, html('</VAR>')]).
 #(code(Text),		[html('<CODE>'),   Text, html('</CODE>')]).
 #(pre(Text),		[html('<PRE>'),    Text, html('</PRE>')]).
+#(pre(Class, Text),	[html(Begin),      Text, html('</PRE>')]) :-
+		format(atom(Begin), '<PRE class="~w">', [Class]).
 #(xmp(Text),		[html('<XMP>'),    Text, html('</XMP>')]).
 #(strong(Text),		[html('<STRONG>'), Text, html('</STRONG>')]).
 #(em(Text),		[html('<EM>'),     Text, html('</EM>')]).
@@ -448,23 +457,36 @@ language_map(table,	'Table').
 #(tt(Text),		[html('<TT>'),	   Text, html('</TT>')]).
 #(sc(Text),		[html('<font size=-1>'), % TBD: upcase Text
 					   Text, html('</font>')]).
+#(div(Class, Text),	[html(Begin), Text, html('</DIV>')]) :-
+	format(atom(Begin), '<DIV class="~w">', [Class]).
+#(span(Class, Text),	[html(Begin), Text, html('</SPAN>')]) :-
+	format(atom(Begin), '<SPAN class="~w">', [Class]).
 #(center(Text),		[html('<CENTER>'), Text, html('</CENTER>')]).
+#(navigate(Text),	[html('<DIV class="navigate">'), Text, html('</DIV>')]).
 #(right(Text),		[html('<RIGHT>'),  Text, html('</RIGHT>')]).
 #(quote(Text),		[html('<BLOCKQUOTE>'), Text, html('</BLOCKQUOTE>')]).
 #(listing(Text),	[html('<P><TABLE WIDTH="90%" ALIGN=center BORDER=6 BGCOLOR="#e0e0e0"><TR><TD NOWRAP>'), Text,
 			 html('</TABLE>')]).
-#(abstract(Text),	[html('<CENTER><H3>Abstract</H3></Center>'),
-			 html('<TABLE WIDTH="90%" ALIGN=center BORDER=2 BGCOLOR="#f0f0f0"><TR><TD>'), Text,
-			 html('</TABLE>')]).
-#(embrace([O,C],Text),	[nospace(OA),	   Text, nospace(CA)]) :-
+#(abstract(Text),	[ html('<DIV class="abstract">'),
+			  html('<DIV class="abstract-title">Abstract</DIV>'),
+			  Text,
+			  html('</DIV>')
+			]).
+#(footnote(Text),	#footnote(Mark, Text)) :-
+	step_counter(footnote, Mark).
+#(footnote(Mark, Text),	[ html('<SUP class="fn">'), Mark,
+			  #span('fn-text', Text),
+			  html('</SUP>')
+			]).
+#(embrace([O,C],Text),	[nospace(OA), Text, nospace(CA)]) :-
 	char_code(OA, O),
 	char_code(CA, C).
 #(embrace(Text),	#embrace("()", Text)).
 #(h(Level, Title),	[html(OpenH), Title, html(CloseH)]) :-
 	h(Level, OpenH, CloseH).
-#(predref(RN, Arity),   #lref(Text, Text)) :-
+#(predref(RN, Arity),   #lref(pred, Text, Text)) :-
 	clean_tt(RN, Name),
-	sformat(Text, '~w/~w', [Name, Arity]).
+	format(string(Text), '~w/~w', [Name, Arity]).
 #(row(Columns),		[html('<TR>'), HtmlCols, html('</TR>')]) :-
 	add_td(Columns, HtmlCols).
 #(label(Lbl, Text, Tag),label(ALabel, Expanded, Tag)) :-
@@ -473,9 +495,11 @@ language_map(table,	'Table').
 #(label(Lbl, Text),	label(ALabel, Expanded, -)) :-
 	string_to_atom(Lbl, ALabel),
 	expand_macros(Text, Expanded).
-#(lref(_Label, Text),	Text) :-
+#(lref(Label, Text), Expanded) :- !,
+	#(lref('', Label, Text), Expanded).
+#(lref(_, _Label, Text),	Text) :-
 	in_anchor, !.
-#(lref(Label, Text),	lref(ALabel, Expanded)) :-
+#(lref(Class, Label, Text),	lref(Class, ALabel, Expanded)) :-
 	canonise_label(Label, ALabel),
 	asserta(in_anchor),
 	expand_macros(Text, Expanded),
@@ -483,16 +507,10 @@ language_map(table,	'Table').
 #(iflref(Label, Text),	iflref(ALabel, Expanded)) :-
 	canonise_label(Label, ALabel),
 	expand_macros(Text, Expanded).
-#(lback(Label, Text),	lback(ALabel, Expanded)) :-
-	string_to_atom(Label, ALabel),
-	expand_macros(Text, Expanded).
-#(lforw(Label, Text),	lforw(ALabel, Expanded)) :-
-	string_to_atom(Label, ALabel),
-	expand_macros(Text, Expanded).
 #(url(_URL, Text),	Text) :-
 	in_anchor, !.
 #(url(URL, Text),	[html(Anchor), Expanded, html('</A>')]) :-
-	sformat(Anchor, '<A HREF="~w">', URL),
+	format(string(Anchor), '<A class="url" href="~w">', URL),
 	asserta(in_anchor),
 	expand_macros(Text, Expanded),
 	retractall(in_anchor).
@@ -539,9 +557,9 @@ cite_references(KeyIn, Functor, Refs) :-
 	make_cite_references(Keys, Functor, Refs).
 
 make_cite_references([], _, []).
-make_cite_references([Key],    F, [#lref(Key, Term)]) :- !,
+make_cite_references([Key],    F, [#lref(cite, Key, Term)]) :- !,
 	Term =.. [F, Key].
-make_cite_references([Key|T0], F, [#lref(Key, Term), nospace(','), ' '|T]) :-
+make_cite_references([Key|T0], F, [#lref(cite, Key, Term), nospace(','), ' '|T]) :-
 	Term =.. [F, Key],
 	make_cite_references(T0, F, T).
 
@@ -633,11 +651,13 @@ do_float(Term, Goal) :-
 
 :- dynamic
 	label/3,			% Label, File, Index
-	next_file/2.			% File, NextFile
+	next_file/2,			% File, NextFile
+	section_label/2.		% Label, SecNr
 
 reset_labels :-
 	retractall(label(_, _, _)),
-	retractall(next_file(_, _)).
+	retractall(next_file(_, _)),
+	retractall(section_label(_, _)).
 
 collect_labels([], _) :- !.
 collect_labels([tell(File)|T], PreviousFile) :- !,
@@ -649,20 +669,18 @@ collect_labels([tell(File)|T], PreviousFile) :- !,
 collect_labels([label(Label, _, Tag)|T], File) :- !,
 	assert(label(Label, File, Tag)),
 	collect_labels(T, File).
-collect_labels([lback(Label, _)|T], File) :- !,
-	assert(label(Label, File, -)),
-	collect_labels(T, File).
-collect_labels([lforw(Label, _)|T], File) :- !,
-	concat('back-to-', Label, BackLabel),
-	assert(label(BackLabel, File, -)),
-	collect_labels(T, File).
 collect_labels([_|T], File) :-
 	collect_labels(T, File).
 
-label_tag(Tag) :-
+label_tag(_Label, Tag) :-
 	current_float(_, Tag), !.
-label_tag(Tag) :-
-	section_tag(Tag).
+label_tag(Label, Tag) :-
+	section_tag(Tag),
+	(   sub_atom(Label, 0, _, _, 'sec:')
+	->  atom_concat('sec:', Tag, SecTag),
+	    assert(section_label(Label, SecTag))
+	;   true
+	).
 
 
 		 /*******************************
@@ -673,7 +691,7 @@ label_tag(Tag) :-
 %
 %	Translate an environment.
 
-env(pcecode(_, Tokens), #listing(#pre(HTML))) :-
+env(pcecode(_, Tokens), #pre(code, HTML)) :-
 	translate(Tokens, pcecode, HTML).
 env(summarylist(_, Summary),
     [ html('<TABLE>'),
@@ -915,9 +933,9 @@ cmd(paragraph({Title}), [#b(+Title), ' ']).
 cmd(subparagraph({Title}), [#b(+Title), ' ']).
 
 cmd(label({Label}), #label(Label, [], Tag)) :-	% \label and \xyzref
-	label_tag(Tag).
-cmd(ref({RefName}), #lref(RefName, ref(RefName))).
-cmd(pageref({RefName}), #lref(RefName, ref(RefName))).
+	label_tag(Label, Tag).
+cmd(ref({RefName}), #lref(ref, RefName, ref(RefName))).
+cmd(pageref({RefName}), #lref(ref, RefName, ref(RefName))).
 	
 cmd(secref({Label}), HTML) :-
 	translate_reference(section, sec, Label, HTML).
@@ -979,11 +997,9 @@ cmd(htmlfiledepth({Depth}), []) :-
 	assert(html_split_level(D)).
 
 cmd(maketitle,					% \maketitle
-    #quote(#quote(#quote(#quote([ #center(#h(1, #thetitle)),
-				  html('<HR>'),
-				  #center(#i(#theauthor)),
-				  html('<HR>')
-				]))))).
+    [ #div(title, #thetitle),
+      #div(author, #theauthor)
+    ]).
 
 cmd(newblock, []).				% BiBTeX \newblock
 cmd(protect, []).				% BiBTeX produced?
@@ -1036,8 +1052,7 @@ cmd(index({Term}), #label(RefName, [])) :-	% \index
 cmd(idx({Term}), #label(RefName, Term)) :-	% \idx
 	translate_index(Term, RefName).
 
-cmd(footnote({TeX}), HTML) :-			% \footnote
-	translate_footnote(TeX, HTML).
+cmd(footnote({TeX}), #(footnote(+TeX))).	% \footnote
 cmd(footnotetext([Num], {Tokens}), [html('<P>'), #embrace(Num), Text]) :-
 	translate(Tokens, normal, Text).
 
@@ -1103,6 +1118,7 @@ cmd(beta, html('&beta;')).
 cmd(alpha, #var(a)).			% \alpha
 cmd(beta, #var(b)).			% \beta
 cmd(tm, #embrace(tm)).			% \tm
+cmd(sum,  html('&Sigma;')).		% \sum
 
 cmd(include({File}), HTML) :-
 	absolute_file_name(tex(File), [ extensions([tex]),
@@ -1164,7 +1180,8 @@ cmd(includegraphics(_Options, {File}), html(Img)) :-
 	ps_extension(Ext),
 	absolute_file_name(includegraphics(File),
 			   [ extensions([Ext]),
-			     access(read)
+			     access(read),
+			     file_errors(fail)
 			   ], PsFile),
 	file_name_extension(Base, Ext, PsFile),
 	file_base_name(Base, GifBase),
@@ -1260,7 +1277,10 @@ cmd(leq, math, '=<').
 cmd(equiv, math, '==').
 cmd(longrightarrow, math, '-->').
 cmd(geq, math, '>=').
+cmd(leq, math, '=<').
 cmd(ge, math, '>=').
+cmd(le, math, '<').
+cmd(mid, math, '|').
 cmd(pm, math, html('&#177;')).
 cmd(langle, math, '<').
 cmd(rangle, math, '>').
@@ -1304,6 +1324,7 @@ cmd(Font, group(Old), group([font(Font)|Old1]), HTML) :-
 html_font(em, 		html('<EM>'), 		html('</EM>')).
 html_font(bf, 		html('<B>'),  		html('</B>')).
 html_font(it, 		html('<I>'),  		html('</I>')).
+html_font(mathit,	html('<I>'),  		html('</I>')).
 html_font(cal, 		html('<I>'),  		html('</I>')).
 html_font(tt, 		html('<TT>'), 		html('</TT>')).
 html_font(sf, 		html('<B>'),  		html('</B>')).
@@ -1323,18 +1344,16 @@ html_font('Huge',       html('<font size=+3>'), html('</font>')).
 		 *	    \REF, ETC.		*
 		 *******************************/
 
-%	translate_reference(+Name, +Tag, +Label, -HTML)
+%%	translate_reference(+Name, +Tag, +Label, -HTML)
 %
 %	Used for the translation of \secref, \figref, etc.
 
 translate_reference(Name, sec, Label,
-	  #lref(fileof(RefName), [Name, ' ', ref(RefName)])) :- !,
-	sformat(RefStr, 'sec:~w', [Label]),
-	string_to_atom(RefStr, RefName).
+	  #lref(sec, fileof(RefName), [Name, ' ', ref(RefName)])) :- !,
+	format(atom(RefName), 'sec:~w', [Label]).
 translate_reference(Name, Tag, Label,
-	  #lref(RefName, [Name, ' ', ref(RefName)])) :-
-	sformat(RefStr, '~w:~w', [Tag, Label]),
-	string_to_atom(RefStr, RefName).
+	  #lref(Tag, RefName, [Name, ' ', ref(RefName)])) :-
+	format(atom(RefName), '~w:~w', [Tag, Label]).
 
 
 		 /*******************************
@@ -1424,44 +1443,6 @@ capitalise([H|T0], [H|T], down) :-
 	capitalise(T0, T, up).
 	
 
-		 /*******************************
-		 *	    FOOTNOTES		*
-		 *******************************/
-
-:- dynamic
-	footnote/2.			% Id, Tokens
-
-reset_footnotes :-
-	retractall(footnote(_, _)).
-
-
-translate_footnote(Tokens,
-		   #lforw(Tag, [' ', #embrace(Value)])) :-
-	step_counter(footnote_counter, Value),
-	sformat(Tag, 'note-~d', [Value]),
-	assert(footnote(Tag, Tokens)).
-
-footnote_html([]) :-
-	\+ footnote(_, _), !.
-footnote_html(HTML) :-
-	HTML0 = [ #tell('Notes'),
-		  #header(notes),
-		  #h(1, #label('document-notes', 'Footnotes')),
-		  html('<DL>'),
-		  FootNodes,
-		  html('</DL>')
-		],
-	findall(FN, footnote_html1(FN), FootNodes),
-	expand_macros(HTML0, HTML).
-
-footnote_html1([ html('<DT>'),
-		 #lback(Tag, Tag),
-		 html('<DD>'),
-	         +TeX
-	       ]) :-
-	footnote(Tag, TeX).
-	
-
 		 /*******************************
 		 *	       SECTION		*
 		 *******************************/
@@ -1517,7 +1498,10 @@ translate_section(Level, -, TeXTitle,
 	[ Footer,
 	  Tell,
 	  Header,
-	  #h(Level, #label(RefName, [Tag, ' ', Title]))
+	  #h(Level, #label(RefName,
+			   [ #span('sec-nr', Tag),
+			     #span('sec-title', Title)
+			   ]))
 	], NodeFile) :- !,
 	translate(TeXTitle, normal, Title),
 	section_tag(OldTag),
@@ -1538,7 +1522,7 @@ translate_section(Level, -, TeXTitle,
 	    Footer = [],
 	    Header = []
 	),
-	sformat(RefName, 'sec:~w', [Tag]),
+	format(string(RefName), 'sec:~w', [Tag]),
 	assert(section(Level, Tag, Title)).
 translate_section(Level, *, Title, #h(Level, +Title), _).
 
@@ -1614,21 +1598,14 @@ tableofcontents([]) :-
 tableofcontents(Sections) :-
 	tableofcontents('', Sections).
 
-tableofcontents(TagPrefix, [Sections, CloseUL]) :-
-	retractall(section_level(_)),
-	(   section(Level, Tag, _Title),
-	    sub_atom(Tag, 0, _, _, TagPrefix)
-	->  L0 is Level - 1
-	;   L0 = 0
-	),
-	asserta(section_level(L0)),
-	findall(S, section_html(TagPrefix, S), Sections),
-	fix_level(L0, CloseUL).
+tableofcontents(TagPrefix, #div(toc, Sections)) :-
+	findall(S, section_html(TagPrefix, S), Sections).
 
 section_html(TagPrefix,
-	     [ FixLevel,
-	       html('<LI>'),
-	       #lref(Ref, [Open, Tag, ' ', Title, Close])
+	     [ #div(Class, #lref(sec, Ref,
+				 [ #span('sec-nr', Tag),
+				   #span('sec-title', Title)
+				 ]))
 	     ]) :-
 	section(Level, Tag, Title),
 	html_split_level(Split),
@@ -1636,33 +1613,10 @@ section_html(TagPrefix,
 	->  Ref = fileof(RefName)
 	;   Ref = RefName
 	),
-	concat(TagPrefix, _, Tag),
-	fix_level(Level, FixLevel),
-	sformat(RefName, 'sec:~w', [Tag]),
-	contents_line_style(Level, Open, Close).
+	atom_concat(TagPrefix, _, Tag),
+	format(string(RefName), 'sec:~w', [Tag]),
+	atom_concat('toc-h', Level, Class).
 
-fix_level(To, []) :-
-	section_level(To), !.
-fix_level(To, HTML) :-
-	(   retract(section_level(From))
-	->  fix_indent(To, From, HTML),
-	    asserta(section_level(To))
-	).
-
-fix_indent(L, L, []) :- !.
-fix_indent(To, From, [html('<UL>')|T]) :-
-	To > From, !,
-	NFrom is From + 1,
-	fix_indent(To, NFrom, T).
-fix_indent(To, From, [html('</UL>')|T]) :-
-	To < From, !,
-	NFrom is From - 1,
-	fix_indent(To, NFrom, T).
-
-	
-contents_line_style(1, html('<STRONG>'), html('</STRONG>')) :- !.
-contents_line_style(2, html('<B>'),      html('</B>')) :- !.
-contents_line_style(_, [],      	 []).
 
 		 /*******************************
 		 *	       INDEX		*
@@ -1704,7 +1658,7 @@ add_to_index(Term) :-
 	section_tag(Tag),
 	add_to_index(Term, Tag).
 
-%	add_to_index(+Term, +Tag)
+%%	add_to_index(+Term, +Tag)
 %	
 %	Add Term to the index using the href Tag. If Tag is of the
 %	format +Tag, it is the primary index for the term, normally
@@ -1769,16 +1723,16 @@ index_terms([Term0|T0], [ html('<DT>'),
 	tex_atom_to_tokens(Term0, Tokens),
 	translate(Tokens, index, _, Term),
 	(   member(+PrimeTag, Tags)
-	->  HtmlTerm = #lref(PrimeTag, Term)
+	->  HtmlTerm = #lref(idx, PrimeTag, Term)
 	;   HtmlTerm = Term
 	),
 	maplist(index_href, Tags, Where),
 	index_terms(T0, TH).
 
 index_href(+(_), []) :- !.
-index_href(Tag:Label, [' ', #lref(Label, Tag)]) :- !.
-index_href(Tag, [' ', #lref(RefName, Tag)]) :-
-	sformat(RefName, 'sec:~w', Tag).
+index_href(Tag:Label, [' ', #lref(idx, Label, Tag)]) :- !.
+index_href(Tag, [' ', #lref(sec, RefName, Tag)]) :-
+	format(string(RefName), 'sec:~w', Tag).
 	
 add_separator(Term, CL, CL, []) :-
 	atom_codes(Term, [CL|_]), !.
@@ -1807,15 +1761,13 @@ node_header([#head([#title(#thetitle),
 		    link(previous),
 		    link(next)]),
 	     #beginbody,
-	     html('<HR>'),
-	     #center( [ body_link(home),
-			body_link(contents),
-			body_link(index),
-			body_link(summary),
-			body_link(previous),
-			body_link(next)
-		      ]),
-	     html('<HR>')
+	     #navigate( [ body_link(home),
+			  body_link(contents),
+			  body_link(index),
+			  body_link(summary),
+			  body_link(previous),
+			  body_link(next)
+		      ])
 	    ]).
 	
 node_header(_, []) :-
@@ -1831,16 +1783,14 @@ node_header(SectionTag,
 		    link(previous),
 		    link(next)]),
 	     #beginbody,
-	     html('<HR>'),
-	     #center( [ body_link(home),
-			body_link(contents),
-			body_link(index),
-			body_link(summary),
-			body_link(up(UpRef)),
-			body_link(previous),
-			body_link(next)
-		      ]),
-	     html('<HR>')
+	     #navigate( [ body_link(home),
+			  body_link(contents),
+			  body_link(index),
+			  body_link(summary),
+			  body_link(up(UpRef)),
+			  body_link(previous),
+			  body_link(next)
+		      ])
 	    ]) :-
 	parent_tag(SectionTag, UpTag), !,
 	sformat(UpRef, 'sec:~w', [UpTag]).
@@ -1859,18 +1809,11 @@ subsection_index(Tag,
 	    ]) :-
 	onefile(false),
 	Tag \== '',
-	concat(Tag, '.', Filter),
+	atom_concat(Tag, '.', Filter),
 	tableofcontents(Filter, SubIndex),
-	submember(#lref(_,_), SubIndex), !. % it is not empty
+	sub_term(#lref(_,_,_), SubIndex), !. % it is not empty
 subsection_index(_, []).
 
-submember(X, [X|_]).
-submember(X, [L|_]) :-
-	is_list(L),
-	submember(X, L).
-submember(X, [_|T]) :-
-	submember(X, T).
-
 
 		 /*******************************
 		 *	     COUNTERS		*
@@ -1899,8 +1842,8 @@ step_counter(Name, NewVal) :-
 		 *******************************/
 
 cmd(rule({'\\linewidth'}, {_H}), [html('<hr>')]) :- !.
-cmd(rule({_W}, {_H}), []) :-
-	format('_W = ~w, H = ~w~n', [_W, _H]).
+cmd(rule({W}, {H}), []) :-
+	format('W = ~w, H = ~w~n', [W, H]).
 
 		 /*******************************
 		 *	      TABLES		*
@@ -2244,10 +2187,10 @@ psfigoption(Term) -->
 	}.
 
 
-		 /*******************************
-		 *	 FIX module/[1,2]	*
-		 *******************************/
-
+%%	fix_predicate_reference(+Ref0, -Ref)
+%
+%	Deal with references such as name/[1,2], send/2-12, etc and find
+%	the first matching predicate from the referenced set.
 
 fix_predicate_reference(Ref0, Ref) :-
 	 atom_codes(Ref0, Chars),
@@ -2471,22 +2414,98 @@ aformat(Atom, Fmt, Args) :-
 :- dynamic
 	pending_par/0.
 
-implicit_par(html('<H1>')).
-implicit_par(html('<H2>')).
-implicit_par(html('<H3>')).
-implicit_par(html('<H4>')).
-implicit_par(html('<PRE>')).
-implicit_par(html('<XMP>')).
-implicit_par(html('<DL>')).
-implicit_par(html('<TABLE>')).
-%implicit_par(html('<CENTER>')).
-implicit_par(html('<BLOCKQUOTE>')).
+is_begin(X, HTML) :-
+	atom(HTML), !,
+	sub_atom(HTML, 0, _, _, <),
+	sub_atom(HTML, 1, L, _, X),
+	N is L+1,
+	(   sub_atom(HTML, N, _, _, ' ')
+	->  true
+	;   sub_atom(HTML, N, _, _, '>')
+	).
+is_begin(X, HTML) :-
+	var(HTML),
+	format(atom(HTML), '<~w>', [X]).
 
-no_par_in('<DL>').			% delete pars here
-%no_par_in('<UL>').
-%no_par_in('<OL>').
+is_end(X, HTML) :-
+	atom(HTML), !,
+	sub_atom(HTML, 0, _, _, '</'),
+	sub_atom(HTML, 2, L, _, X),
+	N is L+2,
+	sub_atom(HTML, N, _, _, '>').
+is_end(X, HTML) :-
+	var(HTML),
+	format(atom(HTML), '</~w>', [X]).
 
 
+
+implicit_par(html(Begin)) :-
+	implicit_par_tag(Tag),
+	is_begin(Tag, Begin), !.
+
+implicit_par_tag('H1').
+implicit_par_tag('H2').
+implicit_par_tag('H3').
+implicit_par_tag('H4').
+implicit_par_tag('PRE').
+implicit_par_tag('XMP').
+implicit_par_tag('DL').
+implicit_par_tag('DT').
+implicit_par_tag('DD').
+implicit_par_tag('TABLE').
+implicit_par_tag('BLOCKQUOTE').
+%implicit_par_tag('CENTER').
+
+:- dynamic
+	is_open/1.
+
+%%	create_pending_open(+Token) is det.
+%%	close_pending_open(+Token) is det.
+%%	close_by(+Open, -CloseBy) is nondet.
+
+create_pending_open(Cmd) :-
+	(   close_by(Open, _)
+	;   create_env(Open)
+	),
+	is_begin(Open, Cmd), !,
+	debug(html, 'Push open of ~w', [Open]),
+	asserta(is_open(Open)).
+create_pending_open(_).
+
+close_pending_open(Cmd) :-
+	pop_implicit_close(Cmd),
+	pop_close(Cmd).
+
+pop_implicit_close(Cmd) :-
+	(   once(is_open(Open)),
+	    close_by(Open, CloseOn),
+	    call(CloseOn, Cmd)
+	->  retract(is_open(Open)),
+	    debug(html, 'Pop open ~w on ~w', [Open, Cmd]),
+	    is_end(Open, Close),
+	    cmd_layout(Close, Pre, Post), !,
+	    put_html_token(html(Close, Pre, Post))
+	;   true
+	).
+
+pop_close(Cmd) :-
+	(   once(is_open(Open)),
+	    is_end(Open, Cmd)
+	->  retract(is_open(Open)),
+	    debug(html, 'Pop open ~w on ~w', [Open, Cmd])
+	;   true
+	).
+
+
+create_env('DL').
+
+close_by('DD', is_begin('DT')).
+close_by('DD', is_end('DL')).
+close_by('DT', is_begin('DD')).
+close_by('DT', is_end('DL')).
+
+%%	write_html(+Tokens) is det.
+
 write_html([]) :- !.			% Unpack lists
 write_html([H|T]) :- !,
 	write_html(H),
@@ -2496,30 +2515,26 @@ write_html('\n') :-
 write_html(' ') :-
 	pending_par, !.
 write_html(html('<P>')) :- !,
-	(   pending_par
+	(   once(is_open(Open)),
+	    Open = 'DL'
+	->  true
+	;   pending_par
 	->  true
 	;   assert(pending_par)
 	).
-write_html(html(Cmd)) :-
-	no_par_in(Cmd), !,
-	(   pending_par
-	->  true
-	;   assert(pending_par)		% Say we have one, do others
-					% are suppressed
-	),
-	cmd_layout(Cmd, Pre, Post), !,
-	put_html_token(html(Cmd, Pre, Post)).
 write_html(Token) :-
 	retract(pending_par), !,
 	(   implicit_par(Token)
 	->  true
-	;   cmd_layout('<P>', Pre, Post),
-	    put_html_token(html('<P>', Pre, Post))
+	;   cmd_layout('<P>', Pre, Post)
+	->  put_html_token(html('<P>', Pre, Post))
 	),
 	write_html(Token).
 write_html(html(Cmd)) :-			% HTML commands
+	close_pending_open(Cmd),
 	cmd_layout(Cmd, Pre, Post), !,
-	put_html_token(html(Cmd, Pre, Post)).
+	put_html_token(html(Cmd, Pre, Post)),
+	create_pending_open(Cmd).
 write_html(ref(Label)) :- !,			% References and labels
 	(   label(Label, _, Ref)
 	->  write_html(Ref)
@@ -2536,7 +2551,7 @@ write_html(label(Label, Text, _)) :- !,
 	).
 write_html(body_link(Link)) :- !,
 	(   translate_ref(Link, Ref, Type)
-	->  sformat(Anchor, '<A HREF="~w">', [Ref]),
+	->  format(string(Anchor), '<A class="nav" href="~w">', [Ref]),
 	    capitalise_atom(Type, Text),
 	    (	link_image(Type, Image)
 	    ->	sformat(Img, '<IMG SRC="~w" BORDER=0 ALT="~w">', [Image, Text]),
@@ -2564,48 +2579,46 @@ write_html(iflref(Label, Text)) :- !,
 	->  write_html(lref(Label, Text))
 	;   true
 	).
-write_html(lref(fileof(Label), Text)) :-
+write_html(lref(Label, Text)) :- !,
+	write_html(lref('', Label, Text)).
+write_html(lref(Class, fileof(Label), Text)) :-
 	(   label(Label, File, _)
-	->  sformat(Anchor, '<A HREF="~w.html">', [File]),
+	->  format(string(Anchor),
+		   '<A class="~w" href="~w.html">', [Class, File]),
 	    write_html([html(Anchor), Text, html('</A>')])
 	;   write_html(lref(Label, Text))
 	).
-write_html(lref(Label, Text)) :-
+write_html(lref(Class, Label, Text)) :-
 	label(Label, File, _), !,
+	(   section_label(Label, TheLabel)
+	->  true
+	;   TheLabel = Label
+	),
 	(   in_anchor
 	->  macro_expand(Text, Expanded),
 	    write_html(Expanded)
 	;   asserta(in_anchor),
 	    (   onefile(false)
-	    ->  sformat(Anchor, '<A HREF="~w.html#~w">', [File, Label])
-	    ;   sformat(Anchor, '<A HREF="#~w">', [Label])
+	    ->  format(string(Anchor),
+		       '<A class="~w" href="~w.html#~w">',
+		       [Class, File, TheLabel])
+	    ;   format(string(Anchor),
+		       '<A class="~w" href="#~w">', [Class, TheLabel])
 	    ),
 	    write_html([html(Anchor), Text, html('</A>')]),
 	    retractall(in_anchor)
 	).
-write_html(lref(Label, Text)) :-
+write_html(lref(pred, Label, Text)) :-
 	fix_predicate_reference(Label, FixedLabel), !,
-	write_html(lref(FixedLabel, Text)).
-write_html(lref(Label, Text)) :- !,
-	format(user_error, 'No label for reference "~w"~n', [Label]),
+	write_html(lref(pred, FixedLabel, Text)).
+write_html(lref(pred, Label, Text)) :- !,
+	format(user_error, 'No description for predicate "~w"~n', [Label]),
+	macro_expand(#span('pred-ext', Text), Expanded),
+	write_html(Expanded).
+write_html(lref(Class, Label, Text)) :- !,
+	format(user_error, 'No label for ~w reference "~w"~n', [Class, Label]),
 	macro_expand(#b(Text), Expanded),
 	write_html(Expanded).
-write_html(lforw(Label, Text)) :- !,
-	concat('back-to-', Label, BackName),
-	(   label(Label, File, _)
-	->  sformat(Anchor, '<A NAME=~w HREF="~w.html#~w">',
-		    [BackName, File, Label]), 
-	    write_html([html(Anchor), Text, html('</A>')])
-	;   format(user_error, 'No target for forward link "~w"~n', [Label])
-	).
-write_html(lback(Label, Text)) :- !,
-	concat('back-to-', Label, BackName),
-	(   label(BackName, File, _)
-	->  sformat(Anchor, '<A NAME=~w HREF="~w.html#~w">',
-		    [Label, File, BackName]), 
-	    write_html([html(Anchor), Text, html('</A>')])
-	;   format(user_error, 'No target for backward link "~w"~n', [Label])
-	).
 write_html(cite(Key)) :- !,
 	(   cite(Key, Cite)
 	->  write_html(Cite)
@@ -2643,6 +2656,7 @@ write_html(_).
 nl_html :-
 	write_html(verb('\n')).
 
+
 %	translate_ref(+Label, -Anchor, -TextLabel)
 
 translate_ref(next, Anchor, next) :-
@@ -2675,7 +2689,7 @@ translate_ref(summary, Anchor, summary) :-
 cmd_layout('<P>',    2, 0).
 cmd_layout('<DL>',   2, 1).
 cmd_layout('</DL>',  1, 2). 
-cmd_layout('<DD>',   0, 1). 
+cmd_layout(DD,   0, 1) :- is_begin('DD', DD).
 cmd_layout('<H1>',   2, 0). 
 cmd_layout('<H2>',   2, 0). 
 cmd_layout('<H3>',   2, 0). 
@@ -2690,8 +2704,13 @@ cmd_layout('</H5>',  0, 2).
 cmd_layout('</H6>',  0, 2). 
 cmd_layout('<HR>',   1, 1). 
 cmd_layout('<BR>',   0, 1). 
+cmd_layout(DIV,      1, 0) :- is_begin('DIV', DIV).
+cmd_layout('</DIV>', 0, 1).
 cmd_layout('<LI>', 	 1, 0).
 cmd_layout('<DT>', 	 1, 0).
+cmd_layout('</DT>', 	 0, 1).
+cmd_layout('<DD>', 	 1, 0).
+cmd_layout('</DD>', 	 0, 1).
 cmd_layout('<UL>', 	 1, 1).
 cmd_layout('</UL>', 	 1, 1).
 cmd_layout('<TR>',       1, 0). 
@@ -2701,8 +2720,8 @@ cmd_layout('<THEAD>',    1, 1).
 cmd_layout('</TABLE>',   0, 2). 
 cmd_layout('<LISTING>',	 2, 0). 
 cmd_layout('</LISTING>', 0, 2). 
-cmd_layout('<PRE>',	 1, 1). 
-cmd_layout('</PRE>', 	 1, 1). 
+cmd_layout(PRE,	 2, 0) :- is_begin('PRE', PRE).
+cmd_layout('</PRE>', 	 0, 2). 
 cmd_layout('<XMP>',	 2, 0). 
 cmd_layout('</XMP>', 	 0, 2). 
 cmd_layout('<HEAD>',	 1, 1). 
