@@ -1,11 +1,11 @@
-/*  $Id: rdf_db.pl,v 1.35 2004/10/19 12:51:47 jan Exp $
+/*  $Id: rdf_db.pl,v 1.84 2006/08/31 09:44:00 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -49,8 +49,12 @@
 	    rdf_update/5,		% +Subject, +Predicate, +Object, +Src, +Act
 	    rdf_set_predicate/2,	% +Predicate, +Property
 	    rdf_predicate_property/2,	% +Predicate, ?Property
+	    rdf_current_predicate/1,	% -Predicate
 
 	    rdf_transaction/1,		% :Goal
+	    rdf_transaction/2,		% :Goal, +Id
+
+	    rdf_monitor/2,		% :Goal, +Options
 
 	    rdf_save_db/1,		% +File
 	    rdf_save_db/2,		% +File, +DB
@@ -64,7 +68,7 @@
 	    rdf_load/1,			% +File
 	    rdf_load/2,			% +File, +Options
 	    rdf_save/1,			% +File
-	    rdf_save/2,			% +File, +DB
+	    rdf_save/2,			% +File, +Options
 	    rdf_unload/1,		% +File
 
 	    rdf_md5/2,			% +DB, -MD5
@@ -79,37 +83,59 @@
 	    rdf_estimate_complexity/4,	% +S,+P,+O,-Count
 
 	    rdf_save_subject/3,		% +Stream, +Subject, +DB
-	    rdf_save_header/2,		% +Out, +DB
+	    rdf_save_header/2,		% +Out, +Options
 	    rdf_save_footer/1,		% +Out
 
 	    rdf_equal/2,		% ?Resource, ?Resource
 
 	    rdf_register_ns/2,		% +Alias, +URI
+	    rdf_register_ns/3,		% +Alias, +URI, +Options
 	    rdf_global_id/2,		% ?NS:Name, ?Global
 	    rdf_global_object/2,	% ?Object, ?NSExpandedObject
 	    rdf_global_term/2,		% Term, WithExpandedNS
 
 	    rdf_match_label/3,		% +How, +String, +Label
 	    rdf_split_url/3,		% ?Base, ?Local, ?URL
+	    rdf_url_namespace/2,	% +URL, ?Base
+	    rdf_quote_uri/2,		% +URI, -Quoted
 
-	    rdf_debug/1			% Set verbosity
+	    rdf_debug/1,		% Set verbosity
+
+	    rdf_new_literal_map/1,	% -Handle
+	    rdf_destroy_literal_map/1,	% +Handle
+	    rdf_reset_literal_map/1,	% +Handle
+	    rdf_insert_literal_map/3,	% +Handle, +Key, +Literal
+	    rdf_delete_literal_map/3,	% +Handle, +Key, +Literal
+	    rdf_delete_literal_map/2,	% +Handle, +Key
+	    rdf_find_literal_map/3,	% +Handle, +KeyList, -Literals
+	    rdf_keys_in_literal_map/3,	% +Handle, +Spec, -Keys
+	    rdf_statistics_literal_map/2, % +Handle, +Name(-Arg...)
+
+	    (rdf_meta)/1,		% +Heads
+	    op(1150, fx, (rdf_meta))
 	  ]).
 :- use_module(library(rdf)).
 :- use_module(library(lists)).
 :- use_module(library(shlib)).
 :- use_module(library(gensym)).
+:- use_module(library(sgml)).
+:- use_module(library(sgml_write)).
+:- use_module(library(option)).
+:- use_module(library(nb_set)).
 
 :- initialization
    load_foreign_library(foreign(rdf_db)).
 
 :- multifile
-	ns/2.
+	ns/2,
+	rdf_meta_specification/2.	% Unbound Head, Head
 :- dynamic
 	ns/2,				% ID, URL
 	rdf_source/4.			% File, ModTimeAtLoad, Triples, MD5
 :- volatile
 	rdf_source/4.
-
+:- discontiguous
+	term_expansion/2.
 
 		 /*******************************
 		 *	     NAMESPACES		*
@@ -123,22 +149,38 @@ ns(dc,   'http://purl.org/dc/elements/1.1/').
 ns(eor,  'http://dublincore.org/2000/03/13/eor#').
 ns(serql,'http://www.openrdf.org/schema/serql#').
 
-%	rdf_register_ns(+Alias, +URI)
+%%	rdf_register_ns(+Alias, +URI) is det.
+%%	rdf_register_ns(+Alias, +URI, +Options) is det.
 %
-%	Register a namespace.  What to do if the Alias already
-%	exists?  Throw a permission error?  Use both URI's as synonyms?
+%	Register a namespace. If force(true) is   provided as an option,
+%	the  existing  namespace  alias  is   replaced.  Otherwise  this
+%	predicate raises a permission error.  Please note that replacing
+%	a namespace is dangerous  as   namespaces  affect preprocessing.
+%	Make sure all code that depends on a namespace is compiled after
+%	changing the registration.
 
 rdf_register_ns(Alias, URI) :-
+	rdf_register_ns(Alias, URI, []).
+
+rdf_register_ns(Alias, URI, _) :-
 	ns(Alias, URI), !.
-rdf_register_ns(Alias, _) :-
+rdf_register_ns(Alias, URI, Options) :-
 	ns(Alias, _),
-	throw(error(permission_error(register, namespace, Alias),
-		    context(_, 'Already defined'))).
-rdf_register_ns(Alias, URI) :-
+	(   option(force(Force), Options, false),
+	    Force == true
+	->  retractall(ns(Alias, _)),
+	    assert(ns(Alias, URI))
+	;   option(keep(Keep), Options, false),
+	    Keep == true
+	->  true
+	;   throw(error(permission_error(register, namespace, Alias),
+			context(_, 'Already defined')))
+	).
+rdf_register_ns(Alias, URI, _) :-
 	assert(ns(Alias, URI)).
 
 
-%	rdf_global_id(?Id, ?GlobalId)
+%%	rdf_global_id(?Id, ?GlobalId)
 %
 %	Convert between NS:Local and global atomic identifier.
 %	To be completed.
@@ -148,7 +190,7 @@ rdf_global_id(NS:Local, Global) :-
 rdf_global_id(Global, Global).
 
 
-%	rdf_global_object(?Object, ?GlobalObject)
+%%	rdf_global_object(?Object, ?GlobalObject)
 %	
 %	Same as rdf_global_id/2,  but  intended   for  dealing  with the
 %	object part of a  triple,  in   particular  the  type  for typed
@@ -174,7 +216,7 @@ global(NS, Local, Global) :-
 
 
 
-%	rdf_global_term(+TermIn, -GlobalTerm)
+%%	rdf_global_term(+TermIn, -GlobalTerm)
 %	
 %	Does rdf_global_id/2 on all terms NS:Local by recursively analysing
 %	the term.
@@ -182,7 +224,8 @@ global(NS, Local, Global) :-
 rdf_global_term(Var, Var) :-
 	var(Var), !.
 rdf_global_term(NS:Local, Global) :-
-	rdf_global_id(NS:Local, Global).
+	rdf_global_id(NS:Local, Global0), !,
+	Global = Global0.
 rdf_global_term([H0|T0], [H|T]) :- !,
 	rdf_global_term(H0, H),
 	rdf_global_term(T0, T).
@@ -194,100 +237,137 @@ rdf_global_term(Term0, Term) :-
 rdf_global_term(Term, Term).
 
 
-%	user:goal_expansion(+NSGoal, -Goal)
-%	
-%	This predicate allows for writing down rdf queries in a friendly
-%	name-space fashion.  
+		 /*******************************
+		 *	      EXPANSION		*
+		 *******************************/
 
 :- multifile
-	user:goal_expansion/2.
+	user:term_expansion/2.
 
-user:goal_expansion(rdf(Subj0, Pred0, Obj0),
-		    rdf(Subj, Pred, Obj)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf_has(Subj0, Pred0, Obj0, RP0),
-		    rdf_has(Subj, Pred, Obj, RP)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj),
-	rdf_global_id(RP0, RP).
-user:goal_expansion(rdf_has(Subj0, Pred0, Obj0),
-		    rdf_has(Subj, Pred, Obj)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf_assert(Subj0, Pred0, Obj0),
-		    rdf_assert(Subj, Pred, Obj)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf_retractall(Subj0, Pred0, Obj0),
-		    rdf_retractall(Subj, Pred, Obj)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf(Subj0, Pred0, Obj0, PayLoad),
-		    rdf(Subj, Pred, Obj, PayLoad)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf_reachable(Subj0, Pred0, Obj0),
-		    rdf_reachable(Subj, Pred, Obj)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf_assert(Subj0, Pred0, Obj0, PayLoad),
-		    rdf_assert(Subj, Pred, Obj, PayLoad)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf_retractall(Subj0, Pred0, Obj0, PayLoad),
-		    rdf_retractall(Subj, Pred, Obj, PayLoad)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdf_update(Subj0, Pred0, Obj0, Action0),
-		    rdf_update(Subj, Pred, Obj, Action)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj),
-	rdf_global_term(Action0, Action).
-user:goal_expansion(rdf_equal(SubjA0, SubjB0),
-		    rdf_equal(SubjA, SubjB)) :-
-	rdf_global_id(SubjA0, SubjA),
-	rdf_global_id(SubjB0, SubjB).
-user:goal_expansion(rdf_source_location(Subj0, Source),
-		    rdf_source_location(Subj, Source)) :-
-	rdf_global_id(Subj0, Subj).
-user:goal_expansion(rdf_subject(Subj0),
-		    rdf_subject(Subj)) :-
-	rdf_global_id(Subj0, Subj).
-user:goal_expansion(rdf_set_predicate(P0, Prop),
-		    rdf_set_predicate(P, Prop)) :-
-	rdf_global_id(P0, P).
-user:goal_expansion(rdf_predicate_property(P0, Prop),
-		    rdf_predicate_property(P, Prop)) :-
-	rdf_global_id(P0, P).
-user:goal_expansion(rdf_estimate_complexity(Subj0, Pred0, Obj0, C),
-		    rdf_estimate_complexity(Subj, Pred, Obj, C)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-					% meta predicates
-user:goal_expansion(rdf_transaction(G0), rdf_transaction(G)) :-
-	expand_goal(G0, G).
+user:term_expansion((:- rdf_meta(Heads)), Clauses) :-
+	mk_clauses(Heads, Clauses).
+
+mk_clauses((A,B), [H|T]) :- !,
+	mk_clause(A, H),
+	mk_clauses(B, T).
+mk_clauses(A, [C]) :-
+	mk_clause(A, C).
+
+mk_clause(Head, rdf_db:rdf_meta_specification(Unbound, Head)) :-
+	valid_rdf_meta_head(Head),
+	functor(Head, Name, Arity),
+	functor(Unbound, Name, Arity).
+
+valid_rdf_meta_head(Head) :-
+	callable(Head), !,
+	Head =.. [_|Args],
+	valid_args(Args).
+valid_rdf_meta_head(Head) :-
+	throw(error(type_error(callable, Head), _)).
+
+valid_args([]).
+valid_args([H|T]) :-
+	valid_arg(H), !,
+	valid_args(T).
+
+valid_arg(:).				% meta argument
+valid_arg(+).				% non-var
+valid_arg(-).				% var
+valid_arg(?).				% either var or non-var
+valid_arg(@).				% not modified
+valid_arg(r).				% RDF resource
+valid_arg(o).				% RDF object
+valid_arg(t).				% term with RDF resources
+valid_arg(A) :-
+	throw(error(type_error(rdf_meta_argument, A), _)).
+
+%%	rdf_meta(+Heads)
+%	
+%	This   directive   is   expanded   using   term-expansion.   The
+%	implementation just throws an error in   case  it is called with
+%	the wrong context.
+
+rdf_meta(Heads) :-
+	throw(error(context_error(nodirective, rdf_meta(Heads)), _)).
 
 
-%	rdf_equal(?Resource1, ?Resource2)
+user:goal_expansion(G, Expanded) :-
+	rdf_meta_specification(G, Spec), !,
+	rdf_expand(G, Spec, Expanded).
+
+rdf_expand(G, Spec, Expanded) :-
+	functor(G, Name, Arity),
+	functor(Expanded, Name, Arity),
+	rdf_expand_args(0, Arity, G, Spec, Expanded).
+
+rdf_expand_args(Arity, Arity, _, _, _) :- !.
+rdf_expand_args(I0, Arity, Goal, Spec, Expanded) :-
+	I is I0 + 1,
+	arg(I, Goal, GA),
+	arg(I, Spec, SA),
+	arg(I, Expanded, EA),
+	rdf_expand_arg(SA, GA, EA),
+	rdf_expand_args(I, Arity, Goal, Spec, Expanded).
+
+rdf_expand_arg(r, A, E) :- !,
+	mk_global(A, E).
+rdf_expand_arg(o, A, E) :- !,
+	rdf_global_object(A, E).
+rdf_expand_arg(t, A, E) :- !,
+	rdf_global_term(A, E).
+rdf_expand_arg(:, A, E) :- !,
+	expand_goal(A, E).
+rdf_expand_arg(_, A, A).
+
+%%	mk_global(+Src, -Resource)
+%	
+%	Realised rdf_global_id(+, -), but adds compiletime checking,
+%	notably to see whether a namespace is not yet defined.
+
+mk_global(X, X) :-
+	var(X), !.
+mk_global(X, X) :-
+	atom(X), !.
+mk_global(NS:Local, Global) :-
+	must_be_atom(NS),
+	must_be_atom(Local),
+	(   ns(NS, Full)
+	->  atom_concat(Full, Local, Global)
+	;   throw(error(existence_error(namespace, NS), _))
+	).
+
+must_be_atom(X) :-
+	atom(X), !.
+must_be_atom(X) :-
+	throw(error(type_error(atom, X), _)).
+
+:- rdf_meta
+	rdf(r,r,o),
+	rdf_has(r,r,o,r),
+	rdf_has(r,r,o),
+	rdf_assert(r,r,o),
+	rdf_retractall(r,r,o),
+	rdf(r,r,o,?),
+	rdf_assert(r,r,o,+),
+	rdf_retractall(r,r,o,?),
+	rdf_reachable(r,r,r),
+	rdf_update(r,r,o,t),
+	rdf_equal(r,r),
+	rdf_source_location(r,-),
+	rdf_subject(r),
+	rdf_set_predicate(r, +),
+	rdf_predicate_property(r, -),
+	rdf_estimate_complexity(r,r,r,-),
+	rdf_transaction(:).
+
+%%	rdf_equal(?Resource1, ?Resource2)
 %	
 %	Simple equality test to exploit goal-expansion
 
 rdf_equal(Resource, Resource).
 
 
-%	rdf_has(?Subject, +Predicate, ?Object)
+%%	rdf_has(?Subject, +Predicate, ?Object)
 %	
 %	Succeeds if the triple rdf(Subject, Predicate, Object) is true
 %	exploiting the rdfs:subPropertyOf predicate.
@@ -300,7 +380,7 @@ rdf_has(Subject, Predicate, Object) :-
 		 *	    COLLECTIONS		*
 		 *******************************/
 
-%	rdf_member_property(?Prop, ?Index)
+%%	rdf_member_property(?Prop, ?Index)
 %	
 %	Deal with the rdf:_1, ... properties.
 
@@ -324,7 +404,7 @@ rdf_member_property(P, N) :-
 		 *	ANONYMOUS SUBJECTS	*
 		 *******************************/
 
-%	rdf_node(-Id)
+%%	rdf_node(-Id)
 %
 %	Generate a unique identifier for a subject.  Obsolete.  New
 %	code should use rdf_bnode/1.
@@ -337,7 +417,7 @@ rdf_node(Value) :-
 	\+ rdf(_, Value, _), !.
 
 
-%	rdf_bnode(-Id)
+%%	rdf_bnode(-Id)
 %
 %	Generate a unique anonymous identifier for a subject.
 
@@ -349,7 +429,7 @@ rdf_bnode(Value) :-
 	\+ rdf(_, Value, _), !.
 
 
-%	rdf_is_bnode(+Id)
+%%	rdf_is_bnode(+Id)
 %	
 %	Tests if a resource is a blank node (i.e. is an anonymous
 %	resource).
@@ -363,7 +443,7 @@ rdf_is_bnode(Id) :-
 		 *	      SOURCE		*
 		 *******************************/
 
-%	rdf_source_location(+Subject, -File:Line)
+%%	rdf_source_location(+Subject, -File:Line)
 %	
 %	Return the source-locations for triples for this subject.
 
@@ -377,9 +457,11 @@ rdf_source_location(Subject, Source) :-
 		 *	     STATISTICS		*
 		 *******************************/
 
-%	rdf_statistics(+Key(-Value))
+%%	rdf_statistics(?KeyValue) is nondet.
 %	
-%	Obtain some statistics
+%	Obtain statistics on the RDF database.
+%	
+%	@param KeyValue	Term of the form Key(Value).
 
 rdf_statistics(sources(Count)) :-
 	predicate_property(rdf_source(_,_,_,_), number_of_clauses(Count)).
@@ -389,9 +471,12 @@ rdf_statistics(properties(Count)) :-
 	rdf_statistics_(predicates(Count)).
 rdf_statistics(triples(Count)) :-
 	rdf_statistics_(triples(Count)).
-%rdf_statistics(triples_by_property(Rel0, Count)) :-
-%	rdf_global_id(Rel0, Rel),
-%	triples_on_relation(Rel, Count).
+rdf_statistics(gc(Count, Time)) :-
+	rdf_statistics_(gc(Count, Time)).
+rdf_statistics(rehash(Count, Time)) :-
+	rdf_statistics_(rehash(Count, Time)).
+rdf_statistics(core(Bytes)) :-
+	rdf_statistics_(core(Bytes)).
 rdf_statistics(lookup(Index, Count)) :-
 	functor(Indexed, indexed, 8),
 	rdf_statistics_(Indexed),
@@ -401,6 +486,8 @@ rdf_statistics(lookup(Index, Count)) :-
 	Count \== 0.
 rdf_statistics(searched_nodes(Count)) :-
 	rdf_statistics_(searched_nodes(Count)).
+rdf_statistics(literals(Count)) :-
+	rdf_statistics_(literals(Count)).
 rdf_statistics(triples_by_file(File, Count)) :-
 	(   var(File)
 	->  rdf_source(File),
@@ -419,29 +506,110 @@ index(rdf(+,-,+), 5).
 index(rdf(-,+,+), 6).
 index(rdf(+,+,+), 7).
 
-triples_on_relation(Rel, Count) :-
-	count_solutions(rdf(_,Rel,_), Count).
 
-%	count_solutions(+Goal, -Count)
+		 /*******************************
+		 *	     PREDICATES		*
+		 *******************************/
+
+%%	rdf_current_predicate(?Predicate)
 %	
-%	Count the number of times Goal succeeds.
+%	True if Predicate is a currently defined predicate.
 
-count_solutions(Goal, Count) :-
-	flag(rdf_db_count_solutions, Old, 0),
-	(   catch(Goal, E, (flag(rdf_db_count_solutions, _, Old),
-			    throw(E))),
-	    flag(rdf_db_count_solutions, C, C+1),
-	    fail
-	;   flag(rdf_db_count_solutions, C, Old)
-	),
-	Count = C.
+rdf_current_predicate(P) :-
+	var(P), !,
+	rdf_current_predicates(All),
+	member(P, All),
+	rdf_predicate_property_(P, triples(N)),
+	N > 0.
+rdf_current_predicate(P) :-
+	rdf_predicate_property_(P, triples(N)),
+	N > 0.
+
+rdf_current_predicate(P, DB) :-
+	rdf_current_predicates(All),
+	member(P, All),
+	once(rdf(_,P,_,DB:_)).
+
+%%	rdf_predicate_property(?Predicate, ?Property)
+%	
+%	Enumerate predicates and their properties
+
+
+rdf_predicate_property(P, Prop) :-
+	var(P), !,
+	rdf_current_predicates(All),
+	member(P, All),
+	rdf_predicate_property_(P, Prop).
+rdf_predicate_property(P, Prop) :-
+	rdf_predicate_property_(P, Prop).
+
+
+		 /*******************************
+		 *	    TRANSACTION		*
+		 *******************************/
+
+%%	rdf_transaction(:Goal)
+%	
+%	Backward compatibility
+
+rdf_transaction(Goal) :-
+	rdf_transaction(Goal, user).
+
+%%	rdf_monitor(:Goal, +Options)
+%
+
+rdf_monitor(Goal, Options) :-
+	monitor_mask(Options, 0xffff, Mask),
+	rdf_monitor_(Goal, Mask).
+
+monitor_mask([], Mask, Mask).
+monitor_mask([H|T], Mask0, Mask) :-
+	update_mask(H, Mask0, Mask1),
+	monitor_mask(T, Mask1, Mask).
+
+update_mask(-X, Mask0, Mask) :- !,
+	monitor_mask(X, M),
+	Mask is Mask0 /\ \M.
+update_mask(+X, Mask0, Mask) :- !,
+	monitor_mask(X, M),
+	Mask is Mask0 \/ M.
+update_mask(X, Mask0, Mask) :-
+	monitor_mask(X, M),
+	Mask is Mask0 \/ M.
+
+%%	monitor_mask(Name, Mask)
+%	
+%	Mask bit for the monitor events.  Note that this must be kept
+%	consistent with the enum broadcast_id defined in rdf_db.c
+
+					% C-defined broadcasts
+monitor_mask(assert,	   0x0001).
+monitor_mask(assert(load), 0x0002).
+monitor_mask(retract,	   0x0004).
+monitor_mask(update,	   0x0008).
+monitor_mask(new_literal,  0x0010).
+monitor_mask(old_literal,  0x0020).
+monitor_mask(transaction,  0x0040).
+monitor_mask(load,	   0x0080).
+monitor_mask(rehash,	   0x0100).
+					% prolog defined broadcasts
+monitor_mask(parse,	   0x1000).
+monitor_mask(reset,	   0x2000).
+monitor_mask(unload,	   0x1000).
+					% mask for all
+monitor_mask(all,	   0xffff).
+
+%rdf_broadcast(Term, MaskName) :-
+%%	monitor_mask(MaskName, Mask),
+%%	rdf_broadcast_(Term, Mask).
 
 
 		 /*******************************
 		 *    QUICK BINARY LOAD/SAVE	*
 		 *******************************/
 
-%	rdf_save_db(+File, [+DB])
+%%	rdf_save_db(+File) is det.
+%%	rdf_save_db(+File, +DB) is det.
 %	
 %	Save triples into File in a   quick-to-load binary format. If DB
 %	is supplied only triples flagged to originate from that database
@@ -457,13 +625,13 @@ rdf_save_db(File, DB) :-
 	call_cleanup(rdf_save_db_(Out, DB), close(Out)).
 
 
-rdf_load_db_no_admin(File) :-
+rdf_load_db_no_admin(File, Id) :-
 	open(File, read, Out, [type(binary)]),
-	call_cleanup(rdf_load_db_(Out), close(Out)).
+	call_cleanup(rdf_load_db_(Out, Id), close(Out)).
 
 
 rdf_load_db(File) :-
-	rdf_load_db_no_admin(File),
+	rdf_load_db_no_admin(File, file(File)),
 	rdf_sources_(Sources),
 	(   member(Src, Sources),
 	    rdf_md5(Src, MD5),
@@ -479,12 +647,12 @@ rdf_load_db(File) :-
 		 *	    LOADING RDF		*
 		 *******************************/
 
-%	rdf_load(+File, +Options)
+%%	rdf_load(+FileOrList, +Options)
 %
 %	Load RDF file.  Options provides additional processing options.
 %	Currently defined options are:
 %	
-%	    result(-Action, -Triples, -MD5)
+%	    # result(-Action, -Triples, -MD5)
 %	    	Return action taken (load, reload, none) and number
 %	    	of triples loaded from the file as well as the MD5
 %	    	digest.
@@ -494,6 +662,10 @@ rdf_load_db(File) :-
 rdf_load(Spec) :-
 	rdf_load(Spec, []).
 
+rdf_load([], _) :- !.
+rdf_load([H|T], Options) :- !,
+	rdf_load(H, Options),
+	rdf_load(T, Options).
 rdf_load(Spec, Options0) :-
 	statistics(cputime, CpuOld),
 	(   select(result(Action, Triples, MD5), Options0, Options1)
@@ -504,7 +676,7 @@ rdf_load(Spec, Options0) :-
 	->  Options2 = Options1
 	;   Options2 = [ blank_nodes(share)|Options1 ]
 	),
-	(   (   Spec = '$stream'(_)
+	(   (   is_stream(Spec)
 	    ->	In = Spec
 	    ;	Spec = stream(In)
 	    )
@@ -521,7 +693,6 @@ rdf_load(Spec, Options0) :-
 	    Load = parsed(ParseTime),
 	    Action = load
 	;   Source = Spec,
-	    Options = Options2,
 	    absolute_file_name(Spec,
 			       [ access(read),
 				 extensions([rdf,rdfs,owl,''])
@@ -536,21 +707,21 @@ rdf_load(Spec, Options0) :-
 	    ;	Action = load
 	    ),
 	    (	Action \== none
-	    ->  atom_concat('file:', File, DefBaseURI),
-		option(base_uri(BaseURI), Options1, DefBaseURI),
+	    ->  add_base_url(Options2, File, Options),
 		retractall(rdf_source(File, _, _, _)),
 		(   cache_file(File, Cache)
 		->  (   time_file(Cache, CacheTime),
 		        time_file(File, FileTime),
 			CacheTime >= FileTime,
-			catch(rdf_load_db_no_admin(Cache), _, fail),
+			catch(rdf_load_db_no_admin(Cache, cache(File)),
+			      _, fail),
 			ignore(memberchk(namespaces([]), Options)) % TBD
 		    ->  Load = cache(ParseTime)
-		    ;   process_rdf(File, assert_triples, Options),
+		    ;   parse_rdf_file(File, Options),
 			Load = parsed(ParseTime),
 			save_cache(File, Cache)
 		    )
-		;   process_rdf(File, assert_triples, Options),
+		;   parse_rdf_file(File, Options),
 		    Load = parsed(ParseTime)
 		),
 		rdf_statistics_(triples(File, Triples)),
@@ -562,35 +733,86 @@ rdf_load(Spec, Options0) :-
 	(   Action \== none
 	->  statistics(cputime, CpuLoaded),
 	    ParseTime is CpuLoaded - CpuOld,
-	    print_message(informational,
+	    (	memberchk(silent(true), Options)
+	    ->	Level = silent
+	    ;	Level = informational
+	    ),
+	    print_message(Level,
 			  rdf(loaded(Source, Triples, Load)))
 	;   true
 	).
 
 
-%	rdf_unload(+Spec)
+parse_rdf_file(File, Options) :-
+	source_descr(File, Id),
+	rdf_transaction(process_rdf(File, assert_triples, Options),
+			parse(Id)).
+
+source_descr(File, file(File)) :-
+	atom(File), !.
+source_descr(Stream, stream(Stream)) :-
+	is_stream(Stream), !.
+source_descr(Descr, Descr).
+
+
+%%	add_base_url(+Options0, +File, -Options)
+%	
+%	Add base_uri(file://File) to the options if  this is not already
+%	there. A base URI is needed  to   avoid  the  case where loading
+%	multiple  files  independently  cause    clashes   of  generated
+%	descriptions.
+
+add_base_url(Options, _File, Options) :-
+	option(base_uri(Base), Options, Var),
+	Base \== Var, !.
+add_base_url(Options, File, [base_uri(Base)|Options]) :-
+	atom_concat('file://', File, Base).
+
+
+%%	rdf_unload(+Spec)
 %	
 %	Remove the triples loaded from the specified source and remove
 %	the source from the database.
 
 rdf_unload(Spec) :-
-	(   Spec = '$stream'(_)
+	(   (   is_stream(Spec)
+	    ;   Spec = stream(_)
+	    )
 	->  throw(error(permission_error(rdf_db, unload, Spec), _))
+	;   atom(Spec),
+	    rdf_statistics_(triples(Spec, _)),
+	    (	rdf(_,_,_,Spec)
+	    ;   rdf(_,_,_,Spec:_)
+	    )
+	->  do_unload(Spec)
 	;   absolute_file_name(Spec,
 			       [ access(read),
-				 extensions([rdf,rdfs,owl,''])
-			       ], File),
-	    rdf_retractall(_,_,_,File:_),
-	    retractall(rdf_source(File, _, _, _))
+				 extensions([rdf,rdfs,owl,'']),
+				 file_errors(fail)
+			       ], File)
+	->  do_unload(File)
+	;   true
 	).
 	
+do_unload(Spec) :-
+	retractall(rdf_source(Spec, _, _, _)),
+	rdf_transaction((rdf_retractall(_,_,_,Spec:_),
+			 rdf_retractall(_,_,_,Spec)),
+			unload(Spec)).
 
-%	rdf_source(?Source)
+
+%%	rdf_source(?Source)
 %	
-%	Query the loaded sources
+%	Query the loaded sources.  
 
 rdf_source(File) :-
-	rdf_source(File, _, _, _).
+	rdf_source(File,_,_,_).		% loaded files
+rdf_source(File) :- 
+	rdf_sources_(Sources),		% other sources
+	member(File, Sources),
+	\+ rdf_source(File,_,_,_),
+	rdf_statistics_(triples(File, Triples)),
+	Triples > 0.
 
 %	rdf_make
 %	
@@ -600,10 +822,10 @@ rdf_source(File) :-
 rdf_make :-
 	forall((rdf_source(File, Time, _Triples, _),
 		Time \== 0),
-	       rdf_load(File)).
+	       catch(rdf_load(File), _, true)).
 
 
-%	cache_file(+Base, -CacheFile)
+%%	cache_file(+Base, -CacheFile)
 %	
 %	Deduce the name of the file used to cache the triples.
 
@@ -621,14 +843,14 @@ cache_file(Base, Cache) :-
 	exists_directory(CacheDir),
 	concat_atom([CacheDir, /, File, '.trp'], Cache).
 
-%	save_cache(+File, +Cache)
+%%	save_cache(+File, +Cache)
 
 save_cache(File, Cache) :-
 	catch(open(Cache, write, CacheStream, [type(binary)]), _, fail), !,
 	rdf_save_db_(CacheStream, File),
 	close(CacheStream).
 
-%	assert_triples(+Triples, +Source)
+%%	assert_triples(+Triples, +Source)
 %
 %	Assert a list of triples into the database. Foir security
 %	reasons we check we aren't inserting anything but nice RDF
@@ -646,21 +868,22 @@ assert_triples([H|_], _) :-
 		 *	       RESET		*
 		 *******************************/
 
-%	rdf_reset_db
+%%	rdf_reset_db
 %	
 %	Remove all triples from the RDF database and reset all its
 %	statistics.
 
 rdf_reset_db :-
-	rdf_reset_db_,
-	retractall(rdf_source(_,_,_,_)).
+	retractall(rdf_source(_,_,_,_)),
+	rdf_transaction(rdf_reset_db_, reset).
 
 
 		 /*******************************
 		 *	     SAVE RDF		*
 		 *******************************/
 
-%	rdf_save(File, [+Options])
+%%	rdf_save(File) is det.
+%%	rdf_save(File, +Options) is det.
 %
 %	Save RDF data to file.  Options is a list of one or more of the
 %	following options:
@@ -676,11 +899,22 @@ rdf_reset_db :-
 %		Call Convertor(-Type, -Content, +RDFObject), providing
 %		the opposite for the convert_typed_literal option of
 %		the RDF parser.
+%		
+%		# encoding(Encoding)
+%		Encoding for the output.  Either utf8 or iso_latin_1
+%		
+%		# document_language(+Lang)
+%		Initial xml:lang saved with rdf:RDF element
 
 :- module_transparent
+	rdf_transaction/1,
+	rdf_monitor/2,
 	rdf_save/2,
 	meta_options/2.
 
+:- thread_local
+	named_anon/2.			% +Resource, -Id
+
 rdf_save(File) :-
 	rdf_save2(File, []).
 
@@ -694,7 +928,9 @@ rdf_save(File, DB) :-
 
 
 rdf_save2(File, Options) :-
-	open(File, write, Out),
+	option(encoding(Encoding), Options, utf8),
+	valid_encoding(Encoding),
+	open(File, write, Out, [encoding(Encoding)]),
 	flag(rdf_db_saved_subjects, OSavedSubjects, 0),
 	flag(rdf_db_saved_triples, OSavedTriples, 0),
 	call_cleanup(rdf_do_save(Out, Options),
@@ -706,6 +942,13 @@ rdf_save2(File, Options) :-
 				  Out)).
 
 
+valid_encoding(Enc) :-
+	(   xml_encoding_name(Enc, _)
+	->  true
+	;   throw(error(domain_error(encoding, Enc), _))
+	).
+
+
 cleanup_save(Reason,
 	     File,
 	     OSavedSubjects,
@@ -714,6 +957,7 @@ cleanup_save(Reason,
 	close(Out),
 	flag(rdf_db_saved_subjects, SavedSubjects, OSavedSubjects),
 	flag(rdf_db_saved_triples, SavedTriples, OSavedTriples),
+	retractall(named_anon(_, _)),
 	(   Reason == exit
 	->  print_message(informational,
 			  rdf(saved(File, SavedSubjects, SavedTriples)))
@@ -745,7 +989,7 @@ db(Options, DB) :-
 	).
 
 
-%	meta_options(+OptionsIn, -OptionsOut)
+%%	meta_options(+OptionsIn, -OptionsOut)
 %	
 %	Do module qualification for options that are module sensitive.
 
@@ -763,17 +1007,17 @@ meta_options([H0|T0], [H|T]) :-
 	meta_options(T0, T).
 
 
-%	rdf_save_header(+Fd, +Options)
+%%	rdf_save_header(+Fd, +Options)
 %
 %	Save XML document header, doctype and open the RDF environment.
 %	This predicate also sets up the namespace notation.
 
 rdf_save_header(Out, Options) :-
 	is_list(Options), !,
-	db(Options, DB),
-	format(Out, '<?xml version=\'1.0\' encoding=\'ISO-8859-1\'?>~n', []),
+	xml_encoding(Out, Encoding),
+	format(Out, '<?xml version=\'1.0\' encoding=\'~w\'?>~n', [Encoding]),
 	format(Out, '<!DOCTYPE rdf:RDF [', []),
-	used_namespaces(NSList, DB),
+	header_namespaces(Options, NSList),
 	(   member(Id, NSList),
 	    ns(Id, NS),
 	    format(Out, '~N    <!ENTITY ~w \'~w\'>', [Id, NS]),
@@ -787,37 +1031,85 @@ rdf_save_header(Out, Options) :-
 	    fail
 	;   true
 	),
+	(   memberchk(document_language(Lang), Options)
+	->  format(Out, '~N    xml:lang="~w"', [Lang])
+	;   true
+	),
 	format(Out, '>~n', []).
 rdf_save_header(Out, FileRef) :-	% compatibility
 	atom(FileRef),
 	rdf_save_header(Out, [db(FileRef)]).
 	
+xml_encoding(Out, Encoding) :-
+	stream_property(Out, encoding(Enc)),
+	(   xml_encoding_name(Enc, Encoding)
+	->  true
+	;   throw(error(domain_error(rdf_encoding, Enc), _))
+	).
 
-%	used_namespaces(-List)
+xml_encoding_name(ascii,       'US-ASCII').
+xml_encoding_name(iso_latin_1, 'ISO-8859-1').
+xml_encoding_name(utf8,        'UTF-8').
+
+
+%%	header_namespaces(Options, -List)
+%	
+%	Get namespaces we will define as entities
+
+header_namespaces(Options, List) :-
+	memberchk(namespaces(NSL0), Options), !,
+	sort([rdf,rdfs|NSL0], List).
+header_namespaces(Options, List) :-
+	db(Options, DB),
+	used_namespace_entities(List, DB).
+	
+%%	used_namespace_entities(-List, ?DB)
 %
 %	Return the list of namespaces used in an RDF database.
 
-used_namespaces(List, DB) :-
+used_namespace_entities(List, DB) :-
 	decl_used_predicate_ns(DB),
-	setof(NS, Full^ns(NS, Full), NS0),
-	used_ns(NS0, List, DB).
+	used_namespaces(List, DB).
 
-used_ns([], [], _).
-used_ns([H|T0], [H|T], DB) :-
-	used_ns(H, DB), !,
-	used_ns(T0, T, DB).
-used_ns([_|T0], T, DB) :-
-	used_ns(T0, T, DB).
+used_namespaces(List, DB) :-
+	empty_nb_set(Set),
+	ns(rdf, RDF),
+	add_nb_set(RDF, Set),
+	(   rdf_db(S, P, O, DB),
+	    add_ns(S, Set),
+	    add_ns(P, Set),
+	    add_ns_obj(O, Set),
+	    fail
+	;   true
+	),
+	nb_set_to_list(Set, FullList),
+	ns_abbreviations(FullList, List).
 
-used_ns(rdf, _) :- !.			% we need rdf:RDF
-used_ns(NS, DB) :-
-	ns(NS, Full),
-	rdf_db(S,P,O,DB),
-	(   sub_atom(S, 0, _, _, Full)
-	;   sub_atom(P, 0, _, _, Full)
-	;   atom(O),
-	    sub_atom(O, 0, _, _, Full)
-	), !.
+ns_abbreviations([], []).
+ns_abbreviations([H0|T0], [H|T]) :-
+	ns(H, H0), !,
+	ns_abbreviations(T0, T).
+ns_abbreviations([_|T0], T) :-
+	ns_abbreviations(T0, T).
+	
+
+add_ns(S, Set) :-
+	rdf_url_namespace(S, Full),
+	Full \== '', !,
+	add_nb_set(Full, Set).
+add_ns(_, _).
+
+add_ns_obj(O, Set) :-
+	atom(O),
+	rdf_url_namespace(O, Full),
+	Full \== '', !,
+	add_nb_set(Full, Set).
+add_ns_obj(literal(type(Type, _)), Set) :-
+	atom(Type), !,
+	rdf_url_namespace(Type, Full),
+	Full \== '', !,
+	add_nb_set(Full, Set).
+add_ns_obj(_,_).
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -830,7 +1122,7 @@ cannot use names holding /, :, etc. as XML identifiers.
 
 decl_used_predicate_ns(DB) :-
 	retractall(predicate_ns(_,_)),
-	(   rdf_db(_,P,_,DB),
+	(   rdf_current_predicate(P, DB),
 	    decl_predicate_ns(P),
 	    fail
 	;   true
@@ -847,7 +1139,7 @@ decl_predicate_ns(Pred) :-
 	xml_codes(LocalCodes), !,
 	(   NSCodes \== []
 	->  atom_codes(NS, NSCodes),
-	    (   between(1, 1000000, N),
+	    (   between(1, infinite, N),
 		atom_concat(ns, N, Id),
 		\+ ns(Id, _)
 	    ->  rdf_register_ns(Id, NS),
@@ -869,22 +1161,22 @@ xml_code(0'-).
 
 
 rdf_save_footer(Out) :-
+	retractall(named_anon(_, _)),
 	format(Out, '</rdf:RDF>~n', []).
 
-%	rdf_save_non_anon_subject(+Out, +Subject, +Options)
+%%	rdf_save_non_anon_subject(+Out, +Subject, +Options)
 %	
 %	Save an object.  Anonymous objects not saved if anon(false)
 %	is present in the Options list.
 
 rdf_save_non_anon_subject(_Out, Subject, Options) :-
-	anonymous_subject(Subject),
+	rdf_is_bnode(Subject),
 	(   memberchk(anon(false), Options)
 	;   db(Options, DB),
 	    rdf_db(_, _, Subject, DB)
 	), !.
 rdf_save_non_anon_subject(Out, Subject, Options) :-
-	db(Options, DB),
-	rdf_save_subject(Out, Subject, DB),
+	rdf_save_subject(Out, Subject, Options),
 	flag(rdf_db_saved_subjects, X, X+1).
 
 
@@ -940,13 +1232,18 @@ xml_is_name(_NS:Atom) :- !,
 xml_is_name(Atom) :-
 	xml_name(Atom).
 
-save_about(_Out, Subject) :-
-	anonymous_subject(Subject), !.
 save_about(Out, Subject) :-
-	rdf_value(Subject, QSubject),
+	rdf_is_bnode(Subject), !,
+	(   named_anon(Subject, NodeID)
+	->  format(Out, ' rdf:nodeID="~w"', [NodeID])
+	;   true
+	).
+save_about(Out, Subject) :-
+	stream_property(Out, encoding(Encoding)),
+	rdf_value(Subject, QSubject, Encoding),
 	format(Out, ' rdf:about="~w"', [QSubject]).
 
-%	save_attributes(+List, +DefNS, +Stream, Element)
+%%	save_attributes(+List, +DefNS, +Stream, Element)
 %
 %	Save the attributes.  Short literal attributes are saved in the
 %	tag.  Others as the content of the description element.  The
@@ -963,7 +1260,7 @@ save_attributes(Atts, DefNS, Out, Element, Indent, Options) :-
 	    format(Out, '~N~*|</~w>~n', [Indent, Element])
 	).
 
-%	split_attributes(+Attributes, -HeadAttrs, -BodyAttr)
+%%	split_attributes(+Attributes, -HeadAttrs, -BodyAttr)
 %	
 %	Split attribute (Name=Value) list into attributes for the head
 %	and body. Attributes can only be in the head if they are literal
@@ -974,7 +1271,7 @@ split_attributes(Atts, HeadAttr, BodyAttr) :-
 	simple_literal_attributes(Singles, HeadAttr, Rest),
 	append(Dupls, Rest, BodyAttr).
 
-%	duplicate_attributes(+Attrs, -Duplicates, -Singles)
+%%	duplicate_attributes(+Attrs, -Duplicates, -Singles)
 %	
 %	Extract attributes that appear more than onces as we cannot
 %	dublicate an attribute in the head according to the XML rules.
@@ -998,7 +1295,7 @@ named_attributes(Name, [H|T], D, R) :-
 	    named_attributes(Name, T, D, RT)
 	).
 
-%	simple_literal_attributes(+Attributes, -Inline, -Body)
+%%	simple_literal_attributes(+Attributes, -Inline, -Body)
 %
 %	Split attributes for (literal) attributes to be used in the
 %	begin-tag and ones that have to go into the body of the description.
@@ -1015,7 +1312,7 @@ in_tag_attribute(_=literal(Text)) :-
 	atom_length(Text, Len),
 	Len < 60.
 
-%	save_attributes(+List, +DefNS, +TagOrBody, +Stream)
+%%	save_attributes(+List, +DefNS, +TagOrBody, +Stream)
 %
 %	Save a list of attributes.
 
@@ -1027,7 +1324,8 @@ save_attributes2([H|T], DefNS, Where, Out, Indent, Options) :-
 save_attribute(tag, Name=literal(Value), DefNS, Out, Indent, _DB) :-
 	AttIndent is Indent + 2,
 	rdf_att_id(Name, DefNS, NameText),
-	xml_quote_attribute(Value, QVal),
+	stream_property(Out, encoding(Encoding)),
+	xml_quote_attribute(Value, QVal, Encoding),
 	format(Out, '~N~*|~w="~w"', [AttIndent, NameText, QVal]).
 save_attribute(body, Name=literal(Literal0), DefNS, Out, Indent, Options) :- !,
 	rdf_id(Name, DefNS, NameText),
@@ -1037,55 +1335,85 @@ save_attribute(body, Name=literal(Literal0), DefNS, Out, Indent, Options) :- !,
 	;   Literal = Literal0
 	),
 	(   Literal = lang(Lang, Value)
-	->  rdf_id(Lang, DefNS, LangText),
-	    format(Out, '~N~*|<~w xml:lang="~w">',
-		   [Indent, NameText, LangText])
+	->  (   memberchk(document_language(Lang), Options)
+	    ->  format(Out, '~N~*|<~w>', [Indent, NameText])
+	    ;   rdf_id(Lang, DefNS, LangText),
+		format(Out, '~N~*|<~w xml:lang="~w">',
+		       [Indent, NameText, LangText])
+	    )
 	;   Literal = type(Type, Value)
-	->  rdf_id(Type, DefNS, TypeText),
-	    format(Out, '~N~*|<~w rdf:dataType="~w">',
-		   [Indent, NameText, TypeText])
-	;   format(Out, '~N~*|<~w>', [Indent, NameText]),
+	->  stream_property(Out, encoding(Encoding)),
+	    rdf_value(Type, QVal, Encoding),
+	    format(Out, '~N~*|<~w rdf:datatype="~w">',
+		   [Indent, NameText, QVal])
+	;   atomic(Literal)
+	->  format(Out, '~N~*|<~w>', [Indent, NameText]),
+	    Value = Literal
+	;   format(Out, '~N~*|<~w rdf:parseType="Literal">',
+		   [Indent, NameText]),
 	    Value = Literal
 	),
-	save_attribute_value(Value, Out),
+	save_attribute_value(Value, Out, Indent),
 	format(Out, '</~w>', [NameText]).
 save_attribute(body, Name=Value, DefNS, Out, Indent, Options) :-
-	anonymous_subject(Value), !,
+	rdf_is_bnode(Value), !,
 	rdf_id(Name, DefNS, NameText),
-	SubIndent is Indent + 2,
-	(   rdf(Value, rdf:type, rdf:'List')
-	->  format(Out, '~N~*|<~w rdf:parseType="Collection">~n',
-		   [Indent, NameText]),
-	    rdf_save_list(Out, Value, DefNS, SubIndent, Options)
-	;   format(Out, '~N~*|<~w>~n',
-		   [Indent, NameText]),
-	    rdf_save_subject(Out, Value, DefNS, SubIndent, Options)
-	),
-	format(Out, '~N~*|</~w>~n', [Indent, NameText]).
+	(   named_anon(Value, NodeID)
+	->  format(Out, '~N~*|<~w rdf:nodeID="~w"/>',
+		   [Indent, NameText, NodeID])
+	;   (   rdf(S1, Name, Value),
+	        rdf(S2, P2, Value),
+		(S1 \== S2 ; Name \== P2)
+	    ->  predicate_property(named_anon(_,_), number_of_clauses(N)),
+		atom_concat('bn', N, NodeID),
+		assert(named_anon(Value, NodeID))
+	    ;	true
+	    ),
+	    SubIndent is Indent + 2,
+	    (   rdf(Value, rdf:type, rdf:'List')
+	    ->  format(Out, '~N~*|<~w', [Indent, NameText]),
+		save_about(Out, Value),
+		format(Out, ' rdf:parseType="Collection">~n', []),
+		rdf_save_list(Out, Value, DefNS, SubIndent, Options)
+	    ;   format(Out, '~N~*|<~w>~n', [Indent, NameText]),
+		rdf_save_subject(Out, Value, DefNS, SubIndent, Options)
+	    ),
+	    format(Out, '~N~*|</~w>~n', [Indent, NameText])
+	).
 save_attribute(body, Name=Value, DefNS, Out, Indent, _DB) :-
-	rdf_value(Value, QVal),
+	stream_property(Out, encoding(Encoding)),
+	rdf_value(Value, QVal, Encoding),
 	rdf_id(Name, DefNS, NameText),
 	format(Out, '~N~*|<~w rdf:resource="~w"/>', [Indent, NameText, QVal]).
 
-save_attribute_value(Value, Out) :-	% strings
+save_attribute_value(Value, Out, _) :-	% strings
 	atom(Value), !,
-	xml_quote_cdata(Value, QVal),
+	stream_property(Out, encoding(Encoding)),
+	xml_quote_cdata(Value, QVal, Encoding),
 	write(Out, QVal).
-save_attribute_value(Value, Out) :-	% numbers
+save_attribute_value(Value, Out, _) :-	% numbers
 	number(Value), !,
 	writeq(Out, Value).		% quoted: preserve floats
-save_attribute_value(Value, _Out) :-
+save_attribute_value(Value, Out, Indent) :-
+	xml_is_dom(Value), !,
+	XMLIndent is Indent+2,
+	xml_write(Out, Value,
+		  [ header(false),
+		    indent(XMLIndent)
+		  ]).
+save_attribute_value(Value, _Out, _) :-
 	throw(error(save_attribute_value(Value), _)).
 
 rdf_save_list(_, List, _, _, _) :-
 	rdf_equal(List, rdf:nil), !.
 rdf_save_list(Out, List, DefNS, Indent, Options) :-
 	rdf_has(List, rdf:first, First),
-	(   anonymous_subject(First)
+	(   rdf_is_bnode(First)
 	->  nl(Out),
 	    rdf_save_subject(Out, First, DefNS, Indent, Options)
-	;   rdf_value(First, QVal),
-	    format(Out, '~N~*|<rdf:Description about="~w"/>',
+	;   stream_property(Out, encoding(Encoding)),
+	    rdf_value(First, QVal, Encoding),
+	    format(Out, '~N~*|<rdf:Description rdf:about="~w"/>',
 		   [Indent, QVal])
 	),
 	flag(rdf_db_saved_triples, X, X+3),
@@ -1095,17 +1423,7 @@ rdf_save_list(Out, List, DefNS, Indent, Options) :-
 	;   true
 	).
 
-%	anonymous_subject(+Subject)
-%	
-%	Test if a resource is anonymous. This is highly dubious.
-%	Probably we need to store this in the database.  The current
-%	release of the RDF parser guarantees that all anonymous ids
-%	start with __.
-
-anonymous_subject(S) :-
-	sub_atom(S, 0, _, _, '__'), !.
-
-%	rdf_id(+Resource, +DefNS, -NSLocal)
+%%	rdf_id(+Resource, +DefNS, -NSLocal)
 %	
 %	Generate a NS:Local name for Resource given the indicated
 %	default namespace.  This call is used for elements.
@@ -1128,12 +1446,22 @@ rdf_att_id(Id, _, NS:Local) :-
 rdf_att_id(Id, _, Id).
 
 
-rdf_value(V, Text) :-
+%%	rdf_value(+Resource, -Text, +Encoding)
+%	
+%	According  to  "6.4  RDF  URI  References"  of  the  RDF  Syntax
+%	specification, a URI reference is  UNICODE string not containing
+%	control sequences, represented as  UTF-8   and  then  as escaped
+%	US-ASCII.
+
+rdf_value(V, Text, Encoding) :-
 	ns(NS, Full),
 	atom_concat(Full, Local, V), !,
-	xml_quote_attribute(Local, QLocal),
+	rdf_quote_uri(Local, QLocal0),
+	xml_quote_attribute(QLocal0, QLocal, Encoding),
 	concat_atom(['&', NS, (';'), QLocal], Text).
-rdf_value(V, V).
+rdf_value(V, Q, Encoding) :-
+	rdf_quote_uri(V, Q0),
+	xml_quote_attribute(Q0, Q, Encoding).
 
 
 		 /*******************************
@@ -1178,3 +1506,19 @@ source(Spec) -->
 in_time(Triples, ParseTime) -->
 	[ ' in ~2f sec; ~D triples'-[ParseTime, Triples]
 	].
+
+		 /*******************************
+		 *	    ENVIRONMENT		*
+		 *******************************/
+
+%	prolog:meta_goal(+Goal, -MetaArgs)
+%	
+%	Used by the XPCE/Prolog cross-referencer as well as PceEmacs for
+%	colouring code.
+
+:- multifile
+	prolog:meta_goal/2.
+
+prolog:meta_goal(rdf_transaction(G),	[G]).
+prolog:meta_goal(rdf_transaction(G,_),	[G]).
+prolog:meta_goal(rdf_monitor(G,_),	[G+1]).
