@@ -1,9 +1,9 @@
-/*  $Id: rdf_edit.pl,v 1.38 2004/09/21 14:22:24 jan Exp $
+/*  $Id: rdf_edit.pl,v 1.44 2006/08/31 09:50:20 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
     Copyright (C): 1985-2002, University of Amsterdam
 
@@ -34,6 +34,7 @@
 	  [ rdfe_assert/3,		% Sub, Pred, Obj
 	    rdfe_assert/4,		% Sub, Pred, Obj, PayLoad
 	    rdfe_retractall/3,		% Sub, Pred, Obj
+	    rdfe_retractall/4,		% Sub, Pred, Obj, PayLoad
 	    rdfe_update/4,		% Sub, Pred, Obj, +Action
 	    rdfe_update/5,		% Sub, Pred, Obj, +PayLoad, +Action
 	    rdfe_load/1,		% +File
@@ -96,50 +97,15 @@ This library provides a number of functions on top of the rdf_db module:
     * Provide undo/redo
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-%	user:goal_expansion(+NSGoal, -Goal)
-%	
-%	This predicate allows for writing down rdf queries in a friendly
-%	name-space fashion.  
-
-:- multifile
-	user:goal_expansion/2.
-
-user:goal_expansion(rdfe_assert(Subj0, Pred0, Obj0),
-		    rdfe_assert(Subj, Pred, Obj)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdfe_assert(Subj0, Pred0, Obj0, PayLoad),
-		    rdfe_assert(Subj, Pred, Obj, PayLoad)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdfe_retractall(Subj0, Pred0, Obj0),
-		    rdfe_retractall(Subj, Pred, Obj)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj).
-user:goal_expansion(rdfe_update(Subj0, Pred0, Obj0, Action0),
-		    rdfe_update(Subj, Pred, Obj, Action)) :-
-	rdf_global_id(Subj0, Subj),
-	rdf_global_id(Pred0, Pred),
-	rdf_global_object(Obj0, Obj),
-	(   compound(Action0)
-	->  Action0 =.. [Name,Res0],
-	    rdf_global_id(Res0, Res),
-	    Action =.. [Name,Res]
-	;   Action = Action0
-	).
-user:goal_expansion(rdfe_delete(Subj0),
-		    rdfe_delete(Subj)) :-
-	rdf_global_id(Subj0, Subj).
-user:goal_expansion(rdfe_transaction(G0),
-		    rdfe_transaction(G)) :-
-	expand_goal(G0, G).
-user:goal_expansion(rdfe_transaction(G0, Name),
-		    rdfe_transaction(G, Name)) :-
-	expand_goal(G0, G).
-
+:- rdf_meta
+	rdfe_assert(r,r,o),
+	rdfe_assert(r,r,o,+),
+	rdfe_retractall(r,r,o),
+	rdfe_update(r,r,o,t),
+	rdfe_delete(r),
+	rdfe_transaction(:),
+	rdfe_transaction(:, +).
+	
 
 		 /*******************************
 		 *     BASIC EDIT OPERATIONS	*
@@ -167,14 +133,14 @@ rdfe_retractall(Subject, Predicate, Object, PayLoad) :-
 	),
 	rdf_retractall(Subject, Predicate, Object, PayLoad).
 	
-%	rdfe_update(+Subject, +Predicate, +Object, +Action)
+%%	rdfe_update(+Subject, +Predicate, +Object, +Action)
 %	
 %	Update an existing triple.  Possible actions are:
 %	
-%		subject(+Subject)
-%		predicate(+Predicate)
-%		object(+Object)
-%		source(+Source)
+%%		subject(+Subject)
+%%		predicate(+Predicate)
+%%		object(+Object)
+%%		source(+Source)
 
 rdfe_update(Subject, Predicate, Object, Action) :-
 	rdfe_current_transaction(TID),
@@ -202,7 +168,7 @@ rdfe_update(Subject, Predicate, Object, PayLoad, Action) :-
 	),
 	journal(update(TID, Subject, Predicate, Object, PayLoad, Action)).
 
-%	rdfe_delete(+Subject)
+%%	rdfe_delete(+Subject)
 %	
 %	Delete a subject and all we know about it. This is a bit tricky.
 %	If we are involved in transitive relations, should we re-joint
@@ -221,7 +187,7 @@ delete(Subject) :-
 		 *	   FILE HANDLING	*
 		 *******************************/
 
-%	rdfe_load(+File)
+%%	rdfe_load(+File)
 %	
 %	Load an RDF file and record this action including version information
 %	to facilitate reliable reload.
@@ -269,7 +235,7 @@ rdfe_unload(Path) :-
 	journal(rdf_unload(TID, Path)).
 
 
-%	ensure_snapshot(+Path)
+%%	ensure_snapshot(+Path)
 %	
 %	Ensure we have a snapshot of Path if we are making a journal, so
 %	we can always reload the snapshot to ensure exactly the same
@@ -295,7 +261,7 @@ ensure_snapshot(Path) :-
 ensure_snapshot(_).
 
 
-%	load_snapshot(+Source, +Path)
+%%	load_snapshot(+Source, +Path)
 %	
 %	Load triples from the given snapshot   file. One of the troubles
 %	is the time-stamp to avoid rdf_make/0   from reloading the file.
@@ -316,7 +282,7 @@ load_snapshot(Source, Path) :-
 	assert(snapshot_file(Path)).
 
 
-%	snapshot_file(+Path, +MD5, +Access, -File)
+%%	snapshot_file(+Path, +MD5, +Access, -File)
 %	
 %	Find existing snapsnot file or location to save a new one.
 
@@ -330,7 +296,7 @@ snapshot_file(Path, MD5, Options, SnapShot) :-
 			   SnapShot).
 
 
-%	rdfe_snapshot_file(-File)
+%%	rdfe_snapshot_file(-File)
 %	
 %	Enumerate the MD5 snapshot files required to restore the current
 %	journal file. Using this  call  we   can  write  a  routine that
@@ -350,7 +316,7 @@ rdfe_snapshot_file(File) :-
 :- volatile
 	system_ns/2.
 
-%	rdfe_register_ns(Id, URI)
+%%	rdfe_register_ns(Id, URI)
 %	
 %	Encapsulation of rdf_register_ns(Id, URI)
 
@@ -394,7 +360,7 @@ save_system_ns :-
 		 *	   TRANSACTIONS		*
 		 *******************************/
 
-%	rdfe_transaction(:Goal)
+%%	rdfe_transaction(:Goal)
 %	
 %	Run Goal, recording all modifications   as a single transaction.
 %	If  Goal  raises  an  exception  or    fails,  all  changes  are
@@ -457,7 +423,7 @@ rdfe_rollback :-
 	journal(rollback(TID)),
 	rollback(TID).
 
-%	rollback(+TID)
+%%	rollback(+TID)
 %	
 %	This is the same as undo/1, but it must not record the undone
 %	actions as rollbacks cannot be `redone'.  Somehow there should
@@ -489,7 +455,7 @@ rollback(Action, Subject, Predicate, Object) :-
 assert_action(TID, Action, Subject, Predicate, Object) :-
 	asserta(undo_log(TID, Action, Subject, Predicate, Object)).
 
-%	undo(+TID)
+%%	undo(+TID)
 %	
 %	Undo a transaction as well as possible transactions nested into
 %	it.
@@ -589,8 +555,8 @@ rdfe_redo :-
 	broadcast(rdf_undo(redo, UnDone)).
 
 
-%	rdfe_can_redo(-TID)
-%	rdfe_can_undo(-TID)
+%%	rdfe_can_redo(-TID) is semidet.
+%%	rdfe_can_undo(-TID) is semidet.
 %	
 %	Check if we can undo and if so return the id of the transaction
 %	that will be un/re-done.  A subsequent call to rdfe_transaction_name
@@ -611,7 +577,7 @@ rdfe_can_undo(Undo) :-			% start undo
 	last_transaction(TID),
 	find_previous_undo(TID, Undo).
 
-%	rdfe_transaction_name(+TID, -Name)
+%%	rdfe_transaction_name(+TID, -Name)
 %	
 %	Return name if the transaction is named.
 
@@ -619,7 +585,7 @@ rdfe_transaction_name(TID, Name) :-
 	transaction_name(TID, Name),
 	Name \== [].
 
-%	rdfe_set_transaction_name(+Name)
+%%	rdfe_set_transaction_name(+Name)
 %	
 %	Set name of the current transaction
 
@@ -627,7 +593,7 @@ rdfe_set_transaction_name(Name) :-
 	current_transaction(TID), !,
 	assert(transaction_name(TID, Name)).
 
-%	rdfe_transaction_member(+TID, -Action)
+%%	rdfe_transaction_member(+TID, -Action)
 %	
 %	Query actions inside a transaction to allow for quick update
 %	of visualisers.
@@ -660,12 +626,12 @@ user_transaction_member(Update, Subject, Predicate, Object,
 	rdf_source_permission/2,	% file, ro/rw
 	rdf_current_default_file/2.	% file, all/fallback
 
-%	rdfe_set_file_property(+File, +Options)
+%%	rdfe_set_file_property(+File, +Options)
 %	
 %	Set properties on the file.  Options is one of
 %	
-%		access(ro/rw)
-%		default(all/fallback)
+%%		access(ro/rw)
+%%		default(all/fallback)
 
 rdfe_set_file_property(File0, access(Access)) :- !,
 	absolute_file_name(File0, File),
@@ -680,7 +646,7 @@ rdfe_set_file_property(File0, default(Type)) :-
 	broadcast(rdf_file_property(File, default(Type))).
 
 
-%	rdfe_get_file_property(?File, ?Option)
+%%	rdfe_get_file_property(?File, ?Option)
 %	
 %	Fetch file properties set with rdf_set_file_property/2.
 
@@ -702,7 +668,7 @@ rdfe_get_file_property(File, default(Default)) :-
 	).
 
 
-%	check_file_protection(-Error)
+%%	check_file_protection(-Error)
 %	
 %	Check modification of all protected files
 
@@ -718,7 +684,7 @@ check_file_protection(Error) :-
 		 *	     MODIFIED		*
 		 *******************************/
 
-%	rdfe_is_modified(?File)
+%%	rdfe_is_modified(?File)
 %	
 %	True if facts have been added, deleted or updated that have File
 %	as `payload'.
@@ -752,7 +718,7 @@ rdfe_clear_modified(File) :-
 		 *	     WATERMARKS		*
 		 *******************************/
 
-%	rdfe_set_watermark(Name)
+%%	rdfe_set_watermark(Name)
 %	
 %	Create a watermark for undo and replay journal upto this point.
 %	The rest of the logic needs to be written later.
@@ -858,7 +824,7 @@ journal_comment(start, Time) :-
 	format(Stream,
 	       '/* Triple20 Journal File\n\n   \
 	       Created: ~w\n   \
-	       Triple20 by Jan Wielemaker <jan@swi.psy.uva.nl>\n\n   \
+	       Triple20 by Jan Wielemaker <wielemak@science.uva.nl>\n\n   \
 	       EDIT WITH CARE!\n\
 	       */~n~n', [String]).
 journal_comment(resume, Time) :-
@@ -880,7 +846,7 @@ rdfe_close_journal :-
 	;   close(Stream)
 	).
 
-%	rdfe_current_journal(-Path)
+%%	rdfe_current_journal(-Path)
 %	
 %	Query the currently open journal
 
@@ -901,7 +867,7 @@ journal(Term) :-
 	(   journal(_, _, Stream)
 	->  write_journal(Term, Stream),
 	    flush_output(Stream)
-	;   report_no_journal
+	;   broadcast(rdf_no_journal(Term))
 	).
 
 write_journal(commit(TID, Time), Stream) :- !,
@@ -910,29 +876,7 @@ write_journal(Term, Stream) :-
 	format(Stream, '~q.~n', [Term]).
 
 
-:- dynamic
-	reported_no_journal/0.
-
-report_no_journal :-
-	reported_no_journal, !.
-report_no_journal :-
-	new(D, dialog('No project')),
-	send(D, append,
-	     text('No project has been created or opened.  Your modifications\n\
-	           are not saved unless you create or open a project.')),
-	send(D, append, button('Ok, all data will be lost',
-			       message(D, return, ok))),
-	send(D, append, button('Remind me next time',
-			       message(D, return, remind))),
-	get(D, confirm_centered, RVal),
-	send(D, destroy),
-	(   RVal == ok
-	->  assert(reported_no_journal)
-	;   true
-	).
-	     
-
-%	rdfe_replay_journal(+File)
+%%	rdfe_replay_journal(+File)
 %	
 %	Replay a journal file. For now  this   is  our cheap way to deal
 %	with save/load. Future versions may be  more clever when dealing
@@ -1004,7 +948,7 @@ replay_actions([H|T]) :-
 	).
 
 
-%	replay_action(+Action)
+%%	replay_action(+Action)
 %	
 %	Replay actions from the journal. Tricky is rdf_load/3. It should
 %	reload the file in the state it  was   in  at  the moment it was
@@ -1059,7 +1003,7 @@ find_file(File, Options, Path) :-
 	make_path(File, PWD, Path),
 	exists_file(Path), !.
 
-%	make_path(+File, +PWD, -Path)
+%%	make_path(+File, +PWD, -Path)
 %	
 %	Return location of File relative to PWD, Parent of PWD, etc. (TBD)
 
