@@ -1,11 +1,11 @@
-/*  $Id: rdfs.pl,v 1.15 2004/09/21 14:22:24 jan Exp $
+/*  $Id: rdfs.pl,v 1.25 2006/08/31 09:44:01 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2005, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -42,6 +42,7 @@
 
 	    rdfs_member/2,		% ?Object, +Set
 	    rdfs_list_to_prolog_list/2,	% +Set, -List
+	    rdfs_assert_list/3,		% +List, -Resource, +DB
 	    rdfs_assert_list/2,		% +List, -Resource
 
 	    rdfs_find/5			% +String, +Dom, +Props, +Method, -Subj
@@ -56,41 +57,20 @@
 		 *	    EXPANSION		*
 		 *******************************/
 
-%	user:goal_expansion(+NSGoal, -Goal)
-%	
-%	This predicate allows for writing down rdf queries in a friendly
-%	name-space fashion.  
-
-:- multifile
-	user:goal_expansion/2.
-
-user:goal_expansion(rdfs_subproperty_of(Sub0, Prop0),
-		    rdfs_subproperty_of(Sub, Prop)) :-
-	rdf_global_id(Sub0, Sub),
-	rdf_global_id(Prop0, Prop).
-user:goal_expansion(rdfs_subclass_of(Sub0, Class0),
-		    rdfs_subclass_of(Sub, Class)) :-
-	rdf_global_id(Sub0, Sub),
-	rdf_global_id(Class0, Class).
-user:goal_expansion(rdfs_class_property(Class0, Prop0),
-		    rdfs_class_property(Class, Prop)) :-
-	rdf_global_id(Class0, Class),
-	rdf_global_id(Prop0, Prop).
-user:goal_expansion(rdfs_individual_of(Resource0, Class0),
-		    rdfs_individual_of(Resource, Class)) :-
-	rdf_global_id(Resource0, Resource),
-	rdf_global_id(Class0, Class).
-user:goal_expansion(rdfs_label(Resource0, Label),
-		    rdfs_label(Resource, Label)) :-
-	rdf_global_id(Resource0, Resource).
+:- rdf_meta
+	rdfs_subproperty_of(r,r),
+	rdfs_subclass_of(r,r),
+	rdfs_class_property(r,r),
+	rdfs_individual_of(r,r),
+	rdfs_label(r,-).
 
 
 		 /*******************************
 		 *	PROPERTY HIERARCHY	*
 		 *******************************/
 
-%	rdfs_subproperty_of(+SubProperty, ?Property)
-%	rdfs_subproperty_of(?SubProperty, +Property)
+%%	rdfs_subproperty_of(+SubProperty, ?Property) is nondet.
+%%	rdfs_subproperty_of(?SubProperty, +Property) is nondet.
 %	
 %	Query the property hierarchy.
 
@@ -102,19 +82,22 @@ rdfs_subproperty_of(SubProperty, Property) :-
 		 *	  CLASS HIERARCHY	*
 		 *******************************/
 
-%	rdfs_subclass_of(+Class, ?Super)
-%	rdfs_subclass_of(?Class, +Super)
+%%	rdfs_subclass_of(+Class, ?Super) is nondet.
+%%	rdfs_subclass_of(?Class, +Super) is nondet.
 %	
 %	Generate  sub/super  classes.  rdf_reachable/3    considers  the
 %	rdfs:subPropertyOf relation as well  as   cycles.  Note  that by
 %	definition all classes are  subclass   of  rdfs:Resource, a case
 %	which is dealt with by the 1st and 3th clauses :-(
+%	
+%	According to production 2.4 "rdfs:Datatype", Each instance of
+%	rdfs:Datatype is a subclass of rdfs:Literal.  
 
 rdfs_subclass_of(Class, Super) :-
 	rdf_equal(rdfs:'Resource', Resource),
 	Super == Resource, !,
 	(   nonvar(Class)
-	->  true
+	->  true			% must check for being a class?
 	;   rdfs_individual_of(Class, rdfs:'Class')
 	).
 rdfs_subclass_of(Class, Super) :-
@@ -125,20 +108,30 @@ rdfs_subclass_of(Class, Super) :-
 	\+ rdf_reachable(Class, rdfs:subClassOf, rdfs:'Resource'),
 	rdfs_individual_of(Class, rdfs:'Class'),
 	rdf_equal(Super, rdfs:'Resource').
+rdfs_subclass_of(Class, Super) :-	% production 2.4
+	(   nonvar(Class)
+	->  rdf_has(Class, rdf:type, CType),
+	    rdf_reachable(CType, rdfs:subClassOf, rdfs:'Datatype'),
+	    \+ rdf_reachable(Class, rdfs:subClassOf, rdfs:'Literal'),
+	    rdf_equal(Super, rdfs:'Literal')
+	;   nonvar(Super)
+	->  rdf_reachable(Super, rdfs:subClassOf, rdfs:'Literal'),
+	    rdfs_individual_of(Class, rdfs:'Datatype')
+	).
 
 
 		 /*******************************
 		 *	    INDIVIDUALS		*
 		 *******************************/
 
-%	rdfs_individual(+Resource, -Class)
-%	rdfs_individual(-Resource, +Class)
+%%	rdfs_individual_of(+Resource, -Class) is nondet.
+%%	rdfs_individual_of(-Resource, +Class) is nondet.
 %	
 %	Generate resources belonging to a class   or  classes a resource
 %	belongs to. We assume everything at the `object' end of a triple
 %	is a class. A validator should confirm this property.
 %	
-%	rdfs_individual(+,  -)  does  not  exploit    domain  and  range
+%%	rdfs_individual_of(+, -) does  not  exploit   domain  and  range
 %	properties, deriving that if rdf(R,  P,   _)  is  present R must
 %	satisfy the domain of P (and similar for range).
 %	
@@ -165,7 +158,8 @@ rdfs_individual_of(Resource, Class) :-
 rdfs_individual_of(_Resource, _Class) :-
 	throw(error(instantiation_error, _)).
 
-%	rdfs_label(?Resource, ?Label)
+%%	rdfs_label(+Resource, -Label).
+%%	rdfs_label(-Resource, +Label).
 %
 %	Convert between class and label.  If the label is generated from
 %	the resource the it uses both rdf:label and its sub-properties,
@@ -174,7 +168,7 @@ rdfs_individual_of(_Resource, _Class) :-
 rdfs_label(Resource, Label) :-
 	rdfs_label(Resource, _, Label).
 
-%	rdfs_label(?Resource, ?Lang, ?Label)
+%%	rdfs_label(?Resource, ?Lang, ?Label)
 %	
 %	Convert between class and label, in a specific language.
 
@@ -184,7 +178,7 @@ rdfs_label(Resource, Lang, Label) :-
 rdfs_label(Resource, Lang, Label) :-
 	rdf_has(Resource, rdfs:label, literal(lang(Lang, Label))).
 
-%	rdfs_ns_label(+Resource, -Label)
+%%	rdfs_ns_label(+Resource, -Label)
 %	
 %	Present label with  namespace  indication.   This  predicate  is
 %	indented  to  provide  meaningful  short   names  applicable  to
@@ -212,7 +206,7 @@ rdfs_ns_label(Resource, Lang, Label) :-
 	).
 
 
-%	take_label(+Resource, ?Lang, -Label)
+%%	take_label(+Resource, ?Lang, -Label)
 %
 %	Get the label to use for a resource in the give Language
 
@@ -229,7 +223,7 @@ label_of(Resource, Lang, Label) :-
 	rdf_has(Resource, LabelP, literal(lang(Lang, Label)), P),
 	P \== LabelP.
 
-%	rdfs_class_property(+Class, ?Property)
+%%	rdfs_class_property(+Class, ?Property)
 %
 %	Enumerate the properties in the domain of Class.
 
@@ -243,20 +237,20 @@ rdfs_class_property(Class, Property) :-
 		 *	     COLLECTIONS	*
 		 *******************************/
 
-%	rdfs_member(?Element, +Set)
+%%	rdfs_member(?Element, +Set)
 %	
 %	As Prolog member on sets.  Operates both on attributes parsed as
 %	parseType="Collection" as well as on Bag, Set and Alt. 
 
 rdfs_member(Element, Set) :-
-	rdfs_individual_of(Set, rdf:'List'), !,
+	rdf_has(Set, rdf:first, _),
 	rdfs_collection_member(Element, Set).
 rdfs_member(Element, Set) :-
 	rdfs_individual_of(Set, rdf:'Container'), !,
 	(   nonvar(Element)
 	->  rdf(Set, Predicate, Element),
 	    rdf_member_property(Predicate, _N)
-	;   between(1, 1000000000, N),
+	;   between(1, inifinite, N),
 	    rdf_member_property(Prop, N),
 	    (	rdf(Set, Prop, Member)
 	    ->	Member = Element
@@ -265,16 +259,13 @@ rdfs_member(Element, Set) :-
 	).
 
 rdfs_collection_member(Element, Set) :-
-	rdf_has(Set, rdf:first, Element),
-	assume(rdfs_individual_of(Set, rdf:'List')).
+	rdf_has(Set, rdf:first, Element).
 rdfs_collection_member(Element, Set) :-
 	rdf_has(Set, rdf:rest, Tail), !,
 	rdfs_collection_member(Element, Tail).
-rdfs_collection_member(_, Set) :-
-	assume(rdf_equal(Set, rdf:nil)),
-	fail.
 
-%	rdfs_list_to_prolog_list(+RDFSList, -PrologList)
+
+%%	rdfs_list_to_prolog_list(+RDFSList, -PrologList)
 %	
 %	Convert ann RDFS list (result from parseType=Collection) into a
 %	Prolog list of elements.
@@ -287,33 +278,37 @@ rdfs_list_to_prolog_list(Set, [H|T]) :-
 	rdfs_list_to_prolog_list(Tail, T).
 
 
-%	rdfs_assert_list(+Resources, -List)
+%%	rdfs_assert_list(+Resources, -List) is det.
+%%	rdfs_assert_list(+Resources, -List, +DB) is det.
 %	
 %	Create an RDF list from the given Resources.
 
-rdfs_assert_list([], Nil) :-
+rdfs_assert_list(Resources, List) :-
+	rdfs_assert_list(Resources, List, user).
+
+rdfs_assert_list([], Nil, _) :-
 	rdf_equal(rdf:nil, Nil).
-rdfs_assert_list([H|T], List) :-
-	rdfs_assert_list(T, Tail),
+rdfs_assert_list([H|T], List, DB) :-
+	rdfs_assert_list(T, Tail, DB),
 	rdf_bnode(List),
-	rdf_assert(List, rdf:rest, Tail),
-	rdf_assert(List, rdf:first, H),
-	rdf_assert(List, rdf:type, rdf:'List').
+	rdf_assert(List, rdf:rest, Tail, DB),
+	rdf_assert(List, rdf:first, H, DB),
+	rdf_assert(List, rdf:type, rdf:'List', DB).
 
 
 		 /*******************************
 		 *     SEARCH IN HIERARCHY	*
 		 *******************************/
 
-%	rdfs_find(+String, +Domain, +Properties, +Method, -Subject)
+%%	rdfs_find(+String, +Domain, ?Properties, +Method, -Subject)
 %	
 %	Search all classes below Domain for a literal property with
 %	that matches String.  Method is one of
 %	
-%		substring
-%		word
-%		prefix
-%		exact
+%		* substring
+%		* word
+%		* prefix
+%		* exact
 %		
 %	domain is defined by owl_satisfy from owl.pl
 %		
@@ -321,6 +316,12 @@ rdfs_assert_list([H|T], List) :-
 %	making the URI-ref fragment name the last resort to determine
 %	the label.
 
+rdfs_find(String, Domain, Fields, Method, Subject) :-
+	var(Fields), !,
+	For =.. [Method,String],
+	rdf_has(Subject, Field, literal(For, _)),
+	owl_satisfies(Domain, Subject),
+	Fields = [Field].		% report where we found it.
 rdfs_find(String, Domain, Fields, Method, Subject) :-
 	globalise_list(Fields, GlobalFields),
 	For =.. [Method,String],
@@ -386,7 +387,7 @@ rdfs_find(String, Domain, Fields, Method, Subject) :-
 	),
 	rdf_match_label(Method, String, Arg).
 	
-%	generate_domain(+Domain, -Resource)
+%%	generate_domain(+Domain, -Resource)
 %	
 %	Generate all resources that satisfy some a domain specification.
 
