@@ -1,11 +1,11 @@
-/*  $Id: rdf_parser.pl,v 1.31 2004/09/21 14:22:24 jan Exp $
+/*  $Id: rdf_parser.pl,v 1.41 2006/08/31 09:44:01 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2004, University of Amsterdam
+    Copyright (C): 2002-2006, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -35,6 +35,10 @@
 	    rdf_name_space/1
 	  ]).
 :- use_module(rewrite).
+:- use_module(library(sgml)).		% xml_name/1
+:- use_module(library(lists)).
+:- use_module(library(url)).
+:- use_module(library(utf8)).
 
 :- op(500, fx, \?).			% Optional (attrs)
 
@@ -44,10 +48,17 @@ goal_expansion(F, T) :- rew_goal_expansion(F, T).
 :- multifile rdf_name_space/1.
 :- dynamic   rdf_name_space/1.
 
+%%	rdf_name_space(?URL) is nondet.
+%
+%	True if URL must be handled  as rdf: Determines special handling
+%	of rdf:about, rdf:resource, etc.
+
+
 rdf_name_space('http://www.w3.org/1999/02/22-rdf-syntax-ns#').
 rdf_name_space('http://www.w3.org/TR/REC-rdf-syntax').
 
-%	xml_to_rdf(+RDFElementOrObject, -RDFTerm, +Options)
+
+%%	xml_to_plrdf(+RDFElementOrObject, -RDFTerm, +Options)
 %
 %	Translate an XML (using namespaces) term into an Prolog term
 %	representing the RDF data.  This term can then be fed into
@@ -62,6 +73,10 @@ xml_to_plrdf(Element, RDF, Options) :-
 xml_to_plrdf(Element, RDF, Options) :-
 	rewrite(\xml_objects(RDF, Options), Element).
 
+%%	element_to_plrdf(+DOM, -RDFTerm, +Options)
+%
+%	Rewrite a single XML element.
+
 element_to_plrdf(Element, RDF, Options) :-
 	rewrite(\nodeElementList(RDF, Options), [Element]).
 
@@ -103,7 +118,7 @@ nodeElementOrError(unparsed(Data), _Options) ::=
 	Data.
 
 nodeElement(container(Type, Id, Elements), Options) ::=
-	\container(Type, Id, Elements, Options), !. % compatibility
+	\container(Type, Id, Elements, Options), !. 	% compatibility
 nodeElement(description(Type, About, BagID, Properties), Options) ::=
 	\description(Type, About, BagID, Properties, Options).
 
@@ -255,7 +270,7 @@ aboutResourceEmptyElt(about(URI), Options) ::=
 aboutResourceEmptyElt(node(URI), _Options) ::=
 	\nodeIDAttr(URI).
 
-%	literal_value(+In, -Value, +Options)
+%%	literal_value(+In, -Value, +Options)
 %	
 %	Translate a literal into its  value.   Notably  if  the value is
 %	plain  CDATA,  remove  the   list.    This   predicate   handles
@@ -267,10 +282,10 @@ literal_value([Value], Literal, Options) :-
 literal_value(Value, literal(Value), _).
 
 
-%	mkliteral(+Atom, -Object, +Options)
+%%	mkliteral(+Atom, -Object, +Options)
 %	
 %	Translate attribute value Atom into an RDF object using the
-%	lang(Lang) option from Options.
+%%	lang(Lang) option from Options.
 
 mkliteral(Text, literal(Val), Options) :-
 	atom(Text),
@@ -280,9 +295,9 @@ mkliteral(Text, literal(Val), Options) :-
 	;   Val = Text
 	).
 
-%	typed_literal(+Type, +Content, -Literal, +Options)
+%%	typed_literal(+Type, +Content, -Literal, +Options)
 %	
-%	Handle a literal attribute with rdf:dataType=Type qualifier. NB:
+%	Handle a literal attribute with rdf:datatype=Type qualifier. NB:
 %	possibly  it  is  faster  to  use  a  global  variable  for  the
 %	conversion hook.
 
@@ -316,7 +331,7 @@ idRefAttr(Id, Options) ::=
 idRefAttr(about(URI), Options) ::=
 	\resourceAttr(URI, Options).
 
-%	an_rdf_object(-Object, +OptionsURI)
+%%	an_rdf_object(-Object, +OptionsURI)
 %
 %	Deals with an object, but there may be spaces around.  I'm still
 %	not sure where to deal with these.  Best is to ask the XML parser
@@ -382,7 +397,7 @@ resourceAttr(URI, Options) ::=
 	\rdf_or_unqualified(resource) = \uri(URI, Options).
 
 typeAttr(Type, Options) ::=
-	\rdf_or_unqualified(dataType) = \uri(Type, Options).
+	\rdf_or_unqualified(datatype) = \uri(Type, Options).
 
 uri(URI, Options) ::=
 	A,
@@ -391,7 +406,7 @@ uri(URI, Options) ::=
 	->  canonical_uri(A, Base, URI)
 	;   sub_atom(A, 0, _, _, #)
 	->  sub_atom(A, 1, _, 0, URI)
-	;   URI = A
+	;   decode_uri(A, URI)
 	}.
 
 globalid(Id, Options) ::=
@@ -415,21 +430,53 @@ make_globalid(In, Options, Id) :-
 	(   memberchk(base_uri(Base), Options),
 	    Base \== []
 	->  (   is_absolute_url(In)
-	    ->	Id = In
-	    ;	concat_atom([Base, In], #, Id)
+	    ->	decode_uri(In, Id)
+	    ;	concat_atom([Base, In], #, Id0),
+		decode_uri(Id0, Id)
 	    )
 	;   sub_atom(In, 0, _, _, #)
 	->  sub_atom(In, 1, _, 0, Id)
-	;   Id = In
+	;   decode_uri(In, Id)
 	).
 
 
-%	canonical_uri(+In, +Base, -Absolute)
+%%	canonical_uri(+In, +Base, -Absolute)
+%	
+%	Make the URI absolute and decode special sequences. For the last
+%	clause, which is the correct order?
 
 canonical_uri('', Base, Base) :- !.	% '' expands to xml:base
-canonical_uri(URI, [], URI) :- !.	% do not use one
+canonical_uri(URI0, [], URI) :- !,	% do not use one
+	decode_uri(URI0, URI).
 canonical_uri(URI, Base, Global) :-	% use our generic library
-	global_url(URI, Base, Global).
+	global_url(URI, Base, Global0),
+	decode_uri(Global0, Global).
+
+
+%%	decode_uri(+Encoded, -Decoded)
+%	
+%	RDF URIs are encoded  Unicode  strings.   First  the  unicode is
+%	translated using UTF-8 and then  the   result  is represented as
+%	US-ASCII using %XX sequences. Here we have to do the reverse.
+%	
+%	NOTE: this encoding is known under the name IRI.
+
+decode_uri(Encoded, Decoded) :-
+	sub_atom(Encoded, _, _, _, '%'), !,
+	atom_codes(Encoded, Codes),
+	unescape_precent(Codes, UTF8),
+	phrase(utf8_codes(Unicodes), UTF8),
+	atom_codes(Decoded, Unicodes).
+decode_uri(Encoded, Encoded).
+
+unescape_precent([], []).
+unescape_precent([0'%,C1,C2|T0], [H|T]) :- !,
+	code_type(C1, xdigit(D1)),
+	code_type(C2, xdigit(D2)),
+	H is D1*16 + D2,
+	unescape_precent(T0, T).
+unescape_precent([H|T0], [H|T]) :-
+	unescape_precent(T0, T).
 
 
 		 /*******************************
@@ -587,15 +634,31 @@ noMoreAttrs ::=
 	| \noMoreAttrs
 	].
 
-%	modify_state(+Element0, +Options0, -Element, -Options)
+%%	modify_state(+Element0, +Options0, -Element, -Options)
 %	
 %	If Element0 contains xml:base = Base, strip it from the
 %	attributes list and update base_uri(_) in the Options
 %	
 %	It Element0 contains xml:lang = Lang, strip it from the
 %	attributes list and update lang(_) in the Options
+%	
+%	Remove all xmlns=_, xmlns:_=_ and xml:_=_.  Only succeed
+%	if something changed.
 
-modify_state(element(Name, Attrs0, Content), Options0,
+modify_state(E0, O0, E, O) :-
+	modify_states([base, lang, xmlns], M, E0, O0, E, O),
+	M \== [].
+
+modify_states([], [], E, O, E, O).
+modify_states([How|TH0], [How|TH], E0, O0, E, O) :-
+	modify_state(How, E0, O0, E1, O1), !,
+	modify_states(TH0, TH, E1, O1, E, O).
+modify_states([_|TH0], TH, E0, O0, E, O) :-
+	modify_states(TH0, TH, E0, O0, E, O).
+
+
+modify_state(base,
+	     element(Name, Attrs0, Content), Options0,
 	     element(Name, Attrs, Content),  Options) :-
 	select(xml:base=Base1, Attrs0, Attrs), !,
 	(   select(base_uri(Base0), Options0, Options1)
@@ -606,7 +669,7 @@ modify_state(element(Name, Attrs0, Content), Options0,
 	remove_fragment(Base1, Base2),
 	canonical_uri(Base2, Base0, Base),
 	Options = [base_uri(Base)|Options1].
-modify_state(element(Name, Attrs0, Content), Options0,
+modify_state(lang, element(Name, Attrs0, Content), Options0,
 	     element(Name, Attrs, Content),  Options) :-
 	select(xml:lang=Lang, Attrs0, Attrs),
 	\+ memberchk(ignore_lang(true), Options0), !,
@@ -615,8 +678,25 @@ modify_state(element(Name, Attrs0, Content), Options0,
 	->  Options = Options1
 	;   Options = [lang(Lang)|Options1]
 	).
+modify_state(xmlns,
+	     element(Name, Attrs0, Content), Options,
+	     element(Name, Attrs, Content), Options) :-
+	clean_xmlns_attr(Attrs0, Attrs),
+	Attrs \== Attrs0.
 
-%	remove_fragment(+URI, -WithoutFragment)
+clean_xmlns_attr([], []).
+clean_xmlns_attr([H=_|T0], T) :-
+	xml_attr(H), !,
+	clean_xmlns_attr(T0, T).
+clean_xmlns_attr([H|T0], [H|T]) :-
+	clean_xmlns_attr(T0, T).
+
+xml_attr(xmlns).
+xml_attr(xmlns:_).
+xml_attr(xml:_).
+
+
+%%	remove_fragment(+URI, -WithoutFragment)
 %	
 %	When handling xml:base, we must delete the possible fragment.
 
