@@ -1,4 +1,4 @@
-/*  $Id: sgml2pl.c,v 1.62 2004/09/21 14:22:24 jan Exp $
+/*  $Id: sgml2pl.c,v 1.75 2006/02/01 11:20:06 jan Exp $
 
     Part of SWI-Prolog
 
@@ -82,6 +82,7 @@ typedef struct _parser_data
   int	      max_warnings;		/* warning limit */
   errormode   error_mode;		/* how to handle errors */
   int	      positions;		/* report file-positions */
+  int         exception;		/* pending exception from callback */
 
   predicate_t on_begin;			/* begin element */
   predicate_t on_end;			/* end element */
@@ -352,7 +353,7 @@ pl_new_dtd(term_t doctype, term_t ref)
   if ( !PL_get_atom_chars(doctype, &dt) )
     return sgml2pl_error(ERR_TYPE, "atom", doctype);
 
-  if ( !(dtd=new_dtd(dt)) )
+  if ( !(dtd=new_dtd((ichar*)dt)) )
     return FALSE;
 
   dtd->references++;
@@ -507,7 +508,7 @@ pl_set_sgml_parser(term_t parser, term_t option)
     { if ( !PL_get_atom_chars(a, &s) )
 	return sgml2pl_error(ERR_TYPE, "atom_or_variable", a);
     
-      p->enforce_outer_element = dtd_add_symbol(p->dtd, s);
+      p->enforce_outer_element = dtd_add_symbol(p->dtd, (ichar*)s);
     }
   } else
     return sgml2pl_error(ERR_DOMAIN, "sgml_parser_option", option);
@@ -607,7 +608,7 @@ pl_get_sgml_parser(term_t parser, term_t option)
 
     PL_get_arg(1, option, a);
     if ( p->enforce_outer_element )
-      return PL_unify_atom_chars(a, p->enforce_outer_element->name);
+      return PL_unify_atom_chars(a, (char*)p->enforce_outer_element->name);
     else
       return TRUE;			/* leave variable */
   } else if ( PL_is_functor(option, FUNCTOR_allowed1) )
@@ -630,7 +631,7 @@ pl_get_sgml_parser(term_t parser, term_t option)
 	{ if ( buf[i] == CDATA_ELEMENT )
 	    PL_put_atom_chars(tmp, "#pcdata");
 	  else
-	    PL_put_atom_chars(tmp, buf[i]->name->name);
+	    PL_put_atom_chars(tmp, (char*)buf[i]->name->name);
   
 	  if ( !PL_unify_list(tail, head, tail) ||
 	       !PL_unify(head, tmp) )
@@ -642,7 +643,7 @@ pl_get_sgml_parser(term_t parser, term_t option)
 	  break;
       }
     } else if ( p->enforce_outer_element )
-    { PL_put_atom_chars(tmp, p->enforce_outer_element->name);
+    { PL_put_atom_chars(tmp, (char*)p->enforce_outer_element->name);
 
       if ( !PL_unify_list(tail, head, tail) ||
 	       !PL_unify(head, tmp) )
@@ -659,7 +660,7 @@ pl_get_sgml_parser(term_t parser, term_t option)
     PL_get_arg(1, option, tail);
 
     for( ; env; env = env->parent)
-    { PL_put_atom_chars(tmp, env->element->name->name);
+    { PL_put_atom_chars(tmp, (char*)env->element->name->name);
 
       if ( !PL_unify_list(tail, head, tail) ||
 	   !PL_unify(head, tmp) )
@@ -674,6 +675,40 @@ pl_get_sgml_parser(term_t parser, term_t option)
 }
 
 
+#if 0
+static void
+writeln(term_t t)
+{ PL_write_term(Serror, t, 1200, PL_WRT_QUOTED|PL_WRT_PORTRAY);
+  Sputcode('\n', Serror);
+}
+#endif
+
+
+static int
+call_prolog(parser_data *pd, predicate_t pred, term_t av)
+{ qid_t qid = PL_open_query(NULL, PL_Q_PASS_EXCEPTION, pred, av);
+  int rc = PL_next_solution(qid);
+
+  if ( !rc && PL_exception(qid) )
+    pd->exception = TRUE;
+  else
+    pd->exception = FALSE;
+
+  PL_close_query(qid);
+
+  return rc;
+}
+
+
+static void
+end_frame(fid_t fid, int ex)
+{ if ( ex )
+    PL_close_foreign_frame(fid);
+  else
+    PL_discard_foreign_frame(fid);
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 put_url(dtd_parser *p, term_t t, const ichar *url)
     Store the url-part of a name-space qualifier in term.  We call
@@ -716,7 +751,7 @@ put_url(dtd_parser *p, term_t t, const ichar *url)
   int i;
 
   if ( !pd->on_urlns )
-  { PL_put_atom_chars(t, url);
+  { PL_put_atom_chars(t, (char*)url);
     return;
   }
 
@@ -725,7 +760,7 @@ put_url(dtd_parser *p, term_t t, const ichar *url)
     { if ( cache[i].canonical )		/* and a canonical value */
 	PL_put_atom(t, cache[i].canonical);
       else
-	PL_put_atom_chars(t, url);
+	PL_put_atom_chars(t, (char*)url);
 
       return;
     }
@@ -743,7 +778,7 @@ put_url(dtd_parser *p, term_t t, const ichar *url)
     term_t av = PL_new_term_refs(3);
     atom_t a;
 
-    PL_put_atom_chars(av+0, url);
+    PL_put_atom_chars(av+0, (char*)url);
     unify_parser(av+2, p);
     if ( PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_urlns, av) &&
 	 PL_get_atom(av+1, &a) )
@@ -751,7 +786,7 @@ put_url(dtd_parser *p, term_t t, const ichar *url)
       cache[0].canonical = a;
       PL_put_atom(t, a);
     } else
-    { PL_put_atom_chars(t, url);
+    { PL_put_atom_chars(t, (char*)url);
     }
     PL_discard_foreign_frame(fid);
   }
@@ -769,12 +804,12 @@ put_attribute_name(dtd_parser *p, term_t t, dtd_symbol *nm)
     { term_t av = PL_new_term_refs(2);
     
       put_url(p, av+0, url);
-      PL_put_atom_chars(av+1, local);
+      PL_put_atom_chars(av+1, (char*)local);
       PL_cons_functor_v(t, FUNCTOR_ns2, av);
     } else
-      PL_put_atom_chars(t, local);
+      PL_put_atom_chars(t, (char*)local);
   } else
-    PL_put_atom_chars(t, nm->name);
+    PL_put_atom_chars(t, (char*)nm->name);
 }
 
 
@@ -790,12 +825,12 @@ put_element_name(dtd_parser *p, term_t t, dtd_element *e)
     { term_t av = PL_new_term_refs(2);
     
       put_url(p, av+0, url);
-      PL_put_atom_chars(av+1, local);
+      PL_put_atom_chars(av+1, (char*)local);
       PL_cons_functor_v(t, FUNCTOR_ns2, av);
     } else
-      PL_put_atom_chars(t, local);
+      PL_put_atom_chars(t, (char*)local);
   } else
-    PL_put_atom_chars(t, e->name->name);
+    PL_put_atom_chars(t, (char*)e->name->name);
 }
 
 
@@ -815,10 +850,16 @@ unify_listval(dtd_parser *p,
 	      term_t t, attrtype type, int len, const char *text)
 { if ( type == AT_NUMBERS && p->dtd->number_mode == NU_INTEGER )
   { char *e;
+#if SIZEOF_LONG == 4 && defined(HAVE_STRTOLL)
+    int64_t v = strtoll(text, &e, 10);
+    if ( e-text == len && errno != ERANGE )
+      return PL_unify_int64(t, v);
+#else
     long v = strtol(text, &e, 10);
 
     if ( e-text == len && errno != ERANGE )
       return PL_unify_integer(t, v);
+#endif
 					/* TBD: Error!? */
   }
 
@@ -826,25 +867,36 @@ unify_listval(dtd_parser *p,
 }
 
 
+static int
+put_att_text(term_t t, sgml_attribute *a)
+{ if ( a->value.textA )
+  { PL_put_atom_chars(t, (char*)a->value.textA);
+    return TRUE;
+  } else if ( a->value.textW )
+  { PL_put_variable(t);
+    PL_unify_wchars(t, PL_ATOM, a->value.number, a->value.textW);
+    return TRUE;
+  } else
+    return FALSE;
+}
+
+
 static void
 put_attribute_value(dtd_parser *p, term_t t, sgml_attribute *a)
 { switch(a->definition->type)
   { case AT_CDATA:
-      PL_put_atom_chars(t, a->value.cdata);
+      put_att_text(t, a);
       break;
     case AT_NUMBER:
-    { if ( a->value.text )
-	PL_put_atom_chars(t, a->value.text);
-      else
+    { if ( !put_att_text(t, a) )
 	PL_put_integer(t, a->value.number);
       break;
     }
-    default:
-    { const ichar *val = a->value.text;
-      const ichar *e;
-
-      if ( a->definition->islist )	/* multi-valued attribute */
+    default:				/* multi-valued attribute */
+    { if ( a->definition->islist && a->value.textA )
       { term_t tail, head = PL_new_term_ref();
+	const ichar *val = a->value.textA;
+	const ichar *e;
 
 	PL_put_variable(t);
 	tail = PL_copy_term_ref(t);
@@ -853,13 +905,13 @@ put_attribute_value(dtd_parser *p, term_t t, sgml_attribute *a)
 	{ if ( e == val )
 	    continue;			/* skip spaces */
 	  PL_unify_list(tail, head, tail);
-	  unify_listval(p, head, a->definition->type, e-val, val);
+	  unify_listval(p, head, a->definition->type, e-val, (char*)val);
 	}
         PL_unify_list(tail, head, tail);
-	unify_listval(p, head, a->definition->type, istrlen(val), val);
+	unify_listval(p, head, a->definition->type, istrlen(val), (char*)val);
         PL_unify_nil(tail);
       } else
-	PL_put_atom_chars(t, val);
+	put_att_text(t, a);
     }
   }
 }
@@ -869,7 +921,6 @@ put_attribute_value(dtd_parser *p, term_t t, sgml_attribute *a)
 Produce a tag-location in the format
 
 	start_location=file:char-char	
-
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
@@ -972,8 +1023,8 @@ on_begin(dtd_parser *p, dtd_element *e, int argc, sgml_attribute *argv)
     unify_attribute_list(p, av+1, argc, argv);
     unify_parser(av+2, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_begin, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_begin, av);
+    end_frame(fid, pd->exception);
   }
 
   return TRUE;
@@ -994,8 +1045,8 @@ on_end(dtd_parser *p, dtd_element *e)
     put_element_name(p, av+0, e);
     unify_parser(av+1, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_end, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_end, av);
+    end_frame(fid, pd->exception);
   }
 
   if ( pd->tail && !pd->stopped )
@@ -1033,14 +1084,14 @@ on_entity(dtd_parser *p, dtd_entity *e, int chr)
     term_t av = PL_new_term_refs(2);
 
     if ( e )
-      PL_put_atom_chars(av+0, e->name->name);
+      PL_put_atom_chars(av+0, (char*)e->name->name);
     else
       PL_put_integer(av+0, chr);
 
     unify_parser(av+1, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_end, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_end, av);
+    end_frame(fid, pd->exception);
   }
 
   if ( pd->tail )
@@ -1066,41 +1117,47 @@ on_entity(dtd_parser *p, dtd_entity *e, int chr)
 
 
 static int
-on_cdata(dtd_parser *p, data_type type, int len, const ochar *data)
+on_data(dtd_parser *p, data_type type, int len,
+	const char *data, const wchar_t *wdata)
 { parser_data *pd = p->closure;
 
   if ( pd->on_cdata )
   { fid_t fid = PL_open_foreign_frame();
     term_t av = PL_new_term_refs(2);
 
-    PL_put_atom_nchars(av+0, len, data);
+    if ( data )
+      PL_put_atom_nchars(av+0, len, data);
+    else
+      PL_unify_wchars(av+0, PL_ATOM, len, wdata);
+
     unify_parser(av+1, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_cdata, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_cdata, av);
+    end_frame(fid, pd->exception);
   }
 
   if ( pd->tail && !pd->stopped )
   { term_t h = PL_new_term_ref();
 
     if ( PL_unify_list(pd->tail, h, pd->tail) )
-    { int rval;
+    { int rval = TRUE;
+      term_t a;
 
       switch(type)
       { case EC_CDATA:
-	  rval = PL_unify_atom_nchars(h, len, data);
+	  a = h;
 	  break;
 	case EC_SDATA:
 	{ term_t d = PL_new_term_ref();
 
-	  PL_put_atom_nchars(d, len, data);
+	  a = d;
 	  rval = PL_unify_term(h, PL_FUNCTOR, FUNCTOR_sdata1, PL_TERM, d);
 	  break;
 	}
 	case EC_NDATA:
 	{ term_t d = PL_new_term_ref();
 
-	  PL_put_atom_nchars(d, len, data);
+	  a = d;
 	  rval = PL_unify_term(h, PL_FUNCTOR, FUNCTOR_ndata1, PL_TERM, d);
 	  break;
 	}
@@ -1109,6 +1166,13 @@ on_cdata(dtd_parser *p, data_type type, int len, const ochar *data)
 	  assert(0);
       }
 			       
+      if ( rval )
+      { if ( data )
+	  rval = PL_unify_atom_nchars(a, len, data);
+	else
+	  rval = PL_unify_wchars(a, PL_ATOM, len, wdata);
+      }
+
       if ( rval )
       { PL_reset_term_refs(h);
 	return TRUE;
@@ -1120,6 +1184,18 @@ on_cdata(dtd_parser *p, data_type type, int len, const ochar *data)
 }
 
 
+static int
+on_cdata(dtd_parser *p, data_type type, int len, const char *data)
+{ return on_data(p, type, len, data, NULL);
+}
+
+
+static int
+on_cwdata(dtd_parser *p, data_type type, int len, const wchar_t *data)
+{ return on_data(p, type, len, NULL, data);
+}
+
+
 static int
 can_end_omitted(dtd_parser *p)
 { sgml_environment *env;
@@ -1177,8 +1253,8 @@ on_error(dtd_parser *p, dtd_error *error)
     PL_put_atom_chars(av+1, error->plain_message);
     unify_parser(av+2, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_error, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_error, av);
+    end_frame(fid, pd->exception);
   } else if ( pd->error_mode != EM_QUIET )
   { fid_t fid = PL_open_foreign_frame();
     predicate_t pred = PL_predicate("print_message", 2, "user");
@@ -1217,14 +1293,14 @@ on_xmlns(dtd_parser *p, dtd_symbol *ns, dtd_symbol *url)
     term_t av = PL_new_term_refs(3);
 
     if ( ns )
-      PL_put_atom_chars(av+0, ns->name);
+      PL_put_atom_chars(av+0, (char*)ns->name);
     else
       PL_put_nil(av+0);
-    PL_put_atom_chars(av+1, url->name);
+    PL_put_atom_chars(av+1, (char*)url->name);
     unify_parser(av+2, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_xmlns, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_xmlns, av);
+    end_frame(fid, pd->exception);
   }
 
   return TRUE;
@@ -1242,11 +1318,11 @@ on_pi(dtd_parser *p, const ichar *pi)
   { fid_t fid = PL_open_foreign_frame();
     term_t av = PL_new_term_refs(2);
 
-    PL_put_atom_chars(av+0, pi);
+    PL_put_atom_chars(av+0, (char*)pi);
     unify_parser(av+1, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_pi, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_pi, av);
+    end_frame(fid, pd->exception);
   }
 
   if ( pd->tail )
@@ -1277,11 +1353,11 @@ on_decl(dtd_parser *p, const ichar *decl)
   { fid_t fid = PL_open_foreign_frame();
     term_t av = PL_new_term_refs(2);
 
-    PL_put_atom_chars(av+0, decl);
+    PL_put_atom_chars(av+0, (char*)decl);
     unify_parser(av+1, p);
 
-    PL_call_predicate(NULL, PL_Q_NORMAL, pd->on_decl, av);
-    PL_discard_foreign_frame(fid);
+    call_prolog(pd, pd->on_decl, av);
+    end_frame(fid, pd->exception);
   }
 
   if ( pd->stopat == SA_DECL )
@@ -1360,6 +1436,7 @@ new_parser_data(dtd_parser *p)
   pd->max_errors = MAX_ERRORS;
   pd->max_warnings = MAX_WARNINGS;
   pd->error_mode = EM_PRINT;
+  pd->exception = FALSE;
   p->closure = pd;
   
   return pd;
@@ -1372,6 +1449,8 @@ pl_open_dtd(term_t ref, term_t options, term_t stream)
   dtd_parser *p;
   parser_data *pd;
   IOSTREAM *s;
+  term_t tail = PL_copy_term_ref(options);
+  term_t option = PL_new_term_ref();
 
   if ( !get_dtd(ref, &dtd) )
     return FALSE;
@@ -1380,6 +1459,29 @@ pl_open_dtd(term_t ref, term_t options, term_t stream)
   pd = new_parser_data(p);
   pd->free_on_close = TRUE;
 
+  while( PL_get_list(tail, option, tail) )
+  { if ( PL_is_functor(option, FUNCTOR_dialect1) )
+    { term_t a = PL_new_term_ref();
+      char *s;
+  
+      PL_get_arg(1, option, a);
+      if ( !PL_get_atom_chars(a, &s) )
+	return sgml2pl_error(ERR_TYPE, "atom", a);
+  
+      if ( streq(s, "xml") )
+	set_dialect_dtd(dtd, DL_XML);
+      else if ( streq(s, "xmlns") )
+	set_dialect_dtd(dtd, DL_XMLNS);
+      else if ( streq(s, "sgml") )
+	set_dialect_dtd(dtd, DL_SGML);
+      else
+	return sgml2pl_error(ERR_DOMAIN, "sgml_dialect", a);
+    } else
+      return sgml2pl_error(ERR_DOMAIN, "dtd_option", option);
+  }
+  if ( !PL_get_nil(tail) )
+    return sgml2pl_error(ERR_TYPE, "list", options);
+
   s = Snew(pd, SIO_OUTPUT, &sgml_stream_functions);
 
   if ( !PL_open_stream(stream, s) )
@@ -1480,6 +1582,7 @@ pl_sgml_parse(term_t parser, term_t options)
     p->on_entity	= on_entity;
     p->on_pi		= on_pi;
     p->on_data          = on_cdata;
+    p->on_wdata         = on_cwdata;
     p->on_error	        = on_error;
     p->on_xmlns		= on_xmlns;
     p->on_decl		= on_decl;
@@ -1572,8 +1675,12 @@ pl_sgml_parse(term_t parser, term_t options)
 
 					/* Parsing input from a stream */
 #define CHECKERROR \
+    { if ( pd->exception ) \
+      { return FALSE; \
+      } \
       if ( pd->errors > pd->max_errors && pd->max_errors >= 0 ) \
-	return sgml2pl_error(ERR_LIMIT, "max_errors", (long)pd->max_errors);
+	return sgml2pl_error(ERR_LIMIT, "max_errors", (long)pd->max_errors); \
+    }
 
   if ( pd->stopat == SA_CONTENT && p->empty_element )
     goto out;
@@ -1581,6 +1688,11 @@ pl_sgml_parse(term_t parser, term_t options)
   if ( in )
   { int eof = FALSE;
 
+    if ( in->encoding == ENC_OCTET )
+      p->encoded = TRUE;		/* parser must decode */
+    else
+      p->encoded = FALSE;		/* already decoded */
+
     if ( !recursive )
     { pd->source = in;
       begin_document_dtd_parser(p);
@@ -1593,10 +1705,10 @@ pl_sgml_parse(term_t parser, term_t options)
       { if ( content_length <= 0 )
 	  c = EOF;
 	else
-	  c = Sgetc(in);
+	  c = Sgetcode(in);
 	ateof = (--content_length <= 0);
       } else
-      { c = Sgetc(in);
+      { c = Sgetcode(in);
 	ateof = Sfeof(in);
       }
 
@@ -1662,7 +1774,7 @@ static void	put_model(term_t t, dtd_model *m);
 static int
 dtd_prop_doctype(dtd *dtd, term_t prop)
 { if ( dtd->doctype )
-    return PL_unify_atom_chars(prop, dtd->doctype);
+    return PL_unify_atom_chars(prop, (char*)dtd->doctype);
   return FALSE;
 }
 
@@ -1693,7 +1805,7 @@ put_model(term_t t, dtd_model *m)
       PL_put_atom(t, ATOM_pcdata);
       goto card;
     case MT_ELEMENT:
-      PL_put_atom_chars(t, m->content.element->name->name);
+      PL_put_atom_chars(t, (char*)m->content.element->name->name);
       goto card;
     case MT_AND:
       f = FUNCTOR_and2;
@@ -1763,7 +1875,7 @@ dtd_prop_elements(dtd *dtd, term_t prop)
   dtd_element *e;
   
   for( e=dtd->elements; e; e=e->next )
-  { PL_put_atom_chars(et, e->name->name);
+  { PL_put_atom_chars(et, (char*)e->name->name);
     if ( !PL_unify_list(tail, head, tail) ||
 	 !PL_unify(head, et) )
       return FALSE;
@@ -1782,7 +1894,7 @@ get_element(dtd *dtd, term_t name, dtd_element **elem)
   if ( !PL_get_atom_chars(name, &s) )
     return sgml2pl_error(ERR_TYPE, "atom", name);
 
-  if ( !(id=dtd_find_symbol(dtd, s)) ||
+  if ( !(id=dtd_find_symbol(dtd, (ichar*)s)) ||
        !(e=id->element) )
     return FALSE;
 
@@ -1824,7 +1936,7 @@ dtd_prop_attributes(dtd *dtd, term_t ename, term_t atts)
     return FALSE;
   
   for(al=e->attributes; al; al=al->next)
-  { PL_put_atom_chars(elem, al->attribute->name->name);
+  { PL_put_atom_chars(elem, (char*)al->attribute->name->name);
 
     if ( !PL_unify_list(tail, head, tail) ||
 	 !PL_unify(head, elem) )
@@ -1896,7 +2008,7 @@ unify_attribute_type(term_t type, dtd_attr *a)
     PL_get_arg(1, type, tail);
 
     for(nl = a->typeex.nameof; nl; nl = nl->next)
-    { PL_put_atom_chars(elem, nl->value->name);
+    { PL_put_atom_chars(elem, (char*)nl->value->name);
       
       if ( !PL_unify_list(tail, head, tail) ||
 	   !PL_unify(head, elem) )
@@ -1936,12 +2048,12 @@ unify_attribute_default(term_t defval, dtd_attr *a)
 	PL_get_arg(1, defval, tmp);
 	switch( a->type )
 	{ case AT_CDATA:
-	    return PL_unify_atom_chars(tmp, a->att_def.cdata);
+	    return PL_unify_atom_chars(tmp, (char*)a->att_def.cdata);
 	  case AT_NAME:
 	  case AT_NMTOKEN:
 	  case AT_NAMEOF:
 	  case AT_NOTATION:
-	    return PL_unify_atom_chars(tmp, a->att_def.name->name);
+	    return PL_unify_atom_chars(tmp, (char*)a->att_def.name->name);
 	  case AT_NUMBER:
 	    return PL_unify_integer(tmp, a->att_def.number);
 	  default:
@@ -1969,7 +2081,7 @@ dtd_prop_attribute(dtd *dtd, term_t ename, term_t aname,
     return FALSE;
   if ( !PL_get_atom_chars(aname, &achars) )
     return sgml2pl_error(ERR_TYPE, "atom", aname);
-  if ( !(asym=dtd_find_symbol(dtd, achars)) )
+  if ( !(asym=dtd_find_symbol(dtd, (ichar*)achars)) )
     return FALSE;
 
   for(al=e->attributes; al; al=al->next)
@@ -1994,15 +2106,13 @@ dtd_prop_entities(dtd *dtd, term_t list)
   dtd_entity *e;
   
   for( e=dtd->entities; e; e=e->next )
-  { PL_put_atom_chars(et, e->name->name);
+  { PL_put_atom_chars(et, (char*)e->name->name);
     if ( !PL_unify_list(tail, head, tail) ||
 	 !PL_unify(head, et) )
       return FALSE;
   }
 
   return PL_unify_nil(tail);
-
-  return FALSE;
 }
 
 
@@ -2015,7 +2125,7 @@ dtd_prop_entity(dtd *dtd, term_t ename, term_t value)
   if ( !PL_get_atom_chars(ename, &s) )
     return sgml2pl_error(ERR_TYPE, "atom", ename);
 
-  if ( !(id=dtd_find_symbol(dtd, s)) ||
+  if ( !(id=dtd_find_symbol(dtd, (ichar*)s)) ||
        !(e=id->entity)  )
     return FALSE;
 
@@ -2036,7 +2146,25 @@ dtd_prop_entity(dtd *dtd, term_t ename, term_t value)
     case ET_LITERAL:
     default:
       if ( e->value )
-	return PL_unify_atom_chars(value, e->value);
+      { const char *wrap;
+
+	switch(e->content)
+	{ case EC_SGML:     wrap = "sgml"; break;
+	  case EC_STARTTAG: wrap = "start_tag"; break;
+	  case EC_ENDTAG:   wrap = "end_tag"; break;
+	  case EC_CDATA:    wrap = NULL; break;
+	  case EC_SDATA:    wrap = "sdata"; break;
+	  case EC_NDATA:    wrap = "ndata"; break;
+	  case EC_PI:       wrap = "pi"; break;
+	  default:	    wrap = NULL; assert(0);
+	}
+
+	if ( wrap )
+	  return PL_unify_term(value, PL_FUNCTOR_CHARS, wrap, 1,
+			       PL_CHARS, e->value);
+	else
+	  return PL_unify_atom_chars(value, (char*)e->value);
+      }
   }
 
   assert(0);
@@ -2052,7 +2180,7 @@ dtd_prop_notations(dtd *dtd, term_t list)
 
   for(n=dtd->notations; n; n=n->next)
   { if ( PL_unify_list(tail, head, tail) &&
-	 PL_unify_atom_chars(head, n->name->name) )
+	 PL_unify_atom_chars(head, (char*)n->name->name) )
       continue;
       
     return FALSE;
