@@ -1,4 +1,4 @@
-/*  $Id: ssllib.c,v 1.2.2.2 2004/12/16 14:05:52 jan Exp $
+/*  $Id: ssllib.c,v 1.8 2006/03/15 08:34:47 jan Exp $
 
     Part of SWI-Prolog
 
@@ -89,7 +89,7 @@ static int ctx_idx;
 
 
 static void
-ssl_error(SSL *ssl, int ssl_ret)
+ssl_error(SSL *ssl, int ssl_ret, int code)
 /*
  * Report about errors occuring in the SSL layer
  */
@@ -97,9 +97,23 @@ ssl_error(SSL *ssl, int ssl_ret)
     char  buf[256];
     char *component[5];
     char *colon;
+    int   error = ERR_get_error();
     int   n;
 
-    (void) ERR_error_string(ERR_get_error(), buf);
+    ssl_deb(1, "ssl_error() ret=%d, code=%d, err=%d\n", ssl_ret, code, error);
+
+    if ( code == SSL_ERROR_SYSCALL && error == 0 )
+    { if ( ssl_ret == 0 )
+      {	/* normal if peer just hangs up the line */
+	ssl_deb(1, "SSL error report: unexpected end-of-file\n");
+	return;
+      } else if ( ssl_ret == -1 )
+      { ssl_deb(0, "SSL error report: syscall error: %s\n", strerror(errno));
+	return;
+      }
+    }
+
+    (void) ERR_error_string(error, buf);
 
     /*
      * Disect the following error string:
@@ -112,8 +126,9 @@ ssl_error(SSL *ssl, int ssl_ret)
         *colon++ = 0;
     }
 
-    ssl_deb(1,
-	    "SSL error report:\n\t%8s: %s\n\t%8s: %s\n\t%8s: %s\n"
+    ssl_deb(0,
+	    "SSL error report:\n\t%8s: %s\n\t%8s: %s\n\t%8s: %s\n\t%8s: %s\n"
+	   ,     "code", component[1]
 	   ,  "library", component[2]
 	   , "function", component[3]
 	   ,   "reason", component[4]
@@ -122,51 +137,51 @@ ssl_error(SSL *ssl, int ssl_ret)
 
 static SSL_SOCK_STATUS
 ssl_inspect_status(SSL *ssl, int sock, int ssl_ret)
-{
+{   int code;
+
     if (ssl_ret > 0) {
         return SSL_SOCK_OK;
     }
 
-    if (ssl_ret == 0) {
-        /* SSL controlled failure, see below */
-        switch (SSL_get_error(ssl, ssl_ret)) {
-            case SSL_ERROR_WANT_READ:
-                if (nbio_wait(sock, REQ_READ) == 0) {
-                    return SSL_SOCK_RETRY;
-                }
-                break;
+    code=SSL_get_error(ssl, ssl_ret);
 
-            case SSL_ERROR_WANT_WRITE:
-                if (nbio_wait(sock, REQ_WRITE) == 0) {
-                    return SSL_SOCK_RETRY;
-                }
-                break;
+    switch (code) {
+	case SSL_ERROR_WANT_READ:
+	    if (nbio_wait(sock, REQ_READ) == 0) {
+		return SSL_SOCK_RETRY;
+	    }
+	    break;
+
+	case SSL_ERROR_WANT_WRITE:
+	    if (nbio_wait(sock, REQ_WRITE) == 0) {
+		return SSL_SOCK_RETRY;
+	    }
+	    break;
 
 #ifdef SSL_ERROR_WANT_CONNECT
-            case SSL_ERROR_WANT_CONNECT:
-                if (nbio_wait(sock, REQ_CONNECT) == 0) {
-                    return SSL_SOCK_RETRY;
-                }
-                break;
+	case SSL_ERROR_WANT_CONNECT:
+	    if (nbio_wait(sock, REQ_CONNECT) == 0) {
+		return SSL_SOCK_RETRY;
+	    }
+	    break;
 #endif
 
 #ifdef SSL_ERROR_WANT_ACCEPT
-            case SSL_ERROR_WANT_ACCEPT:
-                if (nbio_wait(sock, REQ_ACCEPT) == 0) {
-                    return SSL_SOCK_RETRY;
-                }
-                break;
+	case SSL_ERROR_WANT_ACCEPT:
+	    if (nbio_wait(sock, REQ_ACCEPT) == 0) {
+		return SSL_SOCK_RETRY;
+	    }
+	    break;
 #endif
 
-            case SSL_ERROR_ZERO_RETURN:
-                return SSL_SOCK_OK;
+	case SSL_ERROR_ZERO_RETURN:
+	    return SSL_SOCK_OK;
 
-            default:
-                break;
-        }
+	default:
+	    break;
     }
 
-    ssl_error(ssl, ssl_ret);
+    ssl_error(ssl, ssl_ret, code);
     return SSL_SOCK_ERROR;
 }
 
@@ -670,6 +685,23 @@ ssl_exit(PL_SSL *config)
     ssl_deb(1, "Controlled exit\n");
 }
 
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+ERR_print_errors_pl() is like  ERR_print_errors_fp(stderr),   but  deals
+with the fact that on Windows stderr is generally lost, so we use Prolog
+I/O for portability.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static void
+ERR_print_errors_pl()
+{ char errmsg[1024];
+
+  ERR_error_string_n(ERR_get_error(), errmsg, sizeof(errmsg));
+
+  Sdprintf("%s\n", errmsg);
+}
+
+
 PL_SSL *
 ssl_init(PL_SSL_ROLE role)
 /*
@@ -704,7 +736,7 @@ ssl_init(PL_SSL_ROLE role)
     ssl_ctx = SSL_CTX_new(ssl_method);
 
     if (!ssl_ctx) {
-        ERR_print_errors_fp(stderr);
+        ERR_print_errors_pl();
     } else {
         long ctx_mode = 0L;
 
@@ -799,17 +831,17 @@ ssl_config(PL_SSL *config)
         if (SSL_CTX_use_certificate_file( config->pl_ssl_ctx
                                         , config->pl_ssl_certf
                                         , SSL_FILETYPE_PEM) <= 0) {
-            ERR_print_errors_fp(stderr);
+            ERR_print_errors_pl();
             return -2;
         }
         if (SSL_CTX_use_PrivateKey_file( config->pl_ssl_ctx
                                        , config->pl_ssl_keyf
                                        , SSL_FILETYPE_PEM) <= 0) {
-            ERR_print_errors_fp(stderr);
+            ERR_print_errors_pl();
             return -3;
         }
         if (SSL_CTX_check_private_key(config->pl_ssl_ctx) <= 0) {
-            ERR_print_errors_fp(stderr);
+            ERR_print_errors_pl();
             ssl_err("Private key does not match certificate public key\n");
             return -4;
         }
@@ -818,7 +850,7 @@ ssl_config(PL_SSL *config)
     
     (void) SSL_CTX_set_verify( config->pl_ssl_ctx
                              , (config->pl_ssl_peer_cert_required)
-                               ? SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE
+                               ? SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT
                                : SSL_VERIFY_NONE
                              , ssl_cb_cert_verify
                              ) ;
@@ -953,42 +985,40 @@ ssl_ssl(PL_SSL *config, int sock_inst)
             } while (1);
             break;
 
-       case PL_SSL_NONE:
-       case PL_SSL_CLIENT:
-          ssl_deb(1, "setting up SSL client side\n");
-          do {
-             int ssl_ret = SSL_connect(instance->ssl);
-             switch(ssl_inspect_status(instance->ssl, sock_inst, ssl_ret)) {
-                case SSL_SOCK_OK:
-                   /* success */
-                   ssl_deb(1, "established ssl client side\n");
-                   return instance;
-                   
-                case SSL_SOCK_RETRY:
-                   continue;
-                   
-                case SSL_SOCK_ERROR:
-                   if (SSL_get_error(instance->ssl, ssl_ret) == SSL_ERROR_WANT_READ)
-                   {
-                      nbio_wait(sock_inst, REQ_READ);
-                      continue;
-                   }
-                   else if (SSL_get_error(instance->ssl, ssl_ret) == SSL_ERROR_WANT_WRITE)
-                   {
-                      nbio_wait(sock_inst, REQ_WRITE);
-                      continue;
-                   }
-                   else
-                   {
-                      Sdprintf("Unrecoverable error: %d\n", SSL_get_error(instance->ssl, ssl_ret));
-                      Sdprintf("Additionally, get_error returned %d\n", ERR_get_error());
-                      return NULL;
-                   }
-                   
-
-             }
-          } while (1);
-          break;
+	case PL_SSL_NONE:
+	case PL_SSL_CLIENT:
+	   ssl_deb(1, "setting up SSL client side\n");
+	   do {
+	      int ssl_ret = SSL_connect(instance->ssl);
+	      switch(ssl_inspect_status(instance->ssl, sock_inst, ssl_ret)) {
+	         case SSL_SOCK_OK:
+	            /* success */
+	            ssl_deb(1, "established ssl client side\n");
+	            return instance;
+	            
+	         case SSL_SOCK_RETRY:
+	            continue;
+	            
+	         case SSL_SOCK_ERROR:
+	            if (SSL_get_error(instance->ssl, ssl_ret) == SSL_ERROR_WANT_READ)
+	            {
+	               nbio_wait(sock_inst, REQ_READ);
+	               continue;
+	            }
+	            else if (SSL_get_error(instance->ssl, ssl_ret) == SSL_ERROR_WANT_WRITE)
+	            {
+	               nbio_wait(sock_inst, REQ_WRITE);
+	               continue;
+	            }
+	            else
+	            {
+	               Sdprintf("Unrecoverable error: %d\n", SSL_get_error(instance->ssl, ssl_ret));
+	               Sdprintf("Additionally, get_error returned %d\n", ERR_get_error());
+	               return NULL;
+	            }
+	      }
+	   } while (1);
+	   break;
     }
     return NULL;
 }
