@@ -1,11 +1,11 @@
-/*  $Id: swi_help.pl,v 1.22 2004/09/21 14:22:25 jan Exp $
+/*  $Id: swi_help.pl,v 1.28 2006/08/04 14:24:01 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
-    WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    E-mail:        wielemak@science.uva.nl
+    WWW:           http://www.swi-prolog.org/projects/xpce/
+    Copyright (C): 1985-2005, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -454,7 +454,7 @@ mark_titles(TB, From, To) :-
 mark_titles(TB, From, To) :-
 	get(TB, scan, From, line, 0, end, EL),
 %	Len is EL - From,
-%	new(_, fragment(TB, From, Len, title)),
+%%	new(_, fragment(TB, From, Len, title)),
 	(   get(TB, find, EL, @pui_ff, 1, start, Idx)
 	->  mark_titles(TB, Idx, To)
 	;   true
@@ -469,18 +469,20 @@ mark_titles(TB, From, To) :-
 :- dynamic
 	regex_db/2.
 
-:- pce_global(@ul_regex, new(regex('.\b\\(.\\)'))).
+:- pce_global(@ul_regex, new(regex('.\b(.)'))).
 
-regex(bold,	  '\\(\\(.\\)\b\\2\\)+').
-regex(underline,  '\\(\\(.\\)\b_\\)+').
-regex(underline2,  '\\(_\b\\(.\\)\\)+').
-regex(predicate,  '\\(\\w+\\)/\\(\\sd+\\)').
-regex(predicate2, '\\w+/\\[\\sd+[-,]\\sd+\\]').
-regex(function,   'PL_\\w+()').
-regex(section,	  '\\([Ss]ection\\|[Cc]hapter\\)\\s +\\sd+\\(\\.\\sd+\\)*').
-regex(location,	  '\\([a-zA-Z]:\\)?\\(/[-_a-zA-Z0-9~+=.]*\\)+:\\sd+').
-regex(clause,	  '\\sd+-th clause of \\w+:.*/\\sd+').
-regex(method,     '\\w+\\(->\\|<-\\)\\w+').
+%	Bug: send(regex('((.)x\\2)+$'), match, axbaxa) succeeds!
+%	Bug reproduces in TCL.  Submitted to Tcl bugtracking system.
+
+regex(bold,	  '((.)\b\\2)+').
+regex(underline,  '(_\b(.))+').
+regex(predicate,  '(\\w+)/(\\d+)').
+regex(predicate2, '\\w+/\\[\\d+[-,]\\d+\\]').
+regex(function,   'PL_\\w+\\(\\)').
+regex(section,	  '([Ss]ection|[Cc]hapter)\\s+\\d+(\\.\\d+)*').
+regex(location,	  '([a-zA-Z]:)?(/[-_a-zA-Z0-9~+=.]*)+:\\d+').
+regex(clause,	  '\\d+-th clause of \\w+:[^\n]+/\\d+').
+regex(method,     '\\w+(->|<-)\\w+').
 
 regex_object(Id, Re) :-
 	regex_db(Id, Re), !.
@@ -512,7 +514,6 @@ mark_fragment(V, Re:regex) :->
 
 font_style(bold, bold).
 font_style(underline, underline).
-font_style(underline2, underline).
 
 documented(Re, V, button) :-
 	regex_db(predicate, Re), !,
@@ -530,7 +531,6 @@ documented(Re, V, Style) :-
 	send(Re, replace, TB, String).
 documented(_, _, button).
 
-
 jump(V, Caret:[int]) :->
 	"Jump to current fragment"::
 	(   Caret == @default
@@ -567,7 +567,7 @@ drop(V, Id:name) :->
 		 *       MANUAL UTILITIES	*
 		 *******************************/
 
-%	manual_file(-File)
+%%	manual_file(-File)
 %	Find the database file of the manual.  If the manual cannot be
 %	found, display an error message.
 
@@ -583,11 +583,11 @@ manual_file(_File) :-
 		 *     ATOMIC SPECIFICATIONS	*
 		 *******************************/
 
-%	manual_range(+What, -From, -To)
+%%	manual_range(+What, -From, -To)
 %	Find the character range for the given help topic, which is of
 %	the form
 %			name/arity
-%			function()
+%%			function()
 %			n.m..
 
 manual_range(What, Ranges) :-
@@ -647,13 +647,13 @@ section_separator -->
 	"-", !.
 
 name_and_arity(Name, Arity) -->
-	[C1|CT],
+	string(NameCodes),
 	"/",
 	(   integer(Arity)
 	->  []
 	;   "_"
 	), !,
-	{ atom_codes(Name, [C1|CT])
+	{ atom_codes(Name, NameCodes)
 	}.
 
 pl_function(Name) -->
@@ -669,6 +669,12 @@ pl_function(Name) -->
 	}.
 
 
+string([]) -->
+	[].
+string([H|T]) -->
+	[H],
+	string(T).
+
 word([C0|CT]) -->
 	[C0],
 	{ C0 > 0' }, !,
@@ -839,19 +845,19 @@ try_to_edit(Spec) :-
 	send(@emacs, goto_source_location, source_location(File, Line)).
 
 source(File, Line) -->			% path:line
-	[F0|FT],
+	string(FileCodes),
 	":",
 	integer(Line), !,
-	{ atom_codes(File, [F0|FT])
+	{ atom_codes(File, FileCodes)
 	}.
 source(File, Line) -->			% n-th clause of module:name/arity
 	integer(NClause),
 	"-th clause of ",
-	[M0|MT], ":",
-	[N0|NT], "/",
+	string(MCodes), ":",
+	string(NCodes), "/",
 	integer(Arity),
-	{ atom_codes(Name, [N0|NT]),
-	  atom_codes(Module, [M0|MT]),
+	{ atom_codes(Name, NCodes),
+	  atom_codes(Module, MCodes),
 	  functor(Head, Name, Arity),
 	  nth_clause(Module:Head, NClause, CRef),
 	  clause_property(CRef, file(File)),
