@@ -1,4 +1,4 @@
-/*  $Id: interface.h,v 1.112.2.1 2005/02/21 09:47:46 jan Exp $
+/*  $Id: interface.h,v 1.173 2006/08/24 10:29:26 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -31,7 +31,7 @@
 		********************************/
 
 #ifndef PCE_VERSION
-#define PCE_VERSION "6.4.3, February 2005"
+#define PCE_VERSION "6.6.18, July 2006"
 #endif
 
 #ifndef OS_VERSION
@@ -69,7 +69,7 @@ typedef void *		PceClass;	/* PCE's view of a class */
 typedef void *		PceHostData;	/* PCE's view of a host data handle */
 
 #define INT_MASK_SHIFT	1
-#define PCE_MAX_INT	((long)((1L<<(sizeof(Any)*8 - INT_MASK_SHIFT-1))-1))
+#define PCE_MAX_INT	((long)((1L<<(sizeof(void*)*8 - INT_MASK_SHIFT-1))-1))
 #define PCE_MIN_INT	(-(PCE_MAX_INT-1))
 
 typedef long		AnswerMark;	/* Mark on AnswerStack */
@@ -167,8 +167,9 @@ typedef union
 
 __pce_export int    pceToC(PceObject datum, PceCValue *rval);
 __pce_export int    pceToCReference(PceObject datum, PceCValue *rval);
-__pce_export char * pceCharArrayToC(PceObject datum, unsigned int *len);
-__pce_export char * pceStringToC(PceObject datum);
+__pce_export char * pceCharArrayToCA(PceObject datum, unsigned int *len);
+__pce_export wchar_t * pceCharArrayToCW(PceObject datum, unsigned int *len);
+__pce_export int    pceIsString(PceObject datum);
 __pce_export void * pcePointerToC(PceObject datum);
 __pce_export PceHostData  CtoHostData(PceClass class, void *handle, int flags);
 __pce_export void *    getHostDataHandle(PceHostData hd);
@@ -312,6 +313,7 @@ __pce_export int	pceMTdetach(void);
 #define HOST_ATEXIT	10		/* Callback on exit */
 #define HOST_CONSOLE	11		/* Win32: query HWND of console */
 #define HOST_CHECK_INTERRUPT 12		/* Win32: periodic check for ^C */
+#define HOST_ENCODING	13		/* Query default encoding of the host */
 
 #define PCE_METHOD_INFO_HANDLE_ONLY	0x01
 #define PCE_METHOD_INFO_TRACE_ENTER	0x02
@@ -389,13 +391,16 @@ __pce_export int pceInitialise(int handles, const char *home,
 		*           C --> PCE		*
 		********************************/
 
-__pce_export PceObject	cToPceName(const char *);
-__pce_export PceObject	cToPceName_n(const char *, unsigned int len);
+__pce_export PceObject	cToPceName_nA(const char *, unsigned int len);
+__pce_export PceObject	cToPceName_nW(const wchar_t *, unsigned int len);
 __pce_export PceObject	cToPceInteger(long);
 __pce_export PceObject	cToPceReal(double);
-__pce_export PceObject	cToPceString(PceName assoc,
-				     const char *text, unsigned int len,
-				     int translate);
+__pce_export PceObject	cToPceStringA(PceName assoc,
+				      const char *text, unsigned int len,
+				      int translate);
+__pce_export PceObject	cToPceStringW(PceName assoc,
+				      const wchar_t *text, unsigned int len,
+				      int translate);
 __pce_export PceObject	cToPceAssoc(const char *);
 __pce_export PceObject	cToPceReference(unsigned long);
 __pce_export PceObject	cToPcePointer(void *ptr);
@@ -444,6 +449,10 @@ __pce_export int	Cputchar(int chr);
 __pce_export void	Cflush(void);
 __pce_export char *	Cgetline(char *line, int size);
 
+#ifdef PCE_INCLUDED
+	     int	Cputstr(String s);
+#endif
+
 
 		 /*******************************
 		 *	MEMORY ALLOCATION	*
@@ -490,11 +499,12 @@ __pce_export void	pceGetMethod(PceClass class,
 
 #define PCE_OPEN_MAX	64		/* statically allocated max open */
 
-__pce_export int	pceOpen(PceObject obj, int flags);
+__pce_export int	pceOpen(PceObject obj, int flags, void *encoding);
 __pce_export int	pceClose(int handle);
 __pce_export int	pceWrite(int handle, const char *buf, int size);
 __pce_export int	pceRead(int handle, char *buf, int size);
 __pce_export long	pceSeek(int handle, long offset, int whence);
+__pce_export int	pceControl(int handle, int cmd, void *closure);
 __pce_export const char *pceOsError();
 
 					/* flags for pceOpen() (or'ed) */
@@ -508,6 +518,15 @@ __pce_export const char *pceOsError();
 #define PCE_SEEK_CUR	1
 #define PCE_SEEK_END	2
 
+#define PCE_SETENCODING	1		/* pceControl() commands */
+
+		 /*******************************
+		 *	       MISC		*
+		 *******************************/
+
+__pce_export void *	pcePostScriptStream(void);
+
+
 		 /*******************************
 		 *	    CALL-BACK		*
 		 *******************************/
