@@ -1,4 +1,4 @@
-/*  $Id: frame.c,v 1.59 2004/09/21 14:22:25 jan Exp $
+/*  $Id: frame.c,v 1.67 2006/08/11 08:01:15 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -26,13 +26,13 @@
 #include <h/graphics.h>
 
 TileObj		getTileFrame(FrameObj);
-forwards int	get_position_from_center_frame(FrameObj, Point, int *, int *);
-static void	ensure_on_display(FrameObj, int *, int *);
+forwards int	get_position_from_center_frame(FrameObj, Monitor, Point, int *, int *);
+static void	ensure_on_display(FrameObj, Monitor, int *, int *);
 static status	closedFrame(FrameObj, Bool);
 static status	openFrame(FrameObj fr, Point pos, Bool grab, Bool normalise);
 static status	doneMessageFrame(FrameObj fr, Code msg);
-static status	geometryFrame(FrameObj fr, Name spec);
-static status	setFrame(FrameObj fr, Int x, Int y, Int w, Int h);
+static status	geometryFrame(FrameObj fr, Name spec, Monitor mon);
+static status	setFrame(FrameObj fr, Int x, Int y, Int w, Int h, Monitor mon);
 static status	flushFrame(FrameObj fr);
 static status	kindFrame(FrameObj fr, Name kind);
 static status	informTransientsFramev(FrameObj fr, Name selector,
@@ -204,15 +204,15 @@ getConfirmFrame(FrameObj fr, Point pos, Bool grab, Bool normalise)
 
 
 Any
-getConfirmCenteredFrame(FrameObj fr, Point pos, Bool grab)
+getConfirmCenteredFrame(FrameObj fr, Point pos, Bool grab, Monitor mon)
 { int x, y;
   Point p2;
   Any rval;
 
   TRY( send(fr, NAME_create, EAV) );
 
-  get_position_from_center_frame(fr, pos, &x, &y);
-  ensure_on_display(fr, &x, &y);
+  get_position_from_center_frame(fr, mon, pos, &x, &y);
+  ensure_on_display(fr, mon, &x, &y);
   p2 = tempObject(ClassPoint, toInt(x), toInt(y), EAV);
 
   rval = getConfirmFrame(fr, p2, grab, OFF);
@@ -248,19 +248,37 @@ openFrame(FrameObj fr, Point pos, Bool grab, Bool normalise)
   setpos:
 #endif
     if ( normalise == ON )
-    { Size s = getSizeDisplay(fr->display);
-      int dw = valInt(s->w), dh = valInt(s->h);
+    { int mx, my, mw, mh;
       int fw = valInt(fr->area->w), fh = valInt(fr->area->h);
+      Area a;
+      Area tmp = tempObject(ClassArea,
+			    toInt(x), toInt(y), fr->area->w, fr->area->h, EAV);
+      Monitor mon = getMonitorDisplay(fr->display, tmp);
+      
+      considerPreserveObject(tmp);
 
-      if ( valInt(x) + fw > dw ) x = toInt(dw - fw);
-      if ( valInt(y) + fh > dh ) y = toInt(dh - fh);
-      if ( valInt(x) < 0 )       x = ZERO;
-      if ( valInt(y) < 0 )	 y = ZERO;
+      if ( !mon )
+	mon = getMonitorDisplay(fr->display, DEFAULT);
+      if ( !mon )
+	mon = getHeadChain(fr->display->monitors);
+      a = isNil(mon->work_area) ? mon->area : mon->work_area;
+
+      mx = valInt(a->x);
+      my = valInt(a->y);
+      mw = valInt(a->w);
+      mh = valInt(a->h);
+
+      if ( valInt(x) + fw > mx+mw ) x = toInt(mx+mw - fw);
+      if ( valInt(y) + fh > my+mh ) y = toInt(my+mh - fh);
+      if ( valInt(x) < mx ) x = toInt(mx);
+      if ( valInt(y) < my ) y = toInt(my);
     }
 
-    setFrame(fr, x, y, w, h);  
-  }					/* But in Windows `do-it-yourself' */
-#ifdef WIN32
+    setFrame(fr, x, y, w, h, DEFAULT);  
+  } else if ( notNil(fr->geometry) )
+  { ws_x_geometry_frame(fr, fr->geometry, DEFAULT);
+  }					
+#ifdef WIN32				/* But in Windows `do-it-yourself' */
   else if ( notNil(fr->transient_for) )
   { Area pa = fr->transient_for->area;
     int xb, yb, ycap;
@@ -283,15 +301,15 @@ openFrame(FrameObj fr, Point pos, Bool grab, Bool normalise)
 
 
 static status
-openCenteredFrame(FrameObj fr, Point pos, Bool grab)
+openCenteredFrame(FrameObj fr, Point pos, Bool grab, Monitor mon)
 { int x, y;
   int rval;
   Point p2;
 
   TRY( send(fr, NAME_create, EAV) );
 
-  get_position_from_center_frame(fr, pos, &x, &y);
-  ensure_on_display(fr, &x, &y);
+  get_position_from_center_frame(fr, mon, pos, &x, &y);
+  ensure_on_display(fr, DEFAULT, &x, &y);
   p2 = answerObject(ClassPoint, toInt(x), toInt(y), EAV);
   rval = openFrame(fr, p2, grab, OFF);
   doneObject(p2);
@@ -422,7 +440,7 @@ createFrame(FrameObj fr)
   attachWmProtocolsFrame(fr);
 
   if ( isName(fr->geometry) )
-    geometryFrame(fr, fr->geometry);
+    geometryFrame(fr, fr->geometry, DEFAULT);
 
   for_cell(cell, fr->members)
   { updateCursorWindow(cell->value);
@@ -470,7 +488,7 @@ fitFrame(FrameObj fr)
 
   setFrame(fr, DEFAULT, DEFAULT,
 	   add(t->idealWidth, border),
-	   add(t->idealHeight, border));
+	   add(t->idealHeight, border), DEFAULT);
   assign(fr, fitting, OFF);
 
   succeed;
@@ -596,13 +614,57 @@ hiddenFrame(FrameObj fr)
 		*       AREA MANAGEMENENT	*
 		********************************/
 
-static int
-get_position_from_center_frame(FrameObj fr, Point pos, int *x, int *y)
-{ if ( isDefault(pos) )
-  { Size sz = getSizeDisplay(fr->display);
+static Monitor
+CurrentMonitor(FrameObj fr)
+{ DisplayObj d = fr->display;
+  Monitor mon;
 
-    *x = valInt(sz->w) / 2;
-    *y = valInt(sz->h) / 2;
+  if ( isOpenFrameStatus(fr->status) )
+  { if ( notNil(d) && (mon=getMonitorDisplay(d, fr->area)) )
+      return mon;
+  } else if ( notNil(d) && instanceOfObject(EVENT->value, ClassEvent) )
+  { EventObj ev = EVENT->value;
+    Point pt;
+
+    if ( notNil(ev->window) && offFlag(ev->window, F_FREEING|F_FREED) )
+    { pt = getPositionEvent(ev, d);
+    } else
+    { pt = getPointerLocationDisplay(d);
+    }
+
+    if ( pt && (mon=getMonitorDisplay(d, pt)) )
+      return mon;
+  }
+
+  if ( isNil(d) )
+    d = CurrentDisplay(fr);
+
+  if ( notNil(d->monitors) )
+    return getHeadChain(d->monitors);
+
+  return NULL;
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Get XY coordinate of frame if  its  center   must  be  at pos. If Pos is
+DEFAULT it is centered in the given   monitor. If the monitor is default
+too, we deduce the most recent monitor from the event.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int
+get_position_from_center_frame(FrameObj fr, Monitor mon, Point pos,
+			       int *x, int *y)
+{ if ( isDefault(pos) )
+  { if ( isDefault(mon) )
+      mon = CurrentMonitor(fr);
+
+    if ( mon )
+    { *x = valInt(mon->area->x) + valInt(mon->area->w)/2;
+      *y = valInt(mon->area->y) + valInt(mon->area->h)/2;
+    } else
+    { *x = *y = 0;
+    }
   } else
   { *x = valInt(pos->x);
     *y = valInt(pos->y);
@@ -616,17 +678,23 @@ get_position_from_center_frame(FrameObj fr, Point pos, int *x, int *y)
 
 
 static void
-ensure_on_display(FrameObj fr, int *x, int *y)
-{ Size sz = getSizeDisplay(fr->display);
+ensure_on_display(FrameObj fr, Monitor mon, int *x, int *y)
+{ int rm, bm;
 
-  if ( *x + valInt(fr->area->w) > valInt(sz->w) )
-    *x -= *x + valInt(fr->area->w) - valInt(sz->w);
-  if ( *y + valInt(fr->area->w) > valInt(sz->w) )
-    *y -= *y + valInt(fr->area->w) - valInt(sz->w);
-  if ( *x < 0 )
-    *x = 0;
-  if ( *y < 0 )
-    *y = 0;
+  if ( isDefault(mon) )
+    mon = CurrentMonitor(fr);
+
+  rm = valInt(mon->area->x) + valInt(mon->area->w);
+  bm = valInt(mon->area->y) + valInt(mon->area->h);
+
+  if ( *x + valInt(fr->area->w) > rm )
+    *x -= *x + valInt(fr->area->w) - rm;
+  if ( *y + valInt(fr->area->h) > bm )
+    *y -= *y + valInt(fr->area->h) - bm;
+  if ( *x < valInt(mon->area->x) )
+    *x = valInt(mon->area->x);
+  if ( *y < valInt(mon->area->y) )
+    *y = valInt(mon->area->y);
 }
 
 
@@ -656,39 +724,75 @@ getBoundingBoxFrame(FrameObj fr)
 
 
 
+static Monitor
+getMonitorFrame(FrameObj fr)
+{ if ( notNil(fr->display) )
+    answer(getMonitorDisplay(fr->display, fr->area));
+
+  fail;
+}
+
+
+
 static Name
 getGeometryFrame(FrameObj fr)
-{ int x, y, w, h;
+{ int x, y, ww, wh;
 
-  if ( ws_frame_bb(fr, &x, &y, &w, &h) )
-  { int dw, dh;
+  if ( ws_frame_bb(fr, &x, &y, &ww, &wh) ) 	/* outer area */
+  { int mx, my, mw, mh;
     int xn=FALSE, yn=FALSE;
-    Size size;
     char buf[100];
+    Monitor mon;
+    int cw, ch;
 
-    size = getSizeDisplay(fr->display);
-    dw = valInt(size->w);
-    dh = valInt(size->h);
+    cw = valInt(fr->area->w);			/* Client area */
+    ch = valInt(fr->area->h);
 
-    if ( x > (dw - (x+w))*2 )
-    { x = dw - (x+w);
+    if ( (mon=getMonitorFrame(fr)) )
+    { Area a = (isNil(mon->work_area) ? mon->area : mon->work_area);
+
+      mx = valInt(a->x);
+      my = valInt(a->y);
+      mw = valInt(a->w);
+      mh = valInt(a->h);
+
+      DEBUG(NAME_geometry, Cprintf("%s on %s: %d %d %d %d\n", pp(fr), pp(mon),
+				   mx, my, mw, mh));
+    } else
+    { Size size = getSizeDisplay(fr->display);
+      
+      mx = my = 0;
+      mw = valInt(size->w);
+      mh = valInt(size->h);
+    }
+
+    if ( x-mx > ((mx+mw) - (x+ww))*2 )	/* over 2/3th */
+    { x = (mx+mw) - (x+ww);
       xn = TRUE;
+    } else
+    { x -= mx;
     }
-    if ( y > (dh - (y+h))*2 )
-    { y = dh - (y+h);
+    if ( y-my > ((my+mh) - (y+wh))*2 )
+    { y = (my+mh) - (y+wh);
       yn = TRUE;
+    } else
+    { y -= my;
     }
 
-					/* geometry-size denotes client area */
-    w = valInt(fr->area->w);
-    h = valInt(fr->area->h);
-
-    if ( fr->can_resize == OFF )
-      sprintf(buf, "%s%d%s%d",
-	      xn ? "-" : "+", x, yn ? "-" : "+", y);
+    if ( fr->can_resize != OFF )
+      sprintf(buf, "%dx%d", cw, ch);
     else
-      sprintf(buf, "%dx%d%s%d%s%d",
-	      w, h, xn ? "-" : "+", x, yn ? "-" : "+", y);
+      buf[0] = EOS;
+
+    sprintf(buf+strlen(buf),
+	    "%s%d%s%d", xn ? "-" : "+", x, yn ? "-" : "+", y);
+
+    if ( mon && fr->display->monitors->size != ONE )
+    { Int n = getIndexChain(fr->display->monitors, mon);
+
+      if ( n )
+	sprintf(buf+strlen(buf), "@%ld", valInt(n)-1);
+    }
 
     answer(CtoName(buf));
   }
@@ -708,21 +812,30 @@ getImageFrame(FrameObj fr)
 
 
 static status
-geometryFrame(FrameObj fr, Name spec)
+geometryFrame(FrameObj fr, Name spec, Monitor mon)
 { assign(fr, geometry, spec);
 
-  ws_x_geometry_frame(fr, spec);
+  ws_x_geometry_frame(fr, spec, mon);
 
   succeed;
 }
 
 
 static status
-setFrame(FrameObj fr, Int x, Int y, Int w, Int h)
+setFrame(FrameObj fr, Int x, Int y, Int w, Int h, Monitor mon)
 { Area a = fr->area;
   Int ow = a->w;
   Int oh = a->h;
 
+  if ( notDefault(mon) )
+  { if ( notDefault(x) )
+      x = add(x, mon->area->x);
+    if ( notDefault(y) )
+      y = add(y, mon->area->y);
+
+    mon = DEFAULT;
+  }
+
   setArea(a, x, y, w, h);
   if ( valInt(a->w) <= 0 )		/* Window systems don't like that */
     assign(a, w, ONE);
@@ -730,7 +843,7 @@ setFrame(FrameObj fr, Int x, Int y, Int w, Int h)
     assign(a, h, ONE);
 
   if ( createdFrame(fr) )
-  { ws_geometry_frame(fr, x, y, w, h);
+  { ws_geometry_frame(fr, x, y, w, h, mon);
 
     if ( ow != a->w || oh != a->h )
       resizeFrame(fr);
@@ -742,52 +855,52 @@ setFrame(FrameObj fr, Int x, Int y, Int w, Int h)
 
 static status
 xFrame(FrameObj fr, Int x)
-{ return setFrame(fr, x, DEFAULT, DEFAULT, DEFAULT);
+{ return setFrame(fr, x, DEFAULT, DEFAULT, DEFAULT, DEFAULT);
 }
 
 
 static status
 yFrame(FrameObj fr, Int y)
-{ return setFrame(fr, DEFAULT, y, DEFAULT, DEFAULT);
+{ return setFrame(fr, DEFAULT, y, DEFAULT, DEFAULT, DEFAULT);
 }
 
 
 static status
 widthFrame(FrameObj fr, Int w)
-{ return setFrame(fr, DEFAULT, DEFAULT, w, DEFAULT);
+{ return setFrame(fr, DEFAULT, DEFAULT, w, DEFAULT, DEFAULT);
 }
 
 
 static status
 heightFrame(FrameObj fr, Int h)
-{ return setFrame(fr, DEFAULT, DEFAULT, DEFAULT, h);
+{ return setFrame(fr, DEFAULT, DEFAULT, DEFAULT, h, DEFAULT);
 }
 
 
 static status
 sizeFrame(FrameObj fr, Size sz)
-{ return setFrame(fr, DEFAULT, DEFAULT, sz->w, sz->h);
+{ return setFrame(fr, DEFAULT, DEFAULT, sz->w, sz->h, DEFAULT);
 }
 
 
 static status
 positionFrame(FrameObj fr, Point pos)
-{ return setFrame(fr, pos->x, pos->y, DEFAULT, DEFAULT);
+{ return setFrame(fr, pos->x, pos->y, DEFAULT, DEFAULT, DEFAULT);
 }
 
 
 static status
-centerFrame(FrameObj fr, Point pos)
+centerFrame(FrameObj fr, Point pos, Monitor mon)
 { int x, y;
 
-  get_position_from_center_frame(fr, pos, &x, &y);
-  return setFrame(fr, toInt(x), toInt(y), DEFAULT, DEFAULT);
+  get_position_from_center_frame(fr, mon, pos, &x, &y);
+  return setFrame(fr, toInt(x), toInt(y), DEFAULT, DEFAULT, DEFAULT);
 }
 
 
 static status
 areaFrame(FrameObj fr, Area area)
-{ return setFrame(fr, area->x, area->y, area->w, area->h);
+{ return setFrame(fr, area->x, area->y, area->w, area->h, DEFAULT);
 }
 
 
@@ -1517,7 +1630,7 @@ getCatchAllFramev(FrameObj fr, Name name)
 }
 
 
-extern status postscriptFrame(FrameObj fr);
+extern status postscriptFrame(FrameObj fr, Name hb);
 
 		/********************************
 		*             VISUAL		*
@@ -1578,8 +1691,8 @@ getThreadFrame(FrameObj fr)
 
 /* Type declarations */
 
-static char *T_centerADpointD_grabADboolD[] =
-        { "center=[point]", "grab=[bool]" };
+static char *T_openCentered[] =
+        { "center=[point]", "grab=[bool]", "monitor=[monitor]" };
 static char *T_busyCursor[] =
         { "cursor=[cursor]*", "block_input=[bool]" };
 static char *T_icon[] =
@@ -1600,9 +1713,14 @@ static char *T_wmProtocol[] =
 static char *T_convertOldSlot[] =
         { "slot=name", "value=any" };
 static char *T_set[] =
-        { "x=[int]", "y=[int]", "width=[int]", "height=[int]" };
+        { "x=[int]", "y=[int]", "width=[int]", "height=[int]",
+	  "monitor=[monitor]" };
 static char *T_grab_pointer[] =
 	{ "grab=bool", "cursor=[cursor]" };
+static char *T_center[] =
+	{ "center=[point]", "monitor=[monitor]" };
+static char *T_geometry[] =
+	{ "geometry=name", "monitor=[monitor]" };
 
 /* Instance Variables */
 
@@ -1629,7 +1747,7 @@ static vardecl var_frame[] =
      NAME_appearance, "Colourmap for the window's frame"),
   SV(NAME_area, "area", IV_GET|IV_STORE, areaFrame,
      NAME_area, "Area of the opened frame on the display"),
-  SV(NAME_geometry, "name*", IV_NONE|IV_STORE, geometryFrame,
+  IV(NAME_geometry, "name*", IV_NONE,
      NAME_area, "X-window geometry specification"),
   IV(NAME_members, "chain", IV_NONE,
      NAME_organisation, "Windows in the frame"),
@@ -1690,7 +1808,7 @@ static senddecl send_frame[] =
      NAME_appearance, "X-border width"),
   SM(NAME_showLabel, 1, "show=bool", showLabelFrame,
      NAME_appearance, "If @off, sets <->kind to `transient'"),
-  SM(NAME_center, 1, "center=[point]", centerFrame,
+  SM(NAME_center, 2, T_center, centerFrame,
      NAME_area, "Move the frame to make point its center"),
   SM(NAME_height, 1, "height=int", heightFrame,
      NAME_area, "Set height of frame"),
@@ -1698,7 +1816,7 @@ static senddecl send_frame[] =
      NAME_area, "Move the frame on the display"),
   SM(NAME_position, 1, "position=point", positionFrame,
      NAME_area, "Move the frame on the display"),
-  SM(NAME_set, 4, T_set, setFrame,
+  SM(NAME_set, 5, T_set, setFrame,
      NAME_area, "Set XYWH of frame on display"),
   SM(NAME_size, 1, "size=size", sizeFrame,
      NAME_area, "Resize the frame"),
@@ -1708,6 +1826,8 @@ static senddecl send_frame[] =
      NAME_area, "Set X-coordinate of frame"),
   SM(NAME_y, 1, "x=int", yFrame,
      NAME_area, "Set Y-coordinate of frame"),
+  SM(NAME_geometry, 2, T_geometry, geometryFrame,
+     NAME_area, "X-window geometry specification"),
   SM(NAME_busyCursor, 2, T_busyCursor, busyCursorFrame,
      NAME_event, "Define (temporary) cursor for all windows in the frame"),
   SM(NAME_inputWindow, 1, "window", inputWindowFrame,
@@ -1732,7 +1852,7 @@ static senddecl send_frame[] =
      NAME_open, "Inform transients using ->show"),
   SM(NAME_open, 3, T_positionADpointD_grabADboolD_normaliseADboolD, openFrame,
      NAME_open, "->create and map on the display"),
-  SM(NAME_openCentered, 2, T_centerADpointD_grabADboolD, openCenteredFrame,
+  SM(NAME_openCentered, 3, T_openCentered, openCenteredFrame,
      NAME_open, "Open centered around point"),
   SM(NAME_uncreate, 0, NULL, uncreateFrame,
      NAME_open, "Destroy window-system counterpart"),
@@ -1742,7 +1862,7 @@ static senddecl send_frame[] =
      NAME_organisation, "Append a window to the frame"),
   SM(NAME_delete, 1, "member:window", deleteFrame,
      NAME_organisation, "Delete window from the frame"),
-  SM(NAME_Postscript, 0, NULL, postscriptFrame,
+  SM(NAME_Postscript, 1, "{head,body}", postscriptFrame,
      NAME_postscript, "Create PostScript for interior"),
   SM(NAME_bell, 1, "volume=[int]", bellFrame,
      NAME_report, "Ring the bell on display"),
@@ -1801,6 +1921,8 @@ static getdecl get_frame[] =
      NAME_area, "Position on the display"),
   GM(NAME_size, 0, "size", NULL, getSizeFrame,
      NAME_area, "Size on the display"),
+  GM(NAME_monitor, 0, "monitor", NULL, getMonitorFrame,
+     NAME_organisation, "Monitor frame is displayed on"),
   GM(NAME_image, 1, "image", "[{bitmap,pixmap}]", getImageFrame,
      NAME_conversion, "Image with the pixels of the frame"),
   GM(NAME_keyboardFocus, 0, "window", NULL, getKeyboardFocusFrame,
@@ -1815,7 +1937,7 @@ static getdecl get_frame[] =
      NAME_layout, "Find tile managing object"),
   GM(NAME_confirm, 3, "return_value=any", T_positionADpointD_grabADboolD_normaliseADboolD, getConfirmFrame,
      NAME_modal, "Start sub-eventloop until ->return"),
-  GM(NAME_confirmCentered, 2, "return_value=any", T_centerADpointD_grabADboolD, getConfirmCenteredFrame,
+  GM(NAME_confirmCentered, 3, "return_value=any", T_openCentered, getConfirmCenteredFrame,
      NAME_modal, "As <-confirm, but centered around point"),
   GM(NAME_catchAll, 1, "window", "window_name=name", getCatchAllFramev,
      NAME_organisation, "Get named window"),
