@@ -1,4 +1,4 @@
-/*  $Id: xcommon.c,v 1.38 2004/09/21 14:22:25 jan Exp $
+/*  $Id: xcommon.c,v 1.54 2006/08/14 14:13:56 jan Exp $
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -28,6 +28,7 @@
 #include "include.h"
 #include <X11/keysym.h>
 #include <X11/Xproto.h>			/* get request codes */
+#include <locale.h>
 
 #undef roundup
 #define roundup(v, n)		((((v)+(n)-1)/(n))*(n))
@@ -37,6 +38,21 @@
 
 XtAppContext	ThePceXtAppContext;	/* X toolkit application context */
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+The Mac keyboard has a bit different   viewpoint. Alt is primairily used
+for creating non-ascii characters, while the command key is used to give
+commands. The command key is by default passed as `Mod2'.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#ifdef __APPLE__
+#define DefMetaMask Mod2Mask
+#else
+#define DefMetaMask Mod1Mask
+#endif
+
+static unsigned int MetaMask = DefMetaMask;	/* Key-mask for meta */
+  
+
 		 /*******************************
 		 *       X11 APP CONTEXT	*
 		 *******************************/
@@ -73,14 +89,40 @@ x_error_handler(Display *display, XErrorEvent *error)
 }
 
 
-#define USE_XDEFAUL_APP_CONTEXT 1
+#define USE_XDEFAULT_APP_CONTEXT 1
 
-#ifdef USE_XDEFAUL_APP_CONTEXT
+#ifdef USE_XDEFAULT_APP_CONTEXT
 extern XtAppContext _XtDefaultAppContext(void);
 #undef XtCreateApplicationContext
 #define XtCreateApplicationContext _XtDefaultAppContext
 #endif
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(*) XInitThreads() must  be  called  if   multiple  thread  access  Xlib
+functions concurrently. Given the  current  locking,   I  think  this is
+cannot happen as we lock the central  XPCE   lock  both if we receive an
+event and if we send a message from Prolog. 
+
+Nevertheless, XPCE uses this  flag  to   prepare  for  more fine-grained
+locking. It turns out the X11 version  distributed with MacOSX has a but
+that causes a deadlock in handling dead-keys.  Therefore, we do not lock
+for the Mac.  Here is the stack-trace:
+
+#1  0x900019cc in pthread_mutex_lock ()
+#2  0x9ba85b9c in _XLockDisplay ()
+#3  0x9ba61054 in XPutBackEvent ()
+#4  0x9baba0b4 in _XimLocalFilter ()
+#5  0x9ba8548c in XFilterEvent ()
+#6  0x9bde9538 in _XtDefaultDispatcher ()
+#7  0x9bde97b0 in XtDispatchEvent ()
+#8  0x9bdf5934 in XtAppProcessEvent ()
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#ifdef __APPLE__
+static int use_x_init_threads = FALSE;
+#else
+static int use_x_init_threads = TRUE;
+#endif
 
 void *
 pceXtAppContext(void * ctx)
@@ -89,12 +131,14 @@ pceXtAppContext(void * ctx)
     { ThePceXtAppContext = ctx;
       XSetErrorHandler(x_error_handler);
     } else
-    {
+    { 
 #if defined(_REENTRANT) && defined(HAVE_XINITTHREADS)
       if ( XPCE_mt == TRUE )
-	XInitThreads();
-      else
-	XPCE_mt = -1;
+      { if ( use_x_init_threads )
+	  XInitThreads();
+      } else
+      { XPCE_mt = -1;
+      }
 #else
 	XPCE_mt = -1;
 #endif
@@ -106,12 +150,30 @@ pceXtAppContext(void * ctx)
       { errorPce(TheDisplayManager(), NAME_noApplicationContext);
 	fail;
       }
+
+      if ( !XtSetLanguageProc(ThePceXtAppContext, NULL, NULL) )
+      { errorPce(TheDisplayManager(), NAME_noLocaleSupport,
+		 CtoName(setlocale(LC_ALL, NULL)));
+	fail;
+      }
     }
   }
 
   return ThePceXtAppContext;
 }
 
+
+status
+X11ThreadsDisplay(DisplayObj d, Bool val)
+{ if ( ThePceXtAppContext )
+    return errorPce(d, NAME_x11Threads);
+
+  use_x_init_threads = (val == ON ? TRUE : FALSE);
+
+  succeed;
+}
+
+
 		 /*******************************
 		 *	 WIDGET REFERENCE	*
 		 *******************************/
@@ -225,7 +287,7 @@ shift_for_mask(unsigned long mask)
 
 
 status
-postscriptXImage(XImage *im,
+postscriptXImage(XImage *im, XImage *mask,
 		 int fx, int fy, int w, int h,
 		 Display *disp,
 		 Colormap cmap,
@@ -283,7 +345,11 @@ postscriptXImage(XImage *im,
     { for(x = fx; x < w8; x++)
       { int pixval;
 	
-	pixval = (x < w ? psmap[XGetPixel(im, x, y)] : psbright);
+	if ( mask && XGetPixel(mask, x, y) == 0L )
+	  pixval = psbright;
+	else
+	  pixval = (x < w ? psmap[XGetPixel(im, x, y)] : psbright);
+
 	put_value(&stat, pixval);
       }
     } else
@@ -300,32 +366,42 @@ postscriptXImage(XImage *im,
 	int pixval;
 	int r, g, b;
 
-	pixel = XGetPixel(im, x, y);
-	r = (pixel & im->red_mask)   >> r_shift;
-	g = (pixel & im->green_mask) >> g_shift;
-	b = (pixel & im->blue_mask)  >> b_shift;
-	DEBUG(NAME_image, Cprintf(" %x/%x/%x", r, g, b));
-
-	if ( depth == 1 )
-	{ if ( r+g+b > (r_bright+g_bright+b_bright)/2 )
-	    pixval = 1;
-	  else
-	    pixval = 0;
-
-	  put_value(&stat, pixval);
-	} else
-	{ r = rescale(r, r_bright, psbright);
-	  g = rescale(g, g_bright, psbright);
-	  b = rescale(b, b_bright, psbright);
-
-	  if ( iscolor )
-	  { put_value(&stat, r);
-	    put_value(&stat, g);
-	    put_value(&stat, b);
+	if ( mask && XGetPixel(mask, x, y) == 0L )
+	{ if ( iscolor )
+	  { put_value(&stat, psbright);
+	    put_value(&stat, psbright);
+	    put_value(&stat, psbright);
 	  } else
-	  { pixval = (x < w ? INTENSITY(r, g, b) : psbright);
-	  
+	  { put_value(&stat, psbright);
+	  }
+	} else
+	{ pixel = XGetPixel(im, x, y);
+	  r = (pixel & im->red_mask)   >> r_shift;
+	  g = (pixel & im->green_mask) >> g_shift;
+	  b = (pixel & im->blue_mask)  >> b_shift;
+	  DEBUG(NAME_image, Cprintf(" %x/%x/%x", r, g, b));
+  
+	  if ( depth == 1 )
+	  { if ( r+g+b > (r_bright+g_bright+b_bright)/2 )
+	      pixval = 1;
+	    else
+	      pixval = 0;
+  
 	    put_value(&stat, pixval);
+	  } else
+	  { r = rescale(r, r_bright, psbright);
+	    g = rescale(g, g_bright, psbright);
+	    b = rescale(b, b_bright, psbright);
+  
+	    if ( iscolor )
+	    { put_value(&stat, r);
+	      put_value(&stat, g);
+	      put_value(&stat, b);
+	    } else
+	    { pixval = (x < w ? INTENSITY(r, g, b) : psbright);
+	    
+	      put_value(&stat, pixval);
+	    }
 	  }
 	}
       }
@@ -468,32 +544,223 @@ allocNearestColour(Display *display, Colormap map, int depth, Name vt,
   fail;
 }
 
+
 		/********************************
 		*      X-EVENT TRANSLATION	*
 		********************************/
 
-static Any
-keycode_to_name(XEvent *event)
-{ char buf[256];
-  int bytes;
-  KeySym sym;
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This function gets the keyboard event id   from an Xevent. Function keys
+are mapped to XPCE names. Normal keys   must  be mapped to their UNICODE
+character.  This  is  a  bit  complicated.    According   to  the  docs,
+XLookupString only returns ISO-Latin-1 characters.   In practice it also
+appears to return multibyte sequences, especially   UTF-8. It is unclear
+whether all X11 systems do this and whether the output is always UTF-8.
 
-  bytes = XLookupString((XKeyEvent *) event, buf, 256, &sym, NULL);
+Dispite many references, XtSetLanguageProc() and  XLookupString() do not
+ensure proper handling of dead keys (' +   e). Possibly this is due to a
+weird combination of event-related X/Xt  functions   used  in  low level
+XPCE, but even with browsing the X11 sources I could not find it. As the
+long term aim was  to  move   to  XwcLookupString()  anyway, proper -but
+simple- support for XIM has been implemented.
 
-  switch(sym)				/* special ones */
-  { case XK_BackSpace:
-      if ( event->xkey.state & Mod1Mask )
-	return toInt(8+META_OFFSET);
-      return NAME_backspace;
+A good document is:
+
+	http://home.catv.ne.jp/pp/ginoue/im/xim-e.html
+
+One of the issues is the window for an IC. Can/should this always be the
+frame? At the moment it is the  X11 window realising the event receiver.
+Using the frame window for all XIM operations does not work.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#ifdef O_XIM
+static void
+adjustEventMask(Display *d, Window w, XIC ic)
+{ XWindowAttributes winatts;
+  long fevent;
+
+  XGetICValues(ic, XNFilterEvents, &fevent, NULL);
+  XGetWindowAttributes(d, w, &winatts);
+  XSelectInput(d, w, winatts.your_event_mask|fevent);
+}
+
+
+static XIC
+getICWindow(Any obj)
+{ FrameObj fr;
+  FrameWsRef wsfr;
+  DisplayWsXref d;
+  Widget w;
+
+  if ( instanceOfObject(obj, ClassFrame) )
+  { fr = obj;
+    w = widgetFrame(fr);
+    DEBUG(NAME_event, Cprintf("Associating IC with %s ...", pp(fr)));
+  } else if ( instanceOfObject(obj, ClassWindow) )
+  { PceWindow sw = obj;
+    if ( !(fr = getFrameWindow(sw, OFF)) )
+      fail;
+    w = widgetWindow(sw);
+    DEBUG(NAME_event, Cprintf("Associating IC with %s ...", pp(sw)));
+  } else
+    fail;
+
+  if ( !w )
+    fail;				/* possible on MacOS? */
+  d = fr->display->ws_ref;
+
+  if ( (wsfr = fr->ws_ref) && d->im )
+  { XIC ic;
+    Window xwin = XtWindow(w);
+
+    if ( wsfr->ic )
+    { if ( wsfr->icwin != xwin )
+      { XSetICValues(wsfr->ic,
+		     XNClientWindow, xwin,
+		     NULL);
+	wsfr->icwin = xwin;
+	adjustEventMask(d->display_xref, xwin, wsfr->ic);
+	DEBUG(NAME_event, Cprintf("Re-using IC %p (switched window)\n",
+				  wsfr->ic));
+      } else
+      { DEBUG(NAME_event, Cprintf("Re-using IC %p\n", wsfr->ic));
+      }
+
+      return wsfr->ic;
+    }
+
+    ic = XCreateIC(d->im, 
+		   XNInputStyle, XIMPreeditNothing | XIMStatusNothing,
+		   XNClientWindow, xwin,
+		   NULL);
+    if( !ic )
+    { DEBUG(NAME_event, Cprintf("Could not open X Input Context\n"));
+      fail;
+    }
+    adjustEventMask(d->display_xref, xwin, ic);
+
+    DEBUG(NAME_event, Cprintf("Created IC %p\n", ic));
+    wsfr->ic    = ic;
+    wsfr->icwin = xwin;
+
+    return ic;
   }
 
-  if ( bytes == 1 )
-  { int c = buf[0] & 0xff;
+  fail;
+}
 
-    if ( event->xkey.state & Mod1Mask )	/* meta depressed */
-      c += META_OFFSET;
+#endif /*O_XIM*/
 
-    return toInt(c);
+static Any
+keycode_to_name(Any sw, XEvent *event)
+{ char buf[256];
+  int count;
+  KeySym sym;
+#ifdef O_XIM
+  XIC ic = getICWindow(sw);
+
+  if ( ic )
+  { charW wbuf[256];
+    charW *winput = wbuf;
+    Status status;
+    int has_chars = FALSE;
+    int has_sym = FALSE;
+
+  again:
+    count = XwcLookupString(ic, (XKeyPressedEvent*)event,
+			    winput, sizeof(wbuf)/sizeof(wchar_t),
+			    &sym, &status);
+    DEBUG(NAME_event,
+	  { int i;
+	    Cprintf("Got %d wide characters:", count);
+	    for(i=0; i<count; i++)
+	      Cprintf(" %d", wbuf[i]);
+		Cprintf("\n");
+	  });
+
+    switch(status)
+    { case XBufferOverflow:
+	winput = pceMalloc(sizeof(wchar_t)*(count+1));
+        goto again;
+      case XLookupNone:
+	fail;
+      case XLookupKeySym:
+	has_sym = TRUE;
+        break;
+      case XLookupBoth:
+	has_chars = has_sym = TRUE;
+        break;
+      case XLookupChars:
+	has_chars = TRUE;
+        break;
+    }
+
+    if ( has_sym )
+    { switch(sym)
+      { case XK_BackSpace:
+	  if ( event->xkey.state & MetaMask )
+	    return toInt(8+META_OFFSET);
+					/* cannot be a buffer overflow */
+	return NAME_backspace;
+      }
+    }
+
+    if ( count == 1 )
+    { int c = wbuf[0];
+      
+      if ( event->xkey.state & MetaMask )	/* meta depressed */
+	c += META_OFFSET;
+      
+      return toInt(c);
+    }
+
+    if ( !has_sym )
+    { if ( winput != wbuf )		/* TBD: multi-character input */
+	pceFree(winput);
+
+      fail;
+    }
+  } else
+#endif /*O_XIM*/
+  { count = XLookupString((XKeyPressedEvent *) event,
+			  buf, sizeof(buf), &sym, NULL);
+    DEBUG(NAME_event,
+	  { int i;
+	    Cprintf("Got %d characters:", count);
+	    for(i=0; i<count; i++)
+	      Cprintf(" %d", buf[i]&0xff);
+		Cprintf("\n");
+	  });
+
+    switch(sym)				/* special ones */
+    { case XK_BackSpace:
+	if ( event->xkey.state & MetaMask )
+	  return toInt(8+META_OFFSET);
+        return NAME_backspace;
+    }
+
+    if ( count == 1 )
+    { int c = buf[0] & 0xff;
+      
+      if ( event->xkey.state & MetaMask )	/* meta depressed */
+	c += META_OFFSET;
+      
+      return toInt(c);
+    }
+    
+    if ( count > 1 )			/* see above */
+    { char *e;
+      int c;
+
+      if ( (e = utf8_get_char(buf, &c)) && e-buf == count )
+      { DEBUG(NAME_event, Cprintf("\t-->UTF-8 sequence for %d\n", c));
+	
+	if ( event->xkey.state & MetaMask )	/* meta depressed */
+	  c += META_OFFSET;
+
+	return toInt(c);
+      }
+    }
   }
 
   switch(sym)
@@ -586,7 +853,7 @@ state_to_buttons(unsigned int state, Name name)
   if ( state & Button5Mask )	r |= BUTTON_ms_button5;
   if ( state & ShiftMask )	r |= BUTTON_shift;
   if ( state & ControlMask )	r |= BUTTON_control;
-  if ( state & Mod1Mask )	r |= BUTTON_meta;
+  if ( state & MetaMask )	r |= BUTTON_meta;
 
   if ( name == NAME_msLeftDown )
     r |= BUTTON_ms_left;
@@ -648,7 +915,7 @@ CtoEvent(Any window, XEvent *event)	/* window or frame */
       y     = event->xkey.y;
       state = event->xkey.state;
       time  = event->xkey.time;
-      name  = keycode_to_name(event);
+      name  = keycode_to_name(window, event);
       if ( name == FAIL )
         fail;
 
@@ -739,6 +1006,37 @@ CtoEvent(Any window, XEvent *event)	/* window or frame */
 }
 
 
+typedef struct
+{ const char *name;
+  unsigned int mask;
+} modmask;
+
+static const modmask modmasks[] = 
+{ { "mod1", Mod1Mask },
+  { "mod2", Mod2Mask },
+  { "mod3", Mod3Mask },
+  { "mod4", Mod4Mask },
+  { "mod5", Mod5Mask },
+  { NULL, 0 }
+};
+
+
+status
+metaModifierDisplay(DisplayObj d, Name name)
+{ const char *s = strName(name);
+  const modmask *mm = modmasks;
+
+  for( ; mm->name; mm++)
+  { if ( streq(s, mm->name) )
+    { MetaMask = mm->mask;
+      succeed;
+    }
+  }
+  
+  fail;
+}
+
+
 		 /*******************************
 		 *      COMPATIBILITY HACKS	*
 		 *******************************/
