@@ -1,4 +1,4 @@
-/*  $Id: pl-atom.c,v 1.65 2004/09/21 14:22:25 jan Exp $
+/*  $Id: pl-atom.c,v 1.69 2006/07/31 12:10:50 jan Exp $
 
     Part of SWI-Prolog
 
@@ -182,7 +182,7 @@ static PL_blob_t unregistered_blob_atom =
 };
 
 
-static void
+void
 PL_register_blob_type(PL_blob_t *type)
 { PL_LOCK(L_MISC);			/* cannot use L_ATOM */
 
@@ -209,6 +209,22 @@ PL_register_blob_type(PL_blob_t *type)
 }
 
 
+PL_blob_t *
+PL_find_blob_type(const char *name)
+{ PL_blob_t *t;
+
+  PL_LOCK(L_MISC);
+  for(t = GD->atoms.types; t; t = t->next)
+  { if ( streq(name, t->name) )
+      break;
+  }
+  PL_UNLOCK(L_MISC);
+
+  return t;
+}
+
+
+
 int
 PL_unregister_blob_type(PL_blob_t *type)
 { unsigned int i;
@@ -951,7 +967,7 @@ typedef struct match
 
 
 static bool
-allAlpha(register char *s)
+allAlpha(const char *s)
 { for( ; *s; s++)
   { if ( !isAlpha(*s) )
       fail;
@@ -1098,8 +1114,38 @@ static pthread_key_t key;
 
 #define is_signalled() (LD && LD->pending_signals != 0)
 
-char *
-PL_atom_generator(const char *prefix, int state)
+static int
+alnum_text(PL_chars_t *txt)
+{ switch(txt->encoding)
+  { case ENC_ISO_LATIN_1:
+    { const unsigned char *s = (const unsigned char *)txt->text.t;
+      const unsigned char *e = &s[txt->length];
+
+      for(; s<e; s++)
+      { if ( !isAlpha(*s) )
+	  return FALSE;
+      }
+      return TRUE;
+    }
+    case ENC_WCHAR:
+    { const pl_wchar_t *s = (const pl_wchar_t*)txt->text.w;
+      const pl_wchar_t *e = &s[txt->length];
+
+      for(; s<e; s++)
+      { if ( !isAlphaW(*s) )
+	  return FALSE;
+      }
+      return TRUE;
+    }
+    default:
+      assert(0);
+      return FALSE;
+  }
+}
+
+
+static int
+atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 { GET_LD
   long i, mx = entriesBuffer(&atom_array, Atom);
 
@@ -1128,22 +1174,78 @@ PL_atom_generator(const char *prefix, int state)
     if ( is_signalled() )		/* Notably allow windows version */
       PL_handle_signals();		/* to break out on ^C */
 
-    if ( strprefix(a->name, prefix) &&
-	 allAlpha(a->name) &&
-	 strlen(a->name) < ALT_SIZ )
+    if ( get_atom_ptr_text(a, hit) &&
+	 hit->length < ALT_SIZ &&
+	 PL_cmp_text(prefix, 0, hit, 0, prefix->length) == 0 &&
+	 alnum_text(hit) )
     {
 #ifdef O_PLMT
       pthread_setspecific(key, (void *)(i+1));
 #else   
       LD->atoms.generator = i+1;
 #endif
-      return a->name;
+  
+      return TRUE;
+    }
+  }
+
+  return FALSE;
+}
+
+
+char *
+PL_atom_generator(const char *prefix, int state)
+{ PL_chars_t txt, hit;
+
+  PL_init_text(&txt);
+  txt.text.t   = (char *)prefix;
+  txt.encoding = ENC_ISO_LATIN_1;
+  txt.length   = strlen(prefix);
+
+  while ( atom_generator(&txt, &hit, state) )
+  { if ( hit.encoding == ENC_ISO_LATIN_1 )
+      return hit.text.t;		/* text is from atoms, thus static */
+    state = TRUE;
+  }
+
+  return NULL;
+}
+
+
+pl_wchar_t *
+PL_atom_generator_w(const pl_wchar_t *prefix,
+		    pl_wchar_t *buffer,
+		    unsigned int buflen,
+		    int state)
+{ PL_chars_t txt, hit;
+
+  PL_init_text(&txt);
+  txt.text.w   = (pl_wchar_t *)prefix;
+  txt.encoding = ENC_WCHAR;
+  txt.length   = wcslen(prefix);
+
+  for( ; atom_generator(&txt, &hit, state); state = TRUE )
+  { if ( buflen > hit.length+1 )
+    { if ( hit.encoding == ENC_WCHAR )
+      { wcscpy(buffer, hit.text.w);
+      } else
+      { const unsigned char *s = (const unsigned char *)hit.text.t;
+	const unsigned char *e = &s[hit.length];
+	pl_wchar_t *o;
+	
+	for(o=buffer; s<e;)
+	  *o++ = *s++;
+	*o = EOS;
+      }
+
+      return buffer;
     }
   }
 
   return NULL;
 }
 
+
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
