@@ -1,4 +1,4 @@
-/*  $Id: pl-error.c,v 1.31 2004/09/21 14:22:25 jan Exp $
+/*  $Id: pl-error.c,v 1.35 2006/07/27 09:22:42 jan Exp $
 
     Part of SWI-Prolog
 
@@ -48,12 +48,18 @@ put_name_arity(term_t t, functor_t f)
 
 int
 PL_error(const char *pred, int arity, const char *msg, int id, ...)
-{ term_t except, formal, swi;
+{ Definition caller;
+  term_t except, formal, swi;
   va_list args;
   int do_throw = FALSE;
   fid_t fid;
   int rc;
 
+  if ( environment_frame )
+    caller = environment_frame->predicate;
+  else
+    caller = NULL;
+
   if ( id == ERR_FILE_OPERATION && !fileerrors )
     fail;
 
@@ -153,8 +159,12 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     }
     case ERR_UNDEFINED_PROC:
     { Definition def = va_arg(args, Definition);
+      Definition clr = va_arg(args, Definition);
       term_t pred = PL_new_term_ref();
 
+      if ( clr )
+	caller = clr;
+
       unify_definition(pred, def, 0, GP_NAMEARITY);
       PL_unify_term(formal,
 		    PL_FUNCTOR, FUNCTOR_existence_error2,
@@ -185,7 +195,7 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 		      PL_ATOM, ATOM_procedure,
 		      PL_FUNCTOR, FUNCTOR_divide2,
 		        PL_CHARS, name,
-		        PL_INTEGER, arity);
+		        PL_INT, arity);
       break;
     }
     case ERR_FAILED:
@@ -445,7 +455,7 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
   va_end(args);
 
 					/* build SWI-Prolog context term */
-  if ( pred || msg )
+  if ( pred || msg || caller )
   { term_t predterm = PL_new_term_ref();
     term_t msgterm  = PL_new_term_ref();
 
@@ -454,7 +464,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 		    PL_FUNCTOR, FUNCTOR_divide2,
 		      PL_CHARS, pred,
 		      PL_INT, arity);
+    } else if ( caller )
+    { unify_definition(predterm, caller, 0, GP_NAMEARITY);
     }
+
     if ( msg )
     { PL_put_atom_chars(msgterm, msg);
     }
@@ -531,28 +544,23 @@ printMessage(atom_t severity, ...)
 
 int
 PL_get_nchars_ex(term_t t, unsigned int *len, char **s, unsigned int flags)
-{ atom_t expected;
-
-  if ( PL_get_nchars(t, len, s, flags) )
-    return TRUE;
-
-  if ( flags & CVT_LIST )
-    expected = ATOM_text;
-  else if ( flags & CVT_NUMBER )
-    expected = ATOM_atomic;
-  else
-    expected = ATOM_atom;
-
-  return PL_error(NULL, 0, NULL, ERR_TYPE, expected, t);
+{ return PL_get_nchars(t, len, s, flags|CVT_EXCEPTION);
 }
 
 
 int
 PL_get_chars_ex(term_t t, char **s, unsigned int flags)
-{ return PL_get_nchars_ex(t, NULL, s, flags);
+{ return PL_get_nchars(t, NULL, s, flags|CVT_EXCEPTION);
 }
 
 
+int
+PL_get_text_ex(term_t t, PL_chars_t *txt, unsigned int flags)
+{ return PL_get_text(t, txt, flags);
+}
+
+
+
 int
 PL_get_atom_ex(term_t t, atom_t *a)
 { if ( PL_get_atom(t, a) )
