@@ -1,4 +1,4 @@
-/*  $Id: pl-ext.c,v 1.134 2004/09/30 10:41:36 jan Exp $
+/*  $Id: pl-ext.c,v 1.151 2006/08/27 12:31:14 jan Exp $
 
     Part of SWI-Prolog
 
@@ -69,17 +69,6 @@ static const PL_extension foreigns[] = {
   FRG("crash",			0, pl_crash,			0),
 #endif
   FRG("nl",			0, pl_nl,			0),
-  FRG("put",			1, pl_put,			0),
-  FRG("put_byte",		1, pl_put,			0),
-  FRG("put_code",		1, pl_put,			0),
-  FRG("put_char",		1, pl_put,			0),
-  FRG("get0",			1, pl_get_byte,			0),
-  FRG("get_byte",		1, pl_get_byte,			0),
-  FRG("get_code",		1, pl_get_byte,			0),
-  FRG("get_char",		1, pl_get_char,			0),
-  FRG("get",			1, pl_get,			0),
-  FRG("skip",			1, pl_skip,			0),
-  FRG("skip",			2, pl_skip2,			0),
   FRG("get_single_char",	1, pl_get_single_char,		0),
   FRG("$push_input_context",	0, pl_push_input_context,	0),
   FRG("$pop_input_context",	0, pl_pop_input_context,	0),
@@ -117,10 +106,6 @@ static const PL_extension foreigns[] = {
 #ifdef __WIN32__
   FRG("win_exec",		2, pl_win_exec,			0),
   FRG("win_module_file",	2, pl_win_module_file,		0),
-  FRG("win_registry_get_value", 3, pl_get_registry_value,	0),
-#ifdef O_XOS
-  FRG("make_fat_filemap",	1, pl_make_fat_filemap,		0),
-#endif
 #endif
 
   FRG("$on_signal",		4, pl_on_signal,	     META),
@@ -136,8 +121,6 @@ static const PL_extension foreigns[] = {
   FRG("wildcard_match",		2, pl_wildcard_match,		0),
   FRG("$apropos_match",		2, pl_apropos_match,		0),
   FRG("sub_atom",		5, pl_sub_atom,		     NDET),
-  FRG("convert_time",		2, pl_convert_time2,		0),
-  FRG("convert_time",		8, pl_convert_time,		0),
   FRG("sleep",			1, pl_sleep,			0),
   FRG("break",			0, pl_break,			0),
   FRG("$break",			1, pl_break1,			0),
@@ -177,12 +160,8 @@ static const PL_extension foreigns[] = {
   FRG("number_chars",		2, pl_number_chars,		0),
   FRG("number_codes",		2, pl_number_codes,		0),
   FRG("char_code",		2, pl_char_code,		0),
-  FRG("char_type",		2, pl_char_type,	     NDET),
-  FRG("code_type",		2, pl_code_type,	     NDET),
   FRG("downcase_atom",		2, pl_downcase_atom,		0),
   FRG("upcase_atom",		2, pl_upcase_atom,		0),
-  FRG("int_to_atom",		3, pl_int_to_atom,		0),
-  FRG("atom_prefix",		2, pl_atom_prefix,		0),
   FRG("atom_concat",		3, pl_atom_concat,	     NDET),
   FRG("$concat_atom",		2, pl_concat_atom,		0),
   FRG("concat_atom",		3, pl_concat_atom3,		0),
@@ -266,10 +245,6 @@ static const PL_extension foreigns[] = {
   FRG("$fail",			0, pl_fail,		  NOTRACE),
   FRG("abort",			0, pl_abort,			0),
 
-  FRG("between",		3, pl_between,		     NDET),
-  FRG("succ",			2, pl_succ,			0),
-  FRG("plus",			3, pl_plus,			0),
-
   FRG("trace",			0, pl_trace,		  NOTRACE),
   FRG("notrace",		0, pl_notrace,		  NOTRACE),
   FRG("tracing",		0, pl_tracing,		  NOTRACE),
@@ -288,7 +263,6 @@ static const PL_extension foreigns[] = {
   FRG("prolog_choice_attribute",3, pl_prolog_choice_attribute,	0),
   FRG("prolog_skip_level",	2, pl_skip_level,	  NOTRACE),
 
-  FRG("$write_on_string",	2, pl_write_on_string,		0),
   FRG("dwim_match",		3, pl_dwim_match,		0),
   FRG("$dwim_predicate",	2, pl_dwim_predicate,	     NDET),
 
@@ -376,30 +350,12 @@ static const PL_extension foreigns[] = {
   FRG("set_stream",		2, pl_set_stream,		0),
   FRG("current_input",		1, pl_current_input,		0),
   FRG("current_output",		1, pl_current_output,		0),
-  FRG("character_count",	2, pl_character_count,		0),
-  FRG("line_count",		2, pl_line_count,		0),
-  FRG("line_position",		2, pl_line_position,		0),
   FRG("source_location",	2, pl_source_location,		0),
   FRG("at_end_of_stream",	1, pl_at_end_of_stream1,	0),
   FRG("at_end_of_stream",	0, pl_at_end_of_stream0,	0),
-  FRG("peek_byte",		2, pl_peek_byte2,		0),
-  FRG("peek_byte",		1, pl_peek_byte1,		0),
-  FRG("peek_code",		2, pl_peek_code2,		0),
-  FRG("peek_code",		1, pl_peek_code1,		0),
-  FRG("peek_char",		2, pl_peek_char2,		0),
-  FRG("peek_char",		1, pl_peek_char1,		0),
 
   FRG("nl",			1, pl_nl1,			0),
   FRG("tab",			2, pl_tab2,			0),
-  FRG("put",			2, pl_put2,			0),
-  FRG("put_byte",		2, pl_put2,			0),
-  FRG("put_code",		2, pl_put2,			0),
-  FRG("put_char",		2, pl_put2,			0),
-  FRG("get",			2, pl_get2,			0),
-  FRG("get0",			2, pl_get_byte2,		0),
-  FRG("get_byte",		2, pl_get_byte2,		0),
-  FRG("get_code",		2, pl_get_byte2,		0),
-  FRG("get_char",		2, pl_get_char2,		0),
   FRG("read",			2, pl_read2,			0),
   FRG("write",			2, pl_write2,			0),
   FRG("writeq",			2, pl_writeq2,			0),
@@ -466,6 +422,7 @@ predicates, the extensions will be registered.
 
 struct extension_cell
 { PL_extension *extensions;
+  char *module;
   ExtensionCell next;
 };
 
@@ -473,63 +430,52 @@ struct extension_cell
 #define ext_tail		(GD->foreign._ext_tail)
 #define extensions_loaded	(GD->foreign._loaded)
 
-static void
-bindExtensions(const PL_extension *e)
-{ Definition def;
+static char *
+dupStr(const char *str)
+{ if (str)
+  { int len = strlen(str)+1;
+    char *m = PL_malloc(len);
+    memcpy(m, str, len);
+    return m;
+  }
+  return NULL;
+}
 
-  for(; e->predicate_name; e++)
-  { unsigned long flags = TRACE_ME;
-    Module m;
-    atom_t name;
-    char *s;
 
-    for(s=e->predicate_name; isAlpha(*s); s++)
-      ;
+static PL_extension *
+dupExtensions(const PL_extension *e)
+{ int i;
+  PL_extension *dup, *o;
+  int len = 0;
 
-    if ( *s == ':' )			/* module:predicate */
-    { m = PL_new_module(PL_new_atom_nchars(s-e->predicate_name,
-					   e->predicate_name));
-      name = PL_new_atom(s+1);
-    } else
-    { name = PL_new_atom(e->predicate_name);
-      m = (environment_frame ? contextModule(environment_frame)
-	   		     : MODULE_user);
-    }
+  while(e[len++].predicate_name)
+    ;
+  o = dup = PL_malloc(len*sizeof(*e));
 
-    if ( e->flags & PL_FA_NOTRACE )	     flags &= ~TRACE_ME;
-    if ( e->flags & PL_FA_TRANSPARENT )	     flags |= METAPRED;
-    if ( e->flags & PL_FA_NONDETERMINISTIC ) flags |= NONDETERMINISTIC;
-    if ( e->flags & PL_FA_VARARGS )	     flags |= P_VARARG;
-    if ( e->flags & PL_FA_CREF )	     flags |= P_FOREIGN_CREF;
+  for ( i=0; i<len; i++, o++, e++)
+  { o->predicate_name = dupStr(e->predicate_name);
+    o->arity = e->arity;
+    o->function = e->function;
+    o->flags = e->flags;
+  }
 
-    def = lookupProcedure(lookupFunctorDef(name, e->arity), m)->definition;
-    PL_unregister_atom(name);
-    set(def, FOREIGN);
-    if ( m == MODULE_system )
-      set(def, SYSTEM|HIDE_CHILDS);
-    set(def, flags);
-    def->definition.function = e->function;
-    def->indexPattern = 0;
-    def->indexCardinality = 0;
-  }    
+  return dup;
 }
 
 
 void
-PL_register_extensions(const PL_extension *e)
-{ if ( extensions_loaded )
-    bindExtensions(e);
-  else
-  { ExtensionCell cell = malloc(sizeof *cell);
+rememberExtensions(const char *module, const PL_extension *e)
+{ ExtensionCell cell = PL_malloc(sizeof *cell);
 
-    cell->extensions = (PL_extension *) e;
-    cell->next = NULL;
-    if ( ext_tail )
-    { ext_tail->next = cell;
-      ext_tail = cell;
-    } else
-    { ext_head = ext_tail = cell;
-    }
+  cell->extensions = dupExtensions(e);
+  cell->next = NULL;
+  cell->module = dupStr(module);
+  
+  if ( ext_tail )
+  { ext_tail->next = cell;
+    ext_tail = cell;
+  } else
+  { ext_head = ext_tail = cell;
   }
 }
 
@@ -540,7 +486,19 @@ cleanupExtensions(void)
 
   for(c=ext_head; c; c=next)
   { next = c->next;
-    free(c);
+    if (c->module)
+      PL_free(c->module);
+
+    if (c->extensions)
+    { PL_extension *e = c->extensions;
+
+      for(;e->predicate_name; e++)
+	PL_free(e->predicate_name);
+
+      PL_free(c->extensions);
+    }
+
+    PL_free(c);
   }
 
   ext_head = ext_tail = NULL;
@@ -600,6 +558,8 @@ DECL_PLIST(gvar);
 DECL_PLIST(win);
 DECL_PLIST(file);
 DECL_PLIST(btree);
+DECL_PLIST(ctype);
+DECL_PLIST(tai);
 
 void
 initBuildIns(void)
@@ -622,6 +582,8 @@ initBuildIns(void)
   REG_PLIST(wic);
   REG_PLIST(file);
   REG_PLIST(btree);
+  REG_PLIST(ctype);
+  REG_PLIST(tai);
 #ifdef O_ATTVAR
   REG_PLIST(attvar);
 #endif
@@ -640,15 +602,18 @@ initBuildIns(void)
   PROCEDURE_print_message2   = lookupProcedure(FUNCTOR_print_message2, 	  m);
   PROCEDURE_dcall1	     = lookupProcedure(FUNCTOR_dcall1,		  m);
   PROCEDURE_call_cleanup3    = lookupProcedure(FUNCTOR_call_cleanup3,	  m); 
+  PROCEDURE_dthread_init0    = lookupProcedure(FUNCTOR_dthread_init0,	  m);
 #ifdef O_ATTVAR
   PROCEDURE_dwakeup1	     = lookupProcedure(FUNCTOR_dwakeup1,	  m);
 #endif
+  PROCEDURE_exception_hook4  = 
+	PL_predicate("prolog_exception_hook", 4, "user");
 					/* allow debugging in call/1 */
   clear(PROCEDURE_dcall1->definition, HIDE_CHILDS);
   set(PROCEDURE_dcall1->definition, DYNAMIC);
 
   for( ecell = ext_head; ecell; ecell = ecell->next )
-    bindExtensions(ecell->extensions);
+    bindExtensions(ecell->module, ecell->extensions);
 
   extensions_loaded = TRUE;
 }
