@@ -1,4 +1,4 @@
-/*  $Id: pl-feature.c,v 1.37 2004/09/21 14:22:25 jan Exp $
+/*  $Id: pl-feature.c,v 1.48 2006/07/25 08:55:54 jan Exp $
 
     Part of SWI-Prolog
 
@@ -25,6 +25,10 @@
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "pl-ctype.h"
+#include <ctype.h>
+#ifdef __WIN32__
+#include <process.h>			/* getpid() */
+#endif
 
 #define LOCK()   PL_LOCK(L_FEATURE)
 #define UNLOCK() PL_UNLOCK(L_FEATURE)
@@ -63,13 +67,14 @@ want to be able to have a lot and don't harm thread_create/3 too much.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void	setArgvFeature();
+static void	setTZFeature();
 
 typedef struct _feature
 { short		flags;			/* Type | Flags */
   short		index;			/* index in _FEATURE mask */
   union
   { atom_t	a;			/* value as atom */
-    long	i;			/* value as integer */
+    int64_t	i;			/* value as integer */
     record_t	t;			/* value as term */
   } value;
 } feature;
@@ -110,6 +115,9 @@ defFeature(const char *name, int flags, ...)
 
   initFeatureTable();
 
+  if ( type == FT_INT64 )
+    flags = (flags & ~FT_MASK)|FT_INTEGER;
+
   if ( (s = lookupHTable(GD->feature.table, (void *)an)) )
   { f = s->value;
     assert((f->flags & FT_MASK) == (flags & FT_MASK));
@@ -152,6 +160,11 @@ defFeature(const char *name, int flags, ...)
       f->value.i = val;
       break;
     }
+    case FT_INT64:
+    { int64_t val = va_arg(args, int64_t);
+      f->value.i = val;
+      break;
+    }
     case FT_ATOM:
     { const char *text = va_arg(args, const char *);
 
@@ -263,8 +276,25 @@ setWriteAttributes(atom_t a)
 }
 
 
-word
-pl_set_feature(term_t key, term_t value)
+static int
+setEncoding(atom_t a)
+{ IOENC enc = atom_to_encoding(a);
+
+  if ( enc == ENC_UNKNOWN )
+  { term_t value = PL_new_term_ref();
+
+    PL_put_atom(value, a);
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_encoding, value);
+  }
+
+  LD->encoding = enc;
+
+  succeed;
+}
+
+
+static word
+set_feature_unlocked(term_t key, term_t value)
 { atom_t k;
   Symbol s;
   feature *f;
@@ -275,7 +305,6 @@ pl_set_feature(term_t key, term_t value)
   if ( !PL_get_atom(key, &k) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, key);
 
-  LOCK();
 					/* set existing feature */
 #ifdef O_PLMT
   if ( LD->feature.table &&
@@ -343,7 +372,6 @@ pl_set_feature(term_t key, term_t value)
 #endif
       addHTable(GD->feature.table, (void *)k, f);
 
-    UNLOCK();
     succeed;
   }
 
@@ -352,7 +380,8 @@ pl_set_feature(term_t key, term_t value)
     { int val;
 
       if ( !PL_get_bool(value, &val) )
-	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_bool, value);
+      { return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_bool, value);
+      }
       if ( f->index > 0 )
       { unsigned long mask = 1L << (f->index-1);
 
@@ -391,28 +420,33 @@ pl_set_feature(term_t key, term_t value)
 
       if ( !PL_get_atom(value, &a) )
 	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, value);
+
+      if ( k == ATOM_double_quotes )
+      { rval = setDoubleQuotes(a, &m->flags);
+      } else if ( k == ATOM_unknown )
+      { rval = setUnknown(a, &m->flags);
+      } else if ( k == ATOM_write_attributes )
+      { rval = setWriteAttributes(a);
+      } else if ( k == ATOM_encoding )
+      { rval = setEncoding(a);
+      }
+      if ( !rval )
+	fail;
+
       PL_unregister_atom(f->value.a);
       f->value.a = a;
       PL_register_atom(a);
       if ( k == ATOM_float_format )
       { PL_register_atom(a);		/* so it will never be lost! */
 	LD->float_format = PL_atom_chars(a);
-      } else if ( k == ATOM_double_quotes )
-      { rval = setDoubleQuotes(a, &m->flags);
-      } else if ( k == ATOM_unknown )
-      { rval = setUnknown(a, &m->flags);
-      } else if ( k == ATOM_write_attributes )
-      { rval = setWriteAttributes(a);
-      }
+      } 
       break;
     }
     case FT_INTEGER:
     { long i;
 
       if ( !PL_get_long(value, &i) )
-      { rval = PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_integer, value);
-	break;
-      }
+	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_integer, value);
       f->value.i = i;
 #ifdef O_ATOMGC
       if ( k == ATOM_agc_margin )
@@ -429,12 +463,23 @@ pl_set_feature(term_t key, term_t value)
     default:
       assert(0);
   }
-  UNLOCK();
 
   return rval;
 }
 
 
+word
+pl_set_feature(term_t key, term_t value)
+{ word rc;
+
+  LOCK();
+  rc = set_feature_unlocked(key, value);
+  UNLOCK();
+
+  return rc;
+}
+
+
 static int
 unify_feature_value(Module m, atom_t key, feature *f, term_t val)
 { if ( key == ATOM_character_escapes )
@@ -486,7 +531,7 @@ unify_feature_value(Module m, atom_t key, feature *f, term_t val)
     case FT_ATOM:
       return PL_unify_atom(val, f->value.a);
     case FT_INTEGER:
-      return PL_unify_integer(val, f->value.i);
+      return PL_unify_int64(val, f->value.i);
     case FT_TERM:
     { term_t tmp = PL_new_term_ref();
 
@@ -681,8 +726,6 @@ initFeatures()
 { defFeature("iso",  FT_BOOL, FALSE, ISO_FEATURE);
   defFeature("arch", FT_ATOM|FF_READONLY, ARCH);
 #if __WIN32__
-  if ( iswin32s() )
-    defFeature("win32s", FT_BOOL|FF_READONLY, TRUE, 0);
   defFeature("windows",	FT_BOOL|FF_READONLY, TRUE, 0);
 #endif
   defFeature("version",	FT_INTEGER|FF_READONLY, PLVERSION);
@@ -690,7 +733,7 @@ initFeatures()
     defFeature("home", FT_ATOM|FF_READONLY, systemDefaults.home);
   if ( GD->paths.executable )
     defFeature("executable", FT_ATOM|FF_READONLY, GD->paths.executable);
-#ifdef HAVE_GETPID
+#if defined(HAVE_GETPID) || defined(EMULATE_GETPID)
   defFeature("pid", FT_INTEGER|FF_READONLY, getpid());
 #endif
   defFeature("optimise", FT_BOOL, GD->cmdline.optimise, OPTIMISE_FEATURE);
@@ -751,11 +794,20 @@ initFeatures()
   defFeature("editor",		   FT_ATOM, "$EDITOR");
   defFeature("debugger_show_context", FT_BOOL, FALSE, 0);
   defFeature("autoload",  FT_BOOL, TRUE,  AUTOLOAD_FEATURE);
-  defFeature("max_integer",	   FT_INTEGER|FF_READONLY, PLMAXINT);
-  defFeature("min_integer",	   FT_INTEGER|FF_READONLY, PLMININT);
+#ifndef O_GMP
+  defFeature("max_integer",	   FT_INT64|FF_READONLY, PLMAXINT);
+  defFeature("min_integer",	   FT_INT64|FF_READONLY, PLMININT);
+#endif
   defFeature("max_tagged_integer", FT_INTEGER|FF_READONLY, PLMAXTAGGEDINT);
   defFeature("min_tagged_integer", FT_INTEGER|FF_READONLY, PLMINTAGGEDINT);
+#ifdef O_GMP
+  defFeature("bounded",		   FT_BOOL|FF_READONLY,	   FALSE, 0);
+#ifdef __GNU_MP__
+  defFeature("gmp_version",	   FT_INTEGER|FF_READONLY, __GNU_MP__);
+#endif
+#else
   defFeature("bounded",		   FT_BOOL|FF_READONLY,	   TRUE, 0);
+#endif
   if ( (-3 / 2) == -2 )
     defFeature("integer_rounding_function", FT_ATOM|FF_READONLY, "down");
   else
@@ -771,6 +823,7 @@ initFeatures()
   defFeature("unknown", FT_ATOM, "error");
   defFeature("debug", FT_BOOL, FALSE, 0);
   defFeature("verbose", FT_ATOM|FF_KEEP, GD->options.silent ? "silent" : "normal");
+  defFeature("verbose_load", FT_BOOL, TRUE, 0);
   defFeature("allow_variable_name_as_functor", FT_BOOL, FALSE,
 	     ALLOW_VARNAME_FUNCTOR);
   defFeature("toplevel_var_size", FT_INTEGER, 1000);
@@ -780,6 +833,8 @@ initFeatures()
   defFeature("unix", FT_BOOL|FF_READONLY, TRUE, 0);
 #endif
 
+  defFeature("encoding", FT_ATOM, stringAtom(encoding_to_atom(LD->encoding)));
+
   defFeature("tty_control", FT_BOOL|FF_READONLY,
 	     trueFeature(TTY_CONTROL_FEATURE), TTY_CONTROL_FEATURE);
   defFeature("signals", FT_BOOL|FF_READONLY,
@@ -798,7 +853,22 @@ initFeatures()
   }
 #endif
 
+#if defined(HAVE_CONFSTR) && defined(_CS_GNU_LIBPTHREAD_VERSION)
+  { char buf[100];
+
+    if ( confstr(_CS_GNU_LIBPTHREAD_VERSION, buf, sizeof(buf)) )
+    { char *p;
+
+      for(p=buf; *p; p++)
+	*p = tolower(*p);
+
+      defFeature("gnu_libpthread_version", FT_ATOM|FF_READONLY, buf);
+    }
+  }
+#endif
+
   setArgvFeature();
+  setTZFeature();
 }
 
 
@@ -812,9 +882,18 @@ setArgvFeature()
 
   PL_put_nil(l);
   for(n=argc-1; n>= 0; n--)
-  { PL_put_atom_chars(e, argv[n]);
+  { PL_put_variable(e);
+    PL_unify_chars(e, PL_ATOM|REP_FN, -1, argv[n]);
     PL_cons_list(l, e, l);
   }
 
   defFeature("argv", FT_TERM, l);
 }
+
+
+static void
+setTZFeature()
+{ tzset();
+
+  defFeature("timezone", FT_INTEGER|FF_READONLY, timezone);
+}
