@@ -1,4 +1,4 @@
-/*  $Id: pl-fli.c,v 1.119.2.2 2005/01/17 21:13:05 jan Exp $
+/*  $Id: pl-fli.c,v 1.164 2006/01/20 13:07:48 jan Exp $
 
     Part of SWI-Prolog
 
@@ -26,8 +26,15 @@
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "pl-ctype.h"
+#include "pl-utf8.h"
 #include <errno.h>
 
+#include <limits.h>
+#if !defined(LLONG_MAX)
+#define LLONG_MAX 9223372036854775807LL
+#define LLONG_MIN (-LLONG_MAX - 1LL)
+#endif
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 SWI-Prolog  new-style  foreign-language  interface.   This  new  foreign
 interface is a mix of the old  interface using the ideas on term-handles
@@ -284,26 +291,12 @@ PL_new_atom_nchars(unsigned int len, const char *s)
 }
 
 
-const char *
-PL_atom_chars(atom_t a)
-{ return (const char *) stringAtom(a);
-}
-
-
-const char *
-PL_atom_nchars(atom_t a, unsigned int *len)
-{ Atom x = atomValue(a);
-
-  if ( len )
-    *len = x->length;
-
-  return x->name;
-}
-
-
 functor_t
 PL_new_functor(atom_t f,  int a)
-{ return lookupFunctorDef(f, a);
+{ if ( !GD->initialised )
+    initFunctors();
+
+  return lookupFunctorDef(f, a);
 }
 
 
@@ -319,6 +312,246 @@ PL_functor_arity(functor_t f)
 }
 
 
+		 /*******************************
+		 *    WIDE CHARACTER SUPPORT	*
+		 *******************************/
+
+static int	compareUCSAtom(atom_t h1, atom_t h2);
+static int	saveUCSAtom(atom_t a, IOSTREAM *fd);
+static atom_t	loadUCSAtom(IOSTREAM *fd);
+
+static PL_blob_t ucs_atom =
+{ PL_BLOB_MAGIC,
+  PL_BLOB_UNIQUE|PL_BLOB_TEXT|PL_BLOB_WCHAR,
+					/* unique representation of text */
+  "ucs_text",
+  NULL,					/* release */
+  compareUCSAtom,			/* compare */
+  writeUCSAtom,				/* write */
+  NULL,					/* acquire */
+  saveUCSAtom,				/* save load to/from .qlf files */
+  loadUCSAtom
+};
+
+
+static void
+initUCSAtoms()
+{ PL_register_blob_type(&ucs_atom);
+}
+
+
+int
+isUCSAtom(Atom a)
+{ return a->type == &ucs_atom;
+}
+
+
+atom_t
+lookupUCSAtom(const pl_wchar_t *s, unsigned int len)
+{ int new;
+
+  return lookupBlob((const char *)s, len*sizeof(pl_wchar_t),
+		    &ucs_atom, &new);
+}
+
+
+atom_t
+PL_new_atom_wchars(unsigned int len, const wchar_t *s)
+{ PL_chars_t txt;
+
+  if ( !GD->initialised )
+    initAtoms();
+
+  txt.text.w    = (wchar_t*)s;
+  txt.length    = len;
+  txt.encoding  = ENC_WCHAR;
+  txt.storage   = PL_CHARS_HEAP;
+  txt.canonical = FALSE;
+
+  return textToAtom(&txt);
+}
+
+
+int
+get_atom_ptr_text(Atom a, PL_chars_t *text)
+{ if ( false(a->type, PL_BLOB_TEXT) )
+    fail;				/* non-textual atom */
+  if ( a->type == &ucs_atom )
+  { text->text.w   = (pl_wchar_t *) a->name;
+    text->length   = a->length / sizeof(pl_wchar_t);
+    text->encoding = ENC_WCHAR;
+  } else
+  { text->text.t   = a->name;
+    text->length   = a->length;
+    text->encoding = ENC_ISO_LATIN_1;
+  }
+  text->storage   = PL_CHARS_HEAP;
+  text->canonical = TRUE;
+
+  succeed;
+}
+
+
+int
+get_atom_text(atom_t atom, PL_chars_t *text)
+{ Atom a = atomValue(atom);
+
+  return get_atom_ptr_text(a, text);
+}
+
+
+int
+get_string_text(word w, PL_chars_t *text ARG_LD)
+{ if ( isBString(w) )
+  { text->text.t   = getCharsString(w, &text->length);
+    text->encoding = ENC_ISO_LATIN_1;
+  } else
+  { text->text.w   = getCharsWString(w, &text->length);
+    text->encoding = ENC_WCHAR;
+  }
+  text->storage   = PL_CHARS_STACK;
+  text->canonical = TRUE;
+
+  succeed;
+}
+
+
+static int
+compareUCSAtom(atom_t h1, atom_t h2)
+{ Atom a1 = atomValue(h1);
+  Atom a2 = atomValue(h2);
+  const pl_wchar_t *s1 = (const pl_wchar_t*)a1->name;
+  const pl_wchar_t *s2 = (const pl_wchar_t*)a2->name;
+  unsigned len = a1->length < a2->length ? a1->length : a2->length;
+
+  len /= sizeof(pl_wchar_t);
+  
+  for( ; len-- > 0; s1++, s2++)
+  { if ( *s1 != *s2 )
+      return *s1 - *s2;
+  }
+
+  return a1->length - a2->length;
+}
+
+
+static int
+saveUCSAtom(atom_t atom, IOSTREAM *fd)
+{ Atom a = atomValue(atom);
+  const pl_wchar_t *s = (const pl_wchar_t*)a->name;
+  size_t len = a->length/sizeof(pl_wchar_t);
+
+  wicPutStringW(s, len, fd);
+
+  return TRUE;
+}
+
+
+static atom_t
+loadUCSAtom(IOSTREAM *fd)
+{ pl_wchar_t buf[256];
+  pl_wchar_t *w;
+  unsigned len;
+  atom_t a;
+
+  w = wicGetStringUTF8(fd, &len, buf, sizeof(buf)/sizeof(pl_wchar_t));
+  a = lookupUCSAtom(w, len);
+
+  if ( w != buf )
+    PL_free(w);
+
+  return a;
+}
+
+
+int
+PL_unify_wchars(term_t t, int flags, unsigned int len, const pl_wchar_t *s)
+{ PL_chars_t text;
+  int rc;
+
+  if ( len == (unsigned int)-1 )
+    len = wcslen(s);
+
+  text.text.w    = (pl_wchar_t *)s;
+  text.encoding  = ENC_WCHAR;
+  text.storage   = PL_CHARS_HEAP;
+  text.length    = len;
+  text.canonical = FALSE;
+
+  rc = PL_unify_text(t, 0, &text, flags);
+  PL_free_text(&text);
+
+  return rc;
+}
+
+
+int
+PL_unify_wchars_diff(term_t t, term_t tail, int flags,
+		     unsigned int len, const pl_wchar_t *s)
+{ PL_chars_t text;
+  int rc;
+
+  if ( len == (unsigned int)-1 )
+    len = wcslen(s);
+
+  text.text.w    = (pl_wchar_t *)s;
+  text.encoding  = ENC_WCHAR;
+  text.storage   = PL_CHARS_HEAP;
+  text.length    = len;
+  text.canonical = FALSE;
+
+  rc = PL_unify_text(t, tail, &text, flags);
+  PL_free_text(&text);
+
+  return rc;
+}
+
+
+unsigned int
+PL_utf8_strlen(const char *s, unsigned int len)
+{ return utf8_strlen(s, len);
+}
+
+
+		 /*******************************
+		 *	  GET ATOM TEXT		*
+		 *******************************/
+
+const char *
+PL_atom_chars(atom_t a)
+{ return (const char *) stringAtom(a);
+}
+
+
+const char *
+PL_atom_nchars(atom_t a, unsigned int *len)
+{ Atom x = atomValue(a);
+
+  if ( x->type != &ucs_atom )
+  { if ( len )
+      *len = x->length;
+
+    return x->name;
+  } else
+    return NULL;
+}
+
+
+const wchar_t *
+PL_atom_wchars(atom_t a, unsigned int *len)
+{ Atom x = atomValue(a);
+
+  if ( x->type == &ucs_atom )
+  { if ( len )
+      *len = x->length / sizeof(pl_wchar_t);
+
+    return (const wchar_t *)x->name;
+  } else
+    return NULL;
+}
+
+
+
 		 /*******************************
 		 *    QUINTUS WRAPPER SUPPORT   *
 		 *******************************/
@@ -423,7 +656,7 @@ PL_same_compound(term_t t1, term_t t2)
 		 *******************************/
 
 static inline word
-__makeNum(long i ARG_LD)
+__makeNum(int64_t i ARG_LD)
 { word w = consInt(i);
 
   if ( valInt(w) == i )
@@ -433,7 +666,7 @@ __makeNum(long i ARG_LD)
 }
 
 word
-makeNum__LD(long i ARG_LD)
+makeNum__LD(int64_t i ARG_LD)
 { return __makeNum(i PASS_LD);
 }
 
@@ -588,7 +821,7 @@ PL_get_term_value(term_t t, term_value_t *val)
   { case PL_VARIABLE:
       break;
     case PL_INTEGER:
-      val->i = valInteger(w);
+      val->i = valInteger(w);		/* TBD: Handle MPZ integers? */
       break;
     case PL_FLOAT:
       val->f = valReal(w);
@@ -597,7 +830,7 @@ PL_get_term_value(term_t t, term_value_t *val)
       val->a = (atom_t)w;
       break;
     case PL_STRING:
-      val->s = valString(w);
+      val->s = getCharsString(w, NULL);
       break;
     case PL_TERM:
     { FunctorDef fd = valueFunctor(functorTerm(w));
@@ -698,9 +931,12 @@ PL_get_string(term_t t, char **s, unsigned int *len)
   word w = valHandle(t);
 
   if ( isString(w) )
-  { *s = valString(w);
-    *len = sizeString(w);
-    succeed;
+  { char *tmp = getCharsString(w, len);
+
+    if ( tmp )
+    { *s = tmp;
+      succeed;
+    }					/* fails on wide-character string */
   }
   fail;
 }
@@ -710,7 +946,7 @@ PL_get_string(term_t t, char **s, unsigned int *len)
 #define buffer_ring		(LD->fli._buffer_ring)
 #define current_buffer_id	(LD->fli._current_buffer_id)
 
-static Buffer
+Buffer
 findBuffer(int flags)
 { GET_LD
   Buffer b;
@@ -741,7 +977,7 @@ buffer_string(const char *s, int flags)
 }
 
 
-static int
+int
 unfindBuffer(int flags)
 { GET_LD
   if ( flags & BUF_RING )
@@ -753,15 +989,38 @@ unfindBuffer(int flags)
 }
 
 
-int
-PL_get_list_nchars(term_t l,
-		   unsigned int *length, char **s, unsigned int flags)
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+codes_or_chars_to_buffer(term_t l, unsigned flags, int wide)
+
+If l represents a list of codes   or characters, return a buffer holding
+the characters. If wide == TRUE  the   buffer  contains  objects of type
+pl_wchar_t. Otherwise it contains traditional characters.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int
+charCode(word w)
+{ if ( isAtom(w) )
+  { Atom a = atomValue(w);
+
+    if ( a->length == 1 && true(a->type, PL_BLOB_TEXT) )
+      return a->name[0] & 0xff;
+    if ( a->length == sizeof(pl_wchar_t) && a->type == &ucs_atom )
+    { pl_wchar_t *p = (pl_wchar_t*)a->name;
+
+      return p[0];
+    }
+  }
+
+  return -1;
+}
+
+
+Buffer
+codes_or_chars_to_buffer(term_t l, unsigned int flags, int wide)
 { GET_LD
   Buffer b;
   word list = valHandle(l);
   Word arg, tail;
-  char *r;
-  unsigned int len;
   enum { CHARS, CODES } type;
 
   if ( isList(list) )
@@ -769,21 +1028,16 @@ PL_get_list_nchars(term_t l,
     deRef(arg);
     if ( isTaggedInt(*arg) )
     { int i = valInt(*arg);
-      if ( i >= 0 && i < 256 )
+      if ( i >= 0 && (wide || i < 256) )
       { type = CODES;
 	goto ok;
       }
-    } else if ( isAtom(*arg) )
-    { Atom a = atomValue(*arg);
-
-      if ( a->length == 1 && true(a->type, PL_BLOB_TEXT) )
-      { type = CHARS;
-	goto ok;
-      }
+    } else if ( charCode(*arg) >= 0 )
+    { type = CHARS;
+      goto ok;
     }
   } else if ( isNil(list) )
-  { b = findBuffer(flags);
-    goto empty;
+  { return findBuffer(flags);
   }
 
   fail;
@@ -800,49 +1054,62 @@ ok:
     switch(type)
     { case CODES:
 	if ( isTaggedInt(*arg) )
-	{ int i = valInt(*arg);
-	  if ( i >= 0 && i < 256 )
-	    c = i;
+	{ c = valInt(*arg);
 	}
         break;
       case CHARS:
-	if ( isAtom(*arg) )
-	{ Atom a = atomValue(*arg);
-
-	  if ( a->length == 1 && true(a->type, PL_BLOB_TEXT) )
-	    c = a->name[0] & 0xff;
-
-	  break;
-	}
+	c = charCode(*arg);
+        break;
     }
 
-    if ( c == -1 )
-      return unfindBuffer(flags);
+    if ( c < 0 || (!wide && c > 0xff) )
+    { unfindBuffer(flags);
+      return NULL;
+    }
+
+    if ( wide )
+      addBuffer(b, c, pl_wchar_t);
+    else
+      addBuffer(b, c, char);
 
-    addBuffer(b, c, char);
     tail = argTermP(list, 1);
     deRef(tail);
     list = *tail;
   }
   if ( !isNil(list) )
-    return unfindBuffer(flags);
+  { unfindBuffer(flags);
+    return NULL;
+  }
 
-empty:
-  len = entriesBuffer(b, char);
+  return b;
+}
 
-  if ( length )
-    *length = len;
-  addBuffer(b, EOS, char);
-  r = baseBuffer(b, char);
 
-  if ( flags & BUF_MALLOC )
-  { *s = xmalloc(len+1);
-    memcpy(*s, r, len+1);
-    unfindBuffer(flags);
-  } else
-    *s = r;
+int
+PL_get_list_nchars(term_t l,
+		   unsigned int *length, char **s, unsigned int flags)
+{ Buffer b;
 
-  succeed;
+  if ( (b = codes_or_chars_to_buffer(l, flags, FALSE)) )
+  { char *r;
+    int len = entriesBuffer(b, char);
+
+    if ( length )
+      *length = len;
+    addBuffer(b, EOS, char);
+    r = baseBuffer(b, char);
+
+    if ( flags & BUF_MALLOC )
+    { *s = PL_malloc(len+1);
+      memcpy(*s, r, len+1);
+      unfindBuffer(flags);
+    } else
+      *s = r;
+
+    succeed;
+  }
+
+  fail;
 }
 
 
@@ -853,105 +1120,43 @@ PL_get_list_chars(term_t l, char **s, unsigned flags)
 
 
 int
-PL_get_nchars(term_t l, unsigned int *length, char **s, unsigned flags)
-{ GET_LD
-  word w = valHandle(l);
-  char tmp[100];
-  char *r;
-  int type;
-  unsigned int len = ~(unsigned int)0;
-  int malloced = FALSE;
+PL_get_wchars(term_t l, unsigned int *length, pl_wchar_t **s, unsigned flags)
+{ PL_chars_t text;
 
-  DEBUG(7, pl_write(l); pl_nl());
+  if ( !PL_get_text(l, &text, flags) )
+    return FALSE;
 
-  if ( (flags & CVT_ATOM) && isAtom(w) )
-  { Atom a = atomValue(w);
-    if ( false(a->type, PL_BLOB_TEXT) )
-      fail;				/* non-textual atom */
-    type = PL_ATOM;
-    r = a->name;
-    len = a->length;
-  } else if ( (flags & CVT_INTEGER) && isInteger(w) )
-  { type = PL_INTEGER;
-    Ssprintf(tmp, "%ld", valInteger(w) );
-    r = tmp;
-  } else if ( (flags & CVT_FLOAT) && isReal(w) )
-  { char *q;
-    type = PL_FLOAT;
-    Ssprintf(tmp, LD->float_format, valReal(w) );
-    r = tmp;
+  PL_promote_text(&text);
+  PL_save_text(&text, flags);
 
-    q = tmp;				/* See also writePrimitive() */
-    if ( *q == '-' )
-      q++;
-    for(; *q; q++)
-    { if ( !isDigit(*q) )
-	break;
-    }
-    if ( !*q )
-    { *q++ = '.';
-      *q++ = '0';
-      *q = EOS;
-    }
-#ifdef O_STRING
-  } else if ( (flags & CVT_STRING) && isString(w) )
-  { type = PL_STRING;
-    flags |= BUF_RING;			/* always buffer strings */
-    r = valString(w);
-    len = sizeString(w);
-#endif
-  } else if ( (flags & CVT_LIST) &&
-	      (isList(w) || isNil(w)) &&
-	      PL_get_list_nchars(l, length, s, flags) )
-  { DEBUG(7, Sdprintf("--> %s\n", *s));
-    succeed;
-  } else if ( (flags & CVT_VARIABLE) && isVar(w) )
-  { type = PL_VARIABLE;
-    r = varName(l, tmp);
-  } else if ( (flags & CVT_WRITE) )
-  { int size;
-    IOSTREAM *fd;
-    
-    type = PL_STRING;			/* hack to get things below ok */
-    flags |= BUF_RING;
-
-    r = tmp;
-    size = sizeof(tmp);
-    fd = Sopenmem(&r, &size, "w");
-    PL_write_term(fd, l, 1200, 0);
-    Sputc(EOS, fd);
-    Sflush(fd);
-    len = size-1;			/* need to flush first */
-    Sclose(fd);
-    if ( r != tmp )
-      malloced = TRUE;
-  } else
-  { DEBUG(7, Sdprintf("--> fail\n"));
-    fail;
-  }
-    
-  DEBUG(7, Sdprintf("--> %s\n", r));
-  if ( len == ~(unsigned int)0 )
-    len = strlen(r);
   if ( length )
-    *length = len;
+    *length = text.length;
+  *s = text.text.w;
 
-  if ( flags & BUF_MALLOC )
-  { *s = xmalloc(len+1);
-    memcpy(*s, r, len+1);
-  } else if ( ((flags & BUF_RING) && type != PL_ATOM) || /* never atoms */
-	      r == tmp )		/* always buffer tmp */
-  { Buffer b = findBuffer(flags);
+  return TRUE;
+}
 
-    addMultipleBuffer(b, r, len+1, char);
-    *s = baseBuffer(b, char);
+
+int
+PL_get_nchars(term_t l, unsigned int *length, char **s, unsigned flags)
+{ PL_chars_t text;
+
+  if ( !PL_get_text(l, &text, flags) )
+    return FALSE;
+
+  if ( PL_mb_text(&text, flags) )
+  { PL_save_text(&text, flags);
+
+    if ( length )
+      *length = text.length;
+    *s = text.text.t;
+    
+    return TRUE;
   } else
-    *s = r;
+  { PL_free_text(&text);
 
-  if ( malloced )
-    free(r);
-
-  succeed;
+    return FALSE;
+  }
 }
 
 
@@ -987,7 +1192,12 @@ PL_get_integer__LD(term_t t, int *i ARG_LD)
     succeed;
   }
   if ( isBignum(w) )
-  { *i = valBignum(w);
+  { int64_t val = valBignum(w);
+
+    if ( val > INT_MAX || val < INT_MIN )
+      fail;
+    
+    *i = (int)val;
     succeed;
   }
   if ( isReal(w) )
@@ -995,7 +1205,7 @@ PL_get_integer__LD(term_t t, int *i ARG_LD)
     int l;
 
 #ifdef DOUBLE_TO_LONG_CAST_RAISES_SIGFPE
-    if ( !((f >= PLMININT) && (f <= PLMAXINT)) )
+    if ( f > (real)INT_MAX || f < (real)INT_MIN )
       fail;
 #endif
 
@@ -1027,7 +1237,12 @@ PL_get_long__LD(term_t t, long *i ARG_LD)
     succeed;
   }
   if ( isBignum(w) )
-  { *i = valBignum(w);
+  { int64_t val = valBignum(w);
+
+    if ( val > LONG_MAX || val < LONG_MIN )
+      fail;
+    
+    *i = (long)val;
     succeed;
   }
   if ( isReal(w) )
@@ -1035,7 +1250,7 @@ PL_get_long__LD(term_t t, long *i ARG_LD)
     long l;
     
 #ifdef DOUBLE_TO_LONG_CAST_RAISES_SIGFPE
-    if ( !((f >= PLMININT) && (f <= PLMAXINT)) )
+    if ( f > (real)LONG_MAX || f < (real)LONG_MIN )
       fail;
 #endif
 
@@ -1058,6 +1273,39 @@ PL_get_long(term_t t, long *i)
 #define PL_get_long(t, i) PL_get_long__LD(t, i PASS_LD)
 
 
+int
+PL_get_int64(term_t t, int64_t *i)
+{ GET_LD
+  word w = valHandle(t);
+  
+  if ( isTaggedInt(w) )
+  { *i = valInt(w);
+    succeed;
+  }
+  if ( isBignum(w) )
+  { *i = valBignum(w);
+    succeed;
+  }
+  if ( isReal(w) )
+  { real f = valReal(w);
+    int64_t l;
+    
+#ifdef DOUBLE_TO_LONG_CAST_RAISES_SIGFPE
+    if ( !((f >= LLONG_MAX) && (f <= LLONG_MIN)) )
+      fail;
+#endif
+
+    l = (int64_t) f;
+    if ( (real)l == f )
+    { *i = l;
+      succeed;
+    }
+  }
+
+  fail;
+}
+
+
 int
 PL_is_inf(term_t t)
 { GET_LD
@@ -1094,9 +1342,9 @@ PL_get_float(term_t t, double *f)
 
 int
 PL_get_pointer__LD(term_t t, void **ptr ARG_LD)
-{ long p;
+{ int64_t p;
 
-  if ( PL_get_long(t, &p) )
+  if ( PL_get_int64(t, &p) )
   { *ptr = longToPointer((ulong)p);
 
     succeed;
@@ -1334,7 +1582,7 @@ _PL_get_xpce_reference(term_t t, xpceref_t *ref)
       }
       if ( isBignum(*p) )
       { ref->type    = PL_INTEGER;
-	ref->value.i = valBignum(*p);
+	ref->value.i = (long)valBignum(*p);
 
 	goto ok;
       }
@@ -1439,6 +1687,42 @@ PL_is_float(term_t t)
 }
 
 
+static inline int
+isRational(word w ARG_LD)
+{ if ( isTerm(w) )
+  { Functor f = valueTerm(w);
+
+    if ( f->definition == FUNCTOR_rdiv2 )
+    { Word p;
+
+      deRef2(&f->arguments[0], p);
+      if ( !isInteger(*p) )
+	fail;
+      deRef2(&f->arguments[1], p);
+      if ( !isInteger(*p) )
+	fail;
+      if ( *p == consInt(0) )
+	fail;
+
+      return TRUE;
+    }
+  }
+  if ( isInteger(w) )
+    return TRUE;
+
+  return FALSE;
+}
+
+
+int
+PL_is_rational(term_t t)
+{ GET_LD
+  word w = valHandle(t);
+
+  return isRational(w PASS_LD);
+}
+
+
 int
 PL_is_compound(term_t t)
 { GET_LD
@@ -1506,7 +1790,11 @@ PL_is_number(term_t t)
 { GET_LD
   word w = valHandle(t);
 
-  return isNumber(w) ? TRUE : FALSE;
+  if ( isInteger(w) ||
+       isReal(w) )
+    return TRUE;
+
+  return FALSE;
 }
 
 
@@ -1522,7 +1810,7 @@ PL_is_string(term_t t)
 int
 PL_unify_string_chars(term_t t, const char *s)
 { GET_LD
-  word str = globalString((char *)s);
+  word str = globalString(strlen(s), (char *)s);
 
   return unifyAtomic(t, str PASS_LD);
 }
@@ -1530,7 +1818,7 @@ PL_unify_string_chars(term_t t, const char *s)
 int
 PL_unify_string_nchars(term_t t, unsigned int len, const char *s)
 { GET_LD
-  word str = globalNString(len, s);
+  word str = globalString(len, s);
 
   return unifyAtomic(t, str PASS_LD);
 }
@@ -1589,7 +1877,7 @@ PL_put_atom_nchars(term_t t, unsigned int len, const char *s)
 void
 PL_put_string_chars(term_t t, const char *s)
 { GET_LD
-  word w = globalString(s);
+  word w = globalString(strlen(s), s);
 
   setHandle(t, w);
 }
@@ -1598,7 +1886,7 @@ PL_put_string_chars(term_t t, const char *s)
 void
 PL_put_string_nchars(term_t t, unsigned int len, const char *s)
 { GET_LD
-  word w = globalNString(len, s);
+  word w = globalString(len, s);
 
   setHandle(t, w);
 }
@@ -1673,12 +1961,16 @@ PL_put_integer(term_t t, long i)
 #define PL_put_integer(t, i) PL_put_integer__LD(t, i PASS_LD)
 
 
+void
+PL_put_int64(term_t t, int64_t i)
+{ GET_LD
+  setHandle(t, makeNum(i));
+}
+
+
 void
 _PL_put_number__LD(term_t t, Number n ARG_LD)
-{ if ( intNumber(n) )
-    PL_put_integer(t, n->value.i);
-  else
-    PL_put_float(t, n->value.f);
+{ setHandle(t, put_number(n));
 }
 
 
@@ -1867,6 +2159,25 @@ PL_unify_atom_nchars(term_t t, unsigned int len, const char *chars)
 }
 
 
+static atom_t
+uncachedCodeToAtom(int code)
+{ if ( code < 256 )
+  { char tmp[1];
+
+    tmp[0] = code;
+    return lookupAtom(tmp, 1);
+  } else
+  { pl_wchar_t tmp[1];
+    int new;
+
+    tmp[0] = code;
+
+    return lookupBlob((const char *)tmp, sizeof(pl_wchar_t),
+		      &ucs_atom, &new);
+  }
+}
+
+
 atom_t
 codeToAtom(int code)
 { atom_t a;
@@ -1874,14 +2185,25 @@ codeToAtom(int code)
   if ( code == EOF )
     return ATOM_end_of_file;
 
-  code &= 0xff;				/* mask for signedness */
+  assert(code >= 0);
 
-  if ( !(a=GD->atoms.for_code[code]) )
-  { char tmp[1];
+  if ( code < (1<<15) )
+  { int page  = code / 256;
+    int entry = code % 256;
+    atom_t *pv;
 
-    tmp[0] = code;
-    a = lookupAtom(tmp, 1);
-    GD->atoms.for_code[code] = a;
+    if ( !(pv=GD->atoms.for_code[page]) )
+    { pv = PL_malloc(256*sizeof(atom_t));
+      
+      memset(pv, 0, 256*sizeof(atom_t));
+      GD->atoms.for_code[page] = pv;
+    }
+
+    if ( !(a=pv[entry]) )
+    { a = uncachedCodeToAtom(code);
+    }
+  } else
+  { a = uncachedCodeToAtom(code);
   }
   
   return a;
@@ -1954,6 +2276,37 @@ PL_unify_list_chars(term_t l, const char *chars)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+flags: bitwise or of type and representation
+
+	Types:		PL_ATOM, PL_STRING, PL_CODE_LIST, PL_CHAR_LIST
+	Representation: REP_ISO_LATIN_1, REP_UTF8, REP_MB
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+PL_unify_chars(term_t t, int flags, unsigned int len, const char *s)
+{ PL_chars_t text;
+  int rc;
+
+  if ( len == (unsigned int)-1 )
+    len = strlen(s);
+
+  text.text.t    = (char *)s;
+  text.encoding  = ((flags&REP_UTF8) ? ENC_UTF8 : \
+		    (flags&REP_MB)   ? ENC_ANSI : ENC_ISO_LATIN_1);
+  text.storage   = PL_CHARS_HEAP;
+  text.length    = len;
+  text.canonical = FALSE;
+
+  flags &= ~(REP_UTF8|REP_MB|REP_ISO_LATIN_1);
+
+  rc = PL_unify_text(t, 0, &text, flags);
+  PL_free_text(&text);
+
+  return rc;
+}
+
+
 int
 PL_unify_integer__LD(term_t t, long i ARG_LD)
 { return unifyAtomic(t, makeNum(i) PASS_LD);
@@ -1969,6 +2322,14 @@ PL_unify_integer(term_t t, long i)
 #define PL_unify_integer(t, i)	PL_unify_integer__LD(t, i PASS_LD)
 
 
+int
+PL_unify_int64(term_t t, int64_t i)
+{ GET_LD
+
+  return unifyAtomic(t, makeNum(i) PASS_LD);
+}
+
+
 int
 PL_unify_pointer__LD(term_t t, void *ptr ARG_LD)
 { word w = makeNum(pointerToLong(ptr));
@@ -2114,12 +2475,13 @@ PL_unify_termv(term_t t, va_list args)
   tmp_buffer buf;
   int tos = 0;				/* Top-of-stack */
   int rval;
+  int op;
 
   t = PL_copy_term_ref(t);
   initBuffer(&buf);
   
 cont:
-  switch(va_arg(args, int))
+  switch((op=va_arg(args, int)))
   { case PL_VARIABLE:
       rval = TRUE;
       break;
@@ -2139,6 +2501,9 @@ cont:
     case PL_LONG:
       rval = PL_unify_integer(t, va_arg(args, long));
       break;
+    case PL_INT64:
+      rval = PL_unify_int64(t, va_arg(args, int64_t));
+      break;
     case PL_POINTER:
       rval = PL_unify_pointer(t, va_arg(args, void *));
       break;
@@ -2162,6 +2527,71 @@ cont:
       rval = PL_unify_atom_nchars(t, len, s);
       break;
     }
+    case PL_UTF8_CHARS:
+    case PL_UTF8_STRING:
+    { PL_chars_t txt;
+
+      txt.text.t    = va_arg(args, char *);
+      txt.length    = strlen(txt.text.t);
+      txt.storage   = PL_CHARS_HEAP;
+      txt.encoding  = ENC_UTF8;
+      txt.canonical = FALSE;
+
+      rval = PL_unify_text(t, 0, &txt,
+			   op == PL_UTF8_STRING ? PL_STRING : PL_ATOM);
+      break;
+    }
+    case PL_NUTF8_CHARS:
+    case PL_NUTF8_CODES:
+    case PL_NUTF8_STRING:
+    { PL_chars_t txt;
+
+      txt.length    = va_arg(args, unsigned int);
+      txt.text.t    = va_arg(args, char *);
+      txt.storage   = PL_CHARS_HEAP;
+      txt.encoding  = ENC_UTF8;
+      txt.canonical = FALSE;
+
+      rval = PL_unify_text(t, 0, &txt,
+			   op == PL_NUTF8_CHARS ? PL_ATOM : 
+			   op == PL_NUTF8_CODES ? PL_CODE_LIST :
+			   			  PL_STRING);
+      break;
+    }
+    case PL_NWCHARS:
+    case PL_NWCODES:
+    case PL_NWSTRING:
+    { PL_chars_t txt;
+
+      txt.length    = va_arg(args, unsigned int);
+      txt.text.w    = va_arg(args, wchar_t *);
+      txt.storage   = PL_CHARS_HEAP;
+      txt.encoding  = ENC_WCHAR;
+      txt.canonical = FALSE;
+
+      rval = PL_unify_text(t, 0, &txt,
+			   op == PL_NWCHARS ? PL_ATOM : 
+			   op == PL_NWCODES ? PL_CODE_LIST :
+					      PL_STRING);
+      break;
+    }
+    case PL_MBCHARS:
+    case PL_MBCODES:
+    case PL_MBSTRING:
+    { PL_chars_t txt;
+
+      txt.text.t    = va_arg(args, char *);
+      txt.length    = strlen(txt.text.t);
+      txt.storage   = PL_CHARS_HEAP;
+      txt.encoding  = ENC_ANSI;
+      txt.canonical = FALSE;
+
+      rval = PL_unify_text(t, 0, &txt,
+			   op == PL_MBCHARS ? PL_ATOM : 
+			   op == PL_MBCODES ? PL_CODE_LIST :
+					      PL_STRING);
+      break;
+    }
   { functor_t ft;
     int arity;
 
@@ -2676,6 +3106,8 @@ PL_throw(term_t exception)
 		*      REGISTERING FOREIGNS     *
 		*********************************/
 
+#define extensions_loaded	(GD->foreign._loaded)
+
 static void
 notify_registered_foreign(functor_t fd, Module m)
 { if ( GD->initialised )
@@ -2693,62 +3125,40 @@ notify_registered_foreign(functor_t fd, Module m)
 }
 
 
-bool
-PL_register_foreign(const char *name, int arity, Func f, int flags)
-{ Procedure proc;
+static bool
+bindForeign(Module m, const char *name, int arity, Func f, int flags)
+{ GET_LD
+  Procedure proc;
   Definition def;
-  Module m;
   functor_t fdef;
   atom_t aname;
-  const char *s;
 
-  if ( !GD->initialised )
-    initModules();			/* Before PL_initialise()! */
-
-					/* check for module:name */
-  for(s=name; isAlpha(*s); s++)
-    ;
-
-  if ( *s == ':' )
-  { m = PL_new_module(PL_new_atom_nchars(s-name, name));
-    aname = PL_new_atom(s+1);
-  } else
-  { GET_LD
-    aname = PL_new_atom(name);
-    m = (LD && environment_frame ? contextModule(environment_frame)
-			         : MODULE_user);
-  }
+  aname = PL_new_atom(name);
 
   fdef = lookupFunctorDef(aname, arity);
   proc = lookupProcedure(fdef, m);
   def = proc->definition;
 
-  if ( true(def, LOCKED) )
-  { warning("PL_register_foreign(): Attempt to redefine a system predicate: %s",
+  if ( true(def, LOCKED) && !SYSTEM_MODE )
+  { warning("PL_register_foreign: attempt to redefine a system predicate: %s",
 	    procedureName(proc));
     fail;
   }
 
   if ( def->definition.function )
-    warning("PL_register_foreign(): redefined %s", procedureName(proc));
+    warning("PL_register_foreign: redefined %s", procedureName(proc));
   if ( false(def, FOREIGN) && def->definition.clauses != NULL )
     abolishProcedure(proc, m);
 
   def->definition.function = f;
   def->indexPattern = 0;
   def->indexCardinality = 0;
-  def->flags = 0;
+  def->flags = FOREIGN|TRACE_ME;
 
   if ( m == MODULE_system )
-    set(def, FOREIGN|SYSTEM|TRACE_ME);
-  else
-  { GET_LD
-    
-    if ( SYSTEM_MODE )
-      set(def, FOREIGN|SYSTEM|TRACE_ME);
-    else
-      set(def, FOREIGN|TRACE_ME);
-  }
+    set(def, SYSTEM|HIDE_CHILDS);
+  else if ( SYSTEM_MODE )
+    set(def, SYSTEM|HIDE_CHILDS);
 
   if ( (flags & PL_FA_NOTRACE) )	  clear(def, TRACE_ME);
   if ( (flags & PL_FA_TRANSPARENT) )	  set(def, METAPRED);
@@ -2758,53 +3168,80 @@ PL_register_foreign(const char *name, int arity, Func f, int flags)
   notify_registered_foreign(fdef, m);
 
   succeed;
-}  
+}
+
+
+static Module
+resolveModule(const char *module)
+{ if ( !GD->initialised )      /* Before PL_initialise()! */
+    initModules();
+
+  if (module)
+    return PL_new_module(PL_new_atom(module));
+  else
+  { GET_LD
+    return (LD && environment_frame ? contextModule(environment_frame)
+	                            : MODULE_user);
+  }
+}
+
+void
+bindExtensions(const char *module, const PL_extension *ext)
+{ Module m = resolveModule(module);
+
+  for(; ext->predicate_name; ext++)
+  { bindForeign(m, ext->predicate_name, ext->arity,
+		ext->function, ext->flags);
+  }
+}
+
+void
+PL_register_extensions_in_module(const char *module, const PL_extension *e)
+{ if ( extensions_loaded )
+    bindExtensions(module, e);
+  else
+    rememberExtensions(module, e);
+}
 
 
 void
-PL_load_extensions(const PL_extension *ext)
-{ GET_LD
-  const PL_extension *e;
-  Module m;
-
-  m = (environment_frame ? contextModule(environment_frame)
-			 : MODULE_system);
-
-  for(e = ext; e->predicate_name; e++)
-  { unsigned long flags = TRACE_ME;
-    Definition def;
-    Procedure proc;
-    atom_t a = PL_new_atom(e->predicate_name);
-    functor_t fd = lookupFunctorDef(a, e->arity);
-
-    PL_unregister_atom(a);
-
-    if ( e->flags & PL_FA_NOTRACE )	     flags &= ~TRACE_ME;
-    if ( e->flags & PL_FA_TRANSPARENT )	     flags |= METAPRED;
-    if ( e->flags & PL_FA_NONDETERMINISTIC ) flags |= NONDETERMINISTIC;
-    if ( e->flags & PL_FA_VARARGS )	     flags |= P_VARARG;
-
-    proc = lookupProcedure(fd, m);
-    def = proc->definition;
-    if ( true(def, LOCKED) )
-    { warning("PL_load_extensions(): Attempt to redefine system predicate: %s",
-	      procedureName(proc));
-      continue;
-    }
-    if ( def->definition.function )
-      warning("PL_load_extensions(): redefined %s", procedureName(proc));
-    if ( false(def, FOREIGN) && def->definition.clauses != NULL )
-      abolishProcedure(proc, m);
-    set(def, FOREIGN);
-    set(def, flags);
-    def->definition.function = e->function;
-    def->indexPattern = 0;
-    def->indexCardinality = 0;
-
-    notify_registered_foreign(def->functor->functor, m);
-  }    
+PL_register_extensions(const PL_extension *e)
+{ PL_register_extensions_in_module(NULL, e);
 }
 
+
+bool
+PL_register_foreign_in_module(const char *module,
+			      const char *name, int arity, Func f, int flags)
+{ if ( extensions_loaded )
+  { Module m = resolveModule(module);
+    return bindForeign(m, name, arity, f, flags);
+  } else
+  { PL_extension ext[2];
+    ext->predicate_name = (char *)name;
+    ext->arity = arity;
+    ext->function = f;
+    ext->flags = flags;
+    ext[1].predicate_name = NULL;
+    rememberExtensions(module, ext);
+
+    return TRUE;
+  }
+}
+
+
+bool
+PL_register_foreign(const char *name, int arity, Func f, int flags)
+{ return PL_register_foreign_in_module(NULL, name, arity, f, flags);
+}
+
+		    /* deprecated */
+void
+PL_load_extensions(const PL_extension *ext)
+{ PL_register_extensions_in_module(NULL, ext);
+}
+
+
 		 /*******************************
 		 *	 EMBEDDING PROLOG	*
 		 *******************************/
@@ -2820,10 +3257,9 @@ PL_toplevel(void)
 }
 
 
-int
+void
 PL_halt(int status)
-{ if ( !PL_cleanup(status) )
-    fail;
+{ PL_cleanup(status);
 
   exit(status);
 }
@@ -3002,7 +3438,7 @@ cleanupInitialiseHooks(void)
 
 void
 PL_prompt1(const char *s)
-{ prompt1((char *) s);
+{ prompt1(lookupAtom(s, strlen(s)));
 }
 
 
@@ -3033,9 +3469,19 @@ PL_prompt_next(int fd)
 char *
 PL_prompt_string(int fd)
 { if ( fd == 0 )
-    return PrologPrompt();
+  { atom_t a = PrologPrompt();		/* TBD: deal with UTF-8 */
 
-  return "";
+    if ( a )
+    { PL_chars_t txt;
+
+      if ( get_atom_text(a, &txt) )
+      { if ( txt.encoding == ENC_ISO_LATIN_1 )
+	  return txt.text.t;
+      }
+    }
+  }
+
+  return NULL;
 }
 
 
@@ -3320,6 +3766,7 @@ init_c_args()
 { if ( c_argc == -1 )
   { GET_LD
     int i;
+    int opts = 1;
     int argc    = GD->cmdline.argc;
     char **argv = GD->cmdline.argv;
 
@@ -3328,7 +3775,7 @@ init_c_args()
     c_argc = 1;
 
     for(i=1; i<argc; i++)
-    { if ( argv[i][0] == '-' )
+    { if ( opts && argv[i][0] == '-' )
       { switch(argv[i][1])
 	{ case 'x':
 	  case 'g':
@@ -3346,6 +3793,11 @@ init_c_args()
 	  case 'A':
 	  case 'q':
 	    continue;
+         case '-':
+	   if (!argv[i][2])
+	   { opts = 0;
+	     continue;
+	   }
 	}
       }
       c_argv[c_argc++] = argv[i];
@@ -3364,9 +3816,8 @@ PL_query(int query)
       init_c_args();
       return (long) c_argv;
     case PL_QUERY_MAX_INTEGER:
-      return PLMAXINT;
     case PL_QUERY_MIN_INTEGER:
-      return PLMININT;
+      fail;				/* cannot represent (anymore) */
     case PL_QUERY_MAX_TAGGED_INT:
       return PLMAXTAGGEDINT;
     case PL_QUERY_MIN_TAGGED_INT:
@@ -3382,6 +3833,17 @@ PL_query(int query)
 #else
       return 1;
 #endif
+    case PL_QUERY_ENCODING:		
+    { GET_LD
+
+      if ( LD )
+	return LD->encoding;
+      return PL_local_data.encoding;	/* Default: of main thread? */
+    }
+    case PL_QUERY_USER_CPU:		/* User CPU in milliseconds */
+    { double cpu = CpuTime(CPU_USER);
+      return (long)(cpu*1000.0);
+    }
     default:
       sysError("PL_query: Illegal query: %d", query);
       /*NOTREACHED*/
@@ -3448,3 +3910,13 @@ registerForeignLicenses(void)
 
   pre_registered = NULL;
 }
+
+
+		 /*******************************
+		 *	       INIT		*
+		 *******************************/
+
+void
+initForeign(void)
+{ initUCSAtoms();
+}
