@@ -1,4 +1,4 @@
-/*  $Id: pl-funcs.h,v 1.220.2.1 2005/01/17 09:07:42 jan Exp $
+/*  $Id: pl-funcs.h,v 1.258 2006/08/27 12:31:14 jan Exp $
 
     Part of SWI-Prolog
 
@@ -31,13 +31,13 @@ libraries and avoids Prolog picking wrong   symbols. It also reduces ELF
 symbol lookup and relocations.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define COMMON(type) type SO_LOCAL
+#define COMMON(type) SO_LOCAL type
 
 
 /* pl-alloc.c */
 COMMON(void) 		freeHeap__LD(void *mem, size_t n ARG_LD);
 COMMON(word) 		outOfStack(Stack s, stack_overflow_action how);
-COMMON(volatile void)	outOfCore(void);
+COMMON(void)		outOfCore(void) NORETURN;
 COMMON(Word) 		allocGlobal__LD(int words ARG_LD);
 COMMON(Word) 		allocGlobalNoShift__LD(int words ARG_LD);
 COMMON(void *) 		allocHeap__LD(size_t n ARG_LD);
@@ -45,13 +45,14 @@ COMMON(void) 		initMemAlloc(void);
 COMMON(void) 		cleanupMemAlloc(void);
 COMMON(void) 		mergeAllocPool(AllocPool to, AllocPool from);
 COMMON(word) 		globalFunctor(functor_t def);
-COMMON(int) 		sizeString(word w);
-COMMON(word) 		globalString(const char *s);
-COMMON(word) 		globalNString(long len, const char *s);
+COMMON(word) 		globalString(long len, const char *s);
+COMMON(word) 		globalWString(unsigned len, const pl_wchar_t *s);
+COMMON(char *)		getCharsString__LD(word w, unsigned *len ARG_LD);
+COMMON(pl_wchar_t *)	getCharsWString__LD(word w, unsigned *len ARG_LD);
 COMMON(Word) 		newTerm(void);
 COMMON(void) 		doublecpy(void *to, void *from);
 COMMON(word) 		globalReal(real f);
-COMMON(word) 		globalLong(long i ARG_LD);
+COMMON(word) 		globalLong(int64_t i ARG_LD);
 COMMON(double) 		valReal__LD(word w ARG_LD);
 COMMON(word) 		globalIndirect(word in);
 COMMON(int) 		equalIndirect(word r1, word r2);
@@ -116,10 +117,6 @@ COMMON(void) 		_PL_debug_unregister_atom(atom_t a,
 
 /* pl-arith.c */
 
-COMMON(word) 		pl_between(term_t l, term_t h, term_t n,
-				   control_t ctx);
-COMMON(word) 		pl_succ(term_t n1, term_t n2);
-COMMON(word) 		pl_plus(term_t a, term_t b, term_t c);
 COMMON(int) 		ar_compare(Number n1, Number n2, int what);
 COMMON(word) 		pl_current_arithmetic_function(term_t f, control_t h);
 COMMON(void) 		initArith(void);
@@ -129,7 +126,6 @@ COMMON(functor_t) 	functorArithFunction(int n);
 COMMON(bool) 		ar_func_n(code n, int argc, Number *stack);
 COMMON(int) 		valueExpression(term_t p, Number n ARG_LD);
 COMMON(int) 		toIntegerNumber(Number n);
-COMMON(void) 		canoniseNumber(Number n);
 COMMON(int) 		arithChar(Word p ARG_LD);
 
 /* pl-bag.c */
@@ -215,6 +211,8 @@ COMMON(word) 		pl_dwim_predicate(term_t term, term_t dwim,
 /* pl-ext.c */
 COMMON(void) 		initBuildIns(void);
 COMMON(void) 		cleanupExtensions(void);
+COMMON(void)            rememberExtensions(const char *module,
+					   const PL_extension *e);
 
 /* pl-error.c */
 
@@ -223,7 +221,7 @@ COMMON(int) 		PL_error(const char *pred, int arity, const char *msg,
 COMMON(char *) 		tostr(char *buf, const char *fmt, ...);
 COMMON(void) 		printMessage(atom_t severity, ...);
 COMMON(int) 		PL_get_nchars_ex(term_t t, unsigned int *len, char **s,
-				 unsigned int flags);
+					 unsigned int flags);
 COMMON(int) 		PL_get_chars_ex(term_t t, char **s, unsigned int flags);
 COMMON(int) 		PL_get_atom_ex(term_t t, atom_t *a);
 COMMON(int) 		PL_get_integer_ex(term_t t, int *i);
@@ -265,21 +263,11 @@ COMMON(word) 		pl_noprotocol(void);
 COMMON(bool) 		seeString(const char *s);
 COMMON(bool) 		seeingString(void);
 COMMON(bool) 		seenString(void);
-COMMON(bool) 		tellString(char **s, int *size);
+COMMON(bool) 		tellString(char **s, int *size, IOENC enc);
 COMMON(bool) 		toldString(void);
 COMMON(word) 		pl_wait_for_input(term_t streams, term_t avail,
 					  term_t tmo);
-COMMON(word) 		pl_put(term_t c);
-COMMON(word) 		pl_put2(term_t stream, term_t chr);
-COMMON(word) 		pl_get(term_t chr);
-COMMON(word) 		pl_skip(term_t chr);
-COMMON(word) 		pl_skip2(term_t stream, term_t chr);
-COMMON(word) 		pl_get2(term_t stream, term_t chr);
 COMMON(word) 		pl_get_single_char(term_t c);
-COMMON(word) 		pl_get_byte(term_t c);
-COMMON(word) 		pl_get_byte2(term_t stream, term_t c);
-COMMON(word) 		pl_get_char(term_t c);
-COMMON(word) 		pl_get_char2(term_t stream, term_t c);
 COMMON(word) 		pl_seeing(term_t f);
 COMMON(word) 		pl_telling(term_t f);
 COMMON(word) 		pl_tell(term_t f);
@@ -289,10 +277,10 @@ COMMON(word) 		pl_protocol(term_t file);
 COMMON(word) 		pl_protocola(term_t file);
 COMMON(word) 		pl_protocolling(term_t file);
 COMMON(word) 		pl_prompt(term_t old, term_t new);
-COMMON(void) 		prompt1(char *prompt);
+COMMON(void) 		prompt1(atom_t prompt);
 COMMON(word) 		pl_prompt1(term_t prompt);
 COMMON(word) 		pl_tab(term_t n);
-COMMON(char *) 		PrologPrompt(void);
+COMMON(atom_t) 		PrologPrompt(void);
 COMMON(word) 		pl_tab2(term_t stream, term_t n);
 COMMON(word) 		pl_open4(term_t file, term_t mode, term_t stream,
 				 term_t opt);
@@ -313,18 +301,9 @@ COMMON(word) 		pl_set_output(term_t stream);
 COMMON(foreign_t) 	pl_set_stream(term_t stream, term_t attr);
 COMMON(word) 		pl_current_input(term_t stream);
 COMMON(word) 		pl_current_output(term_t stream);
-COMMON(word) 		pl_character_count(term_t stream, term_t count);
-COMMON(word) 		pl_line_count(term_t stream, term_t count);
-COMMON(word) 		pl_line_position(term_t stream, term_t count);
 COMMON(word) 		pl_source_location(term_t file, term_t line);
 COMMON(word) 		pl_at_end_of_stream1(term_t stream);
 COMMON(word) 		pl_at_end_of_stream0();
-COMMON(word) 		pl_peek_byte2(term_t stream, term_t chr);
-COMMON(word) 		pl_peek_byte1(term_t chr);
-COMMON(word) 		pl_peek_code2(term_t stream, term_t chr);
-COMMON(word) 		pl_peek_code1(term_t chr);
-COMMON(word) 		pl_peek_char2(term_t stream, term_t chr);
-COMMON(word) 		pl_peek_char1(term_t chr);
 COMMON(bool) 		unifyTime(term_t t, long time);
 COMMON(word) 		pl_time_file(term_t name, term_t t);
 COMMON(word) 		pl_size_file(term_t name, term_t len);
@@ -358,6 +337,13 @@ COMMON(int) 		PL_unify_char(term_t chr, int c, int mode);
 COMMON(int) 		PL_unify_stream_or_alias(term_t t, IOSTREAM *s);
 COMMON(void) 		pushOutputContext(void);
 COMMON(void) 		popOutputContext(void);
+COMMON(IOENC)		atom_to_encoding(atom_t a);
+COMMON(atom_t)		encoding_to_atom(IOENC enc);
+COMMON(int)		setupOutputRedirect(term_t to, 
+					    redir_context *ctx,
+					    int redir);
+COMMON(int)		closeOutputRedirect(redir_context *ctx);
+COMMON(void)		discardOutputRedirect(redir_context *ctx);
 
 /* pl-flag.c */
 COMMON(void) 		initFlags(void);
@@ -367,7 +353,7 @@ COMMON(void) 		initFeatures(void);
 
 /* pl-fli.c */
 COMMON(word) 		linkVal__LD(Word p ARG_LD);
-COMMON(word) 		makeNum__LD(long i ARG_LD);
+COMMON(word) 		makeNum__LD(int64_t i ARG_LD);
 COMMON(void) 		_PL_put_number__LD(term_t t, Number n ARG_LD);
 COMMON(predicate_t) 	_PL_predicate(const char *name, int arity,
 				      const char *module, predicate_t *bin);
@@ -403,8 +389,18 @@ COMMON(int) 		PL_unify_list__LD(term_t l, term_t h, term_t t ARG_LD);
 COMMON(void) 		PL_cons_list__LD(term_t l, term_t head, term_t tail
 					 ARG_LD);
 COMMON(int)		PL_is_inf(term_t t);
+COMMON(int)		isUCSAtom(Atom a);
+COMMON(atom_t)		lookupUCSAtom(const pl_wchar_t *s, unsigned int len);
+COMMON(Buffer)		codes_or_chars_to_buffer(term_t l, unsigned int flags,
+						 int wide);
+COMMON(Buffer)		findBuffer(int flags);
+COMMON(int)		unfindBuffer(int flags);
 
 COMMON(void) 		registerForeignLicenses(void);
+COMMON(void)            bindExtensions(const char *module,
+				       const PL_extension *ext);
+COMMON(void)		initForeign(void);
+
 
 /* pl-fmt.c */
 COMMON(word) 		pl_format_predicate(term_t chr, term_t descr);
@@ -496,22 +492,22 @@ COMMON(word) 		pl_builtin_op(term_t prec, term_t type, term_t name,
 COMMON(bool) 		initOs(void);
 COMMON(void) 		cleanupOs(void);
 COMMON(char *) 		OsError(void);
-COMMON(long) 		Random(void);
+COMMON(uint64_t)	_PL_Random(void);
 COMMON(char *) 		canonisePath(char *path);
 COMMON(char *) 		OsPath(const char *plpath, char *ospath);
-COMMON(char *) 		PrologPath(const char *ospath, char *plpath);
+COMMON(char *) 		PrologPath(const char *ospath, char *plpath, size_t len);
 COMMON(long) 		LastModifiedFile(char *f);
 COMMON(bool) 		ExistsFile(const char *path);
 COMMON(bool) 		AccessFile(const char *path, int mode);
 COMMON(bool) 		ExistsDirectory(const char *path);
-COMMON(long) 		SizeFile(const char *path);
+COMMON(int64_t)		SizeFile(const char *path);
 COMMON(int) 		RemoveFile(const char *path);
 COMMON(void) 		RemoveTemporaryFiles(void);
 COMMON(bool) 		RenameFile(const char *old, const char *new);
 COMMON(bool) 		SameFile(const char *f1, const char *f2);
 COMMON(bool) 		OpenStream(int fd);
 COMMON(bool) 		MarkExecutable(const char *name);
-COMMON(bool) 		expandVars(const char *pattern, char *expanded);
+COMMON(bool) 		expandVars(const char *pattern, char *expanded, int len);
 COMMON(char *) 		ExpandOneFile(const char *spec, char *file);
 COMMON(char *) 		getwd(char *buf);
 COMMON(char *) 		AbsoluteFile(const char *spec, char *path);
@@ -532,40 +528,30 @@ COMMON(int) 		Unsetenv(char *name);
 COMMON(int) 		System(char *cmd);
 COMMON(char *) 		findExecutable(const char *module, char *buf);
 COMMON(int) 		Pause(real time);
-#if __WIN32__
-COMMON(int) 		iswin32s(void);
-#endif /*__WIN32__*/
 
 /* pl-prims.c */
 COMMON(word) 		pl_notunify(term_t t1, term_t t2);
 COMMON(int) 		compareStandard(Word t1, Word t2, int eq ARG_LD);
 COMMON(int) 		lengthList(term_t list, int errors);
 COMMON(word) 		pl_univ(term_t t, term_t l);
-COMMON(int) 		numberVars(term_t t, functor_t functor,
-				   av_action on_av, int n ARG_LD);
+COMMON(int) 		numberVars(term_t t, nv_options *opts, int n ARG_LD);
 COMMON(word) 		pl_e_free_variables(term_t t, term_t l);
 COMMON(word) 		stringToList(char *s);
 COMMON(word) 		pl_atom_length(term_t w, term_t n);
-COMMON(word) 		pl_int_to_atom(term_t number, term_t base,
-			       term_t atom);
-COMMON(char *) 		formatInteger(bool split, int div, int radix,
-			      bool small, long n, char *out);
-COMMON(word) 		pl_format_number(term_t format, term_t number,
-				 term_t string);
 COMMON(word) 		pl_name(term_t atom, term_t string);
 COMMON(word) 		pl_atom_chars(term_t atom, term_t string);
 COMMON(word) 		pl_atom_codes(term_t atom, term_t string);
 COMMON(word) 		pl_number_chars(term_t number, term_t string);
 COMMON(word) 		pl_number_codes(term_t number, term_t string);
 COMMON(word) 		pl_char_code(term_t atom, term_t chr);
-COMMON(word) 		pl_atom_prefix(term_t atom, term_t prefix);
-COMMON(word) 		pl_atom_concat(term_t a1, term_t a2, term_t a3, control_t ctx);
+COMMON(word) 		pl_atom_concat(term_t a1, term_t a2, term_t a3,
+				       control_t ctx);
 COMMON(word) 		pl_concat_atom(term_t list, term_t atom);
 COMMON(word) 		pl_concat_atom3(term_t list, term_t sep, term_t atom);
 COMMON(word) 		pl_apropos_match(term_t a1, term_t a2);
 COMMON(foreign_t) 	pl_sub_atom(term_t atom,
-			    term_t before, term_t len, term_t after,
-			    term_t sub, control_t h);
+				    term_t before, term_t len, term_t after,
+				    term_t sub, control_t h);
 COMMON(word) 		pl_string_length(term_t str, term_t l);
 COMMON(word) 		pl_string_concat(term_t a1, term_t a2, term_t a3, control_t h);
 COMMON(word) 		pl_string_to_atom(term_t str, term_t a);
@@ -573,7 +559,6 @@ COMMON(word) 		pl_string_to_list(term_t str, term_t list);
 COMMON(word) 		pl_sub_string(term_t str,
 			      term_t offset, term_t length, term_t after,
 			      term_t sub, control_t h);
-COMMON(word) 		pl_write_on_string(term_t goal, term_t string);
 COMMON(word) 		pl_repeat(control_t h);
 COMMON(word) 		pl_fail(void);
 COMMON(word) 		pl_true(void);
@@ -676,9 +661,9 @@ COMMON(void) 		profRedo(struct call_node *node ARG_LD);
 
 /* pl-read.c */
 COMMON(void) 		resetRead(void);
-COMMON(int) 		get_number(const unsigned char *string,
-			   unsigned char **end,
-			   Number value, bool escape);
+COMMON(int) 		str_number(const unsigned char *string,
+				   unsigned char **end,
+				   Number value, bool escape);
 COMMON(word) 		pl_raw_read(term_t term);
 COMMON(word) 		pl_raw_read2(term_t stream, term_t term);
 COMMON(word) 		pl_read(term_t term);
@@ -744,11 +729,6 @@ COMMON(word) 		pl_shell(term_t command, term_t status);
 COMMON(word) 		pl_getenv(term_t var, term_t value);
 COMMON(word) 		pl_setenv(term_t var, term_t value);
 COMMON(word) 		pl_unsetenv(term_t var);
-COMMON(word) 		pl_convert_time(term_t time, term_t year,
-				term_t month, term_t day,
-				term_t hour, term_t minute,
-				term_t second, term_t usec);
-COMMON(word) 		pl_convert_time2(term_t time, term_t string);
 COMMON(word) 		pl_get_time(term_t t);
 COMMON(word) 		pl_sleep(term_t time);
 COMMON(word) 		pl_get_pid(term_t pid);
@@ -787,6 +767,7 @@ COMMON(word) 		pl_prolog_current_frame(term_t fr);
 COMMON(word) 		pl_prolog_frame_attribute(term_t fr, term_t key, term_t val);
 COMMON(foreign_t) 	pl_prolog_choice_attribute(term_t fr, term_t key, term_t val);
 COMMON(void) 		callEventHook(int ev, ...);
+COMMON(void)		PL_put_frame(term_t t, LocalFrame fr);
 
 /* pl-util.c */
 COMMON(char) 		digitName(int n, bool small);
@@ -802,13 +783,20 @@ COMMON(bool) 		strpostfix(const char *string, const char *postfix);
 COMMON(bool) 		stripostfix(const char *string, const char *postfix);
 COMMON(bool) 		scan_options(term_t list, int flags, atom_t name,
 				     const opt_spec *specs, ...);
-COMMON(const char *)  atom_summary(atom_t name, unsigned int maxlen);
-#ifndef HAVE_STRICMP
-COMMON(int) 		stricmp(const char *s1, const char *s2);
+COMMON(const char *)	atom_summary(atom_t name, unsigned int maxlen);
+#ifndef HAVE_STRCASECMP
+COMMON(int) 		strcasecmp(const char *s1, const char *s2);
 #endif
 #ifndef HAVE_STRLWR
 COMMON(char *) 		strlwr(char *s);
 #endif
+#ifndef HAVE_MBSCOLL
+COMMON(int)		mbscoll(const char *s1, const char *s2);
+#endif
+#ifndef HAVE_MBSCASECOLL
+COMMON(int)		mbscasecoll(const char *s1, const char *s2);
+#endif
+
 
 /* pl-wic.c */
 COMMON(bool) 		loadWicFromStream(IOSTREAM *fd);
@@ -829,6 +817,13 @@ COMMON(word) 		pl_qlf_close(void);
 COMMON(word) 		pl_qlf_load(term_t file, term_t module);
 COMMON(word) 		pl_qlf_assert_clause(term_t ref, term_t saveclass);
 
+COMMON(void)		wicPutNum(int64_t n, IOSTREAM *fd);
+COMMON(int64_t)		wicGetNum(IOSTREAM *fd);
+COMMON(void)		wicPutStringW(const pl_wchar_t *w, size_t len, 
+				      IOSTREAM *fd);
+COMMON(pl_wchar_t*)	wicGetStringUTF8(IOSTREAM *fd, unsigned *length,
+					 pl_wchar_t *buf, size_t bufsize);
+
 /* pl-write.c */
 COMMON(char *) 		varName(term_t var, char *buf);
 COMMON(word) 		pl_nl(void);
@@ -846,6 +841,9 @@ COMMON(word) 		pl_write2(term_t stream, term_t term);
 COMMON(word) 		pl_writeq2(term_t stream, term_t term);
 COMMON(word) 		pl_print2(term_t stream, term_t term);
 COMMON(int) 		writeAttributeMask(atom_t name);
+COMMON(int)		writeUCSAtom(IOSTREAM *fd, atom_t atom, int flags);
+COMMON(int)		writeAtomToStream(IOSTREAM *s, atom_t atom);
+COMMON(char *)		format_float(double f, char *buf, const char *format);
 
 /* pl-term.c */
 COMMON(void) 		resetTerm(void);
@@ -857,9 +855,9 @@ COMMON(word) 		pl_tty_size(term_t r, term_t c);
 /* pl-main.c */
 COMMON(int) 		startProlog(int argc, char **argv);
 COMMON(bool) 		sysError(const char *fm, ...);
-COMMON(bool) 		fatalError(const char *fm, ...);
+COMMON(void) 		fatalError(const char *fm, ...) NORETURN;
 COMMON(bool) 		warning(const char *fm, ...);
-COMMON(bool) 		vfatalError(const char *fm, va_list args);
+COMMON(void) 		vfatalError(const char *fm, va_list args) NORETURN;
 COMMON(bool) 		vwarning(const char *fm, va_list args);
 
 /* pl-dde.c */
@@ -888,7 +886,6 @@ COMMON(const)  char *	dlerror(void);
 COMMON(void *) 		dlsym(void *handle, char *symbol);
 COMMON(int) 		dlclose(void *handle);
 #endif /*EMULATE_DLOPEN*/
-COMMON(foreign_t) 	pl_get_registry_value(term_t Key, term_t Name, term_t Value);
 COMMON(void) 		getDefaultsFromRegistry(void);
 
 /* pl-rc.c */
@@ -912,10 +909,9 @@ COMMON(foreign_t) 	pl_open_xterm(term_t title, term_t in, term_t out, term_t err
 
 /* pl-ctype.c */
 
-COMMON(foreign_t) 	pl_char_type(term_t chr, term_t class, control_t h);
-COMMON(foreign_t) 	pl_code_type(term_t chr, term_t class, control_t h);
 COMMON(foreign_t) 	pl_downcase_atom(term_t in, term_t out);
 COMMON(foreign_t) 	pl_upcase_atom(term_t in, term_t out);
+COMMON(IOENC)		initEncoding(void);
 COMMON(void) 		initCharTypes(void);
 COMMON(bool) 		systemMode(bool accept);
 
