@@ -1,4 +1,4 @@
-/*  $Id: pl-gc.c,v 1.97 2004/10/02 15:38:01 jan Exp $
+/*  $Id: pl-gc.c,v 1.106 2006/06/22 09:52:15 jan Exp $
 
     Part of SWI-Prolog
 
@@ -589,18 +589,6 @@ mark_term_refs()
 }
 
 
-static void
-mark_foreign_trail_refs()
-{ GET_LD
-  FliFrame fr = fli_context;
-
-  for( ; fr; fr = fr->parent )
-  { needsRelocation(&fr->mark.trailtop);
-    alien_into_relocation_chain(&fr->mark.trailtop, STG_TRAIL, STG_LOCAL PASS_LD);
-  }
-}
-
-
 #ifdef O_GVAR
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -736,14 +724,17 @@ clearUninitialisedVarsFrame(LocalFrame fr, Code PC)
 	  c = decode(replacedBreak(PC));
 	  goto again;
 #endif
-	case H_INDIRECT:		/* only skip the size of the */
-	case B_INDIRECT:		/* string + header */
+	case H_STRING:			/* only skip the size of the */
+	case B_STRING:			/* string + header */
+	case H_MPZ:
+	case B_MPZ:
 	{ word m = PC[1];
 	  PC += wsizeofInd(m)+1;
 	  break;
 	}
 	case I_EXIT:
 	case I_EXITFACT:
+	case I_EXITCATCH:
 	  return;
 	case C_JMP:
 	  PC += (int)PC[1]+2;
@@ -934,15 +925,95 @@ should it be accessible and otherwise it really is garbage.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static GCTrailEntry
-mark_choicepoints(Choice ch, GCTrailEntry te)
+early_reset_vars(mark *m, Word top, GCTrailEntry te ARG_LD)
+{ GCTrailEntry tm = (GCTrailEntry)m->trailtop;
+  GCTrailEntry te0 = te;
+  int assignments = 0;
+
+  for( ; te >= tm; te-- )		/* early reset of vars */
+  { 
+#if O_DESTRUCTIVE_ASSIGNMENT
+    if ( isTrailVal(te->address) )
+    { Word tard = val_ptr(te[-1].address);
+
+      if ( tard >= top )
+      { te->address = 0;
+	te--;
+	te->address = 0;
+	trailcells_deleted += 2;
+      } else if ( is_marked(tard) )
+      { assignments++;
+      } else
+      { Word gp;
+	DEBUG(3, Sdprintf("Early reset of assignment at %p (*=0x%lx)\n",
+		 tard, *tard));
+	gp = val_ptr(te->address);
+	assert(onGlobal(gp));
+	*tard = *gp;
+	unmark(tard);
+
+	te->address = 0;
+	te--;
+	te->address = 0;
+	trailcells_deleted += 2;
+      }
+    } else
+#endif
+    { Word tard = val_ptr(te->address);
+
+      if ( tard >= top )		/* above local stack */
+      { SECURE(assert(ttag(te[1].address) != TAG_TRAILVAL));
+	te->address = 0;
+	trailcells_deleted++;
+      } else if ( !is_marked(tard) )	/* garbage */
+      { setVar(*tard);
+	DEBUG(3, Sdprintf("Early reset at %p\n", tard));
+	te->address = 0;
+	trailcells_deleted++;
+      }
+    }
+  }
+  
+#if O_DESTRUCTIVE_ASSIGNMENT
+  if ( assignments >= 2 )
+    mergeTrailedAssignments(te0, tm, assignments PASS_LD);
+#endif
+
+  return te;
+}
+
+
+static GCTrailEntry
+mark_foreign_frame(FliFrame fr, GCTrailEntry te)
+{ GET_LD
+
+  SECURE(assert(fr->magic == FLI_MAGIC));
+
+  te = early_reset_vars(&fr->mark, (Word)fr, te PASS_LD);
+
+  DEBUG(3, Sdprintf("Marking foreign frame %p\n", fr));
+  needsRelocation(&fr->mark.trailtop);
+  alien_into_relocation_chain(&fr->mark.trailtop,
+			      STG_TRAIL, STG_LOCAL PASS_LD);
+
+  return te;
+}
+
+
+static GCTrailEntry
+mark_choicepoints(Choice ch, GCTrailEntry te, FliFrame *flictx)
 { GET_LD
 
   for( ; ch; ch = ch->parent )
   { LocalFrame fr = ch->frame;
-    GCTrailEntry tm = (GCTrailEntry)ch->mark.trailtop;
-    GCTrailEntry te0 = te;
     Word top;
-    int assignments = 0;
+
+    while((char*)*flictx > (char*)ch)
+    { FliFrame fr = *flictx;
+
+      te = mark_foreign_frame(fr, te);
+      *flictx = fr->parent;
+    }
 
     if ( ch->type == CHP_CLAUSE )
       top = argFrameP(fr, fr->predicate->functor->arity);
@@ -951,61 +1022,14 @@ mark_choicepoints(Choice ch, GCTrailEntry te)
       top = (Word)ch;
     }
 
-    for( ; te >= tm; te-- )		/* early reset of vars */
-    { 
-#if O_DESTRUCTIVE_ASSIGNMENT
-      if ( isTrailVal(te->address) )
-      { Word tard = val_ptr(te[-1].address);
-
-	if ( tard >= top )
-	{ te->address = 0;
-	  te--;
-	  te->address = 0;
-	  trailcells_deleted += 2;
-	} else if ( is_marked(tard) )
-	{ assignments++;
-	} else
-	{ Word gp;
-	  DEBUG(3, Sdprintf("Early reset of assignment at %p (*=0x%lx)\n",
-		   tard, *tard));
-	  gp = val_ptr(te->address);
-	  assert(onGlobal(gp));
-	  *tard = *gp;
-	  unmark(tard);
-
-	  te->address = 0;
-	  te--;
-	  te->address = 0;
-	  trailcells_deleted += 2;
-	}
-      } else
-#endif
-      { Word tard = val_ptr(te->address);
-
-	if ( tard >= top )		/* above local stack */
-	{ SECURE(assert(ttag(te[1].address) != TAG_TRAILVAL));
-	  te->address = 0;
-	  trailcells_deleted++;
-	} else if ( !is_marked(tard) )	/* garbage */
-	{ setVar(*tard);
-	  DEBUG(3, Sdprintf("Early reset at %p\n", tard));
-	  te->address = 0;
-	  trailcells_deleted++;
-	}
-      }
-    }
-
-#if O_DESTRUCTIVE_ASSIGNMENT
-    if ( assignments >= 2 )
-      mergeTrailedAssignments(te0, tm, assignments PASS_LD);
-#endif
+    te = early_reset_vars(&ch->mark, top, te PASS_LD);
 
     needsRelocation(&ch->mark.trailtop);
-    alien_into_relocation_chain(&ch->mark.trailtop, STG_TRAIL, STG_LOCAL PASS_LD);
+    alien_into_relocation_chain(&ch->mark.trailtop,
+				STG_TRAIL, STG_LOCAL PASS_LD);
     SECURE(trailtops_marked--);
 
-    mark_environments(fr,
-		      ch->type == CHP_JUMP ? ch->value.PC : NULL);
+    mark_environments(fr, ch->type == CHP_JUMP ? ch->value.PC : NULL);
   }
 
   return te;
@@ -1028,16 +1052,20 @@ mark_stacks(LocalFrame fr, Choice ch)
 { GET_LD
   QueryFrame query;
   GCTrailEntry te = (GCTrailEntry)tTop - 1;
+  FliFrame flictx = fli_context;
 
   trailcells_deleted = 0;
 
   for( ; fr; fr = query->saved_environment, ch = query->saved_bfr )
   { query = mark_environments(fr, NULL);
-    te    = mark_choicepoints(ch, te);
+    te    = mark_choicepoints(ch, te, &flictx);
 
     assert(query->magic == QID_MAGIC);
   }
   
+  for( ; flictx; flictx = flictx->parent)
+    te = mark_foreign_frame(flictx, te);
+
   DEBUG(2, Sdprintf("Trail stack garbage: %ld cells\n", trailcells_deleted));
 }
 
@@ -1093,7 +1121,6 @@ mark_phase(LocalFrame fr, Choice ch)
   mark_trail();
 #endif
   mark_stacks(fr, ch);
-  mark_foreign_trail_refs();
 #if O_SECURE
   if ( !scan_global(TRUE) )
     sysError("Global stack corrupted after GC mark-phase");
@@ -1153,6 +1180,13 @@ update_relocation_chain(Word current, Word dest ARG_LD)
     current = valPtr(val);
     tag = tag(val);
     val = get_value(current);
+    DEBUG(3,
+	  { FliFrame f;
+
+	    f = addPointer(current, - offset(fliFrame, mark.trailtop));
+	    if ( onStack(local, f) && f->magic == FLI_MAGIC )
+	      Sdprintf("Updating trail-mark of foreign frame at %p\n", f);
+	  })
     set_value(current, makePtr(dest, tag PASS_LD));
     relocated_cells++;
   } while( is_first(current) );
@@ -1737,9 +1771,12 @@ considerGarbageCollect(Stack s)
     long limit = (char *)s->limit - (char *)s->base;
 
     if ( used > s->factor*s->gced_size + s->small )
-    { gc_status.requested = TRUE;
+    { DEBUG(2, Sdprintf("GC: request on %s, factor=%d, last=%ld, small=%ld\n",
+			s->name, s->factor, s->gced_size, s->small));
+      gc_status.requested = TRUE;
     } else if ( free < limit/8 && used > s->gced_size + limit/32 ) 
-    { gc_status.requested = TRUE;
+    { DEBUG(2, Sdprintf("GC: request on low free\n"));
+      gc_status.requested = TRUE;
     }
 
     DEBUG(1, if ( gc_status.requested)
@@ -1926,6 +1963,11 @@ check_trail()
 
       assert(onGlobal(gp));
       key += checkData(gp);
+#ifdef O_SECURE
+    } else
+    { if ( onStackArea(global, te->address) )
+	assert(onStack(global, te->address));
+#endif
     }
   }
 
@@ -2543,7 +2585,7 @@ growStacks(LocalFrame fr, Choice ch, Code PC, long l, long g, long t)
 
     if ( t )
     { tsize = nextStackSize((Stack) &LD->stacks.trail, t);
-      tb = xrealloc(tb, tsize);
+      tb = PL_realloc(tb, tsize);
       shift_status.trail_shifts++;
     }
 
@@ -2873,21 +2915,27 @@ mark_predicates_in_environments(PL_local_data_t *ld, LocalFrame fr)
       def = fr->predicate;
 
     if ( def &&
-	 true(def, NEEDSCLAUSEGC) &&
-	 false(def, DYNAMIC) )
+	 false(def, DYNAMIC) && 
+	 def->references == 0 )		/* already done */
     { if ( GD->procedures.reloading )
-      { ListCell cell;
+      { ListCell cell;			/* startConsult() */
 
 	for(cell=GD->procedures.reloading->procedures; cell; cell=cell->next)
 	{ Procedure proc = cell->value;
+
 	  if ( proc->definition == def )
-	  { def->references++;
+	  { DEBUG(2, Sdprintf("Marking %s\n", predicateName(def)));
+	    def->references++;
 	    GD->procedures.active_marked++;
 	    break;
 	  }
 	}
-      } else
-	def->references++;
+      } else				/* pl_garbage_collect_clauses() */
+      { if ( true(def, NEEDSCLAUSEGC) )
+	{ DEBUG(2, Sdprintf("Marking %s\n", predicateName(def)));
+	  def->references++;
+	}
+      }
     }
 
     if ( fr->parent )
