@@ -1,4 +1,4 @@
-/*  $Id: pl-global.h,v 1.71 2004/09/28 13:44:17 jan Exp $
+/*  $Id: pl-global.h,v 1.80 2006/07/03 15:19:50 jan Exp $
 
     Part of SWI-Prolog
 
@@ -60,7 +60,7 @@ typedef struct
 { atom_t	file;			/* current source file */
   int	  	line;			/* current line */
   int		linepos;		/* position in the line */
-  long		character;		/* current character location */
+  int64_t	character;		/* current character location */
 } source_location;
 
 		 /*******************************
@@ -154,7 +154,7 @@ typedef struct
     real	gc_time;		/* Time spent on atom-gc */
     PL_agc_hook_t gc_hook;		/* Current hook */
 #endif
-    atom_t	for_code[256];		/* code --> one-char-atom */
+    atom_t     *for_code[256];		/* code --> one-char-atom */
     PL_blob_t  *types;			/* registered atom types */
   } atoms;
 
@@ -197,6 +197,10 @@ typedef struct
     FunctorDef* table;			/* hash-table */
   } functors;
 
+  struct
+  { Code	catch_exit_address;	/* See findCatchExit() */
+  } exceptions;
+
   struct
   { TempFile		_tmpfile_head;
     TempFile		_tmpfile_tail;
@@ -219,6 +223,7 @@ typedef struct
     Procedure	true0;
     Procedure	fail0;
     Procedure	event_hook1;
+    Procedure	exception_hook4;
     Procedure	print_message2;
     Procedure	foreign_registered2;	/* $foreign_registered/2 */
     Procedure	prolog_trace_interception4;
@@ -226,6 +231,7 @@ typedef struct
     Procedure   dcall1;			/* $call/1 */
     Procedure	call_cleanup3;		/* call_cleanup/2 */
     Procedure	undefinterc4;		/* $undefined_procedure/4 */
+    Procedure   dthread_init0;		/* $thread_init/0 */
 #ifdef O_ATTVAR
     Procedure	dwakeup1;		/* system:$wakeup/1 */
     Procedure	portray_attvar1;	/* $attvar:portray_attvar/1 */
@@ -307,6 +313,7 @@ typedef struct PL_local_data
   int		in_arithmetic;		/* doing arithmetic */
   int		autoload_nesting;	/* Nesting level in autoloader */
   void *	glob_info;		/* pl-glob.c */
+  IOENC		encoding;		/* default I/O encoding */
 
   struct
   { int		active;			/* doing pipe I/O */
@@ -315,7 +322,7 @@ typedef struct PL_local_data
 
   struct
   { atom_t	current;		/* current global prompt */
-    char *	first;			/* how to prompt first line */
+    atom_t	first;			/* how to prompt first line */
     int		first_used;		/* did we do the first line? */
     int		next;			/* prompt on next read operation */
   } prompt;
@@ -330,6 +337,7 @@ typedef struct PL_local_data
     term_t	printed;		/* already printed exception */
     term_t	tmp;			/* tmp for errors */
     term_t	pending;		/* used by the debugger */
+    int		in_hook;		/* inside exception_hook() */
   } exception;
 
 #ifdef O_ATTVAR
@@ -347,7 +355,7 @@ typedef struct PL_local_data
 #endif
 
   struct
-  { ulong	inferences;		/* inferences in this thread */
+  { int64_t	inferences;		/* inferences in this thread */
     ulong	last_cputime;		/* milliseconds last CPU time */
     ulong	last_systime;		/* milliseconds last SYSTEM time */
     ulong	last_walltime;		/* milliseconds last Wall time */
@@ -420,17 +428,6 @@ typedef struct PL_local_data
 
   struct				/* Local IO stuff */
   { IOSTREAM *streams[6];		/* handles for standard streams */
-#if 0
-    IOSTREAM *user_input;		/* current user input */
-    IOSTREAM *user_output;		/* current user output */
-    IOSTREAM *user_error;		/* current user error */
-    IOSTREAM *curin;			/* current input stream */
-    IOSTREAM *curout;			/* current output stream */
-    IOSTREAM *din;			/* debugger input */
-    IOSTREAM *dout;			/* debuffer output */
-    IOSTREAM *log;			/* stream used for protocolling */
-    IOSTREAM *term;			/* terminal stream */
-#endif
     struct input_context *input_stack;	/* maintain input stream info */
     struct output_context *output_stack; /* maintain output stream info */
   } IO;
