@@ -1,11 +1,11 @@
-/*  $Id: pl-incl.h,v 1.233 2004/09/28 15:11:31 jan Exp $
+/*  $Id: pl-incl.h,v 1.253 2006/07/31 12:10:50 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -37,6 +37,7 @@
 #define C_LIBS	     ""
 #define C_STATICLIBS ""
 #define C_CC	     "cl"
+#define C_CFLAGS     "/MD /GX"
 #define C_LDFLAGS    ""
 #if defined(_DEBUG)
 #define C_PLLIB	    "libplD.lib"
@@ -121,6 +122,8 @@ handy for it someone wants to add a data type to the system.
       requires O_DESTRUCTIVE_ASSIGNMENT.
   O_CYCLIC
       Provide support for cyclic terms.
+  O_GMP
+      Use GNU gmp library for infinite precision arthmetic
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define PL_KERNEL		1
@@ -144,6 +147,9 @@ handy for it someone wants to add a data type to the system.
 #define O_ATTVAR		1
 #define O_GVAR			1
 #define O_CYCLIC		1
+#ifdef HAVE_GMP_H
+#define O_GMP			1
+#endif
 
 #ifndef DOUBLE_TO_LONG_CAST_RAISES_SIGFPE
 #ifdef __i386__
@@ -268,6 +274,7 @@ void *alloca ();
 #include <string.h>
 #include <stddef.h>
 #include <stdarg.h>
+#include <limits.h>
 
 #ifdef HAVE_SIGNAL
 #include <signal.h>
@@ -280,6 +287,10 @@ void *alloca ();
 #endif
 #endif
 
+#ifdef O_GMP
+#include <gmp.h>
+#endif
+
 #if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
 #include <string.h>
 /* An ANSI string.h and pre-ANSI memory.h might conflict.  */
@@ -327,10 +338,6 @@ A common basis for C keywords.
 #define inline
 #endif
 
-#if defined(__GNUC__) && !defined(register)
-#define register
-#endif
-
 #if defined(__GNUC__) && !defined(__OPTIMIZE__)
 #define _DEBUG 1
 #endif
@@ -339,12 +346,18 @@ A common basis for C keywords.
 #define volatile
 #endif
 
-#ifdef HAVE_HIDDEN_ATTRIBUTE
+#ifdef HAVE_VISIBILITY_ATTRIBUTE
 #define SO_LOCAL __attribute__((visibility("hidden")))
 #else
 #define SO_LOCAL
 #endif
 
+#if defined(__GNUC__) && !defined(NORETURN)
+#define NORETURN __attribute__ ((noreturn))
+#else
+#define NORETURN
+#endif
+
 #if __STRICT_ANSI__
 #undef TAGGED_LVALUE
 #endif
@@ -438,14 +451,16 @@ them.  Descriptions:
 #define WORDBITSIZE		(8 * sizeof(word))
 #define LONGBITSIZE		(8 * sizeof(long))
 #define INTBITSIZE		(8 * sizeof(int))
+#define INT64BITSIZE		(8 * sizeof(int64_t))
 #define WORDS_PER_DOUBLE        ((sizeof(double)+sizeof(word)-1)/sizeof(word))
+#define WORDS_PER_INT64		(sizeof(int64_t)/sizeof(word))
 
 				/* Prolog's integer range */
 #define PLMINTAGGEDINT		(-(long)(1L<<(WORDBITSIZE - LMASK_BITS - 1)))
 #define PLMAXTAGGEDINT		(-PLMINTAGGEDINT - 1)
 #define inTaggedNumRange(n)	(((n)&~PLMAXTAGGEDINT) == 0 || \
 				 ((n)&~PLMAXTAGGEDINT) == ~PLMAXTAGGEDINT)
-#define PLMININT		((long)(-1L<<(WORDBITSIZE-1)))
+#define PLMININT		(((int64_t)-1<<(INT64BITSIZE-1)))
 #define PLMAXINT		(-(PLMININT+1))
 
 #if vax
@@ -470,14 +485,14 @@ sizes  of  the  hash  tables are defined.  Note that these should all be
 #define ARITHHASHSIZE		64	/* arithmetic function table */
 
 #define TABLE_UNLOCKED		0x10000000L /* do not create mutex for table */
-#define TABLE_MASK		0xf0000000L
+#define TABLE_MASK		0xf0000000UL
 
 #define pointerHashValue(p, size) ((((long)(p) >> LMASK_BITS) ^ \
 				    ((long)(p) >> (LMASK_BITS+5)) ^ \
 				    ((long)(p))) & \
 				   ((size)-1))
 
-#define TABLE_REF_MASK		0x1L
+#define TABLE_REF_MASK		0x1UL
 #define isTableRef(p)		((unsigned long)(p) & TABLE_REF_MASK)
 #define makeTableRef(p)		((void*)((unsigned long)(p) | TABLE_REF_MASK))
 #define unTableRef(s, p)	(*((s*)((unsigned long)(p) & ~TABLE_REF_MASK)))
@@ -583,103 +598,109 @@ codes.
 #define H_VAR		((code)11)
 #define B_CONST		((code)12)		/* constant (atomic) */
 #define H_CONST		((code)13)
-#define H_INDIRECT	((code)14)		/* indirect in the head */
-#define B_INTEGER	((code)15)		/* bignum in the head */
-#define H_INTEGER	((code)16)		/* bignum in the body */
-#define B_FLOAT		((code)17)		/* double in the head */
-#define H_FLOAT		((code)18)		/* double in the body */
+#define B_STRING	((code)14)		/* String in body */
+#define H_STRING	((code)15)		/* String in the head */
+#define B_MPZ		((code)16)		/* String in body */
+#define H_MPZ		((code)17)		/* String in the head */
+#define B_INTEGER	((code)18)		/* bignum in the head */
+#define H_INTEGER	((code)19)		/* bignum in the body */
+#define B_INT64		((code)20)		/* bignum in the head */
+#define H_INT64		((code)21)		/* bignum in the body */
+#define B_FLOAT		((code)22)		/* double in the head */
+#define H_FLOAT		((code)23)		/* double in the body */
 
-#define B_FIRSTVAR	((code)19)		/* first occurrence of var */
-#define H_FIRSTVAR	((code)20)
-#define B_VOID		((code)21)		/* anonimous variables */
-#define H_VOID		((code)22)
-#define B_ARGFIRSTVAR	((code)23)		/* body vars nested in funct */
-#define B_ARGVAR	((code)24)
+#define B_FIRSTVAR	((code)24)		/* first occurrence of var */
+#define H_FIRSTVAR	((code)25)
+#define B_VOID		((code)26)		/* anonimous variables */
+#define H_VOID		((code)27)
+#define B_ARGFIRSTVAR	((code)28)		/* body vars nested in funct */
+#define B_ARGVAR	((code)29)
 
-#define H_NIL		((code)25)		/* [] in the head */
-#define B_NIL		((code)26)		/* [] in the body */
-#define H_LIST		((code)27)		/* ./2 in the head */
-#define H_RLIST		((code)28)		/* right-recursive list */
-#define B_LIST		((code)29)		/* ./2 in the body */
-#define B_RLIST		((code)30)		/* right-recursive list */
+#define H_NIL		((code)30)		/* [] in the head */
+#define B_NIL		((code)31)		/* [] in the body */
+#define H_LIST		((code)32)		/* ./2 in the head */
+#define H_RLIST		((code)33)		/* right-recursive list */
+#define B_LIST		((code)34)		/* ./2 in the body */
+#define B_RLIST		((code)35)		/* right-recursive list */
 
-#define B_VAR0		((code)31)		/* B_VAR 0 */
-#define B_VAR1		((code)32)		/* B_VAR 1 */
-#define B_VAR2		((code)33)		/* B_VAR 2 */
+#define B_VAR0		((code)36)		/* B_VAR 0 */
+#define B_VAR1		((code)37)		/* B_VAR 1 */
+#define B_VAR2		((code)38)		/* B_VAR 2 */
 
-#define I_USERCALL0	((code)34)		/* variable in body (call/1) */
-#define I_USERCALLN	((code)35)		/* call/[2...] */
-#define I_CUT		((code)36)		/* ! */
-#define I_APPLY		((code)37)		/* apply/2 */
+#define I_USERCALL0	((code)39)		/* variable in body (call/1) */
+#define I_USERCALLN	((code)40)		/* call/[2...] */
+#define I_CUT		((code)41)		/* ! */
+#define I_APPLY		((code)42)		/* apply/2 */
 
 #if O_COMPILE_ARITH
-#define A_ENTER		((code)38)		/* start arithmetic sequence */
-#define A_INTEGER	((code)39)		/* 32-bit signed int */
-#define A_DOUBLE	((code)40)		/* 64-bit double */
-#define A_VAR0		((code)41)		/* variable-0 */
-#define A_VAR1		((code)42)		/* variable-1 */
-#define A_VAR2		((code)43)		/* variable-2 */
-#define A_VAR		((code)44)		/* variable-n */
-#define A_FUNC0		((code)45)		/* nullary arithmic function */
-#define A_FUNC1		((code)46)		/* unary arithmic function */
-#define A_FUNC2		((code)47)		/* binary arithmic function */
-#define A_FUNC		((code)48)		/* n-ary arithmic function */
-#define A_LT		((code)49)		/* < */
-#define A_GT		((code)50)		/* > */
-#define A_LE		((code)51)		/* =< */
-#define A_GE		((code)52)		/* >= */
-#define A_EQ		((code)53)		/* =:= */
-#define A_NE		((code)54)		/* =\= */
-#define A_IS		((code)55)		/* is */
+#define A_ENTER		((code)43)		/* start arithmetic sequence */
+#define A_INTEGER	((code)44)		/* 32-bit signed int */
+#define A_INT64		((code)45)		/* 64-bit signed int */
+#define A_MPZ		((code)46)		/* GMP mpz_t */
+#define A_DOUBLE	((code)47)		/* 64-bit double */
+#define A_VAR0		((code)48)		/* variable-0 */
+#define A_VAR1		((code)49)		/* variable-1 */
+#define A_VAR2		((code)50)		/* variable-2 */
+#define A_VAR		((code)51)		/* variable-n */
+#define A_FUNC0		((code)52)		/* nullary arithmic function */
+#define A_FUNC1		((code)53)		/* unary arithmic function */
+#define A_FUNC2		((code)54)		/* binary arithmic function */
+#define A_FUNC		((code)55)		/* n-ary arithmic function */
+#define A_LT		((code)56)		/* < */
+#define A_GT		((code)57)		/* > */
+#define A_LE		((code)58)		/* =< */
+#define A_GE		((code)59)		/* >= */
+#define A_EQ		((code)60)		/* =:= */
+#define A_NE		((code)61)		/* =\= */
+#define A_IS		((code)62)		/* is */
 #endif /* O_COMPILE_ARITH */
 
 #if O_COMPILE_OR
-#define C_OR		((code)56)		/* In-clause backtract point */
-#define C_JMP		((code)57)		/* Jump over code */
-#define C_MARK		((code)58)		/* Sub-clause cut mark */
-#define C_CUT		((code)59)		/* cut to corresponding mark */
-#define C_IFTHENELSE	((code)60)		/* if-then-else start */
-#define C_VAR		((code)61)		/* make a variable */
-#define C_END		((code)62)		/* dummy to help decompiler */
-#define C_NOT		((code)63)		/* same as C_IFTHENELSE */
-#define C_FAIL		((code)64)		/* fail */
+#define C_OR		((code)63)		/* In-clause backtract point */
+#define C_JMP		((code)64)		/* Jump over code */
+#define C_MARK		((code)65)		/* Sub-clause cut mark */
+#define C_CUT		((code)66)		/* cut to corresponding mark */
+#define C_IFTHENELSE	((code)67)		/* if-then-else start */
+#define C_VAR		((code)68)		/* make a variable */
+#define C_END		((code)69)		/* dummy to help decompiler */
+#define C_NOT		((code)70)		/* same as C_IFTHENELSE */
+#define C_FAIL		((code)71)		/* fail */
 #endif /* O_COMPILE_OR */
 
-#define B_INDIRECT	((code)65)		/* INDIRECT in body */
-
 #if O_BLOCK
-#define I_CUT_BLOCK	((code)66)		/* !(block) */
-#define B_EXIT		((code)67)		/* exit(block, rval) */
+#define I_CUT_BLOCK	((code)72)		/* !(block) */
+#define B_EXIT		((code)73)		/* exit(block, rval) */
 #endif /*O_BLOCK*/
 
 #if O_INLINE_FOREIGNS
-#define I_CALL_FV0	((code)68)		/* call foreign, no args */
-#define I_CALL_FV1	((code)69)		/* call foreign, 1 var arg */
-#define I_CALL_FV2	((code)70)		/* call foreign, 2 var args */
+#define I_CALL_FV0	((code)74)		/* call foreign, no args */
+#define I_CALL_FV1	((code)75)		/* call foreign, 1 var arg */
+#define I_CALL_FV2	((code)76)		/* call foreign, 2 var args */
 #endif /*O_INLINE_FOREIGNS*/
 
-#define I_FAIL		((code)71)		/* fail */
-#define I_TRUE		((code)72)		/* true */
+#define I_FAIL		((code)77)		/* fail */
+#define I_TRUE		((code)78)		/* true */
 
 #ifdef O_SOFTCUT
-#define C_SOFTIF	((code)73)		/* Start A *-> B ; C */
-#define C_SOFTCUT	((code)74)		/* `Cut' of A *-> B ; C */
+#define C_SOFTIF	((code)79)		/* Start A *-> B ; C */
+#define C_SOFTCUT	((code)80)		/* `Cut' of A *-> B ; C */
 #endif /*O_SOFTCUT*/
 
-#define I_EXITFACT	((code)75)		/* exit from a fact */
-#define D_BREAK		((code)76)		/* Debugger break-point */
+#define I_EXITFACT	((code)81)		/* exit from a fact */
+#define D_BREAK		((code)82)		/* Debugger break-point */
 
 #if O_CATCHTHROW
-#define I_CATCH		((code)77)		/* $catch (catch/3) */
-#define B_THROW		((code)78)		/* throw(Exception) */
+#define I_CATCH		((code)83)		/* $catch (catch/3) */
+#define I_EXITCATCH	((code)84)		/* $exit_catch (catch/3) */
+#define B_THROW		((code)85)		/* throw(Exception) */
 #endif
 
-#define I_CONTEXT	((code)79)		/* Push context module */
-#define C_LCUT		((code)80)		/* ! local in \+ and -> */
-#define I_CALLCLEANUP	((code)81)		/* $call_cleanup */
-#define I_EXITCLEANUP	((code)82)		/* $exit_cleanup */
+#define I_CONTEXT	((code)86)		/* Push context module */
+#define C_LCUT		((code)87)		/* ! local in \+ and -> */
+#define I_CALLCLEANUP	((code)88)		/* $call_cleanup */
+#define I_EXITCLEANUP	((code)89)		/* $exit_cleanup */
 
-#define I_HIGHEST	((code)82)		/* largest WAM code !!! */
+#define I_HIGHEST	((code)89)		/* largest WAM code !!! */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Arithmetic comparison
@@ -871,20 +892,32 @@ typedef struct feature *	Feature; 	/* pl-prims.c */
 		 *	    ARITHMETIC		*
 		 *******************************/
 
+/* the numtype enum requires total ordering.
+*/
+
+typedef enum
+{ V_INTEGER,				/* integer (64-bit) value */
+#ifdef O_GMP    
+  V_MPZ,				/* mpz_t */
+  V_MPQ,				/* mpq_t */
+#endif
+  V_REAL				/* Floating point number (double) */
+} numtype;
+
 typedef struct
-{ int	type;				/* type of number */
-  union { real  f;			/* value as real */
-	  long  i;			/* value as integer */
+{ numtype type;				/* type of number */
+  union { double f;			/* value as real */
+	  int64_t i;			/* value as integer */
 	  word  w[WORDS_PER_DOUBLE];	/* for packing/unpacking the double */
+#ifdef O_GMP
+	  mpz_t mpz;			/* GMP integer */
+	  mpq_t mpq;			/* GMP rational */
+#endif
 	} value;
 } number, *Number;
 
-#define V_INTEGER	0		/* integer (long) value */
-#define V_REAL		1		/* just a real */
-#define V_EXPLICIT_REAL 3		/* explicely specified real */
-
-#define intNumber(n)	((n)->type == V_INTEGER)
-#define floatNumber(n)	((n)->type & V_REAL)
+#define intNumber(n)	((n)->type <  V_REAL)
+#define floatNumber(n)	((n)->type >= V_REAL)
 
 		 /*******************************
 		 *	   GET-PROCEDURE	*
@@ -965,6 +998,7 @@ with one operation, it turns out to be faster as well.
 #define PROC_DEFINED		(DYNAMIC|FOREIGN|MULTIFILE|DISCONTIGUOUS)
 #define P_THREAD_LOCAL		(0x01000000L) /* predicate */
 #define P_FOREIGN_CREF		(0x02000000L) /* predicate */
+#define P_DIRTYREG		(0x04000000L) /* predicate */
 
 #define ERASED			(0x0001) /* clause, record */
 					 /* clause flags */
@@ -1124,6 +1158,20 @@ We assume the compiler will optimise this properly.
 	  } \
 	}
 
+#define cpInt64Data(to, from) \
+	{ Word _f = (Word)(from); \
+	  switch(WORDS_PER_INT64) \
+	  { case 2: \
+	      *(to)++ = *_f++; \
+	    case 1: \
+	      *(to)++ = *_f++; \
+	      from = (void *)_f; \
+	      break; \
+	    default: \
+	      assert(0); \
+	  } \
+	}
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Structure declarations that must be shared across multiple files.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -1231,10 +1279,12 @@ struct clause_ref
 #define CA1_FUNC	2	/* code arg 1 is functor */
 #define CA1_DATA	3	/* code arg 2 is prolog data */
 #define CA1_INTEGER	4	/* long value */
-#define CA1_FLOAT	5	/* next WORDS_PER_DOUBLE are double */
-#define CA1_STRING	6	/* inlined string */
-#define CA1_MODULE	7	/* a module */
-#define CA1_VAR		8	/* a variable(-offset) */
+#define CA1_INT64	5	/* int64 value */
+#define CA1_FLOAT	6	/* next WORDS_PER_DOUBLE are double */
+#define CA1_STRING	7	/* inlined string */
+#define CA1_MODULE	8	/* a module */
+#define CA1_VAR		9	/* a variable(-offset) */
+#define CA1_MPZ	       10	/* GNU mpz number */
 
 typedef struct
 { char		*name;		/* name of the code */
@@ -1863,6 +1913,29 @@ typedef enum
   AV_ERROR
 } av_action;
 
+typedef struct
+{ functor_t functor;			/* Functor to use ($VAR/1) */
+  av_action on_attvar;			/* How to handle attvars */
+  int	    singletons;			/* Write singletons as $VAR('_') */
+} nv_options;
+
+
+		 /*******************************
+		 *	    STREAM I/O		*
+		 *******************************/
+
+typedef struct redir_context
+{ IOSTREAM     *stream;			/* temporary output */
+  int		is_stream;		/* redirect to stream */
+  int		redirected;		/* output is redirected */
+  term_t	term;			/* redirect target */
+  int		out_format;		/* output type */
+  int		out_arity;		/* 2 for difference-list versions */
+  int		size;			/* size of I/O buffer */
+  char	       *data;			/* data written */
+  char		buffer[1024];		/* fast temporary buffer */
+} redir_context;
+
 
 		/********************************
 		*       READ WARNINGS           *
@@ -1895,8 +1968,8 @@ typedef struct
   int		blocked;		/* GC is blocked now */
   bool		active;			/* Currently running? */
   long		collections;		/* # garbage collections */
-  long		global_gained;		/* global stack bytes collected */
-  long		trail_gained;		/* trail stack bytes collected */
+  int64_t	global_gained;		/* global stack bytes collected */
+  int64_t	trail_gained;		/* trail stack bytes collected */
   real		time;			/* time spent in collections */
 } pl_gc_status_t;
 
@@ -1935,6 +2008,8 @@ typedef struct
 #define PROCEDURE_dcall1		(GD->procedures.dcall1)
 #define PROCEDURE_call_cleanup3		(GD->procedures.call_cleanup3)
 #define PROCEDURE_dwakeup1		(GD->procedures.dwakeup1)
+#define PROCEDURE_dthread_init0		(GD->procedures.dthread_init0)
+#define PROCEDURE_exception_hook4	(GD->procedures.exception_hook4)
 
 extern const code_info codeTable[]; /* Instruction info (read-only) */
 
@@ -1991,6 +2066,7 @@ typedef struct debuginfo
 #define FT_BOOL		1		/* boolean feature (true, false) */
 #define FT_INTEGER	2		/* integer feature */
 #define FT_TERM		3		/* term feature */
+#define FT_INT64	4		/* passed as int64_t */
 #define FT_MASK		0x0f		/* mask to get type */
 
 #define FF_READONLY	0x10		/* feature is read-only */
@@ -2091,6 +2167,8 @@ decrease).
 #include "pl-thread.h"			/* thread manipulation */
 #include "pl-global.h"			/* global data */
 #include "pl-funcs.h"			/* global functions */
+#include "pl-text.h"			/* text manipulation */
+#include "pl-gmp.h"			/* GNU-GMP support */
 
 #ifdef __DECC				/* Dec C-compiler: avoid conflicts */
 #undef leave
