@@ -1,4 +1,4 @@
-/*  $Id: pl-index.c,v 1.49 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-index.c,v 1.58 2006/07/31 12:10:50 jan Exp $
 
     Part of SWI-Prolog
 
@@ -100,8 +100,8 @@ static unsigned long mask_mask[] =
 
 
 int
-cardinalityPattern(register unsigned long pattern)
-{ register int result = 0;
+cardinalityPattern(unsigned long pattern)
+{ int result = 0;
 
   for(; pattern; pattern >>= 1)
     if (pattern & 0x1)
@@ -117,7 +117,11 @@ of buckets. This used to be simple, but now that our tag bits are on the
 left side, simply masking will put most things on the same hash-entry as
 it is very common for all clauses of   a predicate to have the same type
 of object. Hence, we now use exclusive or of the real value part and the
-tag-bits.
+tag-bits. 
+
+NOTE: this function must be kept consistent with arg1Key() in pl-comp.c!
+NOTE: This function returns 0 on non-indexable   fields, which is why we
+guarantee that the value is non-0 for indexable values.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static inline int
@@ -135,13 +139,45 @@ indexOfWord(word w ARG_LD)
     { case TAG_VAR:
       case TAG_ATTVAR:
       case TAG_STRING:
-      case TAG_FLOAT:
 	return 0L;
       case TAG_INTEGER:
 	if ( storage(w) != STG_INLINE )
-	  return valBignum(w);
+	{ Word p = valIndirectP(w);
+	  word key;
+
+#if SIZEOF_LONG == 4
+          DEBUG(9, Sdprintf("Index for " INT64_FORMAT " = 0x%x\n",
+			    valBignum(w), p[0]^p[1]));
+	  key = p[0]^p[1];
+#else
+ 	  key = p[0];
+#endif
+	  if ( !key )
+	    key++;
+          return key;
+	}
       case TAG_ATOM:
 	break;				/* atom_t */
+      case TAG_FLOAT:
+      { Word p = valIndirectP(w);
+	word key;
+	
+	switch(WORDS_PER_DOUBLE)
+	{ case 2:
+	    key = p[0]^p[1];
+	    break;
+	  case 1:
+	    key = p[0];
+	    break;
+	  default:
+	    assert(0);
+	    return 0L;
+	}
+
+	if ( !key )
+	  key++;
+	return key;
+      }
       case TAG_COMPOUND:
 	w = *valPtr(w);			/* functor_t */
 	break;
@@ -347,8 +383,7 @@ findClause(ClauseRef cref, Word argv,
 
     return nextClauseArg1(cref, gen, next, key PASS_LD);
   } else if ( def->indexPattern & NEED_REINDEX )
-  { assert(def->hash_info == NULL);
-    reindexDefinition(def);
+  { reindexDefinition(def);
     return findClause(cref, argv, fr, def, next PASS_LD);
   } else
   { return nextClauseMultiIndexed(cref, gen, argv, def, next PASS_LD);
@@ -661,7 +696,6 @@ hashDefinition(Definition def, int buckets)
   }
 
   succeed;
-
 }
 
 word
