@@ -1,4 +1,4 @@
-/*  $Id: pl-itf.h,v 1.235.2.4 2005/02/21 09:47:46 jan Exp $
+/*  $Id: pl-itf.h,v 1.329 2006/08/31 09:44:01 jan Exp $
 
     Part of SWI-Prolog
 
@@ -27,6 +27,12 @@
 
 #include <stdarg.h>
 #include <stdlib.h>			/* get size_t */
+#ifdef WIN32
+typedef __int64 int64_t;
+typedef unsigned __int64 uint64_t;
+#else
+#include <inttypes.h>			/* more portable than stdint.h */
+#endif
 
 #ifdef __cplusplus
 extern "C" {
@@ -44,7 +50,7 @@ before loading this file.  See end of this file.
 /* PLVERSION: 10000 * <Major> + 100 * <Minor> + <Patch> */
 
 #ifndef PLVERSION
-#define PLVERSION 50407
+#define PLVERSION 50618
 #endif
 
 		 /*******************************
@@ -60,7 +66,7 @@ special  declarations  on  exported  symbols.  So,  a  symbol  is  local
 (static), shared between the objects building   an executable or DLL (no
 special declaration) or exported from the executable or DLL.
 
-Both using native Microsoft MSVC as well   as recent CygWin (tested 1.1)
+Both using native Microsoft MSVC as well   as recent Cygwin (tested 1.1)
 compilers support __declspec(...) for exporting symbols.
 
 As SWI-Prolog.h can be included seperately or together with this file we
@@ -73,7 +79,7 @@ duplicated this stuff.
 #define __WIN32__
 #endif
 
-#if (defined(__WIN32__) || defined(__CYGWIN32__)) && !defined(__LCC__)
+#if (defined(__WIN32__) || defined(__CYGWIN__)) && !defined(__LCC__)
 #define HAVE_DECLSPEC
 #endif
 
@@ -123,17 +129,22 @@ typedef void *		PL_engine_t;	/* opaque engine handle */
 typedef unsigned long	functor_t;	/* Name/arity pair */
 typedef unsigned long	PL_atomic_t;	/* same a word */
 typedef unsigned long	foreign_t;	/* return type of foreign functions */
+typedef wchar_t	        pl_wchar_t;	/* Prolog wide character */
 #ifdef __cplusplus
 typedef void * 		pl_function_t;	/* can only pass function as void * */
 #else
 typedef foreign_t	(*pl_function_t)(); /* foreign language functions */
 #endif
 
+#ifndef NORETURN
+#define NORETURN
+#endif
+
 #define fid_t PL_fid_t			/* avoid AIX name-clash */
 
 					/* values for PL_get_term_value() */
 typedef union
-{ long   i;				/* PL_INTEGER */
+{ int64_t i;				/* PL_INTEGER */
   double f;				/* PL_FLOAT */
   char * s;				/* PL_STRING */
   atom_t a;				/* PL_ATOM */
@@ -151,7 +162,6 @@ typedef union
 		 /*******************************
 		 *      TERM-TYPE CONSTANTS	*
 		 *******************************/
-
 					/* PL_unify_term() arguments */
 #define	PL_VARIABLE	 (1)		/* nothing */
 #define PL_ATOM		 (2)		/* const char * */
@@ -176,6 +186,18 @@ typedef union
 #define PL_LONG		 (21)		/* long */
 #define PL_DOUBLE	 (22)		/* double */
 #define PL_NCHARS	 (23)		/* unsigned, const char * */
+#define PL_UTF8_CHARS	 (24)		/* const char * */
+#define PL_UTF8_STRING	 (25)		/* const char * */
+#define PL_INT64	 (26)		/* int64_t */
+#define PL_NUTF8_CHARS	 (27)		/* unsigned, const char * */
+#define PL_NUTF8_CODES	 (29)		/* unsigned, const char * */
+#define PL_NUTF8_STRING	 (30)		/* unsigned, const char * */
+#define PL_NWCHARS	 (31)		/* unsigned, const wchar_t * */
+#define PL_NWCODES	 (32)		/* unsigned, const wchar_t * */
+#define PL_NWSTRING	 (33)		/* unsigned, const wchar_t * */
+#define PL_MBCHARS	 (34)		/* const char * */
+#define PL_MBCODES	 (35)		/* const char * */
+#define PL_MBSTRING	 (36)		/* const char * */
 
 
 		/********************************
@@ -214,7 +236,7 @@ PL_EXPORT(void *)	PL_foreign_context_address(control_t);
 		*      REGISTERING FOREIGNS     *
 		*********************************/
 
-typedef struct _PL_extension
+typedef struct PL_extension
 { char 		*predicate_name;	/* Name of the predicate */
   short		arity;			/* Arity of the predicate */
   pl_function_t	function;		/* Implementing functions */
@@ -229,9 +251,13 @@ typedef struct _PL_extension
 
 extern			PL_extension PL_extensions[]; /* not Win32! */
 PL_EXPORT(void)		PL_register_extensions(const PL_extension *e);
-PL_EXPORT(void)		PL_load_extensions(const PL_extension *e);
+PL_EXPORT(void)		PL_register_extensions_in_module(const char *module, const PL_extension *e);
 PL_EXPORT(int)		PL_register_foreign(const char *name, int arity,
 					    pl_function_t func, int flags);
+PL_EXPORT(int)		PL_register_foreign_in_module(const char *module,
+						      const char *name, int arity,
+						      pl_function_t func, int flags);
+PL_EXPORT(void)		PL_load_extensions(const PL_extension *e);
 
 
 		 /*******************************
@@ -309,8 +335,10 @@ PL_EXPORT(void)		PL_reset_term_refs(term_t r);
 			/* Constants */
 PL_EXPORT(atom_t)	PL_new_atom(const char *s);
 PL_EXPORT(atom_t)	PL_new_atom_nchars(unsigned int len, const char *s);
+PL_EXPORT(atom_t)	PL_new_atom_wchars(unsigned int len, const wchar_t *s);
 PL_EXPORT(const char *)	PL_atom_chars(atom_t a);
 PL_EXPORT(const char *) PL_atom_nchars(atom_t a, unsigned int *len);
+PL_EXPORT(const wchar_t *) PL_atom_wchars(atom_t a, unsigned int *len);
 #ifndef O_DEBUG_ATOMGC
 PL_EXPORT(void)		PL_register_atom(atom_t a);
 PL_EXPORT(void)		PL_unregister_atom(atom_t a);
@@ -360,6 +388,7 @@ PL_EXPORT(int)		PL_is_atom(term_t t);
 PL_EXPORT(int)		PL_is_integer(term_t t);
 PL_EXPORT(int)		PL_is_string(term_t t);
 PL_EXPORT(int)		PL_is_float(term_t t);
+PL_EXPORT(int)		PL_is_rational(term_t t);
 PL_EXPORT(int)		PL_is_compound(term_t t);
 PL_EXPORT(int)		PL_is_functor(term_t t, functor_t f);
 PL_EXPORT(int)		PL_is_list(term_t t);
@@ -419,6 +448,32 @@ PL_EXPORT(int)		PL_unify_list(term_t l, term_t h, term_t t);
 PL_EXPORT(int)		PL_unify_nil(term_t l);
 PL_EXPORT(int)		PL_unify_arg(int index, term_t t, term_t a);
 PL_EXPORT(int)		PL_unify_term(term_t t, ...);
+PL_EXPORT(int)		PL_unify_chars(term_t t, int flags,
+				       unsigned int len, const char *s);
+
+
+		 /*******************************
+		 *    WIDE CHARACTER VERSIONS	*
+		 *******************************/
+
+PL_EXPORT(int)		PL_unify_wchars(term_t t, int type,
+					unsigned int len, const pl_wchar_t *s);
+PL_EXPORT(int)		PL_unify_wchars_diff(term_t t, term_t tail, int type,
+					unsigned int len, const pl_wchar_t *s);
+PL_EXPORT(int)		PL_get_wchars(term_t l,
+				      unsigned int *length, pl_wchar_t **s,
+				      unsigned flags);
+PL_EXPORT(unsigned int)	PL_utf8_strlen(const char *s, unsigned int len);
+
+
+		 /*******************************
+		 *	   WIDE INTEGERS	*
+		 *******************************/
+
+
+PL_EXPORT(int)		PL_get_int64(term_t t, int64_t *i);
+PL_EXPORT(int)		PL_unify_int64(term_t t, int64_t value);
+PL_EXPORT(void)		PL_put_int64(term_t t, int64_t i);
 
 
 		 /*******************************
@@ -440,6 +495,7 @@ PL_EXPORT(int)		PL_get_attr(term_t v, term_t a);
 #define PL_BLOB_UNIQUE	0x01		/* Blob content is unique */
 #define PL_BLOB_TEXT	0x02		/* blob contains text */
 #define PL_BLOB_NOCOPY	0x04		/* do not copy the data */
+#define PL_BLOB_WCHAR	0x08		/* wide character string */
 
 typedef struct PL_blob_t
 { unsigned long		magic;		/* PL_BLOB_MAGIC */
@@ -453,8 +509,15 @@ typedef struct PL_blob_t
   int			(*write)(void *s, atom_t a, int flags);
 #endif
   void			(*acquire)(atom_t a);
+#ifdef SIO_MAGIC
+  int			(*save)(atom_t a, IOSTREAM *s);
+  atom_t		(*load)(IOSTREAM *s);
+#else
+  int			(*save)(atom_t a, void*);
+  atom_t		(*load)(void *s);
+#endif
 					/* private */
-  void *		reserved[12];	/* for future extension */
+  void *		reserved[10];	/* for future extension */
   int			registered;	/* Already registered? */
   int			rank;		/* Rank for ordering atoms */
   struct PL_blob_t *    next;		/* next in registered type-chain */
@@ -472,9 +535,25 @@ PL_EXPORT(int)		PL_get_blob(term_t t, void **blob, unsigned int *len,
 PL_EXPORT(void*)	PL_blob_data(atom_t a,
 				     unsigned int *len,
 				     struct PL_blob_t **type);
+
+PL_EXPORT(void)		PL_register_blob_type(PL_blob_t *type);
+PL_EXPORT(PL_blob_t*)	PL_find_blob_type(const char* name);
 PL_EXPORT(int)		PL_unregister_blob_type(PL_blob_t *type);
 
 
+#ifdef __GNU_MP__
+
+		 /*******************************
+		 *	       GMP		*
+		 *******************************/
+
+PL_EXPORT(int)	PL_get_mpz(term_t t, mpz_t mpz);
+PL_EXPORT(int)	PL_get_mpq(term_t t,  mpq_t mpq);
+PL_EXPORT(int)	PL_unify_mpz(term_t t, mpz_t mpz);
+PL_EXPORT(int)	PL_unify_mpq(term_t t, mpq_t mpq);
+
+#endif /*__GNU_MP__*/
+
 		 /*******************************
 		 *	  FILENAME SUPPORT	*
 		 *******************************/
@@ -490,7 +569,7 @@ PL_EXPORT(int)		PL_unregister_blob_type(PL_blob_t *type);
 
 PL_EXPORT(int)		PL_get_file_name(term_t n, char **name, int flags);
 PL_EXPORT(void)		PL_changed_cwd(void); /* foreign code changed CWD */
-PL_EXPORT(const char *) PL_cwd();
+PL_EXPORT(const char *) PL_cwd(void);
 
 
 		 /*******************************
@@ -572,6 +651,24 @@ PL_EXPORT(void)		_PL_get_arg(int index, term_t t, term_t a);
 #define BUF_RING	0x0100
 #define BUF_MALLOC	0x0200
 
+#define CVT_EXCEPTION	0x10000		/* throw exception on error */
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Output   representation   for   PL_get_chars()     and    friends.   The
+prepresentation type REP_FN is for   PL_get_file_name()  and friends. On
+Windows we use UTF-8 which is translated   by the `XOS' layer to Windows
+UNICODE file functions.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#define REP_ISO_LATIN_1 0x0000		/* output representation */
+#define REP_UTF8 	0x1000
+#define REP_MB		0x2000
+#ifdef WIN32
+#define REP_FN		REP_UTF8
+#else
+#define REP_FN		REP_MB
+#endif
+
 #ifdef SIO_MAGIC			/* defined from <SWI-Stream.h> */
 		 /*******************************
 		 *	  STREAM SUPPORT	*
@@ -635,10 +732,14 @@ PL_EXPORT(int) PL_chars_to_term(const char *chars,
 
 PL_EXPORT(int)		PL_initialise(int argc, char **argv);
 PL_EXPORT(int)		PL_is_initialised(int *argc, char ***argv);
+#ifdef __CYGWIN__
+PL_EXPORT(void)		PL_install_readline(void);
+#else
 install_t		PL_install_readline(void);
+#endif
 PL_EXPORT(int)		PL_toplevel(void);
 PL_EXPORT(int)		PL_cleanup(int status);
-PL_EXPORT(int)		PL_halt(int status);
+PL_EXPORT(void)		PL_halt(int status) NORETURN;
 
 		 /*******************************
 		 *      INPUT/PROMPT/ETC	*
@@ -662,6 +763,10 @@ PL_EXPORT(char *)	PL_prompt_string(int fd);
 PL_EXPORT(void)		PL_write_prompt(int dowrite);
 PL_EXPORT(void)		PL_prompt_next(int fd);
 PL_EXPORT(char *)	PL_atom_generator(const char *prefix, int state);
+PL_EXPORT(pl_wchar_t*)  PL_atom_generator_w(const pl_wchar_t *pref,
+					    pl_wchar_t *buffer,
+					    unsigned int buflen,
+					    int state);
 PL_EXPORT(void)		PL_clock_wait_ticks(long waited);
 
 
@@ -739,6 +844,8 @@ PL_EXPORT(void) PL_on_halt(void (*)(int, void *), void *);
 #define PL_QUERY_MIN_TAGGED_INT	9	/* smallest tagged integer */
 #define PL_QUERY_VERSION        10	/* 207006 = 2.7.6 */
 #define PL_QUERY_MAX_THREADS	11	/* maximum thread count */
+#define PL_QUERY_ENCODING	12	/* I/O encoding */
+#define PL_QUERY_USER_CPU	13	/* User CPU in milliseconds */
 
 PL_EXPORT(long)	PL_query(int);	/* get information from Prolog */
 
@@ -878,7 +985,7 @@ typedef fid_t			bktrk_buf;
 		 *******************************/
 
 PL_EXPORT(PL_atomic_t)	_PL_atomic(term_t t);
-PL_EXPORT(long)		_PL_integer_value(PL_atomic_t t);
+PL_EXPORT(int64_t)	_PL_integer_value(PL_atomic_t t);
 PL_EXPORT(double)	_PL_float_value(PL_atomic_t t);
 PL_EXPORT(char *)	_PL_string_value(PL_atomic_t t);
 PL_EXPORT(char *)	_PL_list_string_value(term_t t);
