@@ -1,11 +1,11 @@
-/*  $Id: pl-main.c,v 1.145 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-main.c,v 1.162 2006/02/22 12:07:06 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2005, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -91,15 +91,17 @@ findHome(char *symbols)
   
   if ( (val  = exec_var("homevar")) &&
        (home = getenv3(val, envbuf, sizeof(envbuf))) &&
-       (home = PrologPath(home, plp)) )
+       (home = PrologPath(home, plp, sizeof(plp))) )
     return store_string(home);
   if ( (val = exec_var("home")) &&
-       (home = PrologPath(home, plp)) )
+       (home = PrologPath(val, plp, sizeof(plp))) )
     return store_string(home);
 
   if ( !(home = getenv3("SWI_HOME_DIR", envbuf, sizeof(envbuf))) )
     home = getenv3("SWIPL", envbuf, sizeof(envbuf));
-  if ( home && (home = PrologPath(home, plp)) && ExistsDirectory(home) )
+  if ( home &&
+       (home = PrologPath(home, plp, sizeof(plp))) &&
+       ExistsDirectory(home) )
     return store_string(home);
 
   if ( (home = symbols) )
@@ -120,7 +122,7 @@ findHome(char *symbols)
 
 #if O_XOS
       { char buf2[MAXPATHLEN];
-	_xos_canonical_filename(buf, buf2);
+	_xos_canonical_filename(buf, buf2, MAXPATHLEN, 0);
 	strcpy(buf, buf2);
       }
 #endif
@@ -142,7 +144,7 @@ findHome(char *symbols)
     }
   }
 
-  if ( (home = PrologPath(PLHOME, plp)) &&
+  if ( (home = PrologPath(PLHOME, plp, sizeof(plp))) &&
        ExistsDirectory(home) )
     return store_string(home);
 
@@ -166,7 +168,7 @@ basename of the running program, taking all the leading alnum characters.
 static char *
 defaultSystemInitFile(char *a0)
 { char plp[MAXPATHLEN];
-  char *base = BaseName(PrologPath(a0, plp));
+  char *base = BaseName(PrologPath(a0, plp, sizeof(plp)));
   char buf[256];
   char *s = buf;
 
@@ -188,18 +190,18 @@ memarea_limit(const char *s)
 { number n;
   unsigned char *q;
 
-  if ( get_number((unsigned char *)s, &q, &n, FALSE) && intNumber(&n) )
+  if ( str_number((unsigned char *)s, &q, &n, FALSE) && intNumber(&n) )
   { switch((int)*q)
     { case 'k':
       case 'K':
       case EOS:
-	return n.value.i K;
+	return (long)n.value.i K;
       case 'm':
       case 'M':
-	return n.value.i K K;
+	return (long)n.value.i K K;
       case 'b':
       case 'B':
-	return n.value.i;
+	return (long)n.value.i;
     }
   }
 
@@ -232,26 +234,36 @@ setupGNUEmacsInferiorMode()
 
 static void
 initPaths()
-{ char plp1[MAXPATHLEN];
-  char plp[MAXPATHLEN];
-  char *symbols = NULL;			/* The executable */
+{ char plp[MAXPATHLEN];
 
-  if ( !(symbols = findExecutable(GD->cmdline.argv[0], plp1)) ||
-       !(symbols = DeRefLink(symbols, plp)) )
-    symbols = GD->cmdline.argv[0];
-
-  DEBUG(2, Sdprintf("rc-module: %s\n", symbols));
-
-  systemDefaults.home	     = findHome(symbols);
+  if ( GD->cmdline.argc > 0 )
+  { char plp1[MAXPATHLEN];
+    char *symbols = NULL;		/* The executable */
 
+    if ( !(symbols = findExecutable(GD->cmdline.argv[0], plp1)) ||
+	 !(symbols = DeRefLink(symbols, plp)) )
+      symbols = GD->cmdline.argv[0];
+  
+    DEBUG(2, Sdprintf("rc-module: %s\n", symbols));
+  
+    systemDefaults.home	       = findHome(symbols);
+  
 #ifdef __WIN32__			/* we want no module but the .EXE */
-  GD->paths.module	     = store_string(symbols);
-  symbols = findExecutable(NULL, plp);
-  DEBUG(2, Sdprintf("Executable: %s\n", symbols));
+    GD->paths.module	       = store_string(symbols);
+    symbols = findExecutable(NULL, plp);
+    DEBUG(2, Sdprintf("Executable: %s\n", symbols));
 #endif
-  GD->paths.executable	     = store_string(symbols);
-  systemDefaults.startup     = store_string(PrologPath(DEFSTARTUP, plp));
-  GD->options.systemInitFile = defaultSystemInitFile(GD->cmdline.argv[0]);
+    GD->paths.executable       = store_string(symbols);
+    GD->options.systemInitFile = defaultSystemInitFile(GD->cmdline.argv[0]);
+  } else
+  { systemDefaults.home	       = findHome(NULL);
+    GD->options.systemInitFile = store_string("none");
+#ifdef __WIN32__			/* we want no module but the .EXE */
+    GD->paths.module	       = store_string("libpl.dll");
+#endif
+  }
+
+  systemDefaults.startup = store_string(PrologPath(DEFSTARTUP, plp, sizeof(plp)));
 
 #ifdef O_XOS
   if ( systemDefaults.home )
@@ -456,7 +468,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 	  switch(*s)
 	  { case 'L':	GD->options.localSize    = size; goto next;
 	    case 'G':	GD->options.globalSize   = size;
-	    		GD->options.argumentSize = size/8;
+	    		GD->options.argumentSize = size/4;
 							 goto next;
 	    case 'T':	GD->options.trailSize    = size; goto next;
 	    case 'A':	GD->options.argumentSize = size; goto next;
@@ -599,6 +611,11 @@ $PATH!
 
 	{pl, '-L0 -f', <file>}
 
+On some unix systems, currently MacOS X  tiger, the arguments are passed
+as { pl, -L0, -f, <file> }, which   must be converted by adding -- after
+the script file. This  mode  is   selected  if  SCRIPT_BREAKDOWN_ARGS is
+defined by configure.
+
 On Windows this is simply passed as below.   We have to analyse the file
 ourselves. Unfortunately this needs to be done  on C as it might contain
 stack-parameters.
@@ -623,6 +640,48 @@ script_argv(int argc, char **argv)
 	    Sdprintf("argv[%d] = '%s'\n", i, argv[i]);
 	});
 
+#ifdef SCRIPT_BREAKDOWN_ARGS
+  for(i=1; i < argc-1; i++)
+  { if ( argv[i][0] == '-' && argv[i][2] == '\0' )
+    { switch(argv[i][1])
+      { case '-':
+	  goto noscript;
+	case 's':
+	case 'f':
+	  if ( (fd=fopen(argv[i+1], "r")) )
+	  { char buf[MAXLINE];
+	    char **av;
+	    int j;
+
+	    if ( !fgets(buf, sizeof(buf), fd) )
+	    { fclose(fd);
+	      goto noscript;
+	    }
+	    if ( !strprefix(buf, "#!") )
+	    { fclose(fd);
+	      goto noscript;
+	    }
+	    fclose(fd);
+
+	    DEBUG(1, Sdprintf("Got script %s\n", argv[i+1]));
+
+	    av = allocHeap(sizeof(char*)*(argc+2));
+	    for(j=0; j<=i+1; j++)
+	      av[j] = argv[j];
+	    av[j] = "--";
+	    for(; j<argc; j++)
+	      av[j+1] = argv[j];
+	    av[j+1] = NULL;
+	    GD->cmdline.argc = argc+1;
+	    GD->cmdline.argv = av;
+
+	    return;
+	  }
+      }
+    }
+  }
+#else /*SCRIPT_BREAKDOWN_ARGS*/
+
 #ifdef __unix__
   if ( argc >= 3 &&
        (strpostfix(argv[1], "-f") || strpostfix(argv[1], "-s")) &&
@@ -641,7 +700,7 @@ script_argv(int argc, char **argv)
     { char tmp[MAXPATHLEN];
       char dir[MAXPATHLEN];
 
-      _xos_canonical_filename(argv[1], tmp);
+      _xos_canonical_filename(argv[1], tmp, MAXPATHLEN, 0);
       if ( IsAbsolutePath(tmp) )
 	chdir(DirName(tmp, dir));
     }
@@ -706,6 +765,7 @@ script_argv(int argc, char **argv)
 
     fclose(fd);
   } else
+#endif /*SCRIPT_BREAKDOWN_ARGS*/
   { noscript:
     GD->cmdline.argc = argc;
     GD->cmdline.argv = argv;
@@ -745,6 +805,7 @@ PL_initialise(int argc, char **argv)
   script_argv(argc, argv);		/* hande #! arguments */
   argc = GD->cmdline.argc;
   argv = GD->cmdline.argv;
+  GD->cmdline._c_argc = -1;
   DEBUG(1,
   { int i;
     
@@ -892,11 +953,7 @@ properly on Linux. Don't bother with it.
 }
 
 
-#ifdef __GNUC__
 typedef const char *cline;
-#else
-typedef char *cline;
-#endif
 
 static int
 usage()
@@ -926,9 +983,15 @@ usage()
     NULL
   };
   const cline *lp = lines;
+  char *prog;
+
+  if ( GD->cmdline.argc > 0 )
+    prog = BaseName(GD->cmdline.argv[0]);
+  else
+    prog = "pl";
 
   for(lp = lines; *lp; lp++)
-    Sfprintf(Serror, *lp, BaseName(GD->cmdline.argv[0]));
+    Sfprintf(Serror, *lp, prog);
 
   return TRUE;
 }
@@ -982,7 +1045,7 @@ runtime_vars(int format)
   { 
 #ifdef O_XOS
     if ( format == FMT_CMD )
-    { _xos_os_existing_filename(systemDefaults.home, base);
+    { _xos_os_filename(systemDefaults.home, base, MAXPATHLEN);
       home = base;
     } else
       home = systemDefaults.home;
@@ -1000,6 +1063,7 @@ runtime_vars(int format)
   printvar("PLARCH",	ARCH, format);
   printvar("PLLIBS",	C_LIBS, format);
   printvar("PLLIB",	C_PLLIB, format);
+  printvar("PLCFLAGS",  C_CFLAGS, format);
   printvar("PLLDFLAGS", C_LDFLAGS, format);
 #ifdef SO_EXT
   printvar("PLSOEXT",	SO_EXT, format);
@@ -1095,14 +1159,9 @@ PL_cleanup(int rval)
 
   Scurout = Soutput;			/* reset output stream to user */
 
-  GD->cleaning = CLN_FOREIGN;
-
-					/* run PL_on_halt() hooks */
-  for(h = GD->os.on_halt_list; h; h = h->next)
-    (*h->function)(rval, h->argument);
-
   GD->cleaning = CLN_PROLOG;
 
+  qlfCleanup();				/* remove errornous .qlf files */
   if ( GD->initialised )
   { fid_t cid = PL_open_foreign_frame();
     predicate_t proc = PL_predicate("$run_at_halt", 0, "system");
@@ -1110,12 +1169,17 @@ PL_cleanup(int rval)
     PL_discard_foreign_frame(cid);
   }
 
+  GD->cleaning = CLN_FOREIGN;
+
+					/* run PL_on_halt() hooks */
+  for(h = GD->os.on_halt_list; h; h = h->next)
+    (*h->function)(rval, h->argument);
+
 #if defined(__WINDOWS__) || defined(__WIN32__)
   if ( rval != 0 && !hasConsole() )
     PlMessage("Exit status is %d", rval);
 #endif
 
-  qlfCleanup();				/* remove errornous .qlf files */
   dieIO();				/* streams may refer to foreign code */
 					/* Standard I/O is only flushed! */
 
@@ -1184,15 +1248,13 @@ sysError(const char *fm, ...)
 }
 
 
-bool
+void
 fatalError(const char *fm, ...)
 { va_list args;
 
   va_start(args, fm);
   vfatalError(fm, args);
-  va_end(args);
-
-  PL_fail;
+/*va_end(args);*/
 }
 
 
@@ -1264,7 +1326,7 @@ action:
 }
 
 
-bool
+void
 vfatalError(const char *fm, va_list args)
 {
 #if defined(__WINDOWS__) || defined(__WIN32__)
@@ -1281,7 +1343,6 @@ vfatalError(const char *fm, va_list args)
 #endif
 
   PL_halt(2);
-  PL_fail;
 }
 
 
