@@ -1,4 +1,4 @@
-/*  $Id: pl-os.c,v 1.157 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-os.c,v 1.182 2006/07/31 12:10:50 jan Exp $
 
     Part of SWI-Prolog
 
@@ -35,6 +35,7 @@
 
 #include "pl-incl.h"
 #include "pl-ctype.h"
+#include "pl-utf8.h"
 #undef abs
 #include <math.h>		/* avoid abs() problem with msvc++ */
 #include <stdio.h>		/* rename() and remove() prototypes */
@@ -74,14 +75,22 @@
 #include <dos.h>
 #endif
 
+#ifdef WIN32
+#define STAT_TYPE struct _stat
+#else
+#define STAT_TYPE struct stat
+#endif
+
 #if OS2 && EMX
 static real initial_time;
 #endif /* OS2 */
 
+#define LOCK()   PL_LOCK(L_OS)
+#define UNLOCK() PL_UNLOCK(L_OS)
+
 static void	initExpand(void);
 static void	cleanupExpand(void);
 static void	initEnviron(void);
-static long	Time(void);
 static char *	Which(const char *program, char *fullname);
 
 #ifndef DEFAULT_PATH
@@ -125,10 +134,7 @@ initOs(void)
   initEnviron();
 
 #ifdef __WIN32__
-  if ( iswin32s() )
-    set(&features, DOS_FILE_NAMES_FEATURE);
-  else
-    set(&features, FILE_CASE_PRESERVING_FEATURE);
+  set(&features, FILE_CASE_PRESERVING_FEATURE);
 #else
   set(&features, FILE_CASE_FEATURE);
   set(&features, FILE_CASE_PRESERVING_FEATURE);
@@ -357,7 +363,7 @@ FreeMemory(void)
 		*********************************/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-    long Random()
+    uint64_t _PL_Random()
 
     Return a random number. Used for arithmetic only. More trouble. On
     some systems (WIN32) the seed of rand() is thread-local, while on
@@ -387,7 +393,7 @@ initRandom(void)
   gettimeofday(&tp, NULL);
   init = tp.tv_sec + tp.tv_usec;
 #else
-  init = Time();
+  init = (long)time((time_t *) NULL);
 #endif
 #endif
 
@@ -400,22 +406,32 @@ initRandom(void)
 #endif
 }
 
-long
-Random(void)
+uint64_t
+_PL_Random(void)
 { if ( !LD->os.rand_initialised )
   { initRandom();
     LD->os.rand_initialised = TRUE;
   }
 
 #ifdef HAVE_RANDOM
-  return random();
-#else
-  { long l = rand();			/* 0<n<2^15-1 */
-  
-    l ^= rand()<<10;
-    l ^= rand()<<20;
+#if SIZEOF_LONG == 4
+  { uint64_t l = random();
+    
+    l ^= (uint64_t)random()<<32;
 
-    return l & (~PLMININT);
+    return l;
+  }
+#else
+  return random();
+#endif
+#else
+  { uint64_t l = rand();			/* 0<n<2^15-1 */
+  
+    l ^= rand()<<15;
+    l ^= rand()<<30;
+    l ^= rand()<<45;
+
+    return l;
   }
 #endif
 }
@@ -491,7 +507,7 @@ TemporaryFile(const char *id)
 #else
   if ( (tmp = _tempnam(tmpdir, id)) )
 #endif
-  { PrologPath(tmp, temp);
+  { PrologPath(tmp, temp, sizeof(temp));
   } else
     Ssprintf(temp, "%s/pl_%s_%d", tmpdir, id, temp_counter++);
 }
@@ -634,9 +650,9 @@ getpagesize()
 */
 
 char *
-PrologPath(char *ospath, char *path)
+PrologPath(char *ospath, char *path, size_t len)
 { char *s = ospath, *p = path;
-  int limit = MAXPATHLEN-1;
+  int limit = len-1;
 
   if (isLetter(s[0]) && s[1] == ':')
   { *p++ = '/';
@@ -678,8 +694,8 @@ OsPath(const char *plpath, char *path)
 
 #ifdef __unix__
 char *
-PrologPath(const char *p, char *buf)
-{ strcpy(buf, p);
+PrologPath(const char *p, char *buf, size_t len)
+{ strncpy(buf, p, len);
 
   return buf;
 }
@@ -694,12 +710,10 @@ OsPath(const char *p, char *buf)
 
 #if O_XOS
 char *
-PrologPath(const char *p, char *buf)
-{ _xos_canonical_filename(p, buf);
-  if ( !trueFeature(FILE_CASE_FEATURE) )
-    strlwr(buf);
+PrologPath(const char *p, char *buf, size_t len)
+{ int flags = (trueFeature(FILE_CASE_FEATURE) ? 0 : XOS_DOWNCASE);
 
-  return buf;
+  return _xos_canonical_filename(p, buf, len, flags);
 }
 
 char *
@@ -715,7 +729,7 @@ LastModifiedFile(char *f)
 { char tmp[MAXPATHLEN];
 
 #if defined(HAVE_STAT) || defined(__unix__)
-  struct stat buf;
+  STAT_TYPE buf;
 
   if ( statfunc(OsPath(f, tmp), &buf) < 0 )
     return -1;
@@ -860,13 +874,13 @@ ExistsDirectory(const char *path)
 }
 
 
-long
+int64_t
 SizeFile(const char *path)
 { char tmp[MAXPATHLEN];
-  struct stat buf;
+  STAT_TYPE buf;
 
 #if defined(HAVE_STAT) || defined(__unix__)
-  if ( statfunc(OsPath(path, tmp), &buf) == -1 )
+  if ( statfunc(OsPath(path, tmp), &buf) < 0 )
     return -1;
 #endif
 
@@ -920,7 +934,7 @@ SameFile(const char *f1, const char *f2)
   { if ( streq(f1, f2) )
       succeed;
   } else
-  { if ( stricmp(f1, f2) == 0 )
+  { if ( strcasecmp(f1, f2) == 0 )
       succeed;
   }
 
@@ -937,28 +951,7 @@ SameFile(const char *f1, const char *f2)
   }
 #endif
 #ifdef O_XOS
-  { char p1[MAXPATHLEN];
-    char p2[MAXPATHLEN];
-    int i;
-  
-    for(i=0; i<=1; i++)
-    { if ( i == 0 )
-      { _xos_limited_os_filename(f1, p1);
-	_xos_limited_os_filename(f2, p2);
-      } else
-      { AbsoluteFile(f1, p1);
-	AbsoluteFile(f2, p2);
-      }
-
-      if ( trueFeature(FILE_CASE_FEATURE) )
-      { if ( streq(p1, p2) )
-	  succeed;
-      } else
-      { if ( stricmp(p1, p2) == 0 )
-	  succeed;
-      }
-    }
-  }
+  return _xos_same_file(f1, f2);
 #endif /*O_XOS*/
     /* Amazing! There is no simple way to check two files for identity. */
     /* stat() and fstat() both return dummy values for inode and device. */
@@ -972,7 +965,7 @@ bool
 MarkExecutable(const char *name)
 {
 #if (defined(HAVE_STAT) && defined(HAVE_CHMOD)) || defined(__unix__)
-  struct stat buf;
+  STAT_TYPE buf;
   mode_t um;
 
   um = umask(0777);
@@ -1121,11 +1114,11 @@ registerParentDirs(const char *path)
     if ( statfunc(OsPath(dirname, tmp), &buf) == 0 )
     { CanonicalDir dn   = malloc(sizeof(*dn));
 
-      dn->next		= canonical_dirlist;
       dn->name		= store_string(dirname);
       dn->inode		= buf.st_ino;
       dn->device	= buf.st_dev;
       dn->canonical	= dn->name;
+      dn->next		= canonical_dirlist;
       canonical_dirlist	= dn;
 
       DEBUG(1, Sdprintf("Registered canonical dir %s\n", dirname));
@@ -1135,16 +1128,60 @@ registerParentDirs(const char *path)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+verify_entry() verifies the path cache for this   path is still safe. If
+not it updates the cache and returns FALSE.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int
+verify_entry(CanonicalDir d)
+{ char tmp[MAXPATHLEN];
+  struct stat buf;
+
+  if ( statfunc(OsPath(d->canonical, tmp), &buf) == 0 )
+  { if ( d->inode  == buf.st_ino &&
+	 d->device == buf.st_dev )
+      return TRUE;
+
+    DEBUG(1, Sdprintf("%s: inode/device changed\n", d->canonical));
+
+    d->inode  = buf.st_ino;
+    d->device = buf.st_dev;
+  } else
+  { DEBUG(1, Sdprintf("%s: no longer exists\n", d->canonical));
+
+    if ( d == canonical_dirlist )
+    { canonical_dirlist = d->next;
+    } else
+    { CanonicalDir cd;
+
+      for(cd=canonical_dirlist; cd; cd=cd->next)
+      { if ( cd->next == d )
+	{ cd->next = d->next;
+	  break;
+	}
+      }
+    }
+
+    free(d);
+  }
+    
+  return FALSE;
+}
+
+
 static char *
 canoniseDir(char *path)
-{ CanonicalDir d;
+{ CanonicalDir d, next;
   struct stat buf;
   char tmp[MAXPATHLEN];
 
   DEBUG(1, Sdprintf("canoniseDir(%s) --> ", path));
 
-  for(d = canonical_dirlist; d; d = d->next)
-  { if ( streq(d->name, path) )
+  for(d = canonical_dirlist; d; d = next)
+  { next = d->next;
+
+    if ( streq(d->name, path) && verify_entry(d) )
     { if ( d->name != d->canonical )
 	strcpy(path, d->canonical);
 
@@ -1163,7 +1200,6 @@ canoniseDir(char *path)
     char dirname[MAXPATHLEN];
     char *e = path + strlen(path);
 
-    dn->next   = canonical_dirlist;
     dn->name   = store_string(path);
     dn->inode  = buf.st_ino;
     dn->device = buf.st_dev;
@@ -1177,17 +1213,19 @@ canoniseDir(char *path)
       DEBUG(2, Sdprintf("Checking %s (dev=%d,ino=%d)\n",
 			dirname, buf.st_dev, buf.st_ino));
 
-      for(d = canonical_dirlist; d; d = d->next)
-      { if ( d->inode == buf.st_ino && d->device == buf.st_dev )
-	{ canonical_dirlist = dn;
+      for(d = canonical_dirlist; d; d = next)
+      { next = d->next;
 
-	  DEBUG(2, Sdprintf("Hit with %s (dev=%d,ino=%d)\n",
+	if ( d->inode == buf.st_ino && d->device == buf.st_dev &&
+	     verify_entry(d) )
+	{ DEBUG(2, Sdprintf("Hit with %s (dev=%d,ino=%d)\n",
 			    d->canonical, d->device, d->inode));
 
 	  strcpy(dirname, d->canonical);
 	  strcat(dirname, e);
 	  strcpy(path, dirname);
 	  dn->canonical = store_string(path);
+	  dn->next = canonical_dirlist;
 	  canonical_dirlist = dn;
 	  DEBUG(1, Sdprintf("(replace) %s\n", path));
 	  registerParentDirs(path);
@@ -1200,6 +1238,7 @@ canoniseDir(char *path)
     } while( e > path );
 
     dn->canonical = dn->name;
+    dn->next = canonical_dirlist;
     canonical_dirlist = dn;
 
     DEBUG(1, Sdprintf("(new, existing) %s\n", path));
@@ -1285,10 +1324,29 @@ canoniseFileName(char *path)
 }
 
 
+static char *
+utf8_strlwr(char *s)
+{ char tmp[MAXPATHLEN];
+  char *o, *i;
+  
+  strcpy(tmp, s);
+  for(i=tmp, o=s; *i; )
+  { int c;
+
+    i = utf8_get_char(i, &c);
+    c = towlower((wint_t)c);
+    o = utf8_put_char(o, c);
+  }
+  *o = EOS;
+
+  return s;
+}
+
+
 char *
 canonisePath(char *path)
 { if ( !trueFeature(FILE_CASE_FEATURE) )
-    strlwr(path);
+    utf8_strlwr(path);
 
   canoniseFileName(path);
 
@@ -1312,10 +1370,10 @@ canonisePath(char *path)
 
 
 static char *
-takeWord(const char **string, char *wrd)
+takeWord(const char **string, char *wrd, int maxlen)
 { const char *s = *string;
   char *q = wrd;
-  int left = MAXPATHLEN-1;
+  int left = maxlen-1;
 
   while( isAlpha(*s) || *s == '_' )
   { if ( --left < 0 )
@@ -1333,10 +1391,9 @@ takeWord(const char **string, char *wrd)
 
 
 bool
-expandVars(const char *pattern, char *expanded)
+expandVars(const char *pattern, char *expanded, int maxlen)
 { int size = 0;
-  char c;
-  char word[MAXPATHLEN];
+  char wordbuf[MAXPATHLEN];
 
   if ( *pattern == '~' )
   { char *user;
@@ -1344,7 +1401,9 @@ expandVars(const char *pattern, char *expanded)
     int l;
 
     pattern++;
-    user = takeWord(&pattern, word);
+    user = takeWord(&pattern, wordbuf, sizeof(wordbuf));
+    LOCK();
+
     if ( user[0] == EOS )		/* ~/bla */
     {
 #ifdef O_XOS
@@ -1353,10 +1412,11 @@ expandVars(const char *pattern, char *expanded)
       if ( !(value = GD->os.myhome) )
       { char envbuf[MAXPATHLEN];
 
-	if ( (value = getenv3("HOME", envbuf, sizeof(envbuf))) )
-	{ value = GD->os.myhome = store_string(PrologPath(value, word));
+	if ( (value = getenv3("HOME", envbuf, sizeof(envbuf))) &&
+	     (value = PrologPath(value, wordbuf, sizeof(wordbuf))) )
+	{ GD->os.myhome = store_string(value);
 	} else
-	{ value = GD->os.myhome = "/";
+	{ value = GD->os.myhome = store_string("/");
 	}
       }
 #endif /*O_XOS*/
@@ -1374,6 +1434,7 @@ expandVars(const char *pattern, char *expanded)
 	    PL_put_atom_chars(name, user);
 	    PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_user, name);
 	  }
+	  UNLOCK();
 	  fail;
 	}
 	if ( GD->os.fred )
@@ -1389,14 +1450,16 @@ expandVars(const char *pattern, char *expanded)
     { if ( fileerrors )
 	PL_error(NULL, 0, NULL, ERR_NOT_IMPLEMENTED_FEATURE, "user_info");
 
+      UNLOCK();
       fail;
     }
 #endif
     size += (l = (int) strlen(value));
-    if ( size >= MAXPATHLEN )
+    if ( size+1 >= maxlen )
       return PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_max_path_length);
     strcpy(expanded, value);
     expanded += l;
+    UNLOCK();
 
 					/* ~/ should not become // */
     if ( expanded[-1] == '/' && pattern[0] == '/' )
@@ -1404,17 +1467,20 @@ expandVars(const char *pattern, char *expanded)
   }
 
   for( ;; )
-  { switch( c = *pattern++ )
+  { int c = *pattern++;
+
+    switch( c )
     { case EOS:
 	break;
       case '$':
 	{ char envbuf[MAXPATHLEN];
-	  char *var = takeWord(&pattern, word);
+	  char *var = takeWord(&pattern, wordbuf, sizeof(wordbuf));
 	  char *value;
 	  int l;
 
 	  if ( var[0] == EOS )
 	    goto def;
+	  LOCK();
 	  value = getenv3(var, envbuf, sizeof(envbuf));
 	  if ( value == (char *) NULL )
 	  { if ( fileerrors )
@@ -1424,20 +1490,26 @@ expandVars(const char *pattern, char *expanded)
 	      PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_variable, name);
 	    }
 
+	    UNLOCK();
 	    fail;
 	  }
 	  size += (l = (int)strlen(value));
-	  if ( size >= MAXPATHLEN )
+	  if ( size+1 >= maxlen )
+	  { UNLOCK();
 	    return PL_error(NULL, 0, NULL, ERR_REPRESENTATION,
 			    ATOM_max_path_length);
+	  }
 	  strcpy(expanded, value);
+	  UNLOCK();
+
 	  expanded += l;
 
 	  continue;
 	}
       default:
       def:
-	if ( ++size >= MAXPATHLEN )
+	size++;
+	if ( size+1 >= maxlen )
 	  return PL_error(NULL, 0, NULL, ERR_REPRESENTATION,
 			  ATOM_max_path_length);
 	*expanded++ = c;
@@ -1447,7 +1519,7 @@ expandVars(const char *pattern, char *expanded)
     break;
   }
 
-  if ( ++size >= MAXPATHLEN )
+  if ( ++size >= maxlen )
     return PL_error(NULL, 0, NULL, ERR_REPRESENTATION,
 		    ATOM_max_path_length);
   *expanded = EOS;
@@ -1461,7 +1533,7 @@ ExpandFile(const char *pattern, char **vector)
 { char expanded[MAXPATHLEN];
   int matches = 0;
 
-  if ( !expandVars(pattern, expanded) )
+  if ( !expandVars(pattern, expanded, sizeof(expanded)) )
     return -1;
   
   vector[matches++] = store_string(expanded);
@@ -1569,7 +1641,7 @@ char *
 AbsoluteFile(const char *spec, char *path)
 { char tmp[MAXPATHLEN];
   char buf[MAXPATHLEN];
-  char *file = PrologPath(spec, buf);
+  char *file = PrologPath(spec, buf, sizeof(buf));
   
   if ( trueFeature(FILEVARS_FEATURE) )
   { if ( !(file = ExpandOneFile(buf, tmp)) )
@@ -1846,12 +1918,6 @@ LocalTime(long int *t, struct tm *r)
 }
 
 
-static long
-Time(void)
-{ return (long)time((time_t *) NULL);
-}
-
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 			TERMINAL IO MANIPULATION
 
@@ -1874,24 +1940,6 @@ ResetStdin()
     GD->os.org_terminal = *Sinput->functions;
 }
 
-static int
-Swrite_protocol(void *handle, char *buf, int size)
-{ int rval;
-#ifdef HAVE_CLOCK
-  long oldclock = clock();
-#endif
-
-  protocol(buf, size);
-
-  rval = (*GD->os.org_terminal.write)(handle, buf, size);
-
-#ifdef HAVE_CLOCK
-  clock_wait_ticks += clock() - oldclock;
-#endif
-
-  return rval;
-}
-
 static int
 Sread_terminal(void *handle, char *buf, int size)
 { long h = (long)handle;
@@ -1899,12 +1947,10 @@ Sread_terminal(void *handle, char *buf, int size)
   source_location oldsrc = LD->read_source;
 
   if ( LD->prompt.next && ttymode != TTY_RAW )
-  { Sfputs(PrologPrompt(), Suser_output);
-    
-    LD->prompt.next = FALSE;
-  }
+    PL_write_prompt(TRUE);
+  else
+    Sflush(Suser_output);
 
-  Sflush(Suser_output);
   PL_dispatch(fd, PL_DISPATCH_WAIT);
   size = (*GD->os.org_terminal.read)(handle, buf, size);
 
@@ -1918,11 +1964,6 @@ Sread_terminal(void *handle, char *buf, int size)
 
   LD->read_source = oldsrc;
 
-#ifdef __WIN32__
-  if ( size > 0 )			/* Why needed only here? */
-    protocol(buf, size);		/* and what is the correct condition */
-#endif
-
   return size;
 }
 
@@ -1934,7 +1975,6 @@ ResetTty()
   if ( !GD->os.iofunctions.read )
   { GD->os.iofunctions       = *Sinput->functions;
     GD->os.iofunctions.read  = Sread_terminal;
-    GD->os.iofunctions.write = Swrite_protocol;
 
     Sinput->functions  = 
     Soutput->functions = 
@@ -2179,11 +2219,16 @@ getenvl(const char *name)
 }
 #endif
 
-#if HAVE_PUTENV
+#if defined(HAVE_PUTENV) || defined(HAVE_SETENV)
 
 int
 Setenv(char *name, char *value)
-{ char *buf = alloca(strlen(name) + strlen(value) + 2);
+{ 
+#ifdef HAVE_SETENV
+  if ( setenv(name, value, TRUE) != 0 )
+    return PL_error("setenv", 2, NULL, ERR_NOMEM);
+#else
+  char *buf = alloca(strlen(name) + strlen(value) + 2);
 
   if ( buf )
   { Ssprintf(buf, "%s=%s", name, value);
@@ -2192,7 +2237,7 @@ Setenv(char *name, char *value)
       return PL_error("setenv", 2, NULL, ERR_NOMEM);
   } else
     return PL_error("setenv", 2, NULL, ERR_NOMEM);
-
+#endif
   succeed;
 }
 
@@ -2204,6 +2249,9 @@ Unsetenv(char *name)
 
   succeed;
 #else
+  if ( !getenv(name) )
+    succeed;
+
   return Setenv(name, "");
 #endif
 }
@@ -2241,7 +2289,7 @@ growEnviron(char **e, int amount)
   }
 
   if ( size < 0 )
-  { register char **env, **e1, **e2;
+  { char **env, **e1, **e2;
 
     for(e1=e, filled=0; *e1; e1++, filled++)
       ;
@@ -2257,7 +2305,7 @@ growEnviron(char **e, int amount)
 
   filled += amount;
   if ( filled + 1 > size )
-  { register char **env, **e1, **e2;
+  { char **env, **e1, **e2;
   
     size += 32;
     env = (char **)realloc(e, size * sizeof(char *));
@@ -2279,7 +2327,7 @@ initEnviron(void)
 
 
 static char *
-matchName(register char *e, register char *name)
+matchName(const char *e, const char *name)
 { while( *name && *e == *name )
     e++, name++;
 
@@ -2436,12 +2484,25 @@ System(char *cmd)
     old_stop = signal(SIGTSTP, SIG_DFL);
 #endif /* SIGTSTP */
 
-    while((n = Wait(&status)) != -1 && n != pid);
-    if (n == -1)
+    for(;;)
+    { 
+#ifdef HAVE_WAITPID
+      n = waitpid(pid, &status, 0);
+#else
+      n = wait(&status);
+#endif
+      if ( n == -1 && errno == EINTR )
+	continue;
+      if ( n != pid )
+	continue;
+      break;
+    }
+
+    if ( n == -1 )
     { term_t tmp = PL_new_term_ref();
       
       PL_put_atom_chars(tmp, cmd);
-      PL_error("shell", 2, NULL, ERR_SHELL_FAILED, tmp);
+      PL_error("shell", 2, MSG_ERRNO, ERR_SHELL_FAILED, tmp);
 
       rval = 1;
     } else if (WIFEXITED(status))
@@ -2618,9 +2679,7 @@ char *command;
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     char *Symbols(char *buf)
 
-    Return the path name of the executable of SWI-Prolog. Used by the -c
-    compiler to generate the #!<path> header line and by the incremental
-    loader, who gives this path to ld, using ld -A <path>.
+    Return the path name of the executable of SWI-Prolog.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifndef __WIN32__			/* Win32 version in pl-nt.c */
@@ -2631,7 +2690,8 @@ findExecutable(const char *av0, char *buffer)
   char buf[MAXPATHLEN];
   char tmp[MAXPATHLEN];
 
-  PrologPath(av0, buf);
+  if ( !av0 || !PrologPath(av0, buf, sizeof(buf)) )
+    return NULL;
   file = Which(buf, tmp);
 
 #if __unix__				/* argv[0] can be an #! script! */
