@@ -1,4 +1,4 @@
-/*  $Id: pl-proc.c,v 1.144 2004/10/11 08:59:48 jan Exp $
+/*  $Id: pl-proc.c,v 1.152 2006/07/27 19:25:05 jan Exp $
 
     Part of SWI-Prolog
 
@@ -139,7 +139,7 @@ resetProcedure(Procedure proc, bool isnew)
        def->definition.clauses == NULL )
     isnew = TRUE;
 
-  def->flags ^= def->flags & ~(SPY_ME|NEEDSCLAUSEGC|P_SHARED);
+  def->flags ^= def->flags & ~(SPY_ME|NEEDSCLAUSEGC|P_SHARED|P_DIRTYREG);
   if ( stringAtom(def->functor->name)[0] != '$' )
     set(def, TRACE_ME);
   def->number_of_clauses = 0;
@@ -807,7 +807,10 @@ assertProcedure(Procedure proc, Clause clause, int where ARG_LD)
 
     addClauseToIndex(def, clause, where PASS_LD);
     if ( def->hash_info->size /2 > def->hash_info->buckets )
-    { set(def, NEEDSREHASH);
+    { if ( false(def, NEEDSREHASH) )
+      { set(def, NEEDSREHASH);
+	DEBUG(2, Sdprintf("Asking re-hash for %s\n", predicateName(def)));
+      }
       if ( true(def, DYNAMIC) && def->references == 0 )
       { gcClausesDefinitionAndUnlock(def); /* does UNLOCKDEF() */
 	return cref;
@@ -1222,12 +1225,15 @@ MT: locked by caller
 
 static void
 registerDirtyDefinition(Definition def)
-{ GET_LD
-  DefinitionChain cell = allocHeap(sizeof(*cell));
+{ if ( false(def, P_DIRTYREG) )
+  { GET_LD
+    DefinitionChain cell = allocHeap(sizeof(*cell));
 
-  cell->definition = def;
-  cell->next = GD->procedures.dirty;
-  GD->procedures.dirty = cell;
+    set(def, P_DIRTYREG);
+    cell->definition = def;
+    cell->next = GD->procedures.dirty;
+    GD->procedures.dirty = cell;
+  }
 }
 
 
@@ -1236,10 +1242,10 @@ pl_garbage_collect_clauses(void)
 { GET_LD
 
   if ( GD->procedures.dirty && !gc_status.blocked )
-  { DefinitionChain c, *cell;
+  { DefinitionChain c, cell, next, last;
     sigset_t set;
 
-    DEBUG(2, Sdprintf("pl_garbage_collect_clauses()\n"));
+    DEBUG(1, Sdprintf("pl_garbage_collect_clauses()\n"));
 
     PL_LOCK(L_THREAD);
     LOCK();
@@ -1250,6 +1256,7 @@ pl_garbage_collect_clauses(void)
     { Definition def = c->definition;
 
       assert(false(def, DYNAMIC));
+      assert(true(def, P_DIRTYREG));
       assert(def->references == 0);
     }
 
@@ -1261,21 +1268,25 @@ pl_garbage_collect_clauses(void)
 
     DEBUG(1, Sdprintf("Marking complete; cleaning predicates\n"));
 
-    for( cell = &GD->procedures.dirty; *cell; )
-    { Definition def = (*cell)->definition;
+    last = NULL;
+    for(cell = GD->procedures.dirty; cell; cell = next)
+    { Definition def = cell->definition;
       
+      next = cell->next;
+
       if ( def->references )
-      { def->references = 0;
-	cell = &(*cell)->next;
-      } else
-      { DefinitionChain next = (*cell)->next;
-
+      { assert(def->references == 1);
 	def->references = 0;
-	DEBUG(2, Sdprintf("gcClausesDefinition(%s)\n", predicateName(def)));
+	last = cell;
+      } else
+      { DEBUG(1, Sdprintf("gcClausesDefinition(%s)\n", predicateName(def)));
 	gcClausesDefinition(def);
-	freeHeap(*cell, sizeof(**cell));
-
-	*cell = next;
+	clear(def, P_DIRTYREG);
+	freeHeap(cell, sizeof(*cell));
+	if ( last )
+	  last->next = next;
+	else
+	  GD->procedures.dirty = next;
       }
     }
 
@@ -1409,8 +1420,11 @@ found:
 					/* TBD: find something better! */
 #ifdef O_PLMT
   PL_LOCK(L_THREAD);
-  if ( GD->statistics.threads_created > 1 )
+  if ( (GD->statistics.threads_created -
+	GD->statistics.threads_finished) == 1 )
+  { assert(false(proc->definition, P_DIRTYREG));
     freeHeap(proc->definition, sizeof(struct definition));
+  }
   PL_UNLOCK(L_THREAD);
 #else
   freeHeap(proc->definition, sizeof(struct definition));
@@ -1529,8 +1543,13 @@ error:
   if ( GD->bootsession )
     sysError("Undefined predicate: %s", predicateName(def));
   else if ( true(module, UNKNOWN_ERROR) )
-    PL_error(NULL, 0, NULL, ERR_UNDEFINED_PROC, def);
-  else
+  { Definition caller;
+    if ( fr->parent )
+      caller = fr->parent->predicate;
+    else
+      caller = NULL;
+    PL_error(NULL, 0, NULL, ERR_UNDEFINED_PROC, def, caller);
+  } else
   { fid_t fid = PL_open_foreign_frame();
     term_t pred = PL_new_term_ref();
 
@@ -1923,6 +1942,8 @@ pl_get_predicate_attribute(term_t pred,
       fail;
 
     def = getProcDefinition(proc);
+    if ( def->number_of_clauses == 0 && false(def, DYNAMIC) )
+      fail;
     return PL_unify_integer(value, def->number_of_clauses);
   } else if ( (att = attribute_mask(key)) )
   { return PL_unify_integer(value, (def->flags & att) ? 1 : 0);
@@ -2153,11 +2174,9 @@ pl_default_predicate(term_t d1, term_t d2)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 reindexDefinition()
-    Rebuilds the clause index for the predicate.  This predicate is prepared
-    for multi-threading just by using enterDefinition()/leaveDefinition().
-    This implies two threads may do the same job at the same time, but I
-    think this is fully safe: they will create the same data and both
-    threads are guaranteed not to continue before this is completed.
+    Rebuilds the clause index for the predicate. This predicate is
+    called whenever NEED_REINDEX is set.  It locks the predicate and
+    checks the flag again to ensure proper multi-threaded behaviour.
 
     We cannot re-index if the predicate is referenced and hashed: other
     predicates are operating on the hashed clauses-lists.  If we are not
@@ -2173,15 +2192,14 @@ reindexDefinition(Definition def)
   int canindex = 0;
   int cannotindex = 0;
 
+  LOCKDEF(def);
+  if ( !(def->indexPattern & NEED_REINDEX) )
+  { UNLOCKDEF(def);
+    return;
+  }
+
   assert(def->references == 1 || !def->hash_info);
-
-  DEBUG(2, if ( def->definition.clauses )
-	   { Procedure proc = def->definition.clauses->clause->procedure;
-
-	     Sdprintf("reindexDefinition(%s)\n", procedureName(proc));
-	   });
-
-  enterDefinition(def);
+  DEBUG(2, Sdprintf("reindexDefinition(%s)\n", predicateName(def)));
   def->indexPattern &= ~NEED_REINDEX;
 
   if ( true(def, AUTOINDEX) || def->indexPattern == 0x1 )
@@ -2200,11 +2218,7 @@ reindexDefinition(Definition def)
 
   if ( true(def, AUTOINDEX) )
   { if ( canindex == 0 )
-    { DEBUG(2, if ( def->definition.clauses )
-	       { Procedure proc = def->definition.clauses->clause->procedure;
-
-		 Sdprintf("not indexed: %s\n", procedureName(proc));
-	       });
+    { DEBUG(2, Sdprintf("not indexed: %s\n", predicateName(def)));
       def->indexPattern = 0x0;
     } else
     { def->indexPattern = 0x1;
@@ -2218,17 +2232,13 @@ reindexDefinition(Definition def)
   def->indexCardinality = cardinalityPattern(def->indexPattern);
   for(cref = def->definition.clauses; cref; cref = cref->next)
     reindexClause(cref->clause, def);
-  leaveDefinition(def);
 
   if ( do_hash )
-  { DEBUG(3,
-	  if ( def->definition.clauses )
-	  { Procedure proc = def->definition.clauses->clause->procedure;
-
-	    Sdprintf("hash(%s, %d)\n", procedureName(proc), do_hash);
-	  });
+  { DEBUG(3, Sdprintf("hash(%s, %d)\n", predicateName(def), do_hash));
     hashDefinition(def, do_hash);
   }
+
+  UNLOCKDEF(def);
 }
 
 
@@ -2620,7 +2630,6 @@ startConsult(SourceFile f)
   if ( f->count++ > 0 )			/* This is a re-consult */
   { ListCell cell, next;
     sigset_t set;
-    int immediate;
     ClauseRef garbage = NULL;
 
     PL_LOCK(L_THREAD);
@@ -2633,47 +2642,51 @@ startConsult(SourceFile f)
 #ifdef O_PLMT
     forThreadLocalData(markPredicatesInEnvironments,
 		       PL_THREAD_SUSPEND_AFTER_WORK);
-
-					/* are we alone? */
-    immediate = ((GD->statistics.threads_created -
-		  GD->statistics.threads_finished) == 1);
-#else
-    immediate = TRUE;
 #endif
     GD->procedures.reloading = NULL;
 
-    for(cell = f->procedures; cell; cell = next)
+					/* remove the clauses */
+    for(cell = f->procedures; cell; cell = cell->next)
     { Procedure proc = cell->value;
       Definition def = proc->definition;
 
-      next = cell->next;
-      if ( def )
-      { DEBUG(2, Sdprintf("removeClausesProcedure(%s), refs = %d\n",
-			  predicateName(def), def->references));
+      DEBUG(2, Sdprintf("removeClausesProcedure(%s), refs = %d\n",
+			predicateName(def), def->references));
 
-	removeClausesProcedure(proc,
-			       true(def, MULTIFILE) ? f->index : 0,
-			       TRUE);
-
-	if ( true(def, NEEDSCLAUSEGC) )
-	{ if ( def->references == 0 )
-	    garbage = cleanDefinition(def, garbage);
-	  else if ( false(def, DYNAMIC) )
-	    registerDirtyDefinition(def);
-	}
-
-	if ( false(def, DYNAMIC) && def->references )
-	{ def->references = 0;
-	  GD->procedures.active_marked--;
-	}
-
-	if ( false(def, MULTIFILE) )
-	  clear(def, FILE_ASSIGNED);
+      removeClausesProcedure(proc,
+			     true(def, MULTIFILE) ? f->index : 0,
+			     TRUE);
+      
+      if ( true(def, NEEDSCLAUSEGC) )
+      { if ( def->references == 0 )
+	  garbage = cleanDefinition(def, garbage);
+	else if ( false(def, DYNAMIC) )
+	  registerDirtyDefinition(def);
       }
+      
+      if ( false(def, MULTIFILE) )
+	clear(def, FILE_ASSIGNED);
+    }
+
+					/* unmark the marked predicates */
+    for(cell = f->procedures; cell; cell = cell->next)
+    { Procedure proc = cell->value;
+      Definition def = proc->definition;
+
+      if ( false(def, DYNAMIC) && def->references )
+      { assert(def->references == 1);
+	def->references = 0;
+	GD->procedures.active_marked--;
+      }
+    }
+
+					/* cleanup the procedure list */
+    for(cell = f->procedures; cell; cell = next)
+    { next = cell->next;
       freeHeap(cell, sizeof(struct list_cell));
     }
-    assert(GD->procedures.active_marked == 0);
     f->procedures = NULL;
+    assert(GD->procedures.active_marked == 0);
 
 #ifdef O_PLMT
     resumeThreads();
