@@ -1,4 +1,4 @@
-/*  $Id: pl-rec.c,v 1.82 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-rec.c,v 1.91 2005/09/05 19:41:05 jan Exp $
 
     Part of SWI-Prolog
 
@@ -26,6 +26,8 @@
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 
+#define WORDS_PER_PLINT (sizeof(int64_t)/sizeof(word))
+
 static RecordList lookupRecordList(word);
 static RecordList isCurrentRecordList(word);
 static void freeRecordRef(RecordRef r);
@@ -134,12 +136,6 @@ typedef struct
   int	     external;			/* Allow for external storage */
 } compile_info, *CompileInfo;
 
-
-typedef struct
-{ word w[WORDS_PER_DOUBLE];
-} fword;
-
-
 #define	PL_TYPE_VARIABLE	(1)	/* variable */
 #define PL_TYPE_ATOM		(2)	/* atom */
 #define PL_TYPE_INTEGER	  	(3)	/* big integer */
@@ -155,6 +151,7 @@ typedef struct
 #define PL_TYPE_ATTVAR		(12)	/* Attributed variable */
 #define PL_REC_ALLOCVAR		(13)	/* Allocate a variable on global */
 #define PL_REC_CYCLE		(14)	/* cyclic reference */
+#define PL_REC_MPZ		(15)	/* GMP integer */
 
 #define addUnalignedBuf(b, ptr, type) \
 	do \
@@ -215,15 +212,13 @@ Add a signed long value. First byte   is  number of bytes, remaining are
 value-bytes, starting at most-significant.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define PLMINLONG   ((long)(-1L<<(LONGBITSIZE-1)))
-
 static void
-addLong(CompileInfo info, long v)
+addInt64(CompileInfo info, int64_t v)
 { int i = sizeof(v);
 
-  if ( v != PLMINLONG )
-  { long absn = (v >= 0 ? v : -v);
-    long mask = -1L << (LONGBITSIZE-9);
+  if ( v != PLMININT )
+  { int64_t absn = (v >= 0 ? v : -v);
+    int64_t mask = (int64_t)-1 << (INT64BITSIZE-9);
 
     for(; i>1; i--, mask >>= 8)
     { if ( absn & mask )
@@ -234,7 +229,7 @@ addLong(CompileInfo info, long v)
   addBuffer(&info->code, i, uchar);
   
   while( --i >= 0 )
-  { int b = (v>>(i*8)) & 0xff;
+  { int b = (int)(v>>(i*8)) & 0xff;
     
     addBuffer(&info->code, b, uchar);
   }
@@ -386,24 +381,40 @@ right_recursion:
       return;
     }
     case TAG_INTEGER:
-    { long val;
+    { int64_t val;
 
       if ( isTaggedInt(w) )
       { val = valInt(w);
 	addOpCode(info, PL_TYPE_TAGGED_INTEGER);
+	addInt64(info, val);
       } else
-      { info->size += sizeof(long)/sizeof(word) + 2;
-	val = valBignum(w);
-	addOpCode(info, PL_TYPE_INTEGER);
+      { number n;
+
+	info->size += wsizeofIndirect(w) + 2;
+
+	get_integer(w, &n);
+	switch(n.type)
+	{ case V_INTEGER:
+	    addOpCode(info, PL_TYPE_INTEGER);
+	    addInt64(info, n.value.i);
+	    break;
+#ifdef O_GMP
+	  case V_MPZ:
+	    addOpCode(info, PL_REC_MPZ);
+	    addMPZToBuffer((Buffer)&info->code, n.value.mpz);
+	    break;
+#endif
+  	  default:
+	    assert(0);
+	}
       }
       
-      addLong(info, val);
       return;
     }
     case TAG_STRING:
     { Word f  = addressIndirect(w);
       int n   = wsizeofInd(*f);
-      int pad = padHdr(*f);		/* see also sizeString() */
+      int pad = padHdr(*f);		/* see also getCharsString() */
       long l  = n*sizeof(word)-pad;
 
       info->size += n+2;
@@ -552,7 +563,8 @@ compileTermToHeap__LD(term_t t, int flags ARG_LD)
 #define	REC_ATOM    0x08		/* Record just contains	atom */
 #define	REC_GROUND  0x10		/* Record is ground */
 #define	REC_VMASK   0xe0		/* Version mask */
-#define	REC_VERSION 0x00		/* Version id */
+#define REC_VSHIFT     5		/* shift for version mask */
+#define	REC_VERSION 0x01		/* Version id */
 
 #define REC_SZMASK  (REC_32|REC_64)	/* SIZE_MASK */
 
@@ -562,7 +574,8 @@ compileTermToHeap__LD(term_t t, int flags ARG_LD)
 #define REC_SZ REC_32
 #endif
 
-#define REC_COMPAT(m)	(((m)&(REC_VMASK|REC_SZMASK)) == (REC_VERSION|REC_SZ))
+#define REC_HDR		(REC_SZ|(REC_VERSION<<REC_VSHIFT))
+#define REC_COMPAT(m)	(((m)&(REC_VMASK|REC_SZMASK)) == REC_HDR)
 
 char *
 PL_record_external(term_t t, unsigned int *len)
@@ -573,7 +586,7 @@ PL_record_external(term_t t, unsigned int *len)
   tmp_buffer hdr;
   int scode, shdr;
   char *rec;
-  int first = REC_SZ|REC_VERSION;
+  int first = REC_HDR;
 #if O_CYCLIC
   Word *m = aTop;
 #endif
@@ -585,7 +598,7 @@ PL_record_external(term_t t, unsigned int *len)
   initBuffer(&info.code);
 
   if ( isInteger(*p) )			/* integer-only record */
-  { long v;
+  { int64_t v;
 
     if ( isTaggedInt(*p) )
       v = valInt(*p);
@@ -594,7 +607,7 @@ PL_record_external(term_t t, unsigned int *len)
     
     first |= (REC_INT|REC_GROUND);
     addOpCode(&info, first);
-    addLong(&info, v);
+    addInt64(&info, v);
 
   ret_primitive:
     scode = sizeOfBuffer(&info.code);
@@ -736,11 +749,11 @@ fetchSizeInt(CopyInfo b)
 }
 
 
-static long
-fetchLong(CopyInfo b)
-{ long val = 0;
+static int64_t
+fetchInt64(CopyInfo b)
+{ int64_t val = 0;
   uint bytes = *b->data++;
-  uint shift = (sizeof(long)-bytes)*8;
+  uint shift = (sizeof(int64_t)-bytes)*8;
 
   while(bytes-- > 0)
     val = (val << 8) | (*b->data++ & 0xff);
@@ -849,20 +862,36 @@ right_recursion:
       return;
     }
     case PL_TYPE_TAGGED_INTEGER:
-    { long val = fetchLong(b);
+    { int64_t val = fetchInt64(b);
 
       *p = consInt(val);
 
       return;
     }
     case PL_TYPE_INTEGER:
-    { *p = consPtr(b->gstore, TAG_INTEGER|STG_GLOBAL);
-      *b->gstore++ = mkIndHdr(1, TAG_INTEGER);
-      *b->gstore++ = fetchLong(b);
-      *b->gstore++ = mkIndHdr(1, TAG_INTEGER);
+    { int i;
+      union
+      { int64_t i64;
+	word    w[WORDS_PER_PLINT];
+      } val;
+
+      val.i64 = fetchInt64(b);
+
+      *p = consPtr(b->gstore, TAG_INTEGER|STG_GLOBAL);
+      *b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
+      for(i=0; i<WORDS_PER_PLINT; i++)
+	*b->gstore++ = val.w[i];
+      *b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
 
       return;
     }
+#ifdef O_GMP
+    case PL_REC_MPZ:
+    { b->data = loadMPZFromCharp(b->data, p, &b->gstore);
+
+      return;
+    }
+#endif
     case PL_TYPE_FLOAT:
     case PL_TYPE_EXT_FLOAT:
     { *p = consPtr(b->gstore, TAG_FLOAT|STG_GLOBAL);
@@ -1025,6 +1054,11 @@ right_recursion:
     { skipLong(b);
       return;
     }
+#ifdef O_GMP
+    case PL_REC_MPZ:
+      b->data = skipMPZOnCharp(b->data);
+      return;
+#endif
     case PL_TYPE_FLOAT:
     case PL_TYPE_EXT_FLOAT:
     { skipBuf(b, double);
@@ -1141,16 +1175,16 @@ unref_cont:
     case TAG_INTEGER:
       if ( isTaggedInt(w) )
       { if ( stag == PL_TYPE_TAGGED_INTEGER )
-	{ long val = valInt(w);
-	  long v2 = fetchLong(info);
+	{ int64_t val = valInt(w);
+	  int64_t v2 = fetchInt64(info);
 
 	  if ( v2 == val )
 	    succeed;
 	}
       } else
       { if ( stag == PL_TYPE_INTEGER )
-	{ long val = valBignum(w);
-	  long v2 = fetchLong(info);
+	{ int64_t val = valBignum(w);
+	  int64_t v2 = fetchInt64(info);
 
 	  if ( v2 == val )
 	    succeed;
@@ -1161,7 +1195,7 @@ unref_cont:
       if ( stag == PL_TYPE_STRING )
       { Word f  = addressIndirect(w);
 	int n   = wsizeofInd(*f);
-	int pad = padHdr(*f);		/* see also sizeString() */
+	int pad = padHdr(*f);		/* see also getCharsString() */
 	uint l  = n*sizeof(word)-pad;
 
 	uint llen = fetchSizeInt(info);
@@ -1343,13 +1377,15 @@ PL_recorded_external(const char *rec, term_t t)
   fetchBuf(&b, &m, uchar);
 
   if ( !REC_COMPAT(m) )
+  { Sdprintf("PL_recorded_external: Incompatible version\n");
     fail;
+  }
 
   if ( m & (REC_INT|REC_ATOM) )		/* primitive cases */
   { if ( m & REC_INT )
-    { long v = fetchLong(&b);
+    { int64_t v = fetchInt64(&b);
       
-      return PL_unify_integer(t, v);
+      return PL_unify_int64(t, v);
     } else
     { atom_t a;
 
@@ -1388,7 +1424,9 @@ PL_erase_external(char *rec)
   b.base = b.data = rec;
   fetchBuf(&b, &m, uchar);
   if ( !REC_COMPAT(m) )
+  { Sdprintf("PL_erase_external(): incompatible version\n");
     fail;
+  }
   if (  m & (REC_INT|REC_ATOM) )
   { if (  m & REC_INT )
       skipLong(&b);
@@ -1486,7 +1524,7 @@ record(term_t key, term_t term, term_t ref, int az)
   RecordList l;
   RecordRef r;
   Record copy;
-  word k;
+  word k = 0L;
 
   DEBUG(3, Sdprintf("record() of ");
 	   PL_write_term(Serror, term, 1200, PL_WRT_ATTVAR_WRITE);
@@ -1540,15 +1578,18 @@ word
 pl_recorded(term_t key, term_t term, term_t ref, control_t h)
 { RecordList rl;
   RecordRef record;
-  word k;
+  word k = 0L;
   term_t copy;
   word rval;
 
   switch( ForeignControl(h) )
   { case FRG_FIRST_CALL:
     { GET_LD
-      if ( PL_get_pointer(ref, (void **)&record) )
-      { LOCK();
+      void *ptr;
+
+      if ( PL_get_pointer(ref, &ptr) )
+      { record = ptr;
+	LOCK();
 	if ( isRecordRef(record) )
 	{ if ( unifyKey(key, record->list->key) )
 	  { GET_LD
@@ -1632,15 +1673,16 @@ pl_recorded(term_t key, term_t term, term_t ref, control_t h)
 word
 pl_erase(term_t ref)
 { GET_LD
+  void *ptr;
   RecordRef record;
   RecordRef prev, r;
   RecordList l;
   word rval;
 
-  if ( !PL_get_pointer(ref, (void **)&record) ||
-       !inCore(record))
+  if ( !PL_get_pointer(ref, &ptr) || !inCore(ptr))
     return PL_error("erase", 1, NULL, ERR_TYPE, ATOM_db_reference, ref);
 
+  record = ptr;
   if ( isClause(record) )
   { Clause clause = (Clause) record;
   
