@@ -1,11 +1,11 @@
-/*  $Id: pl-stream.c,v 1.83 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-stream.c,v 1.130 2006/02/12 15:59:30 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2005, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -30,6 +30,8 @@
 #include <uxnt.h>
 #define MD "config/win32.h"
 #include <winsock2.h>
+#include "pl-mswchar.h"
+#define CRLF_MAPPING 1
 #endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -50,12 +52,17 @@ locking is required.
 #include <config.h>
 #endif
 
-#if O_LARGEFILES
-#define _FILE_OFFSET_BITS 64
+#if _FILE_OFFSET_BITS == 64 || defined(_LARGE_FILES)
+#define O_LARGEFILES 1		/* use for conditional code in Prolog */
+#else
+#undef O_LARGEFILES
 #endif
 
 #define PL_KERNEL 1
+#include <wchar.h>
+typedef wchar_t pl_wchar_t;
 #include "pl-stream.h"
+#include "pl-utf8.h"
 #include <sys/types.h>
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -74,6 +81,7 @@ locking is required.
 #include <string.h>
 #include <fcntl.h>
 #include <stdlib.h>
+#include <limits.h>
 #include <stdarg.h>
 #include <ctype.h>
 #include <sys/stat.h>
@@ -89,6 +97,10 @@ locking is required.
 #include SYSLIB_H
 #endif
 
+#ifndef MB_LEN_MAX
+#define MB_LEN_MAX 6
+#endif
+
 #ifndef FALSE
 #define FALSE 0
 #endif
@@ -104,8 +116,8 @@ int Slinesize = SIO_LINESIZE;		/* Sgets() buffer size */
 
 static int	S__flushbuf(IOSTREAM *s);
 static void	run_close_hooks(IOSTREAM *s);
+static int	S__removebuf(IOSTREAM *s);
 
-#define S__fupdatefilepos(s, c) S___fupdatefilepos(s, c)
 
 #ifdef O_PLMT
 #define SLOCK(s)    if ( s->mutex ) recursiveMutexLock(s->mutex)
@@ -113,7 +125,7 @@ static void	run_close_hooks(IOSTREAM *s);
 inline int
 STRYLOCK(IOSTREAM *s)
 { if ( s->mutex &&
-       recursiveMutexTryLock(s->mutex) == MUTEX_BUSY )
+       recursiveMutexTryLock(s->mutex) == EBUSY )
     return FALSE;
 
   return TRUE;
@@ -129,35 +141,42 @@ extern int 			fatalError(const char *fm, ...);
 extern int 			PL_error(const char *pred, int arity,
 					 const char *msg, int id, ...);
 extern int			PL_handle_signals();
+extern IOENC			initEncoding(void);
 
 		 /*******************************
 		 *	      BUFFER		*
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Note that the  buffer  is  allocated   from  s->unbuffer,  which  starts
+MB_LEN_MAX before s->buffer, so we can always push-back a wide character
+into a multibyte stream. We do  not   do  this for SIO_USERBUF case, but
+this is only used by  the  output   stream  Svfprintf()  where it is not
+needed.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static int
 S__setbuf(IOSTREAM *s, char *buffer, int size)
 { if ( size == 0 )
     size = SIO_BUFSIZE;
 
+  S__removebuf(s);
   s->bufsize = size;
-  
-  if ( s->buffer && !(s->flags & SIO_USERBUF) )
-    free(s->buffer);
 
   if ( buffer )
-  { s->buffer = buffer;
+  { s->unbuffer = s->buffer = buffer;
     s->flags |= SIO_USERBUF;
   } else
-  { if ( !(s->buffer = malloc(s->bufsize)) )
+  { if ( !(s->unbuffer = malloc(s->bufsize+MB_LEN_MAX)) )
     { errno = ENOMEM;
       return -1;
     }
     s->flags &= ~SIO_USERBUF;
+    s->buffer = s->unbuffer + MB_LEN_MAX;
   }
 
-  s->unbuffer = s->buffer;
-  s->limitp   = &s->buffer[s->bufsize];
-  s->bufp     = s->buffer;
+  s->limitp = &s->buffer[s->bufsize];
+  s->bufp   = s->buffer;
 
   return size;
 }
@@ -165,12 +184,15 @@ S__setbuf(IOSTREAM *s, char *buffer, int size)
 
 static int
 S__removebuf(IOSTREAM *s)
-{ if ( s->buffer )
-  { int rval = S__flushbuf(s);
+{ if ( s->buffer && s->unbuffer )
+  { int rval = 0;
+
+    if ( (s->flags & SIO_OUTPUT) && S__flushbuf(s) < 0 )
+      rval = -1;
 
     if ( !(s->flags & SIO_USERBUF) )
-      free(s->buffer);
-    s->bufp = s->limitp = s->buffer = NULL;
+      free(s->unbuffer);
+    s->bufp = s->limitp = s->buffer = s->unbuffer = NULL;
     s->bufsize = 0;
 
     return rval;
@@ -274,29 +296,44 @@ Sunlock(IOSTREAM *s)
 
 static int
 S__flushbuf(IOSTREAM *s)
-{ int rval = 0;
-  int size;
+{ int size;
+  char *from = s->buffer;
 
-  while ( (size = s->bufp - s->buffer) > 0 )
-  { int n = (*s->functions->write)(s->handle, s->buffer, size);
+  while ( (size = s->bufp - from) > 0 )
+  { int n = (*s->functions->write)(s->handle, from, size);
 
-    if ( n >= 0 )
-    { rval += n;
-      s->bufp -= n;
-    } else
+    if ( n > 0 )			/* wrote some */
+    { from += n;
+    } else if ( n < 0 )			/* error */
     { s->flags |= SIO_FERR;
       return -1;
+    } else				/* wrote nothing? */
+    { break;
     }
   }
 
-  return rval;
+  if ( s->bufp - from == 0 )		/* full flush */
+  { int rc = s->bufp - s->buffer;
+
+    s->bufp = s->buffer;
+
+    return rc;
+  } else				/* partial flush */
+  { int rc = from - s->buffer;
+    int left = s->bufp - from;
+
+    memmove(s->buffer, from, left);
+    s->bufp = s->buffer + left;
+
+    return rc;
+  }
 }
 
 
 static int
 S__flushbufc(int c, IOSTREAM *s)
 { if ( s->buffer )
-  { if ( S__flushbuf(s) < 0 )
+  { if ( S__flushbuf(s) <= 0 )		/* == 0: no progress!? */
       c = -1;
     else
       *s->bufp++ = (c & 0xff);
@@ -399,17 +436,26 @@ S__fillbuf(IOSTREAM *s)
       goto error;			/* error */
     }
   } else
-  { int n;
+  { int n, len;
 
     if ( !s->buffer )
     { if ( S__setbuf(s, NULL, 0) < 0 )
 	goto error;
-      s->limitp = s->buffer;
+      s->bufp = s->limitp = s->buffer;
+      len = s->bufsize;
+    } else if ( s->bufp < s->limitp )
+    { len = s->limitp - s->bufp;
+      memmove(s->buffer, s->bufp, s->limitp - s->bufp);
+      s->bufp = s->buffer;
+      s->limitp = &s->bufp[len];
+      len = s->bufsize - len;
+    } else
+    { s->bufp = s->limitp = s->buffer;
+      len = s->bufsize;
     }
 
-    if ( (n=(*s->functions->read)(s->handle, s->buffer, s->bufsize)) > 0 )
-    { s->bufp = s->buffer;
-      s->limitp = &s->buffer[n];
+    if ( (n=(*s->functions->read)(s->handle, s->limitp, len)) > 0 )
+    { s->limitp += n;
       c = char_to_int(*s->bufp++);
       goto ok;
     } else
@@ -441,64 +487,102 @@ ok:
 		 *******************************/
 
 
-inline int
-S___fupdatefilepos(IOSTREAM *s, int c)
-{ IOPOS *p;
+static inline void
+update_linepos(IOSTREAM *s, int c)
+{ IOPOS *p = s->position;
 
-#if 1
-  if ( s->magic != SIO_MAGIC )
-  { if ( s->magic == SIO_CMAGIC )
-      PL_error(NULL, 0, NULL, ERR_CLOSED_STREAM, s);
-					/* PL_error() should not return */
-    fatalError("Did you load a pre-3.1.2 foreign package?"); 
-  }
-#endif
-
-  if ( (p = s->position) )
-  { switch(c)
-    { case '\n':
-	p->lineno++;
-        p->linepos = 0;
-	s->flags &= ~SIO_NOLINEPOS;
+  switch(c)
+  { case '\n':
+      p->lineno++;
+      p->linepos = 0;
+      s->flags &= ~SIO_NOLINEPOS;
 #ifdef __WIN32__
-	if ( s->flags & O_TEXT )
-	  p->charno++;			/* writes one extra! */
+      if ( s->flags & O_TEXT )
+	p->charno++;			/* writes one extra! */
 #endif
-        break;
-      case '\r':
-	p->linepos = 0;
-        s->flags &= ~SIO_NOLINEPOS;
-	break;
-      case '\b':
-	if ( p->linepos > 0 )
-	  p->linepos--;
-	break;
-      case EOF:
-	return c;
-      case '\t':
-	p->linepos |= 7;
-      default:
-	p->linepos++;
-    }
-  
+      break;
+    case '\r':
+      p->linepos = 0;
+      s->flags &= ~SIO_NOLINEPOS;
+      break;
+    case '\b':
+      if ( p->linepos > 0 )
+	p->linepos--;
+      break;
+    case EOF:
+      break;
+    case '\t':
+      p->linepos |= 7;
+    default:
+      p->linepos++;
+  }
+}
+
+
+
+int
+S__fupdatefilepos_getc(IOSTREAM *s, int c)
+{ IOPOS *p = s->position;
+
+  update_linepos(s, c);
+  p->byteno++;
+  p->charno++;
+
+  return c;
+}
+
+
+static inline int
+S__updatefilepos(IOSTREAM *s, int c)
+{ IOPOS *p = s->position;
+
+  if ( p )
+  { update_linepos(s, c);
     p->charno++;
   }
 
   return c;
 }
 
-int
-Sputc(int c, IOSTREAM *s)
+
+static inline int
+get_byte(IOSTREAM *s)
+{ int c = Snpgetc(s);
+
+  if ( s->position )
+    s->position->byteno++;
+
+  return c;
+}
+
+
+static int
+put_byte(int c, IOSTREAM *s)
 { c &= 0xff;
 
   if ( s->bufp < s->limitp )
-    *s->bufp++ = (char)c;
-  else
+  { *s->bufp++ = c;
+  } else
   { if ( S__flushbufc(c, s) < 0 )
     { s->lastc = EOF;
       return -1;
     }
   }
+
+  if ( s->position )
+    s->position->byteno++;
+
+  return c;
+}
+
+
+int
+Sputc(int c, IOSTREAM *s)
+{ c &= 0xff;
+
+  if ( put_byte(c, s) < 0 )
+    return -1;
+
   s->lastc = c;
 
   if ( c == '\n' && (s->flags & SIO_LBUF) )
@@ -512,25 +596,474 @@ Sputc(int c, IOSTREAM *s)
 
 int
 Sfgetc(IOSTREAM *s)
-{ int c;
+{ return Sgetc(s);
+}
 
-  if ( s->bufp < s->limitp )
-    c = (int) *s->bufp++ & 0xff;
-  else
-    c = S__fillbuf(s);
 
-  return S__updatefilepos(s, c);
+static inline void
+unget_byte(int c, IOSTREAM *s)
+{ IOPOS *p = s->position;
+
+  *--s->bufp = c;
+  if ( p )
+  { p->charno--;
+    p->byteno--;
+    s->flags |= (SIO_NOLINENO|SIO_NOLINEPOS);
+  }
 }
 
 
 int
 Sungetc(int c, IOSTREAM *s)
 { if ( s->bufp > s->unbuffer )
-  { *--s->bufp = c;
+  { unget_byte(c, s);
+
     return c;
   }
 
-  return -1;				/* no room */
+  return -1;
+}
+
+
+static int
+reperror(int c, IOSTREAM *s)
+{ if ( c >= 0 && (s->flags & (SIO_REPXML|SIO_REPPL)) )
+  { char buf[16];
+    const char *q;
+
+    if ( (s->flags & SIO_REPPL) )
+    { if ( c <= 0xffff )
+	sprintf(buf, "\\u%04X", c);
+      else
+	sprintf(buf, "\\U%08X", c);
+    } else
+      sprintf(buf, "&#%d;", c);
+
+    for(q = buf; *q; q++)
+    { if ( put_byte(*q, s) < 0 )
+	return -1;
+    }
+	
+    return c;
+  }
+
+  Sseterr(s, SIO_FERR|SIO_CLEARERR, "Encoding cannot represent character");
+  return -1;
+}
+
+
+
+int
+Sputcode(int c, IOSTREAM *s)
+{ if ( c < 0 )
+    return reperror(c, s);
+
+  if ( s->tee && s->tee->magic == SIO_MAGIC )
+    Sputcode(c, s->tee);
+
+#ifdef CRLF_MAPPING
+  if ( c == '\n' && (s->flags&SIO_TEXT) )
+  { if ( Sputcode('\r', s) < 0 )
+      return -1;
+  }
+#endif
+
+  switch(s->encoding)
+  { case ENC_OCTET:
+    case ENC_ISO_LATIN_1:
+      if ( c >= 256 )
+      { if ( reperror(c, s) < 0 )
+	  return -1;
+	break;
+      }
+    simple:
+      if ( put_byte(c, s) < 0 )
+	return -1;
+      break;
+    case ENC_ASCII:
+      if ( c >= 128 )
+      { if ( reperror(c, s) < 0 )
+	  return -1;
+	break;
+      }
+      goto simple;
+    case ENC_ANSI:
+    { char b[MB_LEN_MAX];
+      int n;
+
+      if ( !s->mbstate )
+      { if ( !(s->mbstate = malloc(sizeof(*s->mbstate))) )
+	  return EOF;			/* out of memory */
+	memset(s->mbstate, 0, sizeof(*s->mbstate));
+      }
+
+      if ( (n = wcrtomb(b, (wchar_t)c, s->mbstate)) < 0 )
+      { if ( reperror(c, s) < 0 )
+	  return -1;
+      } else
+      { int i;
+
+	for(i=0; i<n; i++)
+	{ if ( put_byte(b[i]&0xff, s) < 0 )
+	    return -1;
+	}
+      }
+
+      break;
+    }
+    case ENC_UTF8:
+    { char buf[6];
+      char *p, *end;
+      
+      if ( c < 128 )
+	goto simple;
+
+      end = utf8_put_char(buf, c);
+      for(p=buf; p<end; p++)
+      { if ( put_byte(*p&0xff, s) < 0 )
+	  return -1;
+      }
+
+      break;
+    }
+    case ENC_UNICODE_BE:
+      if ( put_byte(c>>8, s) < 0 )
+	return -1;
+      if ( put_byte(c&0xff, s) < 0 )
+	return -1;
+      break;
+    case ENC_UNICODE_LE:
+      if ( put_byte(c&0xff, s) < 0 )
+	return -1;
+      if ( put_byte(c>>8, s) < 0 )
+	return -1;
+      break;
+    case ENC_WCHAR:
+    { pl_wchar_t chr = c;
+      unsigned char *q = (unsigned char *)&chr;
+      unsigned char *e = &q[sizeof(pl_wchar_t)];
+
+      while(q<e)
+      { if ( put_byte(*q++, s) < 0 )
+	  return -1;
+      }
+      
+      break;
+    }
+    case ENC_UNKNOWN:
+      return -1;
+  }
+
+
+  s->lastc = c;
+
+  if ( c == '\n' && (s->flags & SIO_LBUF) )
+  { if ( S__flushbuf(s) < 0 )
+      return -1;
+  }
+
+  return S__updatefilepos(s, c);
+}
+
+
+int
+Scanrepresent(int c, IOSTREAM *s)
+{ switch(s->encoding)
+  { case ENC_OCTET:
+    case ENC_ISO_LATIN_1:
+      if ( c <= 0xff )
+	return 0;
+      return -1;
+    case ENC_ASCII:
+      if ( c < 0x7f )
+	return 0;
+      return -1;
+    case ENC_ANSI:
+    { mbstate_t state;
+      char b[MB_LEN_MAX];
+
+      memset(&state, 0, sizeof(state));
+      if ( wcrtomb(b, (wchar_t)c, &state) >= 0 )
+	return 0;
+      return -1;
+    }
+    case ENC_WCHAR:
+      if ( sizeof(wchar_t) > 2 )
+	return 0;
+    /*FALLTHROUGH*/
+    case ENC_UNICODE_BE:
+    case ENC_UNICODE_LE:
+      if ( c <= 0xffff )
+	return 0;
+      return -1;
+    case ENC_UTF8:
+      return 0;
+    default:
+      assert(0);
+      return -1;
+  }
+}
+
+
+int
+Sgetcode(IOSTREAM *s)
+{ int c;
+
+#ifdef CRLF_MAPPING
+retry:
+#endif
+
+  switch(s->encoding)
+  { case ENC_OCTET:
+    case ENC_ISO_LATIN_1:
+      c = get_byte(s);
+      break;
+    case ENC_ASCII:
+    { c = get_byte(s);
+      if ( c > 128 )
+	Sseterr(s, SIO_WARN, "non-ASCII character");
+      break;
+    }
+    case ENC_ANSI:
+    { char b[1];
+      int rc, n = 0;
+      wchar_t wc;
+
+      if ( !s->mbstate )
+      { if ( !(s->mbstate = malloc(sizeof(*s->mbstate))) )
+	  return EOF;			/* out of memory */
+	memset(s->mbstate, 0, sizeof(*s->mbstate));
+      }
+
+      for(;;)
+      { if ( (c = get_byte(s)) == EOF )
+	{ if ( n == 0 )
+	    return EOF;
+	  else
+	  { Sseterr(s, SIO_WARN, "EOF in multibyte Sequence");
+	    goto mberr;
+	  }
+	}
+	b[0] = c;
+
+	if ( (rc=mbrtowc(&wc, b, 1, s->mbstate)) == 1 )
+	{ c = wc;
+	  goto out;
+	} else if ( rc == -1 )
+	{ Sseterr(s, SIO_WARN, "Illegal multibyte Sequence");
+	  goto mberr;
+	}				/* else -2: incomplete */
+      }
+
+    mberr:
+      c = UTF8_MALFORMED_REPLACEMENT;
+      goto out;
+    }
+    case ENC_UTF8:
+    { c = get_byte(s);
+      if ( c == EOF )
+	break;
+
+      if ( c & 0x80 )
+      { int extra = UTF8_FBN(c);
+	int code;
+
+	code = UTF8_FBV(c,extra);
+	for( ; extra > 0; extra-- )
+	{ int c2 = get_byte(s);
+	  
+	  if ( !ISUTF8_CB(c2) )
+	  { Sseterr(s, SIO_WARN, "Illegal UTF-8 Sequence");
+	    c = UTF8_MALFORMED_REPLACEMENT;
+	    Sungetc(c2, s);
+	    goto out;
+	  }
+	  code = (code<<6)+(c2&0x3f);
+	}
+	c = code;
+      }
+      break;
+    }
+    case ENC_UNICODE_BE:
+    case ENC_UNICODE_LE:
+    { int c1, c2;
+
+      c1 = get_byte(s);
+      if ( c1 == EOF )
+	return EOF;
+      c2 = get_byte(s);
+
+      if ( c2 == EOF )
+      { Sseterr(s, SIO_WARN, "EOF in unicode character");
+	c = UTF8_MALFORMED_REPLACEMENT;
+      } else
+      { if ( s->encoding == ENC_UNICODE_BE )
+	  c = (c1<<8)+c2;
+	else
+	  c = (c2<<8)+c1;
+      }
+
+      break;
+    }
+    case ENC_WCHAR:
+    { pl_wchar_t chr;
+      char *p = (char*)&chr;
+      int n;
+
+      for(n=0; n<sizeof(pl_wchar_t); n++)
+      { int c1 = get_byte(s);
+
+	if ( c1 == EOF )
+	{ if ( n == 0 )
+	  { return EOF;
+	  } else
+	  { Sseterr(s, SIO_WARN, "EOF in UCS character");
+	    c = UTF8_MALFORMED_REPLACEMENT; 
+	    goto out;
+	  }
+	}
+
+	*p++ = c1;
+      }
+
+      c = chr;
+      break;
+    }
+    default:
+      assert(0);
+      c = -1;
+  }
+
+out:
+#ifdef CRLF_MAPPING
+  if ( c == '\r' && (s->flags&SIO_TEXT) )
+    goto retry;
+#endif
+
+  if ( s->tee && s->tee->magic == SIO_MAGIC && c != -1 )
+    Sputcode(c, s->tee);
+
+  return S__updatefilepos(s, c);
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(*) For ENC_ANSI there is  a  problem   as  this  deals with multi-modal
+streams, streams that  may  hold  escape   sequences  to  move  from one
+character set to another: ascii  ...   <esc1>  japanese <esc2> ascii ...
+Suppose now we have two characters   [ascii, japanese]. When reading the
+japanese character the  first  time,  the   system  will  translate  the
+<esc><japanese> and the mode will be   japanese. When pushing back, only
+the japanese character will be put back,   not the escape sequence. What
+to do?
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+Sungetcode(int c, IOSTREAM *s)
+{ switch(s->encoding)
+  { case ENC_OCTET:
+    case ENC_ISO_LATIN_1:
+      if ( c >= 256 )
+	return -1;			/* illegal */
+    simple:
+      if ( s->bufp > s->unbuffer )
+      { unget_byte(c, s);
+        return c;
+      }
+      return -1;			/* no room */
+    case ENC_ASCII:
+      if ( c >= 128 )
+	return -1;			/* illegal */
+      goto simple;
+    case ENC_ANSI:			/* (*) See above */
+    { char b[MB_LEN_MAX];
+      int n;
+
+      if ( !s->mbstate )		/* do we need a seperate state? */
+      { if ( !(s->mbstate = malloc(sizeof(*s->mbstate))) )
+	  return EOF;			/* out of memory */
+	memset(s->mbstate, 0, sizeof(*s->mbstate));
+      }
+
+      if ( (n = wcrtomb(b, (wchar_t)c, s->mbstate)) > 0  &&
+	   s->bufp - s->unbuffer >= n )
+      { int i;
+
+	for(i=n-1; i>=0; i--)
+	{ unget_byte(b[i], s);
+	}
+
+        return c;
+      }
+
+      return -1;
+    }
+    case ENC_UTF8:
+    { if ( (unsigned)c >= 0x8000000 )
+	return -1;
+
+      if ( c < 0x80 )
+      { goto simple;
+      } else
+      { char buf[6];
+	char *p, *end;
+
+	end = utf8_put_char(buf, c);
+	if ( s->bufp - s->unbuffer >= end-buf )
+	{ for(p=end-1; p>=buf; p--)
+	  { unget_byte(*p, s);
+	  }
+
+          return c;
+	}
+
+	return -1;
+      }
+    }
+    case ENC_UNICODE_BE:
+    { if ( c >= 0x10000 )
+	return -1;
+
+      if ( s->bufp-1 > s->unbuffer )
+      { unget_byte(c&0xff, s);
+	unget_byte((c>>8)&0xff, s);
+
+        return c;
+      }
+      return -1;
+    }
+    case ENC_UNICODE_LE:
+    { if ( c >= 0x10000 )
+	return -1;
+
+      if ( s->bufp-1 > s->unbuffer )
+      { unget_byte((c>>8)&0xff, s);
+	unget_byte(c&0xff, s);
+
+        return c;
+      }
+      return -1;
+    }
+    case ENC_WCHAR:
+    { pl_wchar_t chr = c;
+
+      if ( s->bufp-sizeof(chr) >= s->unbuffer )
+      { char *p = (char*)&chr;
+	int n;
+
+	for(n=sizeof(chr); --n>=0; )
+	  unget_byte(p[n], s);
+
+	return c;
+      }
+      return -1;
+    }
+    case ENC_UNKNOWN:
+      return -1;
+  }
+
+  assert(0);
+  return -1;
 }
 
 		 /*******************************
@@ -634,6 +1167,79 @@ Sread_pending(IOSTREAM *s, char *buf, int limit, int flags)
 }
 
 
+		 /*******************************
+		 *               BOM		*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Check the stream for a BOM (Byte Order Marker).  If present (and known),
+update the stream encoding.  Return value is one of
+
+	-1:	error (check errno)
+	 0:	ok.  If BOM, SIO_BOM is added to flags
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+typedef struct
+{ IOENC encoding;
+  int bomlen;
+  const char *bom;
+} bomdef;
+
+static const bomdef bomdefs[] =
+{ { ENC_UTF8,       3, "\357\273\277" }, /* 0xef, 0xbb, 0xbb */
+  { ENC_UNICODE_BE, 2, "\376\377" },	 /* 0xfe, oxff */
+  { ENC_UNICODE_LE, 2, "\377\376" },	 /* 0xff, oxfe */
+  { ENC_UNKNOWN,    0, NULL }
+};
+
+int
+ScheckBOM(IOSTREAM *s)
+{ if ( (s->flags & SIO_NBUF) )
+  { errno = EINVAL;
+    return -1;
+  }
+
+  for(;;)
+  { int avail = s->limitp - s->bufp;
+    const bomdef *bd;
+
+    for(bd=bomdefs; bd->bomlen; bd++)
+    { if ( avail >= bd->bomlen && memcmp(s->bufp, bd->bom, bd->bomlen) == 0 )
+      { s->encoding = bd->encoding;
+	s->bufp += bd->bomlen;
+	s->flags |= SIO_BOM;
+	return 0;
+      }
+    }
+
+    if ( avail >= 4 )			/* longest BOM */
+      return 0;
+
+    if ( S__fillbuf(s) == -1 )
+      return 0;				/* empty stream */
+    s->bufp--;
+  }
+}
+
+
+int
+SwriteBOM(IOSTREAM *s)
+{ switch(s->encoding)
+  { case ENC_UTF8:
+    case ENC_UNICODE_LE:
+    case ENC_UNICODE_BE:
+    { if ( Sputcode(0xfeff, s) != -1 )
+      { s->flags |= SIO_BOM;
+
+	return 0;
+      }
+      return -1;
+    }
+    default:
+      return 0;
+  }
+}
+
 
 		 /*******************************
 		 *               FLAGS		*
@@ -674,10 +1280,53 @@ Sfpasteof(IOSTREAM *s)
 
 void
 Sclearerr(IOSTREAM *s)
-{ s->flags &= ~(SIO_FEOF|SIO_FERR|SIO_FEOF2|SIO_TIMEOUT);
+{ s->flags &= ~(SIO_FEOF|SIO_WARN|SIO_FERR|SIO_FEOF2|SIO_TIMEOUT|SIO_CLEARERR);
+  Sseterr(s, 0, NULL);
 }
 
 
+void
+Sseterr(IOSTREAM *s, int flag, const char *message)
+{ if ( s->message )
+  { free(s->message);
+    s->message = NULL;
+    s->flags &= ~SIO_CLEARERR;
+  }
+  if ( message )
+  { s->flags |= flag;
+    s->message = strdup(message);
+  } else
+  { s->flags &= ~flag;
+  }
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Set the encoding of a stream. The enc   argument is the new encoding. If
+old is not NULL, the old encoding is written to the given location.
+
+Please note that not all code changing  the encoding call Ssetenc at the
+moment.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+Ssetenc(IOSTREAM *s, IOENC enc, IOENC *old)
+{ if ( old )
+    *old = s->encoding;
+  if ( enc == s->encoding )
+    return 0;
+
+  if ( s->functions->control )
+  { if ( (*s->functions->control)(s->handle,
+				  SIO_SETENCODING, 
+				  (void *)&enc) != 0 )
+      return -1;
+  }
+
+  s->encoding = enc;
+  return 0;
+}
+
 		 /*******************************
 		 *	      FLUSH		*
 		 *******************************/
@@ -696,6 +1345,28 @@ Sflush(IOSTREAM *s)
 		 *	      SEEK		*
 		 *******************************/
 
+static int
+Sunit_size(IOSTREAM *s)
+{ switch(s->encoding)
+  { case ENC_UNKNOWN:
+    case ENC_OCTET:
+    case ENC_ASCII:
+    case ENC_ISO_LATIN_1:
+    case ENC_ANSI:
+    case ENC_UTF8:
+      return 1;
+    case ENC_UNICODE_BE:
+    case ENC_UNICODE_LE:
+      return 2;
+    case ENC_WCHAR:
+      return sizeof(wchar_t);
+    default:
+      assert(0);
+      return 1;				/* not reached */
+  }
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Return the size of the underlying data object.  Should be optimized;
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -726,24 +1397,24 @@ Ssize(IOSTREAM *s)
 Maybe we should optimise this to become block-aligned?  Or can we leave
 this to read/write?
 
-The first part checks whether repositioning   the  read/write pointer in
-the buffer suffices to achieve the seek.
+The first part checks whether  repositioning   the  read  pointer in the
+buffer suffices to achieve the seek.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-long
-Sseek(IOSTREAM *s, long pos, int whence)
-{ if ( s->limitp > s->buffer )		/* something there */
-  { long now = Stell(s);
+int64_t
+Sseek64(IOSTREAM *s, int64_t pos, int whence)
+{ if ( (s->flags & SIO_INPUT) && s->limitp > s->buffer ) /* something there */
+  { int64_t now = Stell64(s);
 
     if ( now != -1 )
-    { long rval;
+    { int64_t rval;
       char *nbufp = (char *)-1;
     
       if ( whence == SIO_SEEK_CUR )
-      { nbufp = s->bufp + pos;
+      { nbufp = s->bufp + pos*Sunit_size(s);
 	rval = now + pos;
       } else if ( whence == SIO_SEEK_SET )
-      { nbufp = s->bufp + (pos - now);
+      { nbufp = s->bufp + (pos - now)*Sunit_size(s);
 	rval = pos;
       } else
 	rval = -1;			/* should not happen */
@@ -757,7 +1428,7 @@ Sseek(IOSTREAM *s, long pos, int whence)
     }
   }
 
-  if ( !s->functions->seek )
+  if ( !s->functions->seek && !s->functions->seek64 )
   { errno = ESPIPE;
     return -1;
   }
@@ -765,13 +1436,24 @@ Sseek(IOSTREAM *s, long pos, int whence)
   Sflush(s);
     
   s->bufp   = s->buffer;
-  s->limitp = s->buffer;
+  if ( (s->flags & SIO_INPUT) )
+    s->limitp = s->buffer;
 
   if ( whence == SIO_SEEK_CUR )
-  { pos += Stell(s);
+  { pos += Stell64(s);
     whence = SIO_SEEK_SET;
   }
-  pos = (*s->functions->seek)(s->handle, pos, whence);
+  
+  pos *= Sunit_size(s);
+  if ( s->functions->seek64 )
+    pos = (*s->functions->seek64)(s->handle, pos, whence);
+  else if ( pos <= LONG_MAX )
+    pos = (*s->functions->seek)(s->handle, (long)pos, whence);
+  else
+  { errno = EINVAL;
+    return -1;
+  }
+  pos /= Sunit_size(s);
 
 update:
   s->flags &= ~SIO_FEOF;		/* not on eof of file anymore */
@@ -786,19 +1468,38 @@ update:
 
 
 long
-Stell(IOSTREAM *s)
+Sseek(IOSTREAM *s, long pos, int whence)
+{ int64_t p2 = Sseek64(s, pos, whence);
+
+  if ( p2 <= LONG_MAX )
+    return (long) p2;
+
+  errno = EINVAL;
+  return -1;
+}
+
+
+
+int64_t
+Stell64(IOSTREAM *s)
 { if ( s->position )
   { return s->position->charno;
-  } else if ( s->functions->seek )
-  { long pos = (*s->functions->seek)(s->handle, 0L, SIO_SEEK_CUR);
+  } else if ( s->functions->seek || s->functions->seek64 )
+  { int64_t pos;
+
+    if ( s->functions->seek64 )
+      pos = (*s->functions->seek64)(s->handle, 0L, SIO_SEEK_CUR);
+    else
+      pos = (*s->functions->seek)(s->handle, 0L, SIO_SEEK_CUR);
+    pos /= Sunit_size(s);
 
     if ( s->buffer )			/* open */
-    { if ( s->flags & SIO_INPUT )
-      { pos -= s->limitp - s->buffer;
-	pos += s->bufp - s->buffer;
-      } else
-      { pos += s->bufp - s->buffer;
-      }
+    { int64_t off = s->bufp - s->buffer;
+
+      if ( s->flags & SIO_INPUT )
+	off -= s->limitp - s->buffer;
+
+      pos += off/Sunit_size(s);
     }
 
     return pos;
@@ -809,10 +1510,29 @@ Stell(IOSTREAM *s)
 }
 
 
+long
+Stell(IOSTREAM *s)
+{ int64_t pos = Stell64(s);
+
+  if ( pos <= LONG_MAX )
+    return (long) pos;
+
+  errno = EINVAL;
+  return -1;
+}
+
+
 		 /*******************************
 		 *	      CLOSE		*
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(*) Sclose() can be called recursively. For example if an XPCE object is
+only referenced from an open stream,  the close-function will delete the
+object, which in turn calls the  ->unlink   which  may wish to close the
+associated stream.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 Sclose(IOSTREAM *s)
 { int rval = 0;
@@ -822,19 +1542,16 @@ Sclose(IOSTREAM *s)
     return -1;
   }
 
+  if ( (s->flags&SIO_CLOSING) )		/* recursive (*) */
+    return rval;
+
   SLOCK(s);
   while(s->locks > 0)			/* remove buffer-locks */
     rval = Sunlock(s);
 
-  if ( s->buffer )
-  { if ( (s->flags & SIO_OUTPUT) && S__flushbuf(s) < 0 )
-      rval = -1;
-
-    if ( !(s->flags & SIO_USERBUF) )
-      free(s->buffer);
-
-    s->buffer = NULL;
-  }
+  rval = S__removebuf(s);
+  if ( s->mbstate )
+    free(s->mbstate);
 
   s->flags |= SIO_CLOSING;
 #ifdef __WIN32__
@@ -919,7 +1636,7 @@ int
 Sfputs(const char *q, IOSTREAM *s)
 { 
   for( ; *q; q++)
-  { if ( Sputc(*q, s) < 0 )
+  { if ( Sputcode(*q&0xff, s) < 0 )
       return EOF;
   }
 
@@ -969,8 +1686,14 @@ Svprintf(const char *fm, va_list args)
 }
 
 
+#define NEXTCHR(s, c)	if ( utf8 ) \
+			{ (s) = utf8_get_char((s), &(c)); \
+			} else \
+			{ c = *(s)++; c &= 0xff; \
+			}
+
 #define OUTCHR(s, c)	do { printed++; \
-			     if ( Sputc((c), (s)) < 0 ) goto error; \
+			     if ( Sputcode((c), (s)) < 0 ) goto error; \
 			   } while(0)
 #define valdigit(c)	((c) - '0')
 #define A_LEFT	0			/* left-aligned field */
@@ -1004,8 +1727,9 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	int has_arg1 = FALSE, has_arg2 = FALSE;
 	int arg1=0, arg2=0;
 	char fbuf[100], *fs = fbuf, *fe = fbuf;
-	int islong = FALSE;
+	int islong = 0;
 	int pad = ' ';
+	int utf8 = FALSE;
 
 	for(;;)
 	{ switch(*fm)
@@ -1020,6 +1744,7 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 
 	if ( *fm == '*' )
 	{ has_arg1++;
+	  fm++;
 	  arg1 = va_arg(args, int);
 	} else if ( isdigit(char_to_int(*fm)) )
 	{ if ( *fm == '0' )
@@ -1042,7 +1767,15 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	}
 
 	if ( *fm == 'l' )
-	{ islong++;
+	{ islong++;			/* 1: %ld */
+	  fm++;
+	}
+	if ( *fm == 'l' )
+	{ islong++;			/* 2: %lld */
+	  fm++;
+	}
+	if ( *fm == 'U' )		/* %Us: UTF-8 string */
+	{ utf8 = TRUE;
 	  fm++;
 	}
 
@@ -1057,21 +1790,42 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  case 'u':
 	  case 'x':
 	  case 'X':
-	  { long v;
+	  { long v = 0;			/* make compiler silent */
+	    int64_t vl = 0;
 	    char fmbuf[8], *fp=fmbuf;
 
-	    if ( islong )
-	      v = va_arg(args, long);
-	    else
-	      v = va_arg(args, int);
+	    switch( islong )
+	    { case 0:
+		v = va_arg(args, int);
+	        break;
+	      case 1:
+		v = va_arg(args, long);
+	        break;
+	      case 2:
+	        vl = va_arg(args, int64_t);
+		break;
+	    }
 
 	    *fp++ = '%';
 	    if ( modified )
 	      *fp++ = '#';
 	    *fp++ = 'l';
-	    *fp++ = *fm;
-	    *fp   = '\0';
-	    sprintf(fs, fmbuf, v);
+	    if ( islong < 2 )
+	    { *fp++ = *fm;
+	      *fp   = '\0';
+	      sprintf(fs, fmbuf, v);
+	    } else
+	    {
+#ifdef WIN32
+	      strcat(fp-1, "I64");	/* Synchronise with INT64_FORMAT! */
+	      fp += strlen(fp);
+#else
+	      *fp++ = 'l';
+#endif
+	      *fp++ = *fm;
+	      *fp   = '\0';
+	      sprintf(fs, fmbuf, vl);
+	    }
 	    fe = &fs[strlen(fs)];
 
 	    break;
@@ -1116,7 +1870,9 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  if ( align == A_LEFT )
 	  { int w = 0;
 	    while(*fs)
-	    { OUTCHR(s, *fs++);
+	    { int c;
+	      NEXTCHR(fs, c);
+	      OUTCHR(s, c);
 	      w++;
 	    }
 	    while(w < arg1)
@@ -1131,13 +1887,19 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	    else
 	      w = strlen(fs);
 
+	    if ( utf8 )
+	      w = utf8_strlen(fs, w);
+
 	    w = arg1 - w;
 	    while(w > 0 )
 	    { OUTCHR(s, pad);
 	      w--;
 	    }
 	    while(*fs)
-	      OUTCHR(s, *fs++);
+	    { int c;
+	      NEXTCHR(fs, c);
+	      OUTCHR(s, c);
+	    }
 	  }
 	} else
 	{ if ( fs == fbuf )		/* unaligned field, just output */
@@ -1145,7 +1907,10 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	      OUTCHR(s, *fs++);
 	  } else
 	  { while(*fs)
-	      OUTCHR(s, *fs++);
+	    { int c;
+	      NEXTCHR(fs, c);
+	      OUTCHR(s, c);
+	    }
 	  }
 	}
 	fm++;
@@ -1199,6 +1964,8 @@ Svsprintf(char *buf, const char *fm, va_list args)
   s.handle    = NULL;
   s.functions = NULL;
   s.mutex     = NULL;
+  s.encoding  = ENC_ISO_LATIN_1;
+  s.tee	      = NULL;
   
   if ( (rval = Svfprintf(&s, fm, args)) >= 0 )
     *s.bufp = '\0';
@@ -1567,6 +2334,17 @@ Sseek_file(void *handle, long pos, int whence)
 }
 
 
+#ifdef O_LARGEFILES
+static int64_t
+Sseek_file64(void *handle, int64_t pos, int whence)
+{ long h = (long) handle;
+
+					/* cannot do EINTR according to man */
+  return lseek((int)h, pos, whence);
+}
+#endif
+
+
 static int
 Sclose_file(void *handle)
 { long h = (long) handle;
@@ -1596,6 +2374,8 @@ Scontrol_file(void *handle, int action, void *arg)
       }
       return -1;
     }
+    case SIO_SETENCODING:
+      return 0;
     default:
       return -1;
   }
@@ -1607,7 +2387,12 @@ IOFUNCTIONS Sfilefunctions =
   Swrite_file,
   Sseek_file,
   Sclose_file,
-  Scontrol_file
+  Scontrol_file,
+#ifdef O_LARGEFILES
+  Sseek_file64
+#else
+  NULL
+#endif
 };
 
 
@@ -1628,6 +2413,7 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
   s->functions     = functions;
   s->timeout       = -1;		/* infinite */
   s->posbuf.lineno = 1;
+  s->encoding      = ENC_ISO_LATIN_1;
   if ( flags & SIO_RECORDPOS )
     s->position = &s->posbuf;
 #ifdef O_PLMT
@@ -1654,26 +2440,32 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
 Open a file. In addition to the normal  arguments, "lr" means get a read
 (shared-) lock on the file and  "lw"   means  get  an write (exclusive-)
 lock.  How much do we need to test here?
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+Note that the low-level open  is  always   binary  as  O_TEXT open files
+result in lost and corrupted data in   some  encodings (UTF-16 is one of
+them).  Sgetcode()  and  Sputcode()  do  the  LF  <->  CRLF  mapping  of
+CRLF_MAPPING is defined.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 IOSTREAM *
 Sopen_file(const char *path, const char *how)
 { int fd;
-  int oflags = 0, flags = SIO_FILE|SIO_TEXT|SIO_RECORDPOS;
+  int oflags = O_BINARY;
+  int flags = SIO_FILE|SIO_TEXT|SIO_RECORDPOS;
   int op = *how++;
   long lfd;
   enum {lnone=0,lread,lwrite} lock = lnone;
   IOSTREAM *s;
+  IOENC enc = ENC_UNKNOWN;
 
   for( ; *how; how++)
   { switch(*how)
     { case 'b':				/* binary */
 	flags &= ~SIO_TEXT;
-        oflags = O_BINARY;
+	enc = ENC_OCTET;
         break;
       case 'r':				/* no record */
-	flags &= SIO_RECORDPOS;
+	flags &= ~SIO_RECORDPOS;
         break;
       case 'l':				/* lock r: read, w: write */
 	if ( *++how == 'r' )
@@ -1692,7 +2484,7 @@ Sopen_file(const char *path, const char *how)
   }
 
 #if O_LARGEFILES && defined(O_LARGEFILE)
-  oflags |= O_LARGEFILE
+  oflags |= O_LARGEFILE;
 #endif
 
   switch(op)
@@ -1760,6 +2552,8 @@ Sopen_file(const char *path, const char *how)
 
   lfd = (long)fd;
   s = Snew((void *)lfd, flags, &Sfilefunctions);
+  if ( enc != ENC_UNKNOWN )
+    s->encoding = enc;
   if ( lock )
     s->flags |= SIO_ADVLOCK;
 
@@ -1816,14 +2610,16 @@ Sfileno(IOSTREAM *s)
 		      ((void *)(n)), &Sfilefunctions, \
 		      0, NULL, \
 		      (void (*)(void *))0, NULL, \
-		      -1 \
+		      -1, \
+		      0, \
+		      ENC_ISO_LATIN_1 \
 		    }
 
 #define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY)
 #define STDIO_STREAMS \
   STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */ \
-  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT), 		/* Soutput */ \
-  STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT)		/* Serror */
+  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT|SIO_REPPL), 	/* Soutput */ \
+  STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT|SIO_REPPL)	/* Serror */
 
 
 IOSTREAM S__iob[] =
@@ -1842,10 +2638,13 @@ SinitStreams()
 
   if ( !done++ )
   { int i;
+    IOENC enc = initEncoding();
 
     for(i=0; i<=2; i++)
     { if ( !isatty(i) )
 	S__iob[i].flags &= ~SIO_ISATTY;
+      if ( S__iob[i].encoding == ENC_ISO_LATIN_1 )
+	S__iob[i].encoding = enc;
 #ifdef O_PLMT
       S__iob[i].mutex = malloc(sizeof(recursiveMutex));
       recursiveMutexInit(S__iob[i].mutex);
@@ -1907,14 +2706,18 @@ IOFUNCTIONS Spipefunctions =
 
 IOSTREAM *
 Sopen_pipe(const char *command, const char *type)
-{ FILE *fd = popen(command, type);	/* HACK for now */
+{ char mode[2];
+  FILE *fd;
 
 #if 0
   Sdprintf("Opening \"%s\", mode \"%s\" --> %p (%d)\n",
 	   command, type, fd, errno);
 #endif
 
-  if ( fd )
+  mode[0] = type[0];
+  mode[1] = '\0';
+
+  if ( (fd = popen(command, mode)) )
   { int flags;
 
     if ( *type == 'r' )
@@ -2215,6 +3018,7 @@ Sopen_string(IOSTREAM *s, char *buf, int size, const char *mode)
   s->unbuffer  = buf;
   s->handle    = s;			/* for Sclose_string() */
   s->functions = &Sstringfunctions;
+  s->encoding  = ENC_ISO_LATIN_1;
 
   switch(*mode)
   { case 'r':
@@ -2237,13 +3041,6 @@ Sopen_string(IOSTREAM *s, char *buf, int size, const char *mode)
   return s;
 }
 
-#undef S__fupdatefilepos
-
-int
-S__fupdatefilepos(IOSTREAM *s, int c)
-{ return S___fupdatefilepos(s, c);
-}
-
 		 /*******************************
 		 *	       HOOKS		*
 		 *******************************/
@@ -2289,6 +3086,16 @@ Sclosehook(void (*hook)(IOSTREAM *s))
 		 *	       CLEANUP		*
 		 *******************************/
 
+void
+Sreset(void)
+{ IOSTREAM *s = Sinput;
+
+  if ( s && s->magic == SIO_MAGIC )
+  { s->bufp = s->limitp = s->buffer;
+  }
+} 
+
+
 void
 Scleanup(void)
 { close_hook *p, *next;
@@ -2302,8 +3109,10 @@ Scleanup(void)
   close_hooks = NULL;
 
   for(i=0; i<=2; i++)
-  { if ( S__iob[i].buffer )
-      free(S__iob[i].buffer);
-    S__iob[i] = S__iob0[i];
+  { IOSTREAM *s = &S__iob[i];
+
+    s->bufp = s->buffer;		/* avoid actual flush */
+    S__removebuf(s);
+    *s = S__iob0[i];			/* re-initialise */
   }
 }
