@@ -1,4 +1,4 @@
-/*  $Id: pl-sys.c,v 1.30 2004/09/21 14:22:26 jan Exp $
+/*  $Id: pl-sys.c,v 1.34 2006/07/14 15:38:05 jan Exp $
 
     Part of SWI-Prolog
 
@@ -32,13 +32,13 @@ word
 pl_shell(term_t command, term_t status)
 { char *cmd;
 
-  if ( PL_get_chars(command, &cmd, CVT_ALL) )
+  if ( PL_get_chars_ex(command, &cmd, CVT_ALL|REP_FN) )
   { int rval = System(cmd);
 
     return PL_unify_integer(status, rval);
   }
-  
-  return warning("shell/1: instantiation fault");
+    
+  fail;
 }
 
 
@@ -46,7 +46,7 @@ word
 pl_getenv(term_t var, term_t value)
 { char *n;
 
-  if ( PL_get_chars(var, &n, CVT_ALL) )
+  if ( PL_get_chars_ex(var, &n, CVT_ALL|REP_FN) )
   { int len = getenvl(n);
 
     if ( len >= 0 )
@@ -56,7 +56,7 @@ pl_getenv(term_t var, term_t value)
       { char *s;
 
 	if ( (s=getenv3(n, buf, len+1)) )
-	  return PL_unify_atom_chars(value, s);
+	  return PL_unify_chars(value, PL_ATOM|REP_FN, -1, s);
       } else
 	return PL_error("getenv", 2, NULL, ERR_NOMEM);
     }
@@ -64,7 +64,7 @@ pl_getenv(term_t var, term_t value)
     fail;
   }
 
-  return warning("getenv/2: instantiation fault");
+  fail;
 }  
 
 
@@ -72,13 +72,13 @@ word
 pl_setenv(term_t var, term_t value)
 { char *n, *v;
 
-  if ( PL_get_chars(var, &n, CVT_ALL|BUF_RING) &&
-       PL_get_chars(value, &v, CVT_ALL) )
+  if ( PL_get_chars_ex(var, &n, CVT_ALL|REP_FN|BUF_RING) &&
+       PL_get_chars_ex(value, &v, CVT_ALL|REP_FN) )
   { Setenv(n, v);
     succeed;
   }
 
-  return warning("setenv/2: instantiation fault");
+  fail;
 }
 
 
@@ -86,66 +86,13 @@ word
 pl_unsetenv(term_t var)
 { char *n;
 
-  if ( PL_get_chars(var, &n, CVT_ALL) )
+  if ( PL_get_chars_ex(var, &n, CVT_ALL|REP_FN) )
   { Unsetenv(n);
 
     succeed;
   }
 
-  return warning("unsetenv/1: instantiation fault");
-}
-
-
-word
-pl_convert_time(term_t time, term_t year, term_t month,
-		term_t day, term_t hour, term_t minute,
-		term_t second, term_t usec)
-{ double tf;
-
-  if ( PL_get_float(time, &tf) && tf <= PLMAXINT && tf >= PLMININT )
-  { struct tm tm;
-    long t    = (long) tf;
-    long us   = (long)((tf - (double) t) * 1000.0);
-
-    LocalTime(&t, &tm);
-
-    if ( PL_unify_integer(year,   tm.tm_year + 1900) &&
-	 PL_unify_integer(month,  tm.tm_mon + 1) &&
-	 PL_unify_integer(day,    tm.tm_mday) &&
-	 PL_unify_integer(hour,   tm.tm_hour) &&
-	 PL_unify_integer(minute, tm.tm_min) &&
-	 PL_unify_integer(second, tm.tm_sec) &&
-	 PL_unify_integer(usec,   us) )
-      succeed;
-    else
-      fail;
-  }
-
-  return PL_error("convert_time", 8, NULL, ERR_TYPE, ATOM_time_stamp, time);
-}
-
-
-word
-pl_convert_time2(term_t time, term_t string)
-{ double tf;
-
-  if ( PL_get_float(time, &tf) && tf <= PLMAXINT && tf >= PLMININT )
-  { time_t t  = (time_t)(long)tf;
-    char *s = ctime(&t);
-
-    if ( s )
-    { char *e = s + strlen(s);
-      while(e>s && e[-1] == '\n')
-	e--;
-      *e = EOS;
-
-      return PL_unify_string_chars(string, s);
-    }
-
-    return warning("convert_time/2: %s", OsError());
-  }
-  
-  return PL_error("convert_time", 2, NULL, ERR_TYPE, ATOM_time_stamp, time);
+  fail;
 }
 
 
