@@ -1,11 +1,11 @@
-/*  $Id: pl-thread.c,v 1.151 2004/10/02 15:38:01 jan Exp $
+/*  $Id: pl-thread.c,v 1.181 2006/08/16 07:15:29 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -73,6 +73,14 @@ APPROACH
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #include <errno.h>
+#if defined(__linux__) && defined(HAVE_GETTID)
+#include <linux/unistd.h>
+_syscall0(pid_t,gettid)
+#endif
+
+#ifdef HAVE_SYS_SYSCALL_H
+#include <sys/syscall.h>
+#endif
 
 #ifdef HAVE_SEMA_INIT			/* Solaris */
 #include <synch.h>
@@ -224,7 +232,8 @@ counting_mutex _PL_mutexes[] =
   COUNT_MUTEX_INITIALIZER("L_INIT"),
   COUNT_MUTEX_INITIALIZER("L_TERM"),
   COUNT_MUTEX_INITIALIZER("L_GC"),
-  COUNT_MUTEX_INITIALIZER("L_FOREIGN")
+  COUNT_MUTEX_INITIALIZER("L_FOREIGN"),
+  COUNT_MUTEX_INITIALIZER("L_OS")
 };
 
 
@@ -252,6 +261,16 @@ initMutexes()
     simpleMutexInit(&m->mutex);
 }
 
+static void
+deleteMutexes()
+{ counting_mutex *m;
+  int n = sizeof(_PL_mutexes)/sizeof(*m);
+  int i;
+
+  for(i=0, m=_PL_mutexes; i<n; i++, m++)
+    simpleMutexDelete(&m->mutex);
+}
+
 
 BOOL WINAPI
 DllMain(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved)
@@ -263,6 +282,8 @@ DllMain(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved)
       initMutexes();
       break;
     case DLL_PROCESS_DETACH:
+      deleteMutexes();
+      break;
     case DLL_THREAD_ATTACH:
     case DLL_THREAD_DETACH:
       break;
@@ -289,7 +310,7 @@ PRED_IMPL("mutex_statistics", 0, mutex_statistics, 0)
   { if ( cm->count == 0 )
       continue;
 
-    Sdprintf("%-32s %8d", cm->name, cm->count);
+    Sdprintf("%-32Us %8d", cm->name, cm->count); /* %Us: UTF-8 string */
 #ifdef O_CONTENTION_STATISTICS
     Sdprintf(" %8d", cm->collisions);
 #endif
@@ -311,6 +332,10 @@ PRED_IMPL("mutex_statistics", 0, mutex_statistics, 0)
 }
 
 
+#if defined(PTHREAD_CPUCLOCKS) || defined(LINUX_CPU_CLOCKS)
+#define ThreadCPUTime(info, which)	ThreadCPUTime(info)
+#endif
+
 		 /*******************************
 		 *	  LOCAL PROTOTYPES	*
 		 *******************************/
@@ -323,11 +348,11 @@ static void	unaliasThread(atom_t name);
 static void	run_thread_exit_hooks();
 static void	free_thread_info(PL_thread_info_t *info);
 static void	set_system_thread_id(PL_thread_info_t *info);
-static int	get_message_queue(term_t t, message_queue **queue,
-				  int create);
+static int	get_message_queue(term_t t, message_queue **queue);
 static void	cleanupLocalDefinitions(PL_local_data_t *ld);
 static int	unify_thread(term_t id, PL_thread_info_t *info);
 static pl_mutex *mutexCreate(atom_t name);
+static double   ThreadCPUTime(PL_thread_info_t *info, int which);
 
 
 		 /*******************************
@@ -386,6 +411,10 @@ static int
 initialise_thread(PL_thread_info_t *info)
 { assert(info->thread_data);
 
+  LOCK();
+  GD->statistics.threads_created++;
+  UNLOCK();
+
   TLD_set(PL_ldata, info->thread_data);
 
   if ( !info->local_size    ) info->local_size    = GD->options.localSize;
@@ -397,15 +426,16 @@ initialise_thread(PL_thread_info_t *info)
 			 info->global_size,
 			 info->trail_size,
 			 info->argument_size) )
+  { PL_local_data_t *ld = info->thread_data;
+
+    memset(&ld->stacks, 0, sizeof(ld->stacks));
+
     fail;
+  }
 
   initPrologLocalData();
   info->thread_data->magic = LD_MAGIC;
 
-  LOCK();
-  GD->statistics.threads_created++;
-  UNLOCK();
-
   return TRUE;
 }
 
@@ -421,6 +451,7 @@ free_prolog_thread(void *data)
 { PL_local_data_t *ld = data;
   PL_thread_info_t *info;
   int acknowlege;
+  double time;
 
   if ( !threads_ready )
     return;				/* Post-mortem */
@@ -439,7 +470,8 @@ free_prolog_thread(void *data)
   
   DEBUG(2, Sdprintf("Destroying data\n"));
   ld->magic = 0;
-  freeStacks(ld);
+  if ( ld->stacks.global.base )		/* otherwise assume they are not */
+    freeStacks(ld);			/* initialised */
   freeLocalData(ld);
 
   if ( ld->feature.table )
@@ -447,11 +479,12 @@ free_prolog_thread(void *data)
   /*PL_unregister_atom(ld->prompt.current);*/
 
   freeThreadSignals(ld);
+  time = ThreadCPUTime(info, CPU_USER);
 
   LOCK();
   destroy_message_queue(&ld->thread.messages);
   GD->statistics.threads_finished++;
-  GD->statistics.thread_cputime += CpuTime(CPU_USER);
+  GD->statistics.thread_cputime += time;
 
   info->thread_data = NULL;
   ld->thread.info = NULL;		/* avoid a loop */
@@ -606,7 +639,9 @@ exitPrologThreads()
     }
   }
   if ( canceled )
-  { Sdprintf("%d threads wouldn't die\n", canceled);
+  { printMessage(ATOM_informational,
+		 PL_FUNCTOR_CHARS, "threads_not_died", 1,
+		   PL_INT, canceled);
   } else
   { DEBUG(1, Sdprintf("done\n"));
   }
@@ -638,7 +673,7 @@ aliasThread(int tid, atom_t name)
 		    ERR_PERMISSION, ATOM_thread, ATOM_create, obj);
   }
 
-  addHTable(threadTable, (void *)name, (void *)tid);
+  addHTable(threadTable, (void *)name, (void *)(long)tid);
   PL_register_atom(name);
   threads[tid].name = name;
 
@@ -807,13 +842,11 @@ system_thread_id(PL_thread_info_t *info)
 #endif
 }
 
-
-
 static void
 set_system_thread_id(PL_thread_info_t *info)
 { 
-#ifdef __linux__
-  info->pid = getpid();
+#ifdef HAVE_GETTID
+  info->pid = gettid();
 #else
 #ifdef WIN32
   info->w32id = GetCurrentThreadId();
@@ -842,13 +875,13 @@ start_thread(void *closure)
 
   blockSignal(SIGINT);			/* only the main thread processes */
 					/* Control-C */
+  set_system_thread_id(info);		/* early to get exit code ok */
+
   if ( !initialise_thread(info) )
   { info->status = PL_THREAD_NOMEM;
     return (void *)TRUE;
   }
     
-  set_system_thread_id(info);
-
   pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
   pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
 
@@ -859,8 +892,14 @@ start_thread(void *closure)
   UNLOCK();
 
   goal = PL_new_term_ref();
-  PL_recorded(info->goal, goal);
-  rval  = callProlog(info->module, goal, PL_Q_CATCH_EXCEPTION, &ex);
+  PL_put_atom(goal, ATOM_dthread_init);
+
+  rval = callProlog(MODULE_system, goal, PL_Q_CATCH_EXCEPTION, &ex);
+
+  if ( rval )
+  { PL_recorded(info->goal, goal);
+    rval  = callProlog(info->module, goal, PL_Q_CATCH_EXCEPTION, &ex);
+  }
 
   if ( !rval && info->detached )
   { if ( ex )
@@ -924,7 +963,7 @@ pl_thread_create(term_t goal, term_t id, term_t options)
 
   ldnew = info->thread_data;
 
-  if ( !scan_options(options, 0,
+  if ( !scan_options(options, 0, /*OPT_ALL,*/
 		     ATOM_thread_option, make_thread_options,
 		     &info->local_size,
 		     &info->global_size,
@@ -937,7 +976,7 @@ pl_thread_create(term_t goal, term_t id, term_t options)
     fail;
   }
 
-#define MK_KBYTES(v) if ( v < (PLMAXINT/1024) ) v *= 1024
+#define MK_KBYTES(v) if ( v < (LONG_MAX/1024) ) v *= 1024
 
   MK_KBYTES(info->local_size);
   MK_KBYTES(info->global_size);
@@ -959,11 +998,14 @@ pl_thread_create(term_t goal, term_t id, term_t options)
   PL_register_atom(LD->prompt.current);
   ldnew->prompt			 = LD->prompt;
   if ( LD->prompt.first )
-    ldnew->prompt.first		 = store_string(LD->prompt.first);
+  { ldnew->prompt.first		 = LD->prompt.first;
+    PL_register_atom(ldnew->prompt.first);
+  }
   ldnew->modules		 = LD->modules;
   ldnew->IO			 = LD->IO;
   ldnew->_fileerrors		 = LD->_fileerrors;
   ldnew->float_format		 = LD->float_format;
+  ldnew->encoding		 = LD->encoding;
   ldnew->_debugstatus		 = LD->_debugstatus;
   ldnew->_debugstatus.retryFrame = NULL;
   ldnew->feature.mask		 = LD->feature.mask;
@@ -1006,7 +1048,7 @@ get_thread(term_t t, PL_thread_info_t **info, int warn)
     { Symbol s;
 
       if ( (s = lookupHTable(threadTable, (void *)name)) )
-	i = (int)s->value;
+	i = (int)(long)s->value;
     }
   }
 
@@ -1785,12 +1827,18 @@ dispatch_cond_wait(message_queue *queue)
     timeout.tv_sec  = now.tv_sec;
     timeout.tv_nsec = (now.tv_usec+250000) * 1000;
 
+    if ( timeout.tv_nsec >= 1000000000 ) /* some platforms demand this */
+    { timeout.tv_nsec -= 1000000000;
+      timeout.tv_sec += 1;
+    }
+
     rc = pthread_cond_timedwait(&queue->cond_var, &queue->mutex, &timeout);
 #ifdef O_DEBUG
     if ( LD && LD->thread.info )	/* can be absent during shutdown */
     { switch( LD->thread.info->ldata_status )
       { case LDATA_IDLE:
 	case LDATA_ANSWERED:
+	case LDATA_SIGNALLED:
 	  break;
 	default:
 	  Sdprintf("%d: ldata_status = %d\n",
@@ -1965,7 +2013,7 @@ pl_thread_send_message(term_t queue, term_t msg)
 { message_queue *q;
 
   LOCK();
-  if ( !get_message_queue(queue, &q, TRUE) )
+  if ( !get_message_queue(queue, &q) )
   { UNLOCK();
     fail;
   }
@@ -2000,7 +2048,7 @@ unify_queue(term_t t, message_queue *q)
   else
     return PL_unify_term(t,
 			 PL_FUNCTOR, FUNCTOR_dmessage_queue1,
-			 PL_INTEGER, valInt(q->id));
+			 PL_LONG, valInt(q->id));
 }
 
 
@@ -2046,7 +2094,7 @@ unlocked_message_queue_create(term_t queue)
 */
 
 static int
-get_message_queue(term_t t, message_queue **queue, int create)
+get_message_queue(term_t t, message_queue **queue)
 { atom_t name;
   word id = 0;
   int tid;
@@ -2085,20 +2133,11 @@ get_message_queue(term_t t, message_queue **queue, int create)
   { Symbol s = lookupHTable(threadTable, (void *)id);
 
     if ( s )
-    { tid = (int)s->value;
+    { tid = (int)(long)s->value;
       goto thread_queue;
     }
   }
 
-  if ( create && isTextAtom(id) )
-  { message_queue *new;
-
-    if ( (new = unlocked_message_queue_create(t)) )
-    { *queue = new;
-      return TRUE;
-    }
-  }
-
   return PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_message_queue, t);
 }
 
@@ -2121,7 +2160,7 @@ PRED_IMPL("message_queue_destroy", 1, message_queue_destroy, 0)
   Symbol s;
 
   LOCK();
-  if ( !get_message_queue(A1, &q, FALSE) )
+  if ( !get_message_queue(A1, &q) )
   { UNLOCK();
     fail;
   }
@@ -2146,6 +2185,28 @@ PRED_IMPL("message_queue_destroy", 1, message_queue_destroy, 0)
 }
 
 
+static 
+PRED_IMPL("message_queue_size", 2, message_queue_size, 0)
+{ message_queue *q;
+  thread_message *m;
+  int rc, n;
+
+  LOCK();
+  rc = get_message_queue(A1, &q);
+  UNLOCK();
+  if ( !rc )
+    fail;
+
+
+  simpleMutexLock(&q->mutex);
+  for(n=0, m=q->head; m; m = m->next)
+    n++;
+  simpleMutexUnlock(&q->mutex);
+
+  return PL_unify_integer(A2, n);
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 thread_get_message(+Queue, -Message)
 thread_get_message(-Message)
@@ -2156,9 +2217,12 @@ thread_get_message(-Message)
 static
 PRED_IMPL("thread_get_message", 2, thread_get_message, 0)
 { message_queue *q;
-
-					/* Needs lock? */
-  if ( !get_message_queue(A1, &q, TRUE) )
+  int rc;
+  
+  LOCK();
+  rc=get_message_queue(A1, &q);
+  UNLOCK();
+  if ( !rc )
     fail;
   
   return get_message(q, A2);
@@ -2168,8 +2232,12 @@ PRED_IMPL("thread_get_message", 2, thread_get_message, 0)
 static
 PRED_IMPL("thread_peek_message", 2, thread_peek_message, 0)
 { message_queue *q;
-
-  if ( !get_message_queue(A1, &q, TRUE) )
+  int rc;
+  
+  LOCK();
+  rc=get_message_queue(A1, &q);
+  UNLOCK();
+  if ( !rc )
     fail;
   
   return peek_message(q, A2);
@@ -2229,7 +2297,7 @@ int
 recursiveMutexInit(recursiveMutex *m)
 { 
 #ifdef RECURSIVE_MUTEXES
-  pthread_mutexattr_t *attr;
+  pthread_mutexattr_t *attr = NULL;
   int rc;
   
   if ( (rc=recursive_attr(&attr)) )
@@ -2381,7 +2449,7 @@ unify_mutex(term_t t, pl_mutex *m)
   else
     return PL_unify_term(t,
 			 PL_FUNCTOR, FUNCTOR_dmutex1,
-			 PL_INTEGER, valInt(m->id));
+			 PL_LONG, valInt(m->id));
 }
 
 
@@ -2534,10 +2602,10 @@ pl_mutex_trylock(term_t mutex)
 
   if ( self == m->owner )
   { m->count++;
-  } else if ( MUTEX_OK(rval = pthread_mutex_trylock(&m->mutex)) )
+  } else if ( (rval = pthread_mutex_trylock(&m->mutex)) == 0 )
   { m->count = 1;
     m->owner = self;
-  } else if ( rval == MUTEX_BUSY )
+  } else if ( rval == EBUSY )
   { fail;
   } else
   { assert(0);
@@ -2565,6 +2633,13 @@ PL_mutex_unlock(struct pl_mutex *m)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+The error message of this  predicate  is   not  thread-safe.  I.e. it is
+possible the message is wrong. This can   only be fixed by modifying the
+API of PL_mutex_unlock(), which is asking a  bit too much for this small
+error.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 foreign_t
 pl_mutex_unlock(term_t mutex)
 { pl_mutex *m;
@@ -2573,8 +2648,11 @@ pl_mutex_unlock(term_t mutex)
     fail;
 
   if ( !PL_mutex_unlock(m) )
-    return PL_error("mutex_unlock", 1, MSG_ERRNO, ERR_PERMISSION,
+  { char *msg = m->owner ? "not owner" : "not locked";
+
+    return PL_error("mutex_unlock", 1, msg, ERR_PERMISSION,
 		    ATOM_mutex, ATOM_unlock, mutex);
+  }
 
   succeed;
 }
@@ -2606,10 +2684,10 @@ pl_mutex_destroy(term_t mutex)
 { pl_mutex *m;
   Symbol s;
 
-  LOCK();
   if ( !get_mutex(mutex, &m, FALSE) )
     fail;
 
+  LOCK();
   if ( m->owner )
   { char msg[100];
     
@@ -2733,6 +2811,7 @@ PL_thread_attach_engine(PL_thread_attr_t *attr)
   ldnew->IO			 = ldmain->IO;
   ldnew->_fileerrors		 = ldmain->_fileerrors;
   ldnew->float_format		 = ldmain->float_format;
+  ldnew->encoding		 = ldmain->encoding;
   ldnew->_debugstatus		 = ldmain->_debugstatus;
   ldnew->_debugstatus.retryFrame = NULL;
   ldnew->feature.mask		 = ldmain->feature.mask;
@@ -2758,6 +2837,7 @@ PL_thread_attach_engine(PL_thread_attr_t *attr)
       return -1;
     }
   }
+  PL_call_predicate(MODULE_system, PL_Q_NORMAL, PROCEDURE_dthread_init0, 0);
 
   return info->pl_tid;
 }
@@ -2902,8 +2982,6 @@ PL_destroy_engine(PL_engine_t e)
 		 *	     STATISTICS		*
 		 *******************************/
 
-static void	sync_statistics(PL_thread_info_t *info, atom_t key);
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 thread_statistics(+Thread, +Key, -Value)
     Same as statistics(+Key, -Value) but operates on another thread.
@@ -2918,6 +2996,7 @@ PRED_IMPL("thread_statistics", 3, thread_statistics, 0)
 { PL_thread_info_t *info;
   PL_local_data_t *ld;
   int rval;
+  atom_t k;
 
   LOCK();
   if ( !get_thread(A1, &info, TRUE) )
@@ -2932,22 +3011,23 @@ PRED_IMPL("thread_statistics", 3, thread_statistics, 0)
 		    ATOM_statistics, ATOM_thread, A1);
   }
 
-  if ( LD != ld )
-  { atom_t k;
+  if ( !PL_get_atom(A2, &k) )
+    k = 0;
 
-    if ( !PL_get_atom(A2, &k) )
-      k = 0;
-
-    sync_statistics(info, k);
-    if ( k == ATOM_heapused )
-      ld = LD;
-  }
+  if ( k == ATOM_heapused )
+    ld = LD;
+  else if ( k == ATOM_cputime || k == ATOM_runtime )
+    ld->statistics.user_cputime = ThreadCPUTime(info, CPU_USER);
+  else if ( k == ATOM_system_time )
+    ld->statistics.system_cputime = ThreadCPUTime(info, CPU_SYSTEM);
 
   if ( LD == ld )		/* self: unlock first to avoid deadlock */
-    UNLOCK();
+  { UNLOCK();
+    return pl_statistics_ld(A2, A3, ld PASS_LD);
+  }
+
   rval = pl_statistics_ld(A2, A3, ld PASS_LD);
-  if ( LD != ld )
-    UNLOCK();
+  UNLOCK();
 
   return rval;
 }
@@ -2962,35 +3042,225 @@ PRED_IMPL("thread_statistics", 3, thread_statistics, 0)
 #define nano * 0.0000001
 #define ntick 1.0			/* manual says 100.0 ??? */
 
-static void
-sync_statistics(PL_thread_info_t *info, atom_t key)
-{ if ( key == ATOM_cputime || key == ATOM_runtime || key == ATOM_system_time )
-  { double t;
-    FILETIME created, exited, kerneltime, usertime;
-    HANDLE win_thread = pthread_getw32threadhandle_np(info->tid);
+static double
+ThreadCPUTime(PL_thread_info_t *info, int which)
+{ double t;
+  FILETIME created, exited, kerneltime, usertime;
+  HANDLE win_thread = pthread_getw32threadhandle_np(info->tid);
 
-    if ( GetThreadTimes(win_thread,
-			&created, &exited, &kerneltime, &usertime) )
-    { FILETIME *p;
+  if ( GetThreadTimes(win_thread,
+		      &created, &exited, &kerneltime, &usertime) )
+  { FILETIME *p;
 
-      if ( key == ATOM_system_time )
-	p = &kerneltime;
-      else
-	p = &usertime;
+    if ( which == CPU_SYSTEM )
+      p = &kerneltime;
+    else
+      p = &usertime;
 
-      t = (double)p->dwHighDateTime * (4294967296.0 * ntick nano);
-      t += (double)p->dwLowDateTime  * (ntick nano);
+    t = (double)p->dwHighDateTime * (4294967296.0 * ntick nano);
+    t += (double)p->dwLowDateTime  * (ntick nano);
 
-      if ( key == ATOM_system_time )
-	info->thread_data->statistics.system_cputime = t;
-      else
-	info->thread_data->statistics.user_cputime = t;
-    }
+    return t;
   }
+
+  return 0.0;
 }
 
 #else /*WIN32*/
 
+#define timespec_to_double(ts) \
+	(double)ts.tv_sec + (double)ts.tv_nsec/1000000000ull
+
+#ifdef PTHREAD_CPUCLOCKS
+
+static double
+ThreadCPUTime(PL_thread_info_t *info, int which)
+{ if ( info->tid )
+  { clockid_t clock_id;
+    struct timespec ts;
+
+    pthread_getcpuclockid(info->tid, &clock_id);
+    if (clock_gettime(clock_id, &ts) == 0)
+      return timespec_to_double(ts);
+  }
+
+  return 0.0;
+}
+
+#else /*PTHREAD_CPUCLOCKS*/
+
+#ifdef LINUX_CPUCLOCKS
+
+static double
+ThreadCPUTime(PL_thread_info_t *info, int which)
+{
+  struct timespec ts;
+
+  if (syscall(__NR_clock_gettime, CLOCK_THREAD_CPUTIME_ID, &ts) == 0)
+    return timespec_to_double(ts);
+
+  return 0.0;
+}
+
+#else /*LINUX_CPUCLOCKS*/
+
+#ifdef LINUX_PROCFS
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Unfortunately POSIX threads does not define a   way  to get the CPU time
+per thread. Some systems do have mechanisms  for that. POSIX does define
+clock_gettime(CLOCK_THREAD_CPUTIME_ID), but it certainly doesn't work in
+Linux 2.6.8.
+
+Autoconf detects the presense of  /proc  and   we  read  the values from
+there. This is rather slow. To make things not  too bad we use a pool of
+open handles to entries in the /proc  table. All this junk should really
+be moved into a file trying to implement  thread CPU time for as many as
+possible platforms. If you happen to know  such a library, please let me
+know.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#include <fcntl.h>
+
+#define CACHED_PROCPS_ENTRIES 5
+
+typedef struct
+{ int tid;				/* process id */
+  int fd;				/* file descriptor */
+  int offset;				/* start of numbers */
+  int usecoount;
+} procps_entry;
+
+static procps_entry procps_entries[CACHED_PROCPS_ENTRIES]; /* cached entries */
+
+static void
+close_procps_entry(procps_entry *e)
+{ if ( e->tid )
+  { close(e->fd);
+    memset(e, 0, sizeof(*e));
+  }
+}
+
+
+static procps_entry*
+reclaim_procps_entry()
+{ procps_entry *e, *low;
+  int i; int lowc;
+
+  low=e=procps_entries;
+  lowc=low->usecoount;
+
+  for(e++, i=1; i<CACHED_PROCPS_ENTRIES; e++, i++)
+  { if ( e->usecoount < lowc )
+    { lowc = e->usecoount;
+      low = e;
+    }
+  }
+
+  for(e=procps_entries, i=0; i<CACHED_PROCPS_ENTRIES; e++, i++)
+    e->usecoount = 0;
+
+  close_procps_entry(low);
+
+  return low;
+}
+
+
+static procps_entry *
+open_procps_entry(procps_entry *e, int tid)
+{ char fname[256];
+  int fd;
+
+  sprintf(fname, "/proc/self/task/%d/stat", tid);
+  if ( (fd=open(fname, O_RDONLY)) >= 0 )
+  { char buffer[1000];
+    int pos;
+
+    pos = read(fd, buffer, sizeof(buffer)-1);
+    if ( pos > 0 )
+    { char *bp;
+
+      buffer[pos] = EOS;
+      if ( (bp=strrchr(buffer, ')')) )
+      { e->tid = tid;
+	e->fd = fd;
+	e->offset = (bp-buffer)+4;
+	e->usecoount = 1;
+	
+	return e;
+      }
+    }
+  }
+
+  return NULL;
+}
+
+
+static procps_entry*
+get_procps_entry(int tid)
+{ int i;
+  procps_entry *e;
+
+  for(e=procps_entries, i=0; i<CACHED_PROCPS_ENTRIES; e++, i++)
+  { if ( e->tid == tid )
+    { e->usecoount++;
+
+      return e;
+    }
+  }
+
+  for(e=procps_entries, i=0; i<CACHED_PROCPS_ENTRIES; e++, i++)
+  { if ( e->tid == 0 )
+      return open_procps_entry(e, tid);
+  }
+
+  e = reclaim_procps_entry();
+  return open_procps_entry(e, tid);
+}
+
+
+static double
+ThreadCPUTime(PL_thread_info_t *info, int which)
+{ procps_entry *e;
+
+  if ( (e=get_procps_entry(info->pid)) )
+  { char buffer[1000];
+    char *s;
+    long long ticks;
+    int i, n, nth = 10;			/* user time */
+
+    if ( which == CPU_SYSTEM )
+      nth++;
+
+    lseek(e->fd, e->offset, SEEK_SET);
+    n = read(e->fd, buffer, sizeof(buffer)-1);
+    if ( n >= 0 )
+      buffer[n] = EOS;
+					/* most likely does not need reuse */
+    if ( info->status != PL_THREAD_RUNNING )
+      close_procps_entry(e);
+    
+    for(s=buffer, i=0; i<nth; i++, s++)
+    { while(*s != ' ')
+	s++;
+    }
+
+    ticks = atoll(s);
+    return (double)ticks/100.0;
+  }
+
+  return 0.0;
+}
+
+#else /*LINUX_PROCFS*/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Code that probably only works  on   Linux  2.4  systems, where CpuTime()
+returns the per-thread time. This isn't very   nice as the time is store
+in LD in addition to being returned, but it  is ok for now and Linux 2.4
+is almost dead anyway.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static void
 SyncUserCPU(int sig)
 { LD->statistics.user_cputime = CpuTime(CPU_USER);
@@ -3005,17 +3275,19 @@ SyncSystemCPU(int sig)
 }
 
 
-static void
-sync_statistics(PL_thread_info_t *info, atom_t key)
-{ if ( key == ATOM_cputime || key == ATOM_runtime || key == ATOM_system_time )
+static double
+ThreadCPUTime(PL_thread_info_t *info, int which)
+{ if ( info->thread_data == LD )
+  { return CpuTime(which);
+  } else
   { struct sigaction old;
     struct sigaction new;
 
     sem_init(sem_mark_ptr, USYNC_THREAD, 0);
     memset(&new, 0, sizeof(new));
-    if ( key == ATOM_cputime || key == ATOM_runtime )
+    if ( which == CPU_USER )
       new.sa_handler = SyncUserCPU;
-    else /*if ( key == ATOM_system_time )*/
+    else 
       new.sa_handler = SyncSystemCPU;
 
     new.sa_flags   = SA_RESTART;
@@ -3026,9 +3298,17 @@ sync_statistics(PL_thread_info_t *info, atom_t key)
     }
     sem_destroy(&sem_mark);
     sigaction(SIG_FORALL, &old, NULL);
+
+    if ( which == CPU_USER )
+      return LD->statistics.user_cputime;
+    else
+      return LD->statistics.system_cputime;
   }
 }
 
+#endif /*LINUX_PROCFS*/
+#endif /*LINUX_CPUCLOCKS*/
+#endif /*PTHREAD_CPUCLOCKS*/
 #endif /*WIN32*/
 
 
@@ -3566,5 +3846,6 @@ BeginPredDefs(thread)
   PRED_DEF("message_queue_destroy", 1, message_queue_destroy, 0)
   PRED_DEF("thread_setconcurrency", 2, thread_setconcurrency, 0)
   PRED_DEF("mutex_statistics", 0, mutex_statistics, 0)
+  PRED_DEF("message_queue_size", 2, message_queue_size, 0)
 #endif
 EndPredDefs
