@@ -1,4 +1,4 @@
-/*  $Id: pl-trace.c,v 1.102.2.1 2005/01/17 09:07:43 jan Exp $
+/*  $Id: pl-trace.c,v 1.108 2006/08/07 12:36:19 jan Exp $
 
     Part of SWI-Prolog
 
@@ -64,7 +64,7 @@ PL_unify_frame(term_t t, LocalFrame fr)
 }
 
 
-static void
+void
 PL_put_frame(term_t t, LocalFrame fr)
 { if ( fr )
   { assert(fr >= lBase && fr < lTop);
@@ -331,7 +331,7 @@ ok:
 Give a trace on the skipped goal for a redo.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-  { Code pc2;
+  { Code pc2 = NULL;
 
     if ( port == REDO_PORT && debugstatus.skiplevel == VERY_DEEP &&
 	 (fr = redoFrame(frame, &pc2)) != NULL )
@@ -418,7 +418,7 @@ again:
     if ( action == ACTION_AGAIN )
       goto again;
   } else
-    Sputc('\n', Sdout);
+    Sputcode('\n', Sdout);
 
 out:
   unblockGC(PASS_LD1);
@@ -532,7 +532,7 @@ traceAction(char *cmd, int port, LocalFrame frame, Choice bfr, bool interactive)
   char *s;
 
 #define FeedBack(msg)	{ if (interactive) { if (cmd[1] != EOS) \
-					       Sputc('\n', Sdout); \
+					       Sputcode('\n', Sdout); \
 					     else \
 					       Sfputs(msg, Sdout); } }
 #define Warn(msg)	{ if (interactive) \
@@ -1664,11 +1664,17 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
   } else if ( key == ATOM_pc )
   { if ( fr->programPointer &&
 	 fr->parent &&
-	 false(fr->parent->predicate, FOREIGN) )
-      PL_put_integer(result,
-		     fr->programPointer - fr->parent->clause->clause->codes);
-    else
-      fail;
+	 false(fr->parent->predicate, FOREIGN) &&
+	 fr->parent->predicate != PROCEDURE_dcall1->definition )
+    { long pc = fr->programPointer - fr->parent->clause->clause->codes;
+
+      if ( pc < 0 )
+	trap_gdb();
+
+      PL_put_integer(result, pc);
+    } else
+    { fail;
+    }
   } else if ( key == ATOM_hidden )
   { atom_t a;
 
@@ -1708,7 +1714,7 @@ pl_prolog_frame_attribute(term_t frame, term_t what,
 
 foreign_t
 pl_prolog_choice_attribute(term_t choice, term_t what, term_t value)
-{ Choice ch;
+{ Choice ch = NULL;
   atom_t key;
 
   if ( !PL_get_choice(choice, &ch) ||
@@ -1761,8 +1767,9 @@ callEventHook(int ev, ...)
     arg = PL_new_term_ref();
 
     if ( exception_term )
-      ex = PL_copy_term_ref(exception_term);
-    else
+    { ex = PL_copy_term_ref(exception_term);
+      exception_term = 0;
+    } else
       ex = 0;
 
     va_start(args, ev);
