@@ -1,11 +1,11 @@
-/*  $Id: pl-wic.c,v 1.111.2.1 2005/02/19 16:43:54 jan Exp $
+/*  $Id: pl-wic.c,v 1.132 2006/06/22 09:55:50 jan Exp $
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2006, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -22,7 +22,7 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-#define O_DEBUG 1
+/*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
@@ -31,20 +31,23 @@
 #include <unistd.h>
 #endif
 
-static char *	getString(IOSTREAM *);
-static long	getNum(IOSTREAM *);
+static char *	getString(IOSTREAM *, unsigned *len);
+static int64_t	getInt64(IOSTREAM *);
+static long	getLong(IOSTREAM *);
+static int	getInt(IOSTREAM *);
 static real	getReal(IOSTREAM *);
 static bool	loadWicFd(IOSTREAM *);
 static bool	loadPredicate(IOSTREAM *, int skip ARG_LD);
 static bool	loadImport(IOSTREAM *, int skip ARG_LD);
-static void	putString(char *, IOSTREAM *);
-static void	putAtom(atom_t, IOSTREAM *);
-static void	putNum(long, IOSTREAM *);
+static void	saveXRBlobType(PL_blob_t *type, IOSTREAM *fd);
+static void	putString(const char *, unsigned len, IOSTREAM *);
+static void	putNum(int64_t, IOSTREAM *);
 static void	putReal(real, IOSTREAM *);
 static void	saveWicClause(Clause, IOSTREAM *);
 static void	closeProcedureWic(IOSTREAM *);
 static bool	compileFile(char *);
 static word	loadXRc(int c, IOSTREAM *fd ARG_LD);
+static atom_t   getBlob(IOSTREAM *fd ARG_LD);
 static bool	loadStatement(int c, IOSTREAM *fd, int skip ARG_LD);
 static bool	loadPart(IOSTREAM *fd, Module *module, int skip ARG_LD);
 static bool	loadInModule(IOSTREAM *fd, int skip ARG_LD);
@@ -134,14 +137,17 @@ Below is an informal description of the format of a `.qlf' file:
 		      | 'X' 				% end of list
 <XR>		::=	XR_REF     <num>		% XR id from table
 			XR_ATOM    <len><chars>		% atom
+			XR_BLOB	   <blob><private>	% typed atom (blob)
 			XR_INT     <num>		% number
 			XR_FLOAT   <word>*		% real (double)
 			XR_STRING  <string>		% string
+			XR_STRING_UTF8  <utf-8 string>	% wide string
 			XR_FUNCTOR <XR/name> <num>	% functor
 			XR_PRED    <XR/fdef> <XR/module>% predicate
 			XR_MODULE  <XR/name>		% module
 			XR_FILE	   's'|'u' <XR/atom> <time>
 				   '-'
+			XR_BLOB_TYPE <len><chars>	% blob type-name
 <term>		::=	<num>				% # variables in term
 			<theterm>
 <theterm>	::=	<XR/atomic>			% atomic data
@@ -167,8 +173,8 @@ between  16  and  32  bits  machines (arities on 16 bits machines are 16
 bits) as well as machines with different byte order.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define LOADVERSION 38			/* load all versions later >= X */
-#define VERSION 38			/* save version number */
+#define LOADVERSION 43			/* load all versions later >= X */
+#define VERSION 43			/* save version number */
 #define QLFMAGICNUM 0x716c7374		/* "qlst" on little-endian machine */
 
 #define XR_REF     0			/* reference to previous */
@@ -180,6 +186,9 @@ bits) as well as machines with different byte order.
 #define XR_STRING  6			/* string */
 #define XR_FILE	   7			/* source file */
 #define XR_MODULE  8			/* a module */
+#define XR_BLOB	   9			/* a typed atom (blob) */
+#define XR_BLOB_TYPE 10			/* name of atom-type declaration */
+#define XR_STRING_UTF8 11		/* Wide character string */
 
 #define PRED_SYSTEM	 0x01		/* system predicate */
 #define PRED_HIDE_CHILDS 0x02		/* hide my childs */
@@ -309,43 +318,82 @@ qlfLoadError(IOSTREAM *fd, char *ctx)
 }
 
 
-static char *getstr_buffer;
-static char *getstr_buffer_end;
-static int   getstr_buffer_size = 512;
+static char *getstr_buffer = NULL;
+static int  getstr_buffer_size = 0;
 
 static char *
-getString(IOSTREAM *fd)
+getString(IOSTREAM *fd, unsigned *length)
 { char *s;
-  Char c;
+  int len = getInt(fd);
+  int i;
 
-  if ( getstr_buffer == NULL )
-  { if ( !(getstr_buffer = malloc(getstr_buffer_size)) )
+  if ( getstr_buffer_size < len+1 )
+  { int size = ((len+1+1023)/1024)*1024;
+
+    if ( getstr_buffer )
+      getstr_buffer = realloc(getstr_buffer, size);
+    else
+      getstr_buffer = malloc(size);
+
+    if ( getstr_buffer )
+      getstr_buffer_size = size;
+    else
       outOfCore();
-    getstr_buffer_end = &getstr_buffer[getstr_buffer_size-1];
   }
 
-  for( s = getstr_buffer; (*s = c = Getc(fd)) != EOS; s++ )
-  { if ( s == getstr_buffer_end )
-    { if ( !(getstr_buffer = realloc(getstr_buffer, getstr_buffer_size+512)) )
-	outOfCore();
-      s = &getstr_buffer[getstr_buffer_size-1];
-      getstr_buffer_size += 512;
-      getstr_buffer_end = &getstr_buffer[getstr_buffer_size-1];
-    }
+  for( i=0, s = getstr_buffer; i<len; i++ )
+  { int c = Getc(fd);
+
     if ( c == EOF )
       fatalError("Unexpected EOF on intermediate code file at offset %d",
 		 Stell(fd));
+
+    *s++ = c;
   }
+  *s = EOS;
+
+  if ( length )
+    *length = (unsigned) len;
 
   return getstr_buffer;
 }
 
 
+pl_wchar_t *
+wicGetStringUTF8(IOSTREAM *fd, unsigned *length,
+		 pl_wchar_t *buf, size_t bufsize)
+{ size_t i, len = (size_t)wicGetNum(fd);
+  IOENC oenc = fd->encoding;
+  pl_wchar_t *tmp, *o;
+  
+  if ( length )
+    *length = len;
+
+  if ( len < bufsize )
+    tmp = buf;
+  else
+    tmp = PL_malloc(len*sizeof(pl_wchar_t));
+
+  fd->encoding = ENC_UTF8;
+  for(i=0, o=tmp; i<len; i++)
+  { int c = Sgetcode(fd);
+
+    if ( c < 0 )
+      fatalError("Unexpected EOF in UCS atom");
+    *o++ = c;
+  }
+  fd->encoding = oenc;
+
+  return tmp;
+}
+
+
+
 static atom_t
-getAtom(IOSTREAM *fd ARG_LD)
+getAtom(IOSTREAM *fd, PL_blob_t *type ARG_LD)
 { char buf[1024];
   char *tmp, *s;
-  int len = getNum(fd);
+  int len = getInt(fd);
   int i;
   atom_t a;
 
@@ -362,7 +410,13 @@ getAtom(IOSTREAM *fd ARG_LD)
 		 Stell(fd));
     *s++ = c;
   }
-  a = lookupAtom(tmp, len);
+  if ( type )
+  { int new;
+
+    a = lookupBlob(tmp, len, type, &new);
+  } else
+  { a = lookupAtom(tmp, len);
+  }
 
   if ( tmp != buf )
     freeHeap(tmp, len);
@@ -371,6 +425,14 @@ getAtom(IOSTREAM *fd ARG_LD)
 }
 
 
+static PL_blob_t *
+getBlobType(IOSTREAM *fd)
+{ const char *name = getString(fd, NULL);
+
+  return PL_find_blob_type(name);
+}
+
+
 static char *
 getMagicString(IOSTREAM *fd, char *buf, int maxlen)
 { char *s;
@@ -387,16 +449,16 @@ getMagicString(IOSTREAM *fd, char *buf, int maxlen)
 }
 
 
-static long
-getNum(IOSTREAM *fd)
-{ long first = Getc(fd);
+static int64_t
+getInt64(IOSTREAM *fd)
+{ int64_t first = Getc(fd);
   int bytes, shift, b;
 
   if ( !(first & 0xc0) )		/* 99% of them: speed up a bit */    
-  { first <<= (LONGBITSIZE-6);
-    first >>= (LONGBITSIZE-6);
+  { first <<= (INT64BITSIZE-6);
+    first >>= (INT64BITSIZE-6);
 
-    DEBUG(4, Sdprintf("getNum() --> %ld\n", first));
+    DEBUG(4, Sdprintf("getInt64() --> " INT64_FORMAT "\n", first));
     return first;
   }
 
@@ -409,28 +471,44 @@ getNum(IOSTREAM *fd)
       first |= Getc(fd) & 0xff;
     }
 
-    shift = (sizeof(long)-1-bytes)*8 + 2;
+    shift = (sizeof(first)-1-bytes)*8 + 2;
   } else
   { int m;
 
-    bytes = first;
-    first = 0L;
+    bytes = (int)first;
+    first = (int64_t)0;
 
     for(m=0; m<bytes; m++)
     { first <<= 8;
       first |= Getc(fd) & 0xff;
     }
-    shift = (sizeof(long)-bytes)*8;
+    shift = (sizeof(first)-bytes)*8;
   }
 
   first <<= shift;
   first >>= shift;
   
-  DEBUG(4, Sdprintf("getNum() --> %ld\n", first));
+  DEBUG(4, Sdprintf("getInt64() --> " INT64_FORMAT "\n", first));
   return first;
 }
 
 
+static long
+getLong(IOSTREAM *fd)
+{ int64_t val = getInt64(fd);
+
+  return (long)val;
+}
+
+
+static int
+getInt(IOSTREAM *fd)
+{ int64_t val = getInt64(fd);
+
+  return (int)val;
+}
+
+
 #ifdef WORDS_BIGENDIAN
 static const int double_byte_order[] = { 7,6,5,4,3,2,1,0 };
 #else
@@ -459,8 +537,8 @@ getReal(IOSTREAM *fd)
 }
 
 
-unsigned long
-getLong(IOSTREAM *s)
+word
+getWord(IOSTREAM *s)
 { unsigned long v;
 
   v  = (Sgetc(s) & 0xff) << 24;
@@ -486,24 +564,36 @@ loadXRc(int c, IOSTREAM *fd ARG_LD)
 
   switch( c )
   { case XR_REF:
-    { long xr  = getNum(fd);
+    { long xr  = getLong(fd);
       word val = lookupXrId(xr);
 
       return val;
     }
     case XR_ATOM:
     { id = ++loadedXRTableId;
-      xr = getAtom(fd PASS_LD);
+      xr = getAtom(fd, NULL PASS_LD);
       DEBUG(3, Sdprintf("XR(%d) = '%s'\n", id, stringAtom(xr)));
       break;
     }
+    case XR_BLOB:
+    { id = ++loadedXRTableId;
+      xr = getBlob(fd PASS_LD);
+      DEBUG(3, Sdprintf("XR(%d) = <blob>\n", id));
+      break;
+    }
+    case XR_BLOB_TYPE:
+    { id = ++loadedXRTableId;
+      xr = (word)getBlobType(fd);
+      DEBUG(3, Sdprintf("XR(%d) = <blob-type>%s", id, ((PL_blob_t*)xr)->name));
+      break;
+    }
     case XR_FUNCTOR:
     { atom_t name;
       int arity;
 
       id = ++loadedXRTableId;
       name = loadXR(fd);
-      arity = getNum(fd);
+      arity = getInt(fd);
       xr = (word) lookupFunctorDef(name, arity);
       DEBUG(3, Sdprintf("XR(%d) = %s/%d\n", id, stringAtom(name), arity));
       break;
@@ -528,12 +618,31 @@ loadXRc(int c, IOSTREAM *fd ARG_LD)
       break;
     }
     case XR_INT:
-      return makeNum(getNum(fd));
+      return makeNum(getInt64(fd));
     case XR_FLOAT:
       return globalReal(getReal(fd));
 #if O_STRING
     case XR_STRING:
-      return globalString(getString(fd));
+    { char *s;
+      unsigned len;
+
+      s = getString(fd, &len);
+
+      return globalString(len, s);
+    }
+    case XR_STRING_UTF8:
+    { pl_wchar_t *w;
+      unsigned len;
+      pl_wchar_t buf[256];
+      word s;
+      
+      w = wicGetStringUTF8(fd, &len, buf, sizeof(buf)/sizeof(pl_wchar_t));
+      s = globalWString(len, w);
+      if ( w != buf )
+	PL_free(w);
+
+      return s;
+    }
 #endif
     case XR_FILE:
     { int c;
@@ -544,7 +653,7 @@ loadXRc(int c, IOSTREAM *fd ARG_LD)
       { case 'u':
 	case 's':
 	{ atom_t name   = loadXR(fd);
-	  word   time   = getNum(fd);
+	  word   time   = getLong(fd);
 	  const char *s = stringAtom(name);
 	  SourceFile sf = lookupSourceFile(qlfFixSourcePath(s));
 
@@ -552,6 +661,7 @@ loadXRc(int c, IOSTREAM *fd ARG_LD)
 	  { sf->time   = time;
 	    sf->system = (c == 's' ? TRUE : FALSE);
 	  }
+	  sf->count++;
 	  xr = (word)sf;
 	  break;
 	}
@@ -577,12 +687,24 @@ loadXRc(int c, IOSTREAM *fd ARG_LD)
 }
 
 
+static atom_t
+getBlob(IOSTREAM *fd ARG_LD)
+{ PL_blob_t *type = (PL_blob_t*)loadXR(fd);
+  
+  if ( type->load )
+  { return (*type->load)(fd);
+  } else
+  { return getAtom(fd, type PASS_LD);
+  }
+}
+
+
 static void
 do_load_qlf_term(IOSTREAM *fd, term_t vars[], term_t term ARG_LD)
 { int c = Qgetc(fd);
 
   if ( c == 'v' )
-  { int id = getNum(fd);
+  { int id = getInt(fd);
     
     if ( vars[id] )
       PL_unify(term, vars[id]);
@@ -614,7 +736,7 @@ loadQlfTerm(term_t term, IOSTREAM *fd ARG_LD)
 
   DEBUG(3, Sdprintf("Loading from %d ...", Stell(fd)));
 
-  if ( (nvars = getNum(fd)) )
+  if ( (nvars = getInt(fd)) )
   { term_t *v;
     int n;
 
@@ -682,14 +804,16 @@ loadWicFd(IOSTREAM *fd)
 
   s = getMagicString(fd, mbuf, sizeof(mbuf));
   if ( !s || !streq(s, saveMagic) )
-    return fatalError("Not a SWI-Prolog saved state");
+  { fatalError("Not a SWI-Prolog saved state");
+    fail;				/* NOTREACHED */
+  }
 
-  if ( (saved_version=getNum(fd)) < LOADVERSION )
+  if ( (saved_version=getInt(fd)) < LOADVERSION )
   { fatalError("Saved state has incompatible save version");
     fail;
   }
 
-  saved_wsize = getNum(fd);
+  saved_wsize = getInt(fd);
   if ( saved_wsize != sizeof(word)*8 )
   { fatalError("Saved state has incompatible (%d) word-length", saved_wsize);
     fail;
@@ -707,7 +831,7 @@ loadWicFd(IOSTREAM *fd)
 	popPathTranslation();
 	succeed;
       case 'W':
-	{ char *name = store_string(getString(fd) );
+	{ char *name = store_string(getString(fd, NULL) );
 
 	  loadWicFile(name);
 	  continue;
@@ -750,13 +874,18 @@ loadStatement(int c, IOSTREAM *fd, int skip ARG_LD)
       int     oln = source_line_no;
 
       source_file_name = (currentSource ? currentSource->name : NULL_ATOM);
-      source_line_no   = getNum(fd);
+      source_line_no   = getInt(fd);
       
       loadQlfTerm(goal, fd PASS_LD);
-      DEBUG(2, Sdprintf("%s:%d: Directive: ",
+      DEBUG(2,
+	    if ( source_file_name )
+	    { Sdprintf("%s:%d: Directive: ",
 			PL_atom_chars(source_file_name), source_line_no);
-	       pl_write(goal);
-	       Sdprintf("\n"));
+	    } else
+	    { Sdprintf("Directive: ");
+	    }
+	    pl_write(goal);
+	    Sdprintf("\n"));
       if ( !skip )
       { if ( !callProlog(MODULE_user, goal, PL_Q_NODEBUG, NULL) )
 	{ Sfprintf(Serror,
@@ -794,7 +923,7 @@ loadPredicateFlags(Definition def, IOSTREAM *fd, int skip ARG_LD)
   { if ( !skip && SYSTEM_MODE )
       set(def, SYSTEM|HIDE_CHILDS|LOCKED);
   } else
-  { int	flags = getNum(fd);
+  { int	flags = getInt(fd);
 
     if ( !skip )
     { unsigned long lflags = 0L;
@@ -816,9 +945,13 @@ loadPredicate(IOSTREAM *fd, int skip ARG_LD)
   Definition def;
   Clause clause;
   functor_t f = (functor_t) loadXR(fd);
+  SourceFile csf = NULL;
 
   proc = lookupProcedure(f, LD->modules.source);
-  DEBUG(3, Sdprintf("Loading %s ", procedureName(proc)));
+  DEBUG(2, Sdprintf("Loading %s%s",
+		    procedureName(proc),
+		    skip ? " (skip)" : ""));
+
   def = proc->definition;
   if ( !skip && currentSource )
   { if ( def->definition.clauses )
@@ -832,7 +965,7 @@ loadPredicate(IOSTREAM *fd, int skip ARG_LD)
   for(;;)
   { switch(Getc(fd) )
     { case 'X':
-      { unsigned long pattern = getNum(fd);
+      { unsigned long pattern = getLong(fd);
 
 	if ( (def->indexPattern & ~NEED_REINDEX) != pattern )
 	{ if ( def->references == 0 && !def->hash_info )
@@ -841,32 +974,35 @@ loadPredicate(IOSTREAM *fd, int skip ARG_LD)
 	    Sdprintf("Cannot change indexing of %s\n", predicateName(def));
 	}
 
-	DEBUG(3, Sdprintf("ok\n"));
+	DEBUG(2, Sdprintf("ok\n"));
 	succeed;
       }
       case 'C':
       { Code bp, ep;
-	int ncodes = getNum(fd);
+	int ncodes = getInt(fd);
 
-	DEBUG(3, Sdprintf("."));
+	DEBUG(2, Sdprintf("."));
 	clause = (Clause) allocHeap(sizeofClause(ncodes));
-	clause->code_size = (unsigned short) ncodes;
-	clause->line_no = (unsigned short) getNum(fd);
-	if ( load_state->saved_version < 32 )
-	  clause->source_no = (currentSource ? currentSource->index : 0);
-	else
+	clause->code_size = (unsigned int) ncodes;
+	clause->line_no = (unsigned short) getInt(fd);
+
 	{ SourceFile sf = (void *) loadXR(fd);
 	  int sno = (sf ? sf->index : 0);
 
 	  clause->source_no = sno;
+	  if ( sf && sf != csf )
+	  { addProcedureSourceFile(sf, proc);
+	    csf = sf;
+	  }
 	}
+
 	clearFlags(clause);
-	clause->prolog_vars = (unsigned short) getNum(fd);
-	clause->variables   = (unsigned short) getNum(fd);
+	clause->prolog_vars = (unsigned short) getInt(fd);
+	clause->variables   = (unsigned short) getInt(fd);
 #ifdef O_SHIFT_STACKS
 	clause->marks = clause->variables - clause->prolog_vars;
 #endif
-	if ( getNum(fd) == 0 )		/* 0: fact */
+	if ( getLong(fd) == 0 )		/* 0: fact */
 	  set(clause, UNIT_CLAUSE);
 	clause->procedure = proc;
 	GD->statistics.codes += clause->code_size;
@@ -875,11 +1011,12 @@ loadPredicate(IOSTREAM *fd, int skip ARG_LD)
 	ep = bp + clause->code_size;
 
 	while( bp < ep )
-	{ code op = getNum(fd);
+	{ code op = getInt(fd);
 	  int n = 0;
 	  int narg = codeTable[op].arguments;
 	  
-	  DEBUG(3, Sdprintf("\t%s (%d args)\n", codeTable[op].name, narg));
+	  DEBUG(3, Sdprintf("\t%s (%d args) from %ld\n",
+			    codeTable[op].name, narg, Stell(fd)));
 	  *bp++ = encode(op);
 	  switch(codeTable[op].argtype)
 	  { case CA1_PROC:
@@ -901,9 +1038,17 @@ loadPredicate(IOSTREAM *fd, int skip ARG_LD)
 	      n++;
 	      break;
 	    case CA1_INTEGER:
-	      *bp++ = getNum(fd);
+	      *bp++ = getLong(fd);
 	      n++;
 	      break;
+	    case CA1_INT64:
+	    { int64_t val = getInt64(fd);
+	      Word p = (Word)&val;
+	      
+	      cpInt64Data(bp, p);
+	      n += WORDS_PER_INT64;
+	      break;
+	    }
 	    case CA1_FLOAT:
 	    { union
 	      { word w[WORDS_PER_DOUBLE];
@@ -916,7 +1061,7 @@ loadPredicate(IOSTREAM *fd, int skip ARG_LD)
 	      break;
 	    }
 	    case CA1_STRING:		/* <n> chars */
-	    { int l = getNum(fd);
+	    { int l = getInt(fd);
 	      int lw = (l+sizeof(word))/sizeof(word);
 	      int pad = (lw*sizeof(word) - l);
 	      char *s = (char *)&bp[1];
@@ -930,9 +1075,33 @@ loadPredicate(IOSTREAM *fd, int skip ARG_LD)
 	      n++;
 	      break;
 	    }
+	    case CA1_MPZ:
+#ifdef O_GMP
+	    DEBUG(3, Sdprintf("Loading MPZ from %ld\n", Stell(fd)));
+	    { int mpsize = getInt(fd);
+	      int l      = abs(mpsize)*sizeof(mp_limb_t);
+	      int wsz	 = (l+sizeof(word)-1)/sizeof(word);
+	      word m     = mkIndHdr(wsz+1, TAG_INTEGER);
+	      char *s;
+
+	      *bp++     = m;
+	      *bp++     = mpsize;
+	      s         = (char*)bp;
+	      bp[wsz-1] = 0L;
+	      bp       += wsz;
+
+	      while(--l >= 0)
+		*s++ = Getc(fd);
+	      DEBUG(3, Sdprintf("Loaded MPZ to %ld\n", Stell(fd)));
+	      n++;
+	      break;
+	    }
+#else
+  	      fatalError("No support for MPZ numbers");
+#endif
 	  }
 	  for( ; n < narg; n++ )
-	    *bp++ = getNum(fd);
+	    *bp++ = getInt(fd);
 	}
 
 	if ( skip )
@@ -998,8 +1167,8 @@ the module where it is a multifile one.
 
 static bool
 qlfLoadSource(IOSTREAM *fd)
-{ char *str = getString(fd);
-  long time = getNum(fd);
+{ char *str = getString(fd, NULL);
+  long time = getLong(fd);
   int issys = (Qgetc(fd) == 's') ? TRUE : FALSE;
   atom_t fname;
 
@@ -1014,7 +1183,7 @@ qlfLoadSource(IOSTREAM *fd)
   if ( GD->bootsession )		/* (**) */
     currentSource->count++;
   else
-  startConsult(currentSource);
+    startConsult(currentSource);
   PL_unregister_atom(fname);		/* locked with sourceFile */
 
   succeed;
@@ -1168,26 +1337,63 @@ of a predicate together.
 static Table savedXRTable;		/* saved XR entries */
 static long  savedXRTableId;		/* next id */
 
+#define STR_NOLEN (~(unsigned)0)
+
 static void
-putString(char *s, IOSTREAM *fd)
-{ while(*s)
+putString(const char *s, unsigned len, IOSTREAM *fd)
+{ const char *e;
+
+  if ( len == STR_NOLEN )
+    len = strlen(s);
+  e = &s[len];
+
+  putNum(len, fd);
+  while(s<e)
   { Sputc(*s, fd);
     s++;
   }
+}
 
-  Sputc(EOS, fd);
+
+static void
+putStringW(const pl_wchar_t *s, unsigned len, IOSTREAM *fd)
+{ const pl_wchar_t *e;
+  IOENC oenc = fd->encoding;
+
+  if ( len == STR_NOLEN )
+    len = wcslen(s);
+  e = &s[len];
+
+  putNum(len, fd);
+  fd->encoding = ENC_UTF8;
+  while(s<e)
+  { Sputcode(*s, fd);
+    s++;
+  }
+  fd->encoding = oenc;
 }
 
 
 static void
 putAtom(atom_t w, IOSTREAM *fd)
 { Atom a = atomValue(w);
-  const char *s = a->name;
-  const char *e = s+a->length;
+  static PL_blob_t *text_blob;
 
-  putNum(a->length, fd);
-  for( ; s<e; s++ )
-    Sputc(*s, fd);
+  if ( !text_blob )
+    text_blob = PL_find_blob_type("text");
+
+  if ( a->type != text_blob )
+  { Sputc(XR_BLOB, fd);
+    saveXRBlobType(a->type, fd);
+    if ( a->type->save )
+    { (*a->type->save)(a->atom, fd);
+    } else
+    { putString(a->name, a->length, fd);
+    }
+  } else
+  { Sputc(XR_ATOM, fd);
+    putString(a->name, a->length, fd);
+  }
 }
 
 
@@ -1205,30 +1411,30 @@ First byte:  bits 8&7  bits 1-6 (low order)
 #define PLMINLONG   ((long)(-1L<<(LONGBITSIZE-1)))
 
 static void
-putNum(long n, IOSTREAM *fd)
+putNum(int64_t n, IOSTREAM *fd)
 { int m;
-  long absn = (n >= 0 ? n : -n);
+  int64_t absn = (n >= 0 ? n : -n);
 
   DEBUG(8, Sdprintf("0x%x at %ld\n", (unsigned long)n, Stell(fd)));
 
-  if ( n != PLMINLONG )
+  if ( n != PLMININT )
   { if ( absn < (1L << 5) )
-    { Sputc((n & 0x3f), fd);
+    { Sputc((int)(n & 0x3f), fd);
       return;
     } else if ( absn < (1L << 13) )
-    { Sputc((((n >> 8) & 0x3f) | (1 << 6)), fd);
-      Sputc((n & 0xff), fd);
+    { Sputc((int)(((n >> 8) & 0x3f) | (1 << 6)), fd);
+      Sputc((int)(n & 0xff), fd);
       return;
     } else if ( absn < (1L << 21) )
-    { Sputc((((n >> 16) & 0x3f) | (2 << 6)), fd);
-      Sputc(((n >> 8) & 0xff), fd);
-      Sputc((n & 0xff), fd);
+    { Sputc((int)(((n >> 16) & 0x3f) | (2 << 6)), fd);
+      Sputc((int)((n >> 8) & 0xff), fd);
+      Sputc((int)(n & 0xff), fd);
       return;
     }
   }
 
   for(m = sizeof(n); ; m--)
-  { int b = (absn >> (((m-1)*8)-1)) & 0x1ff;
+  { int b = (int)(absn >> (((m-1)*8)-1)) & 0x1ff;
 
     if ( b == 0 )
       continue;
@@ -1238,7 +1444,7 @@ putNum(long n, IOSTREAM *fd)
   Sputc(m | (3 << 6), fd);
 
   for( ; m > 0; m--)
-  { int b = (n >> ((m-1)*8)) & 0xff;
+  { int b = (int)(n >> ((m-1)*8)) & 0xff;
     
     Sputc(b, fd);
   }
@@ -1330,8 +1536,17 @@ saveXR__LD(word xr, IOSTREAM *fd ARG_LD)
     return;
 #if O_STRING
   } else if ( isString(xr) )
-  { Sputc(XR_STRING, fd);
-    putString(valString(xr), fd);
+  { char *s;
+    pl_wchar_t *w;
+    unsigned len;
+
+    if ( (s = getCharsString(xr, &len)) )
+    { Sputc(XR_STRING, fd);
+      putString(s, len, fd);
+    } else if ( (w=getCharsWString(xr, &len)) )
+    { Sputc(XR_STRING_UTF8, fd);
+      putStringW(w, len, fd);
+    }
     return;
 #endif /* O_STRING */
   }
@@ -1341,7 +1556,6 @@ saveXR__LD(word xr, IOSTREAM *fd ARG_LD)
 
   if ( isAtom(xr) )
   { DEBUG(3, Sdprintf("XR(%d) = '%s'\n", savedXRTableId, stringAtom(xr)));
-    Sputc(XR_ATOM, fd);
     putAtom(xr, fd);
     return;
   }
@@ -1351,6 +1565,16 @@ saveXR__LD(word xr, IOSTREAM *fd ARG_LD)
 #define saveXR(xr, s) saveXR__LD(xr, s PASS_LD)
 
 
+static void
+saveXRBlobType(PL_blob_t *type, IOSTREAM *fd)
+{ if ( savedXRPointer(type, fd) )
+    return;
+
+  Sputc(XR_BLOB_TYPE, fd);
+  putString(type->name, STR_NOLEN, fd);
+}
+
+
 static void
 saveXRModule(Module m, IOSTREAM *fd ARG_LD)
 { if ( savedXRPointer(m, fd) )
@@ -1443,6 +1667,7 @@ static void
 saveQlfTerm(term_t t, IOSTREAM *fd ARG_LD)
 { int nvars;
   fid_t cid;
+  nv_options options;
 
   cid = PL_open_foreign_frame();
 
@@ -1450,7 +1675,10 @@ saveQlfTerm(term_t t, IOSTREAM *fd ARG_LD)
 	Sdprintf("Saving ");
 	PL_write_term(Serror, t, 1200, 0);
 	Sdprintf(" from %d ... ", Stell(fd)));
-  nvars = numberVars(t, FUNCTOR_var1, AV_SKIP, 0 PASS_LD);
+  options.functor = FUNCTOR_var1;
+  options.on_attvar = AV_SKIP;
+  options.singletons = FALSE;		/* TBD: TRUE may be better! */
+  nvars = numberVars(t, &options, 0 PASS_LD);
   putNum(nvars, fd);
   do_save_qlf_term(valTermRef(t), fd PASS_LD);	/* TBD */
   DEBUG(3, Sdprintf("to %d\n", Stell(fd)));
@@ -1492,6 +1720,9 @@ saveWicClause(Clause clause, IOSTREAM *fd)
     int n = 0;
 
     putNum(op, fd);
+    DEBUG(3, Sdprintf("\t%s (%d args) at %ld\n", 
+		      codeTable[op].name, codeTable[op].arguments,
+		      Stell(fd)));
     switch(codeTable[op].argtype)
     { case CA1_PROC:
       { Procedure p = (Procedure) *bp++;
@@ -1522,6 +1753,15 @@ saveWicClause(Clause clause, IOSTREAM *fd)
 	n++;
 	break;
       }
+      case CA1_INT64:
+      { int64_t val;
+	Word p = (Word)&val;
+
+	cpInt64Data(p, bp);
+	n += WORDS_PER_INT64;
+	putNum(val, fd);
+	break;
+      }
       case CA1_FLOAT:
       { union
 	{ word w[WORDS_PER_DOUBLE];
@@ -1546,6 +1786,24 @@ saveWicClause(Clause clause, IOSTREAM *fd)
 	n++;
 	break;
       }
+#ifdef O_GMP
+      case CA1_MPZ:
+      { word m = *bp++;
+	int wn = wsizeofInd(m);
+	int mpsize = (int)*bp;
+	int l = abs(mpsize)*sizeof(mp_limb_t);
+	char *s = (char*)&bp[1];
+	bp += wn;
+
+	DEBUG(3, Sdprintf("Saving MPZ from %ld\n", Stell(fd)));
+	putNum(mpsize, fd);
+	while(--l >= 0)
+	  Sputc(*s++&0xff, fd);
+	DEBUG(3, Sdprintf("Saved MPZ to %ld\n", Stell(fd)));
+	n++;
+	break;
+      }
+#endif
     }
     for( ; n < codeTable[op].arguments; n++ )
       putNum(*bp++, fd);
@@ -1608,17 +1866,27 @@ openProcedureWic(Procedure proc, IOSTREAM *fd, atom_t sclass ARG_LD)
 }
 
 
+static bool
+putMagic(const char *s, IOSTREAM *fd)
+{ for(; *s; s++)
+    Sputc(*s, fd);
+  Sputc(EOS, fd);
+
+  succeed;
+}
+
+
 static bool
 writeWicHeader(IOSTREAM *fd)
 { wicFd = fd;
 
-  putString(saveMagic, fd);
+  putMagic(saveMagic, fd);
   putNum(VERSION, fd);
   putNum(sizeof(word)*8, fd);	/* bits-per-word */
   if ( systemDefaults.home )
-    putString(systemDefaults.home,  fd);
+    putString(systemDefaults.home, STR_NOLEN, fd);
   else
-    putString("<no home>",  fd);
+    putString("<no home>",  STR_NOLEN, fd);
 
   currentProc    = (Procedure) NULL;
   currentSource  = (SourceFile) NULL;
@@ -1759,7 +2027,7 @@ qlfSourceInfo(IOSTREAM *s, long offset, term_t list ARG_LD)
 
   if ( Sseek(s, offset, SIO_SEEK_SET) != offset )
     return warning("%s: seek failed: %s", wicFile, OsError());
-  if ( Getc(s) != 'F' || !(str=getString(s)) )
+  if ( Getc(s) != 'F' || !(str=getString(s, NULL)) )
     return warning("QLF format error");
   
   return PL_unify_list(list, head, list) &&
@@ -1797,19 +2065,19 @@ qlfInfo(const char *file,
   }
   TRY(PL_unify_integer(version, lversion));
 
-  saved_wsize = getNum(s);		/* word-size of file */
+  saved_wsize = getInt(s);		/* word-size of file */
   TRY(PL_unify_integer(wsize, saved_wsize));
 
   pushPathTranslation(s, file, 0);
 
   if ( Sseek(s, -4, SIO_SEEK_END) < 0 )	/* 4 bytes of putLong() */
     return warning("qlf_info/4: seek failed: %s", OsError());
-  nqlf = getLong(s);
+  nqlf = getWord(s);
   DEBUG(1, Sdprintf("Found %d sources at", nqlf));
   qlfstart = (long *)allocHeap(sizeof(long) * nqlf);
   Sseek(s, -4 * (nqlf+1), SIO_SEEK_END);
   for(i=0; i<nqlf; i++)
-  { qlfstart[i] = getLong(s);
+  { qlfstart[i] = getWord(s);
     DEBUG(1, Sdprintf(" %d", qlfstart[i]));
   }
   DEBUG(1, Sdprintf("\n"));
@@ -1865,11 +2133,11 @@ qlfOpen(atom_t name)
 
   mkWicFile = wicFile;
 
-  putString(qlfMagic, wicFd);
+  putMagic(qlfMagic, wicFd);
   putNum(VERSION, wicFd);
   putNum(sizeof(word)*8, wicFd);
 
-  putString(absname, wicFd);
+  putString(absname, STR_NOLEN, wicFd);
 
   currentProc    = (Procedure) NULL;
   currentSource  = (SourceFile) NULL;
@@ -1909,7 +2177,7 @@ qlfVersion(IOSTREAM *s)
     return warning("%s: not a SWI-Prolog .qlf file", wicFile);
   }
 
-  return getNum(s);
+  return getInt(s);
 }
 
 
@@ -1923,7 +2191,7 @@ pushPathTranslation(IOSTREAM *fd, const char *absloadname, int flags)
   new->previous = load_state;
   load_state = new;
   
-  abssavename = getString(fd);
+  abssavename = getString(fd, NULL);
   if ( absloadname && !streq(absloadname, abssavename) )
   { char load[MAXPATHLEN];
     char save[MAXPATHLEN];
@@ -2001,7 +2269,7 @@ qlfLoad(char *file, Module *module ARG_LD)
 	      wicFile, lversion, VERSION);
     fail;
   }
-  saved_wsize = getNum(fd);
+  saved_wsize = getInt(fd);
   if ( saved_wsize != sizeof(word)*8 )
   { warning("QLF file %s has incompatible (%d) word-length",
 	    file, saved_wsize);
@@ -2026,9 +2294,11 @@ qlfLoad(char *file, Module *module ARG_LD)
 
 static bool
 qlfSaveSource(SourceFile f, IOSTREAM *fd ARG_LD)
-{ sourceMark(fd PASS_LD);
+{ Atom a = atomValue(f->name);
+
+  sourceMark(fd PASS_LD);
   Sputc('F', fd);
-  putString(stringAtom(f->name), fd);
+  putString(a->name, a->length, fd);
   putNum(f->time, fd);
   Sputc(f->system ? 's' : 'u', fd);
 
@@ -2133,8 +2403,8 @@ pl_qlf_start_file(term_t name)
   { GET_LD
     atom_t a;
 
-    if ( !PL_get_atom(name, &a) )
-      return warning("qlf_start_file/1: argument must be an atom");
+    if ( !PL_get_atom_ex(name, &a) )
+      fail;
   
     return qlfStartFile(lookupSourceFile(a), wicFd PASS_LD);
   }
@@ -2155,13 +2425,12 @@ pl_qlf_end_part()
 
 word
 pl_qlf_open(term_t file)
-{ GET_LD
-  atom_t a;
+{ atom_t a;
 
-  if ( PL_get_atom(file, &a) )
+  if ( PL_get_atom_ex(file, &a) )
     return qlfOpen(a);
 
-  return warning("qlf_open/1: instantiation fault");
+  fail;
 }
 
 
@@ -2276,9 +2545,8 @@ pl_qlf_assert_clause(term_t ref, term_t saveclass)
 
     if ( !get_clause_ptr_ex(ref, &clause) )
       fail;
-    if ( !PL_get_atom(saveclass, &sclass) )
-      return PL_error("qlf_assert_clause", 2, NULL,
-		      ERR_DOMAIN, ATOM_save_class, 2);
+    if ( !PL_get_atom_ex(saveclass, &sclass) )
+      fail;
 
     openProcedureWic(clause->procedure, s, sclass PASS_LD);
     saveWicClause(clause, s);
@@ -2439,7 +2707,10 @@ compileFileList(IOSTREAM *fd, int argc, char **argv)
 void
 qlfCleanup()
 { if ( mkWicFile )
-  { warning("Removing incomplete Quick Load File %s", mkWicFile);
+  { printMessage(ATOM_warning,
+		 PL_FUNCTOR_CHARS, "qlf", 1,
+		   PL_FUNCTOR_CHARS, "removed_after_error", 1,
+		     PL_CHARS, mkWicFile);
     RemoveFile(mkWicFile);
     mkWicFile = NULL;
   }
@@ -2451,6 +2722,28 @@ qlfCleanup()
   }
 }
 
+		 /*******************************
+		 *	 PUBLIC FUNCTIONS	*
+		 *******************************/
+
+void
+wicPutNum(int64_t n, IOSTREAM *fd)
+{ putNum(n, fd);
+}
+
+
+int64_t
+wicGetNum(IOSTREAM *fd)
+{ return getInt64(fd);
+}
+
+
+void
+wicPutStringW(const pl_wchar_t *w, size_t len, IOSTREAM *fd)
+{ putStringW(w, len, fd);
+} 
+
+
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
