@@ -1,4 +1,4 @@
-/*  $Id: pl-write.c,v 1.63.2.2 2005/02/21 08:20:04 jan Exp $
+/*  $Id: pl-write.c,v 1.92 2006/08/28 13:32:19 jan Exp $
 
     Part of SWI-Prolog
 
@@ -26,6 +26,19 @@
 #include "pl-incl.h"
 #include "pl-ctype.h"
 #include <stdio.h>			/* sprintf() */
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+#ifdef HAVE_FLOAT_H
+#include <float.h>
+#endif
+#ifdef HAVE_IEEEFP_H
+#include <ieeefp.h>
+#endif
+
+#ifdef fpclassify
+#define HAVE_FPCLASSIFY 1
+#endif
 
 typedef struct visited
 { Word address;				/* we have done this address */
@@ -94,7 +107,7 @@ varName(term_t t, char *name)
 #define AT_SPECIAL	5
 
 static int
-atomType(atom_t a)
+atomType(atom_t a, IOSTREAM *fd)
 { Atom atom = atomValue(a);
   char *s = atom->name;
   int len = atom->length;
@@ -103,7 +116,7 @@ atomType(atom_t a)
     return AT_QUOTE;
 
   if ( isLower(*s) )
-  { for(++s; --len > 0 && isAlpha(*s); s++)
+  { for(++s; --len > 0 && isAlpha(*s) && Scanrepresent(*s, fd)==0; s++)
       ;
     return len == 0 ? AT_LOWER : AT_QUOTE;
   }
@@ -115,7 +128,7 @@ atomType(atom_t a)
   { if ( len >= 2 && s[0] == '/' && s[1] == '*' )
       return AT_QUOTE;
 
-    for(++s; --len > 0 && isSymbol(*s); s++)
+    for(++s; --len > 0 && isSymbol(*s) && Scanrepresent(*s, fd)==0; s++)
       ;
 
     return len == 0 ? AT_SYMBOL : AT_QUOTE;
@@ -125,11 +138,6 @@ atomType(atom_t a)
   if ( len == 1 && *s != '%' )
   { if ( isSolo(*s) )
       return AT_SOLO;
-    switch( *s )
-    { /*case ',':*/
-      case '|':
-	return AT_SOLO;
-    }
   }
 
   if ( a == ATOM_nil || a == ATOM_curl )
@@ -145,21 +153,34 @@ atomType(atom_t a)
 
 static bool
 Putc(int c, IOSTREAM *s)
-{ return Sputc(c, s) == EOF ? FALSE : TRUE;
+{ return Sputcode(c, s) == EOF ? FALSE : TRUE;
 }
 
 
 static bool
 PutString(const char *str, IOSTREAM *s)
-{ return Sfputs(str, s) == EOF ? FALSE : TRUE;
+{ const unsigned char *q = (const unsigned char *)str;
+
+  for( ; *q != EOS; q++ )
+  { if ( Sputcode(*q, s) == EOF )
+      return FALSE;
+  }
+
+  return TRUE;
 }
 
 
 static bool
 PutStringN(const char *str, unsigned int length, IOSTREAM *s)
-{ int len = (int)length;
+{ unsigned int i;
+  const unsigned char *q = (const unsigned char *)str;
 
-  return Sfwrite(str, 1, len, s) == len ? TRUE : FALSE;
+  for(i=0; i<length; i++, q++)
+  { if ( Sputcode(*q, s) == EOF )
+      return FALSE;
+  } 
+
+  return TRUE;
 }
 
 
@@ -174,8 +195,8 @@ PutOpenToken(int c, IOSTREAM *s)
   { s->lastc = EOF;
     return TRUE;
   } else if ( s->lastc != EOF &&
-	      ((isAlpha(s->lastc) && isAlpha(c)) ||
-	       (isSymbol(s->lastc) && isSymbol(c)) ||
+	      ((isAlphaW(s->lastc) && isAlphaW(c)) ||
+	       (isSymbolW(s->lastc) && isSymbolW(c)) ||
 	       (s->lastc != '(' && !isBlank(s->lastc) && c == '(')) )
   { return Putc(' ', s);
   }
@@ -222,6 +243,68 @@ PutCloseBrace(IOSTREAM *s)
 }
 
 
+static bool
+putQuoted(int c, int quote, int flags, IOSTREAM *stream)
+{ if ( (flags & PL_WRT_CHARESCAPES) )
+  { if ( !(c < 0xff && isControl(c)) && c != quote && c != '\\' )
+    { TRY(Putc(c, stream));
+    } else
+    { char esc[8];
+
+      esc[1] = EOS;
+
+      if ( c == quote )
+      { esc[0] = c;
+      } else
+      { switch(c)
+	{ case 7:
+	    esc[0] = 'a';
+	    break;
+	  case '\b':
+	    esc[0] = 'b';
+	    break;
+	  case '\t':
+	    esc[0] = 't';
+	    break;
+	  case '\n':
+	    esc[0] = 'n';
+	    break;
+	  case 11:
+	    esc[0] = 'v';
+	    break;
+	  case '\r':
+	    esc[0] = 'r';
+	    break;
+	  case '\f':
+	    esc[0] = 'f';
+	    break;
+	  case '\\':
+	    esc[0] = '\\';
+	    break;
+	  default:
+	    if ( c <= 0xff )
+	      Ssprintf(esc, "%03o\\", c);
+	    else
+	      assert(0);			/* to be done */
+	}
+      }
+      if ( !Putc('\\', stream) ||
+	   !PutString(esc, stream) )
+	fail;
+    }
+  } else
+  { if ( c == quote )
+    { TRY(Putc(c, stream) && Putc(c, stream));
+    } else
+    { return Putc(c, stream);
+    }
+  }
+
+  return TRUE;
+}
+
+
+
 static bool
 writeQuoted(IOSTREAM *stream, const char *text, int len, int quote,
 	    write_options *options)
@@ -230,59 +313,7 @@ writeQuoted(IOSTREAM *stream, const char *text, int len, int quote,
   TRY(Putc(quote, stream));
 
   while(len-- > 0)
-  { int c = *s++;
-
-    if ( true(options, PL_WRT_CHARESCAPES) )
-    { if ( !isControl(c) && c != quote && c != '\\' )
-      { TRY(Putc(c, stream));
-      } else
-      { char esc[8];
-
-	esc[1] = EOS;
-
-	if ( c == quote )
-	{ esc[0] = c;
-	} else
-	{ switch(c)
-	  { case 7:
-	      esc[0] = 'a';
-	      break;
-	    case '\b':
-	      esc[0] = 'b';
-	      break;
-	    case '\t':
-	      esc[0] = 't';
-	      break;
-	    case '\n':
-	      esc[0] = 'n';
-	      break;
-	    case 11:
-	      esc[0] = 'v';
-	      break;
-	    case '\r':
-	      esc[0] = 'r';
-	      break;
-	    case '\f':
-	      esc[0] = 'f';
-	      break;
-	    case '\\':
-	      esc[0] = '\\';
-	      break;
-	    default:
-	      Ssprintf(esc, "%03o\\", c);
-	  }
-	}
-	if ( !Putc('\\', stream) ||
-	     !PutString(esc, stream) )
-	  fail;
-      }
-    } else
-    { if ( c == quote )
-      { TRY(Putc(c, stream) && Putc(c, stream));
-      } else
-      { TRY(Putc(c, stream));
-      }
-    }
+  { TRY(putQuoted(*s++, quote, options->flags, stream));
   }
 
   return Putc(quote, stream);
@@ -371,7 +402,7 @@ writeAtom(atom_t a, write_options *options)
     return writeBlob(a, options);
 
   if ( true(options, PL_WRT_QUOTED) )
-  { switch( atomType(a) )
+  { switch( atomType(a, options->out) )
     { case AT_LOWER:
       case AT_SYMBOL:
       case AT_SOLO:
@@ -390,18 +421,210 @@ writeAtom(atom_t a, write_options *options)
 }
 
 
-#if !defined(HAVE_ISNAN) && defined(NaN)
-#define isnan(f)  ((f) == NaN)
-#define HAVE_ISNAN
+int
+writeAtomToStream(IOSTREAM *s, atom_t atom)
+{ write_options options;
+
+  memset(&options, 0, sizeof(options));
+  options.out = s;
+  options.module = MODULE_user;
+
+  return writeAtom(atom, &options);
+}
+
+
+int
+writeUCSAtom(IOSTREAM *fd, atom_t atom, int flags)
+{ Atom a = atomValue(atom);
+  pl_wchar_t *s = (pl_wchar_t*)a->name;
+  int len = a->length/sizeof(pl_wchar_t);
+  pl_wchar_t *e = &s[len];
+
+  if ( flags & PL_WRT_QUOTED )
+  { pl_wchar_t quote = L'\'';
+    int rc;
+
+    if ( isLowerW(*s) )
+    { pl_wchar_t *q;
+
+      for(q=s; q<e; q++)
+      { if ( !isAlphaW(*q) || Scanrepresent(*q, fd) < 0 )
+	  break;
+      }
+      if ( q == e )
+        goto unquoted;
+    }
+
+    TRY(Putc(quote, fd));
+
+    while(s < e)
+    { TRY(putQuoted(*s++, quote, flags, fd));
+    }
+
+    rc = Putc(quote, fd);
+
+    return rc;
+  }
+
+unquoted:
+  if ( s < e && !PutOpenToken(s[0], fd) )
+    return FALSE;
+  for( ; s<e; s++)
+  { if ( !Putc(*s, fd) )
+      return FALSE;
+  }
+
+  return TRUE;
+}
+
+
+#if O_STRING
+
+static inline int
+get_chr_from_text(const PL_chars_t *t, int index)
+{ switch(t->encoding)
+  { case ENC_ISO_LATIN_1:
+      return t->text.t[index]&0xff;
+    case ENC_WCHAR:
+      return t->text.w[index];
+    default:
+      assert(0);
+    return 0;
+  }
+}
+
+
+static int
+writeString(term_t t, write_options *options)
+{ PL_chars_t txt;
+
+  PL_get_text(t, &txt, CVT_STRING);
+  
+  if ( true(options, PL_WRT_QUOTED) )
+  { int quote;
+    unsigned int i;
+
+    if ( true(options, PL_WRT_BACKQUOTED_STRING) )
+      quote = '`';
+    else
+      quote = '"';
+      
+    TRY(Putc(quote, options->out));
+
+    for(i=0; i<txt.length; i++)
+    { int chr = get_chr_from_text(&txt, i);
+
+      TRY(putQuoted(chr, quote, options->flags, options->out));
+    }
+
+    return Putc(quote, options->out);
+  } else
+  { unsigned int i;
+
+    for(i=0; i<txt.length; i++)
+    { int chr = get_chr_from_text(&txt, i);
+
+      TRY(Putc(chr, options->out));
+    }
+  }
+
+  succeed;
+}
+
+#endif /*O_STRING*/
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Formatting a float. This is very  complicated   as  we must write floats
+such that it can be read as a float. This means using the conventions of
+the C locale and if the float happens to be integer as <int>.0.
+
+Switching the locale is no option as  locale handling is not thread-safe
+and may have unwanted  consequences  for   embedding.  There  is  a long
+discussion on the very same topic on  the Python mailinglist. Many hacks
+are proposed, none is very satisfactory.   Richard  O'Keefe suggested to
+use ecvt(), fcvt() and gcvt(). These  are   not  thread-safe.  The GNU C
+library provides *_r() variations that  can   do  the  trick. An earlier
+patch used localeconv() to find the  decimal   point,  but  this is both
+complicated and not thread-safe.
+
+Finally, with help of Richard we decided  to replace the first character
+that is not a digit nor [eE], as this must be the decimal point. 
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+char *
+format_float(double f, char *buf, const char *format)
+{ char *q;
+
+  sprintf(buf, format, f);
+
+  q = buf;
+  if ( *q == '-' )			/* skip -?[0-9]* */
+    q++;
+  while(*q && (isDigit(*q) || *q <= ' '))
+    q++;
+
+  switch( *q )
+  { case '\0':
+      *q++ = '.';
+      *q++ = '0';
+      *q = EOS;
+      break;
+    case 'e':
+    case 'E':
+      break;
+    default:
+      *q = '.';
+  }
+
+  return buf;
+}
+
+
+static bool
+WriteNumber(Number n, write_options *options)
+{ switch(n->type)
+  { case V_INTEGER:
+    { char buf[32];
+
+      sprintf(buf, INT64_FORMAT, n->value.i);
+      return PutToken(buf, options->out);
+    }
+#ifdef O_GMP
+    case V_MPZ:
+    { char tmp[1024];
+      char *buf;
+      size_t sz = mpz_sizeinbase(n->value.mpz, 10) + 2;
+      bool rc;
+
+      if ( sz <= sizeof(tmp) )
+	buf = tmp;
+      else
+	buf = PL_malloc(sz);
+
+      mpz_get_str(buf, 10, n->value.mpz);
+      rc = PutToken(buf, options->out);
+      if ( buf != tmp )
+	PL_free(buf);
+	
+      return rc;
+    }
+    case V_MPQ:				/* should not get here */
 #endif
+    case V_REAL:
+      assert(0);
+  }
+
+  fail;
+}
+
+
 
 static bool
 writePrimitive(term_t t, write_options *options)
 { double f;
-  char *s;
   atom_t a;
-  unsigned int n;
-  char buf[16];
+  char buf[32];
   IOSTREAM *out = options->out;
 
 #if O_ATTVAR
@@ -416,19 +639,52 @@ writePrimitive(term_t t, write_options *options)
     return writeAtom(a, options);
 
   if ( PL_is_integer(t) )		/* beware of automatic conversion */
-  { long i;
-    char buf[32];
+  { number n;
 
-    PL_get_long(t, &i);
-    sprintf(buf, "%ld", i);
-    return PutToken(buf, out);
+    PL_get_number(t, &n);
+
+    return WriteNumber(&n, options);
   }
 
   if ( PL_get_float(t, &f) )
   { char *s = NULL;
 
-#ifdef HUGE_VAL
-    if ( f == HUGE_VAL )
+#ifdef HAVE_FPCLASSIFY
+    switch(fpclassify(f))
+    { case FP_NAN:
+	s = (true(options, PL_WRT_QUOTED) ? "'$NaN'" : "NaN");
+        break;
+      case FP_INFINITE:
+	s = (true(options, PL_WRT_QUOTED) ? "'$Infinity'" : "Infinity");
+        break;
+    }
+#else
+#ifdef HAVE_FPCLASS
+    switch(fpclass(f))
+    { case FP_SNAN:
+      case FP_QNAN:
+	s = (true(options, PL_WRT_QUOTED) ? "'$NaN'" : "NaN");
+        break;
+      case FP_NINF:
+      case FP_PINF:
+	s = (true(options, PL_WRT_QUOTED) ? "'$Infinity'" : "Infinity");
+        break;
+    }
+#else
+#ifdef HAVE__FPCLASS
+    switch(_fpclass(f))
+    { case _FPCLASS_SNAN:
+      case _FPCLASS_QNAN:
+	s = (true(options, PL_WRT_QUOTED) ? "'$NaN'" : "NaN");
+        break;
+      case _FPCLASS_NINF:
+      case _FPCLASS_PINF:
+	s = (true(options, PL_WRT_QUOTED) ? "'$Infinity'" : "Infinity");
+        break;
+    }
+#else
+#ifdef HAVE_ISINF
+    if ( isinf(f) )
     { s = (true(options, PL_WRT_QUOTED) ? "'$Infinity'" : "Infinity");
     } else
 #endif
@@ -437,49 +693,24 @@ writePrimitive(term_t t, write_options *options)
     { s = (true(options, PL_WRT_QUOTED) ? "'$NaN'" : "NaN");
     }
 #endif
+#endif /*HAVE__FPCLASS*/
+#endif /*HAVE_FPCLASS*/
+#endif /*HAVE_FPCLASSIFY*/
 
-    if ( s ) {
-      return PutToken(s, out);
+    if ( s )
+    { return PutToken(s, out);
     } else
     { char buf[100];
-      char *q;
 
-      sprintf(buf, LD->float_format, f);
-					/* make sure to write float */
-					/* such that it reads as a float */
-      if ( true(options, PL_WRT_QUOTED) )
-      { q = buf;
-	if ( *q == '-' )
-	  q++;
-	for(; *q; q++)
-	{ if ( !isDigit(*q) )
-	    break;
-	}
-        if ( !*q )
-	{ *q++ = '.';
-	  *q++ = '0';
-	  *q = EOS;
-	}
-      }
+      format_float(f, buf, LD->float_format);
 
       return PutToken(buf, out);
     }
   }
 
 #if O_STRING
-  if ( PL_get_string(t, &s, &n) )
-  { if ( true(options, PL_WRT_QUOTED) )
-    { int q;
-
-      if ( true(options, PL_WRT_BACKQUOTED_STRING) )
-	q = '`';
-      else
-	q = '"';
-      
-      return writeQuoted(out, s, n, q, options);
-    } else
-      return PutStringN(s, n, out);
-  }
+  if ( PL_is_string(t) )
+    return writeString(t, options);
 #endif /* O_STRING */
 
   assert(0);
@@ -492,7 +723,7 @@ pl_nl1(term_t stream)
 { IOSTREAM *s;
 
   if ( getOutputStream(stream, &s) )
-  { Sputc('\n', s);
+  { Sputcode('\n', s);
     return streamStatus(s);
   }
 
@@ -594,7 +825,37 @@ writeTerm2(term_t t, int prec, write_options *options)
   if ( !PL_get_name_arity(t, &functor, &arity) )
   { return writePrimitive(t, options);
   } else
-  { if ( false(options, PL_WRT_IGNOREOPS) )
+  { if ( functor == ATOM_isovar &&			/* $VAR/1 */
+	 true(options, PL_WRT_NUMBERVARS) )
+    { int n;
+      atom_t a;
+      term_t arg = PL_new_term_ref();
+
+      PL_get_arg(1, t, arg);
+      if ( PL_get_integer(arg, &n) && n >= 0 )
+      { int i = n % 26;
+	int j = n / 26;
+	char buf[16];
+
+	if ( j == 0 )
+	{ buf[0] = i+'A';
+	  buf[1] = EOS;
+	} else
+	{ sprintf(buf, "%c%d", i+'A', j);
+	}
+
+	return PutToken(buf, out);
+      }
+      if ( PL_get_atom(arg, &a) )
+      { write_options o2 = *options;
+	clear(&o2, PL_WRT_QUOTED);
+	
+	return writeAtom(a, &o2);
+      }	    
+    }
+
+
+    if ( false(options, PL_WRT_IGNOREOPS) )
     { term_t arg = PL_new_term_ref();
 
       if ( arity == 1 )
@@ -617,33 +878,6 @@ writeTerm2(term_t t, int prec, write_options *options)
 	  succeed;
 	}
   
-	if ( functor == ATOM_isovar &&			/* $VAR/1 */
-	     true(options, PL_WRT_NUMBERVARS) )
-	{ int n;
-	  atom_t a;
-
-	  PL_get_arg(1, t, arg);
-	  if ( PL_get_integer(arg, &n) && n >= 0 )
-	  { int i = n % 26;
-	    int j = n / 26;
-	    char buf[16];
-
-	    if ( j == 0 )
-	    { buf[0] = i+'A';
-	      buf[1] = EOS;
-	    } else
-	    { sprintf(buf, "%c%d", i+'A', j);
-	    }
-
-	    return PutToken(buf, out);
-	  }
-	  if ( PL_get_atom(arg, &a) )
-	  { write_options o2 = *options;
-	    clear(&o2, PL_WRT_QUOTED);
-	    
-	    return writeAtom(a, &o2);
-	  }	    
-	}
 					  /* op <term> */
 	if ( currentOperator(options->module, functor, OP_PREFIX,
 			     &op_type, &op_pri) )
@@ -844,7 +1078,13 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
 
   options.out = s;
   PutOpenToken(EOF, s);			/* reset this */
-  writeTerm(term, 1200, &options);
+  if ( (options.flags & PL_WRT_QUOTED) && !(s->flags&SIO_REPPL) )
+  { s->flags |= SIO_REPPL;
+    writeTerm(term, 1200, &options);
+    s->flags &= ~SIO_REPPL;
+  } else
+  { writeTerm(term, 1200, &options);
+  }
 
   return streamStatus(s);
 }
@@ -914,7 +1154,19 @@ pl_print2(term_t stream, term_t term)
 
 word
 pl_write_canonical2(term_t stream, term_t term)
-{ return do_write2(stream, term, PL_WRT_QUOTED|PL_WRT_IGNOREOPS);
+{ fid_t fid = PL_open_foreign_frame();
+  nv_options options;
+  word rc;
+
+  options.functor = FUNCTOR_isovar1;
+  options.on_attvar = AV_SKIP;
+  options.singletons = TRUE;
+  numberVars(term, &options, 0 PASS_LD);
+  rc = do_write2(stream, term,
+		 PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS);
+  PL_discard_foreign_frame(fid);
+
+  return rc;
 }
 
 word
