@@ -1,12 +1,30 @@
-/*  $Id: test.pl,v 1.76.2.1 2005/01/17 09:07:43 jan Exp $
+/*  $Id: test.pl,v 1.129 2006/08/18 11:12:59 jan Exp $
 
-    E-mail: jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Copyright (C) 1996 University of Amsterdam. All rights reserved.
+    Author:        Jan Wielemaker and Anjo Anjewierden
+    E-mail:        wielemak@science.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2005, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 :- set_prolog_flag(optimise, true).
 %:- set_prolog_flag(trace_gc, true).
+:- use_module(library(lists)).
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 SWI-Prolog test file.  A test is a clause of the form:
@@ -26,9 +44,46 @@ available test sets. The public goals are:
 % Required to get this always running regardless of user LANG setting.
 % Without this the tests won't run on machines with -for example- LANG=ja
 % according to NIDE Naoyuki, nide@ics.nara-wu.ac.jp.  Thanks!
+%
+% NOTE: In 5.5.x we should not need this any longer
 
-:- getenv('LANG', _) -> setenv('LANG', 'C'); true.
+% :- getenv('LANG', _) -> setenv('LANG', 'C'); true.
 
+% The test-suite Tests/library/test_date.pl depends on  the timezone. As
+% correct results are only  provided  for   the  CET  (Central European)
+% timezine we use this. Timezone cannot be  changed at runtime, so we do
+% this early.
+
+:- setenv('TZ', 'CET').
+
+		 /*******************************
+		 *    CONDITIONAL COMPILATION	*
+		 *******************************/
+
+:- dynamic
+	include_code/1.
+
+term_expansion((:- if(G)), []) :-
+	(   catch(G, E, (print_message(error, E), fail))
+	->  asserta(include_code(true))
+	;   asserta(include_code(false))
+	).
+term_expansion((:- else), []) :-
+	(   retract(include_code(X))
+	->  (   X == true
+	    ->  X2 = false 
+	    ;   X2 = true
+	    ),
+	    asserta(include_code(X2))
+	;   throw(error(context_error(no_if), _))
+	).
+term_expansion((:- endif), []) :-
+	retract(include_code(_)), !.
+
+term_expansion(_, []) :-
+	include_code(X), !,
+	X == false.
+	    
 
 		 /*******************************
 		 *	      SYNTAX		*
@@ -56,27 +111,36 @@ syntax(op-7) :-
 	op(500, xf, op2),
 	catch(atom_to_term("op1 a op2", op1(op2(a)), []), E, true),
 	E = error(syntax_error(operator_clash), _).
+syntax(op-8) :-				% assume 200 fy and 500 yfx
+	atom_to_term("- - a", -(-(a)), []).
 syntax(atom-1) :-
 	atom_codes('\003\\'\n\x80\', X),
-	X = [3, 39, 10, 128].
+	X == [3, 39, 10, 128].
 syntax(char-1) :-
-	10 = 0'\n.
-syntax(char-2) :-
-	52 = 0'\x34.
+	10 == 0'\n.
 syntax(char-2) :-
+	52 == 0'\x34.
+syntax(char-3) :-
 	"\\" =:= 0'\\.
+
 syntax(string-1) :-
-	'\c ' = ''.
+	'\c ' == ''.
 syntax(string-2) :-
-	'x\c y' = xy.
-syntax(number-1) :-			% check integer overflow translation
-	Chars = "41234567891",
-	name(X, Chars),
-	sformat(S, '~0f', [X]),
-	string_to_list(S, Chars).
+	'x\c y' == xy.
+syntax(quote-1) :-			% not really clear this is ok
+	'\x' == x.
+syntax(quote-2) :-
+	'\x61' == a.
+syntax(quote-3) :-
+	'\x61\' == a.
+syntax(quote-4) :-
+	char_code('\'', 39).
 syntax(number-2) :-
 	catch(atom_to_term('2\'', _, _), E, true),
 	E = error(syntax_error(illegal_number), _).
+syntax(zero-1) :-
+	term_to_atom(T, 'hello(\000\"\000\x")'),
+	T == hello([0, 120]).
 
 
 		 /*******************************
@@ -97,6 +161,38 @@ write_test(q-6) :-
 	term_to_atom('*/*', X), X == '*/*'.
 
 
+		 /*******************************
+		 *	      FORMAT		*
+		 *******************************/
+
+format_test(intD-1) :-
+	format(atom(X), '~D', [1000]),
+	X == '1,000'.
+format_test(intD-2) :-
+	format(atom(X), '~2D', [1000]),
+	X == '10.00'.
+format_test(intD-3) :-
+	format(atom(X), '~2D', [100000]),
+	X == '1,000.00'.
+format_test(intr-1) :-
+	format(atom(X), '~16r', [1000]),
+	X == '3e8'.
+format_test(intR-1) :-
+	format(atom(X), '~16R', [1000]),
+	X == '3E8'.
+format_test(oncodes-1) :-
+	format(codes(C), 'hello ~w', [world]),
+	C == "hello world".
+format_test(oncodes-2) :-
+	format(codes(C,T), 'hello ~w', [world]),
+	append("hello world", T2, C2),
+	C-T =@= C2-T2.
+format_test(onstring-1) :-
+	format(string(S), 'hello ~w', [world]),
+	string(S),
+	string_to_list(S, "hello world").
+
+
 		 /*******************************
 		 *	       UNIFY		*
 		 *******************************/
@@ -110,6 +206,30 @@ unify(cycle-2) :-			% Kuniaki Mukai
 	X = f(X), Y=f(Y), X=f(Y).
 
 
+		 /*******************************
+		 *    UNIFY-WITH-OCCURS-CHECK	*
+		 *******************************/
+
+occurs_check(simple-1) :-
+	\+ unify_with_occurs_check(A, list(A)).
+occurs_check(simple-2) :-
+	unify_with_occurs_check(_A, _B).
+occurs_check(attvar-1) :-		% test wakeup
+	freeze(X, X = Y),
+	unify_with_occurs_check(X, a), 
+	Y == a.
+occurs_check(attvar-2) :-		% test occurs-check
+	freeze(A, true),
+	\+ unify_with_occurs_check(A, list(A)).
+occurs_check(attvar-3) :-
+	freeze(A, true),
+	unify_with_occurs_check(A, A).
+occurs_check(attvar-4) :-
+	freeze(A, true),
+	freeze(B, true),
+	unify_with_occurs_check(A, B).
+
+
 		 /*******************************
 		 *	       CYCLIC		*
 		 *******************************/
@@ -119,6 +239,14 @@ cyclic(hash_term-1) :-
 	integer(T).
 cyclic(streq-1) :-
 	X = [X], Y = [Y], X =@= Y.
+cyclic(test-1) :-
+	X = f(X), cyclic_term(X).
+cyclic(test-2) :-
+	acyclic_term(f(x)).
+cyclic(test-3) :-
+	acyclic_term(_).
+cyclic(test-4) :-
+	X = f(a), acyclic_term(a(X, X)).
 
 
 		 /*******************************
@@ -163,7 +291,9 @@ arithmetic(arith-4) :-
 arithmetic(arith-5) :-
 	1 =:= integer(0.5).
 arithmetic(arith-6) :-
-	4.5 =:= abs(-4.5).
+	4.5 =:= abs(-4.5),
+	4 =:= abs(4),
+	4 =:= abs(-4).
 arithmetic(arith-7) :-
 	5.5 =:= max(1, 5.5).
 arithmetic(arith-8) :-
@@ -173,16 +303,31 @@ arithmetic(arith-9) :-
 arithmetic(arith-10) :-
 	-4000 =:= integer(10000 * float_fractional_part(-1e10 - 0.4)).
 arithmetic(arith-11) :-
-	current_prolog_flag(iso, ISO),
-	set_prolog_flag(iso, true),
 	1.0 is sin(pi/2),
-	set_prolog_flag(iso, false),
-	1   is sin(pi/2),
-	set_prolog_flag(iso, ISO).
+	\+ 1 is sin(pi/2).
 arithmetic(arith-12) :-
 	1.0 is float(sin(pi/2)).
 arithmetic(arith-13) :-
 	1.0 =:= sin(pi/2).
+arithmetic(sign-1) :-
+	-1 =:= sign(-1),   0 =:= sign(0),   1 =:= sign(1),
+	-1 =:= sign(-1.5), 0 =:= sign(0.0), 1 =:= sign(pi).
+arithmetic(floor-1) :-
+	0 is floor(0.0),
+	0 is floor(0.9),
+	-1 is floor(-0.1),
+	-1 is floor(-0.9).
+arithmetic(ceil-1) :-
+	0 is ceil(0.0),
+	1 is ceil(0.9),
+	0 is ceil(-0.1),
+	0 is ceil(-0.9).
+arithmetic(truncate-1) :-
+	1 is truncate(1.1),
+	1 is truncate(1.9),
+	-1 is truncate(-1.1),
+	-1 is truncate(-1.9).
+
 
 		 /*******************************
 		 *	    BIG NUMBERS		*
@@ -199,7 +344,7 @@ arithmetic(cmp-1) :-
 		 *******************************/
 
 foverflow(X) :-
-	X2 is X * 1000,
+	X2 is X * pi * pi,
 	foverflow(X2),
 	1 = 1.			% avoid tail-recursion to force termination
 
@@ -223,14 +368,14 @@ floattest(float-4) :-
 floattest(float-5) :-
 	clause(ftest(X), true),
 	X == 4.5.
-floattest(float-5) :-
+floattest(float-6) :-
 	clause(ftest, ftest(X)),
 	X == 4.5.
-floattest(float-6) :-
-	catch(foverflow(2), E, true),
-	E = error(evaluation_error(float_overflow), _).
 floattest(float-7) :-
-	catch(foverflow(-2), E, true),
+	catch(foverflow(2.2), E, true),
+	E = error(evaluation_error(float_overflow), _).
+floattest(float-8) :-
+	catch(foverflow(-2.2), E, true),
 	E = error(evaluation_error(float_overflow), _).
 
 
@@ -242,6 +387,8 @@ floattest(float-7) :-
 :- arithmetic_function(twice/1).
 :- arithmetic_function(mean/2).
 :- arithmetic_function(euler/0).
+:- arithmetic_function(fail/0).
+:- arithmetic_function(except/0).
 
 ten(10).
 twice(X, R) :-
@@ -251,6 +398,11 @@ mean(X1, X2, R) :-
 
 euler(2.71828).
 
+fail(_) :- fail.
+
+except(_) :-
+	throw(error(foobar)).
+
 arithmetic_functions(func-1) :-
 	A is ten, A =:= 10.
 arithmetic_functions(func-2) :-
@@ -261,6 +413,239 @@ arithmetic_functions(func-4) :-
         Exp = 6*euler*7*1,		% test functions corrupting stack
         EE is Exp,
 	EE =:= 6*euler*7*1.
+arithmetic_functions(fail-1) :-
+	catch(_ is fail, E, true),
+	compound(E), E = error(E2, _),
+	compound(E2), E2 = failure_error(_).
+arithmetic_functions(except-1) :-
+	catch(_ is except, E, true),
+	E == error(foobar).
+
+
+		 /*******************************
+		 *     UNBOUNDED ARITHMETIC	*
+		 *******************************/
+
+:- dynamic
+	gmp_clause/2.
+
+fac(1,1) :- !.
+fac(X,N) :-
+	X > 1,
+	X2 is X - 1,
+	fac(X2, N0),
+	N is N0 * X.
+
+oefac(X, Fac) :-			% produce large pos and neg ints
+	fac(X, F0),
+	odd_even_neg(X, F0, Fac).
+
+odd_even_neg(X, V0, V) :-
+	(   X mod 2 =:= 0
+	->  V = V0
+	;   V is -V0
+	).
+
+:- if(current_prolog_flag(bounded, false)). % GMP implies rational
+
+ratp(C, X, X ) :-
+	C =< 0, !.
+ratp(Count, In, Out) :-
+        succ(Count0, Count),
+        T is In + (In rdiv 2),
+	ratp(Count0, T, Out).
+
+:- arithmetic_function(idiv/2).
+
+idiv(Dd,Dr,Iq):-
+        Q is Dd/Dr,
+        rational(Q,Qt,Qn),
+        Iq is Qt//Qn.
+
+
+gmp(neg-1) :-				% check conversion of PLMININT
+	A is -9223372036854775808,
+	-A =:= 9223372036854775808.
+gmp(neg-2) :-
+	A is -(1<<63+1), A == -9223372036854775809.
+gmp(abs-1) :-
+	A is -9223372036854775808,
+	abs(A) =:= 9223372036854775808.
+gmp(sign-1) :-
+	-1 =:= sign(-5 rdiv 3),
+	0 =:= sign(0 rdiv 1), 
+	1 =:= sign(2 rdiv 7),
+	fac(60, X),
+	-1 =:= sign(-X),
+	1 =:= sign(X).
+gmp(floor-1) :-
+	A is floor(1e20),
+	integer(A),
+	1e20 =:= float(A).
+gmp(floor-2) :-
+	0 is floor(9 rdiv 10),
+	-1 is floor(-1 rdiv 10),
+	-1 is floor(-9 rdiv 10).
+gmp(ceil-1) :-
+	A is ceil(1e20),
+	integer(A),
+	1e20 =:= float(A).
+gmp(ceil-2) :-
+	1 is ceil(9 rdiv 10),
+	0 is ceil(-1 rdiv 10),
+	0 is ceil(-9 rdiv 10).
+gmp(msb-0) :-
+	catch(0 =:= msb(0), E, true),
+	E = error(domain_error(not_less_than_one, 0), _).
+gmp(msb-1) :-
+	10 =:= msb(1<<10).
+gmp(msb-2) :-
+	100 =:= msb(1<<100).
+gmp(lsb-0) :-
+	catch(0 =:= lsb(0), E, true),
+	E = error(domain_error(not_less_than_one, 0), _).
+gmp(lsb-1) :-
+	10 =:= lsb(1<<10).
+gmp(lsb-2) :-
+	100 =:= lsb(1<<100).
+gmp(popcount-1) :-
+	1 =:= popcount(1<<5).
+gmp(popcount-2) :-
+	1 =:= popcount(1<<100).
+gmp(shift-1) :-
+	A is 1<<54, B is A<<8,
+	B =:= 4611686018427387904.
+gmp(shift-2) :-
+	A is 1<<55, B is A<<8,
+	B =:= 9223372036854775808.
+gmp(fac-1) :-
+	fac(25, X),
+	X == 15511210043330985984000000. % computed with bc
+gmp(arith-1) :-
+	A = 12345678901234567890123456789,
+	B = 7070707070707070707,
+	X is A * (A * B * A * B * A * B),
+	integer(X),
+	Y is B * A * B * A * B * A,
+	integer(Y),
+	R is X / Y,
+	R == A. 
+gmp(pow-1) :-
+	A is 10**50, integer(A).
+gmp(pow-2) :-
+	A is -10**3, integer(A), A = -1000.
+gmp(pow-3) :-
+	A is 0**0, A = 1.
+gmp(pow-4) :-
+	A is 1.5**2, A = 2.25.
+gmp(pow-5) :-
+	A is -100**5,
+	A < 0,
+	abs(A) =:= 100**5.
+gmp(integer-1) :-			% rounding integer conversion
+	0 =:= integer(1 rdiv 3),
+	1 =:= integer(2 rdiv 3),
+	0 =:= integer(-1 rdiv 3),
+       -1 =:= integer(-2 rdiv 3).
+gmp(integer-2) :-
+	0 =:= integer(1/3),
+	1 =:= integer(2/3),
+	0 =:= integer(-1/3),
+       -1 =:= integer(-2/3).
+gmp(rational-1) :-				% IEEE can represent 0.25
+	rational(1/4) =:= 1 rdiv 4.
+gmp(rational-2) :-
+	A is 2 rdiv 4,
+	rational(A, 1, 2).
+gmp(rationalize-1) :-
+	A is rationalize(0.0), A == 0,
+	B is rationalize(0.1), B == 1 rdiv 10,
+	C is rationalize(10.0), C == 10,
+	D is rationalize(-0.1), D == -1 rdiv 10.
+gmp(rationalize-2) :-
+	pi =:= float(rationalize(pi)).
+gmp(number-1) :-
+	A is 1 rdiv 3,
+	rational(A).
+gmp(float-1) :-
+	1/3 =:= float(1 rdiv 3).
+gmp(cmp-1) :-
+	0 < 3,
+	-1 < 0,
+	-1 < 1 rdiv 3,
+	1 > 0.3,
+	fac(100, F100), fac(200, F200), F100 < F200,
+	pi > 5 rdiv 2.
+gmp(clause-1) :-
+	Clause = (gmp_clause(X,Y) :-
+		       X is Y + 3353461138769748319272960000),
+	assert(Clause, Ref),
+	clause(H,B,Ref),
+	erase(Ref),
+	(H:-B) =@= Clause.
+gmp(comp-1) :-
+	retractall(gmp_clause(_,_)),
+	forall(between(1, 50, X),
+	       (   oefac(X, Fac),
+		   assert(gmp_clause(Fac, X))
+	       )),
+	forall(between(1, 50, X),
+	       (   oefac(X, Fac),
+		   gmp_clause(Fac, Y),
+		   X == Y,
+		   clause(gmp_clause(Fac, Z), true),
+		   X == Z
+	       )),
+	retractall(gmp_clause(_,_)).
+gmp(comp-2) :-
+	X is ((1 rdiv 2)*2) rdiv 3,
+	X == 1 rdiv 3.
+gmp(rec-1) :-
+	forall(recorded(gmp_fac, _, R), erase(R)),
+	forall(between(1, 50, X),
+	       (   oefac(X, Fac),
+		   recordz(gmp_fac, X-Fac)
+	       )),
+	forall(between(1, 50, X),
+	       (   oefac(X, Fac),
+		   recorded(gmp_fac, X-Y),
+		   Fac == Y
+	       )),
+	forall(recorded(gmp_fac, _, R), erase(R)).
+gmp(number_codes-1) :-
+	fac(25, X),
+	number_codes(X, Codes),		% write
+	number_codes(Y, Codes),		% read
+	X == Y.
+gmp(atom_number-1) :-
+	fac(100, X),
+	atom_number(Atom, X),		% write
+	atom_number(Atom, Y),		% read
+	X == Y.
+gmp(hex-1) :-
+	atom_number('0xFFFFFFFFFFFFFFFF', 18446744073709551615).
+gmp(fmtd-1) :-
+	format(atom(X), '~d', [12345678901234567890123456]),
+	X == '12345678901234567890123456'.
+gmp(fmtd-2) :-
+	format(atom(X), '~2d', [12345678901234567890123456]),
+	X == '123456789012345678901234.56'.
+gmp(fmtD-1) :-
+	format(atom(X), '~D', [12345678901234567890123456]),
+	X == '12,345,678,901,234,567,890,123,456'.
+gmp(fmtD-2) :-
+	format(atom(X), '~2D', [12345678901234567890123456]),
+	X ==  '123,456,789,012,345,678,901,234.56'.
+gmp(fmtf-1) :-
+	ratp(999, 1, X),
+	format(atom(S), '~5f', [X]),
+	sub_atom(S, _, _, 0, '935376.65824').
+gmp(idiv-1) :-
+	Qi is idiv(3 rdiv 2,2 rdiv 5),
+	Qi == 3.
+
+:- endif.
+
 
 		 /*******************************
 		 *	     CHARACTERS		*
@@ -274,6 +659,34 @@ chars(chars-2) :-
 	A == 97.
 
 
+		 /*******************************
+		 *	 WIDE CHARACTERS	*
+		 *******************************/
+
+wchar_string("abc").				% ISO Latin-1
+wchar_string([1097, 1098, 1099]).		% UCS
+wchar_string([97, 98, 99, 1097, 1098, 1099]).   % Mixed
+wchar_string([1097, 1098, 1099, 97, 98, 99]).	% Mixed
+
+wchars(cmp-1) :-
+	forall(( wchar_string(S1),
+		 wchar_string(S2)),
+	       ( atom_codes(A1, S1),
+		 atom_codes(A2, S2),
+		 compare(Diff, A1, A2),
+		 compare(Diff, S1, S2))).
+wchars(cmp-2) :-
+	forall(( wchar_string(S1),
+		 wchar_string(S2)),
+	       ( string_to_list(A1, S1),
+		 string_to_list(A2, S2),
+		 compare(Diff, A1, A2),
+		 (   compare(Diff, S1, S2)
+		 ->  true
+		 ;   format(user_error, 'CMP ~w ~w FAILED~n', [S1, S2])
+		 ))).
+
+
 		 /*******************************
 		 *	   META CALLING		*
 		 *******************************/
@@ -291,18 +704,18 @@ meta(call-4) :-
 	length(X, 100000), call((is_list(X) -> true ; fail)).
 meta(call-5) :-
 	call((X=a;X=b)), X = b.
-meta(call-5) :-
-	call((foo:hello(X)->true)), X = world.
 meta(call-6) :-
-	call((X=a,x(X)=Y)), Y == x(a).
+	call((foo:hello(X)->true)), X = world.
 meta(call-7) :-
+	call((X=a,x(X)=Y)), Y == x(a).
+meta(call-8) :-
 	string_to_list(S, "hello world"),
 	call((string(S), true)).
-meta(call-8) :-
-	call((foo:true, true)).
 meta(call-9) :-
-	call((A=x, B=x, A==B)).		% avoid I_CALL_FVX for dynamic call
+	call((foo:true, true)).
 meta(call-10) :-
+	call((A=x, B=x, A==B)).		% avoid I_CALL_FVX for dynamic call
+meta(call-11) :-
 	A = (	member(_,[1,2,3]),
 		flag(a, F, F+1),
 		(   F >= 999999
@@ -313,9 +726,9 @@ meta(call-10) :-
 	flag(a, Old, 0),
 	forall(A, true),
 	flag(a, 3, Old).
-meta(call-11) :-
+meta(call-12) :-
 	catch(call(1), E, true),
-	E =@= error(type_error(callable, 1), _).
+	error(E, type_error(callable, 1)).
 meta(apply-1) :-
 	apply(=, [a,a]).
 meta(apply-2) :-
@@ -337,6 +750,19 @@ cleanup_2(b).
 cleanup_3 :-
 	fail.
 
+ex(error("Nice term")).
+
+genex1(1) :- ex(Ex), throw(Ex).
+genex1(2).
+
+genex2(1).
+genex2(2) :- ex(Ex), throw(Ex).
+genex2(3).
+
+genex3(1).
+genex3(2).
+genex3(3) :- ex(Ex), throw(Ex).
+
 cleanup(clean-1) :-
 	retractall(clean_rval(_)),
 	call_cleanup(cleanup_1, R, assert(clean_rval(R))),
@@ -366,7 +792,15 @@ cleanup(clean-8) :-
 	call_cleanup(bagof(x, cleanup_1, _Xs), Reason,
 		     assert(clean_rval(Reason))),
 	retract(clean_rval(exit)).
-
+cleanup(findall-1) :-
+	catch(findall(X, genex1(X), _), E, true),
+	ex(Ex), E == Ex.
+cleanup(findall-2) :-
+	catch(findall(X, genex2(X), _), E, true),
+	ex(Ex), E == Ex.
+cleanup(findall-3) :-
+	catch(findall(X, genex3(X), _), E, true),
+	ex(Ex), E == Ex.
 
 
 		 /*******************************
@@ -390,13 +824,6 @@ dl_fail(N) :-
 
 :- arithmetic_function(fac/1).
 
-fac(1, 1) :- !.
-fac(N, V) :-
-	NN is N - 1,
-	fac(NN, V0),
-	V is N*V0.
-
-
 depth_limit(depth-1) :-
 	call_with_depth_limit(dl_det(1), 10, 1),
 	deterministic(true).
@@ -482,11 +909,11 @@ list(sort-4) :-
 	msort([e,b,c,e], [b,c,e,e]).
 list(sort-5) :-
 	keysort([e-2,b-5,c-6,e-1], [b-5,c-6,e-2,e-1]).
-list(sort-5) :-
-	sort([a,g,b], [a,b|G]), G == [g].
 list(sort-6) :-
-	sort([X], [Y]), X == Y.
+	sort([a,g,b], [a,b|G]), G == [g].
 list(sort-7) :-
+	sort([X], [Y]), X == Y.
+list(sort-8) :-
 	sort([_X, _Y], [_,_]).
 
 
@@ -565,6 +992,9 @@ atom_handling(concat-4) :-
 atom_handling(concat-5) :-
 	findall(X-Y, atom_concat(X, Y, 'abc'), Pairs),
 	Pairs == [''-abc, a-bc, ab-c, abc-''].
+atom_handling(concat-6) :-
+	atom_concat(X, a, a),
+	X == ''.
 
 atom_handling(number-1) :-
 	atom_number('42', X), X == 42.
@@ -574,6 +1004,11 @@ atom_handling(number-3) :-
 	atom_number(X, 1.0), X == '1.0'.
 atom_handling(number-4) :-
 	atom_number(X, 42), X == '42'.
+atom_handling(number-5) :-
+	A is 1<<42,
+	number_codes(A, Codes),
+	number_codes(A2, Codes),
+	A == A2.
 
 atom_handling(sub_atom-1) :-
 	\+ sub_atom(a, _, _, 3, _).
@@ -750,6 +1185,53 @@ record(erase-2) :-
 	Xs = [].
 
 
+		 /*******************************
+		 *	    ASSERT/RETRACT	*
+		 *******************************/
+
+%	compiler
+%	
+%	This suite tests whether all data-types can be compiled properly
+%	and handled by the decompiler.
+
+:- dynamic
+	compiler_test/1.
+
+compiler(assert-1) :-
+	mkterm(T0),			% verify head compile mode
+	assert(compiler_test(T0)),
+	compiler_test(T),
+	T0 =@= T,
+	retractall(compiler_test(_)).
+compiler(assert-2) :-
+	mkterm(T0),			% verify head compile mode
+	assert(compiler_test(T0)),
+	compiler_test(T0),
+	retractall(compiler_test(_)).
+compiler(assert-3) :-			% verify body compile mode
+	mkterm(T0),
+	assert((compiler_test(X) :- unify(X, T0))),
+	compiler_test(T),
+	T0 =@= T,
+	retractall(compiler_test(_)).
+compiler(assert-4) :-
+	mkterm(T0),
+	assert(compiler_test(T0)),
+	retract(compiler_test(T)),
+	T0 =@= T.
+compiler(assert-5) :-
+	mkterm(T0),
+	assert(compiler_test(T0)),
+	retract(compiler_test(T0)).
+compiler(assert-6) :-
+	numlist(0, 100, L),
+	append(L, [_], A),
+	T =.. [x|A],
+	assert(T),
+	T.
+
+unify(X, X).
+
 		 /*******************************
 		 *	       FLAGS		*
 		 *******************************/
@@ -876,7 +1358,7 @@ control(softcut-2) :-
 	findall(A, softcut2(A), [1]).
 control(block-1) :-
 	catch(block(myblock, do_block, _), E, true),
-	E =@= error(existence_error(block, notmyblock), _).
+	error(E, existence_error(block, notmyblock)).
 control(block-2) :-
 	block(notmyblock, do_block, X),
 	X == ok.
@@ -933,20 +1415,31 @@ throwit :-
 catchme :-
 	catch(throwit, _, true).
 
+undef :-
+	'this is not defined'.
+
 exception(call-1) :-
 	catch(do_exception_1, E, true),
-	E =@= error(instantiation_error, _).
+	error(E, instantiation_error).
 exception(call-2) :-
 	\+ catch(do_exception_1, _, fail).
 exception(call-3) :-
 	catch(rethrow(do_exception_1), E, true),
-	E =@= error(instantiation_error, _).
+	error(E, instantiation_error).
 exception(call-4) :-
 	catch(throwit, foo(X), X = a),
 	X = a.
 exception(call-5) :-
 	catch(throwit, _, catchme).
 
+exception(context-1) :-
+	catch(functor(_,_,_), E, true),
+	error_pred(E, functor/3).
+exception(context-2) :-
+	catch(undef, E, true),
+	error_pred(E, undef/0).
+
+
 		 /*******************************
 		 *	  RESOURCE ERRORS	*
 		 *******************************/
@@ -1013,23 +1506,31 @@ gc(agc-2) :-
 gc_data(a).
 
 		 /*******************************
-		 *            FLOATS		*
+		 *	INTEGER OVERFLOW	*
 		 *******************************/
 
-floatconv(float-1) :-
-	A is 5.5/5.5, integer(A).
-floatconv(float-2) :-
+intoverflow(add-1) :-
 	current_prolog_flag(max_integer, MI),
-	ToHigh is MI + 10000,		% +1 may fail on 64-bit systems
+	ToHigh is MI + 10000,
 	float(ToHigh).
-floatconv(float-3) :-
+intoverflow(syntax-1) :-
 	(   current_prolog_flag(max_integer, 2147483647)
 	->  term_to_atom(X, 2147483648)
 	;   current_prolog_flag(max_integer, 9223372036854775807)
 	->  term_to_atom(X, 9223372036854775808)
 	),
 	float(X).
-
+intoverflow(syntax-2) :-
+	(   current_prolog_flag(max_integer, 2147483647)
+	->  Chars = "41234567891",
+	    Float = 41234567891.0
+	;   Chars = "9223372036854775900",
+	    Float = 9223372036854775900.0
+	),
+	name(X, Chars),
+	float(X),
+	X =:= Float.
+	
 
 		 /*******************************
 		 *	ATTRIBUTED VARIABLES	*
@@ -1057,6 +1558,8 @@ u(arith, X) :-
 	X is 3.
 u(args, X) :-
 	u_args(X, args).
+u(functor, X) :-
+	functor(X, foo, 1).
 
 test_wakeup(How) :-
 	freeze(X, Y=ok),
@@ -1105,6 +1608,8 @@ avar(wakeup-5) :-
 	test_wakeup(arith).
 avar(wakeup-6) :-
 	test_wakeup(args).
+avar(wakeup-7) :-
+	test_wakeup(functor).
 avar(type-1) :-
 	put_attr(X, test, a),
 	avar(X).
@@ -1173,7 +1678,7 @@ avar(order-1) :-			% attributes do not change standard
 
 nogvar(Var) :-
 	catch(nb_getval(Var, _), E, true),
-	E =@= error(existence_error(variable, Var), _).
+	error(E, existence_error(variable, Var)).
 
 gvar(set-1) :-
 	nb_setval(gnu, gnat),
@@ -1260,13 +1765,14 @@ copy_term(av-3) :-
 	get_attr(Y, test, A),
 	A = f(Z),
 	Y == Z.
-copy_term(av-3) :-
+copy_term(av-4) :-
 	freeze(X, true),
 	freeze(X, Done = true),
 	copy_term(X, Y),
 	X = ok,
 	Done == true,
-	get_attr(Y, freeze, (true, D2=true)),
+	get_attr(Y, freeze, Att),
+	Att = (user:true, user:(D2=true)),
 	var(D2),
 	Y = ok,
 	D2  == true.
@@ -1274,13 +1780,51 @@ copy_term(reset-1) :-			% reset cycle resetting for shared
 	A = [a:b,c:d|_],		% terms.
 	copy_term(A,_B),
 	A = [a:b,c:d|_].
-copy_term(nat-1) :-
+copy_term(nat-1) :-			% shared variables
 	put_attr(X, foo, y),
-	copy_term_nat(x(X,X), Y),
-	Y = x(A,B),
-	A == B,
-	var(A),
-	\+ attvar(A).
+	T = x(X,X),
+	copy_term_nat(T, Y),
+	Y = x(A,B), A == B, var(A), \+ attvar(A),
+	T = x(C,D), C == D, var(C), get_attr(C, foo, y).
+copy_term(nat-2) :-			% cyclic term
+	put_attr(X, foo, x),
+	T = t(T,X),
+	copy_term_nat(T, Y),
+	A = t(A,B),
+	Y = A,
+	var(B), \+ attvar(B),
+	arg(2, T, X2),
+	get_attr(X2, foo, x).
+
+
+		 /*******************************
+		 *	     HASH-TERM		*
+		 *******************************/
+
+%	NOTE: these numbers are for 32-bit platforms. Different value of
+%	max_tagged_integer is main reason why the result is different on
+%	64-bit platforms. What to do? Use the   32-bit  limit also on 64
+%	bit platforms?  For now the test is disabled on 64-bit machines.
+
+hash_term(simple-1) :-
+	hash_term(aap, 480).
+hash_term(simple-2) :-			% small int
+	hash_term(42, 42).
+hash_term(simple-3) :-			% not tagged int
+	hash_term(2000000000, 3511296).
+hash_term(simple-4) :-
+	A is pi,
+	hash_term(A, 5049571).
+hash_term(simple-5) :-
+	string_to_list(S, "hello world"),
+	hash_term(S, 16017976).
+hash_term(compound-1) :-
+	hash_term(hello(world), 259300). 
+hash_term(compound-2) :-
+	X = x(a),
+	hash_term(hello(X, X), 92021). 
+hash_term(compound-3) :-
+	hash_term(hello(x(a), x(a)), 92021). 
 
 
 		 /*******************************
@@ -1306,6 +1850,18 @@ term_atom(term_to_atom-2) :-
 	termtest(1000).
 
 
+		 /*******************************
+		 *	        OS		*
+		 *******************************/
+
+os(getenv-1) :-
+	getenv('PATH', _).		% should be around on most OSes
+os(setenv-1) :-
+	setenv(pltestsetenv, yes),
+	getenv(pltestsetenv, X),
+	X == yes.
+
+
 		 /*******************************
 		 *		I/O		*
 		 *******************************/
@@ -1393,6 +1949,9 @@ popen(cat-2) :-
 					% otherwise
 	;   E = error(io_error(write, _), context(_, 'Broken pipe'))
 	->  true
+	;   E = error(io_error(write, _), _),
+	    current_prolog_flag(windows, true)
+	->  true
 	;   format(user_error, 'Wrong exception: ~p~n', [E]),
 	    fail
 	).
@@ -1425,7 +1984,9 @@ collect_data(C, Fd, [C|T]) :-
 timeout(pipe-1) :-
 	(   current_prolog_flag(pipe, true),
 	    \+ current_prolog_flag(windows, true) % cannot wait on pipes
-	->  open(pipe('echo xx && sleep 2 && echo xx.'), read, In),
+	->  open(pipe('echo xx && sleep 2 && echo xx.'), read, In,
+		 [ bom(false)
+		 ]),
 	    set_stream(In, timeout(1)),
 	    wait_for_input([In], [In], infinite),
 	    catch(read(In, _), E1, true),
@@ -1461,6 +2022,108 @@ file(dir-1) :-
 file(dir-2) :-
 	testfile(File),
 	\+ exists_directory(File).
+file(cwd-1) :-
+	working_directory(CWD, CWD),
+	exists_directory(CWD),
+	same_file(CWD, '.').
+file(absfile-2) :-			% canoniseDir() caching issues
+	X = 'pl-test-x',
+	Y = 'pl-test-y',
+	atom_concat(X, '/file', XF),
+	atom_concat(Y, '/file', YF),
+	make_directory(X),
+	touch(XF),
+	absolute_file_name(XF, Abs1),
+	atom_concat(_, XF, Abs1),
+	delete_file(XF),
+	delete_directory(X),
+	make_directory(Y),
+	touch(YF),
+	absolute_file_name(YF, Abs2),
+	delete_file(YF),
+	delete_directory(Y),
+	atom_concat(_, YF, Abs2).
+file(ext-2) :-
+	\+ file_name_extension(foo, _, 'bar.pl'). 	% Bug#69
+file(open-1) :-
+	catch(open(foobar, read, _, [lock(qqq)]), E, true),
+	E =@= error(domain_error(lock, qqq), context(system:open/4, _)).
+
+
+touch(File) :-
+	open(File, update, Out),
+	close(Out).
+
+		 /*******************************
+		 *	UNICODE FILENAMES	*
+		 *******************************/
+
+unicode_file(mkdir-1) :-			% create Cyrillic directory
+	atom_codes(Dir, [1074, 1086, 1079, 1076, 1091, 1093, 1072]),
+	catch(delete_directory(Dir), _, true),
+	make_directory(Dir),
+	exists_directory(Dir),
+	working_directory(Old, Dir),
+	working_directory(O2, '..'),
+	same_file(Old, '.'),
+	same_file(O2, Dir),
+	delete_directory(Dir).
+unicode_file(file-1) :-				% create Cyrillic file
+	atom_codes(File, [1074, 1086, 1079, 1076, 1091, 1093, 1072]),
+	Term = hello(world),
+	catch(delete_file(File), _, true),
+	open(File, write, Out),
+	format(Out, '~q.~n', [Term]),
+	close(Out),
+	exists_file(File),
+	open(File, read, In),
+	read(In, Read),
+	close(In),
+	Read =@= Term,
+	delete_file(File).
+unicode_file(absfile-1) :-
+	atom_codes(File, [1074, 1086, 1079, 1076, 1091, 1093, 1072]),
+	absolute_file_name(File, Path),
+	file_directory_name(Path, Dir),
+	same_file(Dir, '.'),
+	file_base_name(Path, Base),
+	Base == File.
+unicode_file(ext-1) :-
+	atom_codes(File, [1074, 1086, 1079, 1076, 0'., 1091, 1093, 1072]),
+	file_name_extension(Base, Ext, File),
+	atom_codes(Base, [1074, 1086, 1079, 1076]),
+	atom_codes(Ext, [1091, 1093, 1072]).
+
+
+		 /*******************************
+		 *		SEEK		*
+		 *******************************/
+
+seek(write-1) :-
+	tmp_file(seek, File),
+	open(File, write, S, [type(binary)]),
+	Max = 999,
+	forall(between(0, Max, _),
+	       format(S, '1234567890~n', [])),
+	forall(between(0, Max, N),
+	       (   Pos is N * 11 + 6,
+		   seek(S, Pos, bof, _),
+		   format(S, 'x', [])
+	       )),
+	close(S),
+
+	open(File, read, In, [type(binary)]),
+	forall(between(0, Max, N),
+	       must_read("123456x890\n", In)),
+	close(In),
+	delete_file(File).
+
+must_read([], _) :- !.
+must_read([H|T], In) :-
+	get_byte(In, Byte),
+	H == Byte,
+	must_read(T, In).
+
 
 
 		 /*******************************
@@ -1482,6 +2145,15 @@ ctype(code_type-5) :-
 	code_type(48, digit(W)),
 	W == 0.
 
+		 /*******************************
+		 *	       WCTYPE		*
+		 *******************************/
+
+wctype(code_type-6) :-
+	char_code(X, 1080),
+	code_type(X, alnum).
+
+
 		 /*******************************
 		 *	      CONSULT		*
 		 *******************************/
@@ -1546,6 +2218,34 @@ thread(at_exit-1) :-
 	retract(at_exit_called).
 	
 
+		 /*******************************
+		 *	 MUTEX HANDLING		*
+		 *******************************/
+
+mutex(trylock-1) :-
+	gensym(mutex, Mutex),
+	thread_self(Main),
+	thread_create(( mutex_lock(Mutex),
+			thread_send_message(Main, locked),
+			thread_get_message(_)),
+		      Id, []),
+	thread_get_message(locked),
+	\+ mutex_trylock(Mutex),
+	thread_send_message(Id, done),
+	thread_join(Id, true).
+mutex(unlock-1) :-
+	gensym(mutex, Mutex),
+	mutex_lock(Mutex),
+	mutex_unlock(Mutex),
+	catch(mutex_unlock(Mutex), E, true),
+	E == error(permission_error(mutex, unlock, Mutex),
+		   context(mutex_unlock/1, 'not locked')).
+mutex(destroy-1) :-
+	gensym(mutex, Mutex),
+	mutex_create(Mutex),
+	mutex_destroy(Mutex).
+
+
 		 /*******************************
 		 *	      SCRIPTS		*
 		 *******************************/
@@ -1621,11 +2321,16 @@ script_failed(File, Except) :-
 
 testset(syntax).
 testset(write_test).
+testset(format_test).
 testset(unify).
+testset(occurs_check).
 testset(arithmetic).
 testset(arithmetic_functions).
 testset(floattest).
+testset(gmp) :-
+	current_prolog_flag(bounded, false).
 testset(chars).
+testset(wchars).
 testset(depth_limit) :-
 	current_predicate(_, user:call_with_depth_limit(_,_,_)).
 testset(type_test).
@@ -1633,6 +2338,8 @@ testset(meta).
 testset(avar).
 testset(gvar).
 testset(copy_term).
+testset(hash_term) :-			% 32-bit platforms
+	current_prolog_flag(max_integer, 2147483647).
 testset(cyclic).
 testset(cleanup).
 testset(term).
@@ -1643,30 +2350,67 @@ testset(string_handling).
 testset(proc).
 testset(cl).
 testset(record).
+testset(compiler).
 testset(flag).
 testset(update).
 testset(gc).
-testset(floatconv) :-
-	current_prolog_flag(iso, false).
+testset(intoverflow) :-
+	current_prolog_flag(iso, false),
+	current_prolog_flag(max_integer, _).
 testset(control).
 testset(exception).
 testset(term_atom).
+testset(os).
 testset(io).
 testset(popen) :-
 	current_prolog_flag(pipe, true).
 testset(timeout).
 testset(file).
+testset(unicode_file) :-
+	unicode_file_locale.
+testset(seek).
 testset(load_program).
 testset(ctype).
+testset(wctype) :-
+	wide_character_types.
 testset(thread) :-
 	current_prolog_flag(threads, true).
+testset(mutex) :-
+	current_prolog_flag(threads, true).
 testset(resource).
 
+%	unicode_file_locale/0
+%	
+%	True if out filesystem can   handle Unicode filenames. Difficult
+%	to have a good test.
+
+unicode_file_locale :-
+	current_prolog_flag(encoding, utf8), !.
+unicode_file_locale :-
+	catch(file_name_extension(_,_,[1050]), E, true),
+	(   var(E)
+	->  true
+	;   E \= error(representation_error(encoding), _)
+	).
+
+%	wide_character_types
+%	
+%	True if the  character  classification   routines  work  on wide
+%	characters. Hard to say when this is  the case. On some machines
+%	the wide character versions always work,  on others only for the
+%	codepages covered by the locale.
+
+wide_character_types :-
+	current_prolog_flag(encoding, utf8), !.
+
 %	testdir(Dir)
 %	
 %	Enumerate directories holding tests.
 
 testdir('Tests/attvar').
+testdir('Tests/library').
+testdir('Tests/charset').
+testdir('Tests/clp').
 testdir('Tests/thread') :-
 	current_prolog_flag(threads, true).
 
@@ -1711,14 +2455,15 @@ runtest(Name) :-
 	format('Running test set "~w" ', [Name]),
 	flush,
 	functor(Head, Name, 1),
-	nth_clause(Head, _N, R),
-	clause(Head, _, R),
+	findall(Head-R, nth_clause_head(Head, R), Heads),
+	unique_heads(Heads),
+	member(Head-R, Heads),
 	(   catch(Head, Except, true)
 	->  (   var(Except)
-	    ->  put(.), flush
+	    ->  put(.), flush_output
 	    ;   Except = blocked(Reason)
 	    ->  assert(blocked(Head, Reason)),
-		put(!), flush
+		put(!), flush_output
 	    ;   test_failed(R, Except)
 	    )
 	;   test_failed(R, fail)
@@ -1727,6 +2472,27 @@ runtest(Name) :-
 runtest(_) :-
 	format(' done.~n').
 	
+nth_clause_head(Head, R) :-
+	nth_clause(Head, _N, R),
+	clause(Head, _, R).
+
+unique_heads(Heads) :-
+	keysort(Heads, Sorted),
+	check_uniqye(Sorted).
+
+check_uniqye([]).
+check_uniqye([Head-R1,Head-R2|T]) :- !,
+	clause_property(R1, line_count(Line1)),
+	clause_property(R1, file(File1)),
+	clause_property(R2, line_count(Line2)),
+	clause_property(R2, file(File2)),
+	format('~N~w:~d: test ~w duplicated at ~w:~d~n',
+	       [File2, Line2, Head, File1, Line1]),
+	check_uniqye([Head-R1|T]).
+check_uniqye([_|T]) :-
+	check_uniqye(T).
+
+
 test_failed(R, Except) :-
 	clause(Head, _, R),
 	functor(Head, Name, 1),
@@ -1745,3 +2511,23 @@ test_failed(R, Except) :-
 blocked(Reason) :-
 	throw(blocked(Reason)).
 
+
+%	error(+Exception, +Expected)
+%	
+%	Check whether the correct exception  is thrown, disregarding the
+%	2nd context argument.
+
+error(error(Ex, _Ctx), Expected) :-
+	Ex =@= Expected, !.
+error(error(Ex, _Ctx), Expected) :-
+	format('~NWrong exception: ~p (expected ~p)~n', [Ex, Expected]),
+	fail.
+
+error_pred(error(_, context(Pred, _)), Pred).
+error_pred(error(_, context(Module:Pred, _)), Pred) :-
+	hidden_module(Module).
+
+hidden_module(user) :- !.
+hidden_module(system) :- !.
+hidden_module(M) :-
+	sub_atom(M, 0, _, _, $).
