@@ -1,4 +1,4 @@
-/*  $Id: console.c,v 1.28 2004/09/21 14:22:26 jan Exp $
+/*  $Id: console.c,v 1.41 2005/11/21 16:09:53 jan Exp $
 
     Part of SWI-Prolog
 
@@ -80,9 +80,14 @@ static void initHeapDebug(void);
 #endif
 
 #include <windows.h>
+#include <tchar.h>
 #ifndef WM_MOUSEWHEEL			/* sometimes not defined */
 #define WM_MOUSEWHEEL 0x020A
 #endif
+#ifndef WM_UNICHAR
+#define WM_UNICHAR 0x109
+#define UNICODE_NOCHAR 0xFFFF
+#endif
 
 #include <stdlib.h>
 #include <io.h>
@@ -98,7 +103,7 @@ static void initHeapDebug(void);
 #include <stdio.h>
 
 #ifndef isletter
-#define isletter(c) (isalpha(c) || (c) == '_')
+#define isletter(c) (_istalpha(c) || (c) == '_')
 #endif
 
 #ifndef MAXPATHLEN
@@ -148,7 +153,7 @@ static void initHeapDebug(void);
 
 #define Control(x) ((x) - '@')
 
-#define streq(s, q) (strcmp((s), (q)) == 0)
+#define streq(s, q) (_tcscmp((s), (q)) == 0)
 
 #include "console_i.h"			/* internal package stuff */
 
@@ -162,7 +167,7 @@ static void initHeapDebug(void);
        RlcData  _rlc_stdio = NULL;	/* the main buffer */
 static int      _rlc_show;		/* initial show */
 static char	_rlc_word_chars[CHAR_MAX]; /* word-characters (selection) */
-static const char *	_rlc_program;		/* name of the program */
+static const TCHAR *	_rlc_program;		/* name of the program */
 static HANDLE   _rlc_hinstance;		/* Global instance */
 static HICON    _rlc_hicon;		/* Global icon */
 
@@ -186,7 +191,7 @@ static void	rlc_init_text_dimensions(RlcData b, HFONT f);
 static void	rlc_save_font_options(HFONT f, rlc_console_attr *attr);
 static void	rlc_get_options(rlc_console_attr *attr);
 static HKEY	rlc_option_key(rlc_console_attr *attr, int create);
-static void	rlc_progbase(char *path, char *base);
+static void	rlc_progbase(TCHAR *path, TCHAR *base);
 static int	rlc_add_queue(RlcData b, RlcQueue q, int chr);
 static int	rlc_add_lines(RlcData b, int here, int add);
 static void	rlc_start_selection(RlcData b, int x, int y);
@@ -196,12 +201,12 @@ static void	rlc_copy(RlcData b);
 static void	rlc_destroy(RlcData b);
 static void	rlc_request_redraw(RlcData b);
 static void	rlc_redraw(RlcData b);
-static int	rlc_breakargs(char *program, char *line, char **argv);
+static int	rlc_breakargs(TCHAR *line, TCHAR **argv);
 static void	rlc_resize(RlcData b, int w, int h);
 static void	rlc_adjust_line(RlcData b, int line);
-static int	text_width(RlcData b, HDC hdc, const char *text, int len);
+static int	text_width(RlcData b, HDC hdc, const TCHAR *text, int len);
 static void	rlc_queryfont(RlcData b);
-static void     rlc_do_write(RlcData b, char *buf, int count);
+static void     rlc_do_write(RlcData b, TCHAR *buf, int count);
 static void     rlc_reinit_line(RlcData b, int line);
 static void	rlc_free_line(RlcData b, int line);
 static int	rlc_between(RlcData b, int f, int t, int v);
@@ -232,7 +237,7 @@ static DWORD WINAPI window_loop(LPVOID arg);	/* console window proc */
 
 #ifdef O_DEBUG
 #include <stdarg.h>
-static void Dprintf(const char *fmt, ...);
+static void Dprintf(const TCHAR *fmt, ...);
 static void Dprint_lines(RlcData b, int from, int to);
 #define DEBUG(Code) Code
 
@@ -247,8 +252,8 @@ expect to be true about the data is indeed the case.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
-rlc_assert(const char *msg)
-{ MessageBox(NULL, msg, "Console assertion failed", MB_OK|MB_TASKMODAL);
+rlc_assert(const TCHAR *msg)
+{ MessageBox(NULL, msg, _T("Console assertion failed"), MB_OK|MB_TASKMODAL);
 }
 
 void
@@ -275,18 +280,18 @@ rlc_check_assertions(RlcData b)
 #endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-rlc_long_name(char *buffer)
+rlc_long_name(TCHAR *buffer)
 	Translate a filename, possibly holding 8+3 abbreviated parts into
 	the `real' filename.  I couldn't find a direct call for this.  If
 	you have it, I'd be glad to receive a better implementation.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-rlc_long_name(char *file)
-{ char buf[MAXPATHLEN];
-  char *i = file;
-  char *o = buf;
-  char *ok = buf;
+rlc_long_name(TCHAR *file)
+{ TCHAR buf[MAXPATHLEN];
+  TCHAR *i = file;
+  TCHAR *o = buf;
+  TCHAR *ok = buf;
   int changed = 0;
 
   while(*i)
@@ -303,9 +308,9 @@ rlc_long_name(char *file)
 
       *o = '\0';
       if ( (h=FindFirstFile(buf, &data)) != INVALID_HANDLE_VALUE )
-      { strcpy(ok, data.cFileName);
+      { _tcscpy(ok, data.cFileName);
 	FindClose(h);
-	o = ok + strlen(ok);
+	o = ok + _tcslen(ok);
 	changed++;
       }
     }
@@ -316,7 +321,7 @@ rlc_long_name(char *file)
 
   if ( changed )
   { *o = '\0';
-    strcpy(file, buf);
+    _tcscpy(file, buf);
   }
 }
 
@@ -330,16 +335,16 @@ it.
 In old versions this was fixed to "RlcConsole"
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static char *
+static TCHAR *
 rlc_window_class(HICON icon)
-{ static char winclassname[32];
+{ static TCHAR winclassname[32];
   static WNDCLASS wndClass;
   HINSTANCE instance = _rlc_hinstance;
 
   if ( !winclassname[0] )
-  { if ( !GetEnvironmentVariable("PLTERM_CLASS",
+  { if ( !GetEnvironmentVariable(TEXT("PLTERM_CLASS"),
 				 winclassname, sizeof(winclassname)) )
-      sprintf(winclassname, "PlTerm-%d", instance);
+      _stprintf(winclassname, TEXT("PlTerm-%d"), instance);
 
     wndClass.lpszClassName	= winclassname;
     wndClass.style		= CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS;
@@ -362,15 +367,26 @@ rlc_window_class(HICON icon)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+There are two ways to get the commandline.   It  is passed to WinMain as
+8-bit string. This version does *not* include  the command itself. It is
+also available through GetCommandLine(), which  does include the command
+itself and returns LPTSTR (Unicode/ANSI). We assume the latter.
+
+Nevertheless, for backward compatibility as well  as easy to extract the
+full pathname of the  executable,  we   replace  argv[0]  with  the long
+filename version of the current module, so argv[0] is guaranteed to be a
+full path refering to the .exe file.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
 rlc_main(HANDLE hInstance, HANDLE hPrevInstance,
-	 LPSTR lpszCmdLine, int nCmdShow,
+	 LPTSTR lpszCmdLine, int nCmdShow,
 	 RlcMain mainfunc, HICON icon)
-{ char *	    argv[100];
+{ TCHAR *	    argv[100];
   int		    argc;
-  char		    program[MAXPATHLEN];
-  char	 	    progbase[100];
+  TCHAR		    program[MAXPATHLEN];
+  TCHAR	 	    progbase[100];
   RlcData           b;
   rlc_console_attr  attr;
 
@@ -382,14 +398,14 @@ rlc_main(HANDLE hInstance, HANDLE hPrevInstance,
 
   GetModuleFileName(hInstance, program, sizeof(program));
   rlc_long_name(program);
-  rlc_progbase(program, progbase);
+  argc = rlc_breakargs(lpszCmdLine, argv);
+  argv[0] = program;
+  rlc_progbase(argv[0], progbase);
 
   memset(&attr, 0, sizeof(attr));
   _rlc_program = attr.title = progbase;
   _rlc_stdio = b = rlc_create_console(&attr);
 
-  argc = rlc_breakargs(program, lpszCmdLine, argv);
-
   if ( mainfunc )
     return (*mainfunc)(b, argc, argv);
   else
@@ -397,24 +413,25 @@ rlc_main(HANDLE hInstance, HANDLE hPrevInstance,
 }
 
 
+
 rlc_console
 rlc_create_console(rlc_console_attr *attr)
 { RlcData b;
   MSG msg;
-  const char *title;
+  const TCHAR *title;
 
   rlc_get_options(attr);
 
   if ( attr->title )
     title = attr->title;
   else
-    title = "Untitled";
+    title = _T("Untitled");
 
   b = rlc_make_buffer(attr->width, attr->savelines);
   b->create_attributes = attr;
-  strcpy(b->current_title, title);
+  _tcscpy(b->current_title, title);
   if ( attr->key )
-  { b->regkey_name = strdup(attr->key);
+  { b->regkey_name = _tcsdup(attr->key);
   }
 
   rlc_init_text_dimensions(b, NULL);
@@ -499,18 +516,18 @@ rlc_iswin32s()
 
 
 static void
-rlc_progbase(char *path, char *base)
-{ char *s;
-  char *e;
+rlc_progbase(TCHAR *path, TCHAR *base)
+{ TCHAR *s;
+  TCHAR *e;
 
-  if ( !(s=strrchr(path, '\\')) )
+  if ( !(s=_tcsrchr(path, '\\')) )
     s = path;				/* takes the filename part */
   else
     s++;
-  if ( !(e = strchr(s, '.')) )
-    strcpy(base, s);
+  if ( !(e = _tcschr(s, '.')) )
+    _tcscpy(base, s);
   else
-  { memmove(base, s, e-s);
+  { _tcsncpy(base, s, e-s);
     base[e-s] = '\0';
   }
 }
@@ -530,14 +547,14 @@ rlc_kill_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
   return DefWindowProc(hwnd, message, wParam, lParam);
 }
 
-static char *
+static TCHAR *
 rlc_kill_window_class()
-{ static char winclassname[32];
+{ static TCHAR winclassname[32];
   static WNDCLASS wndClass;
   HINSTANCE instance = _rlc_hinstance;
 
   if ( !winclassname[0] )
-  { sprintf(winclassname, "Console-hidden-win%d", instance);
+  { _stprintf(winclassname, _T("Console-hidden-win%d"), instance);
 
     wndClass.style		= 0;
     wndClass.lpfnWndProc	= (LPVOID) rlc_kill_wnd_proc;
@@ -560,7 +577,7 @@ rlc_kill_window_class()
 static void
 _rlc_create_kill_window(RlcData b)
 { b->kill_window = CreateWindow(rlc_kill_window_class(),
-				"Console hidden window",
+				_T("Console hidden window"),
 				0,
 				0, 0, 32, 32,
 				NULL, NULL, _rlc_hinstance, NULL);
@@ -574,16 +591,21 @@ _rlc_create_kill_window(RlcData b)
 #define MAXREGSTRLEN 1024
 
 static void
-reg_save_int(HKEY key, const char *name, int value)
+reg_save_int(HKEY key, const TCHAR *name, int value)
 { DWORD val = value;
 
-  RegSetValueEx(key, name, 0,
-		REG_DWORD_LITTLE_ENDIAN, (LPBYTE)&val, sizeof(val));
+  if ( RegSetValueEx(key, name, 0,
+		     REG_DWORD_LITTLE_ENDIAN,
+		     (LPBYTE)&val, sizeof(val)) != ERROR_SUCCESS )
+    DEBUG(MessageBox(NULL, _T("Failed to save int setting"),
+		     _T("Error"), MB_OK));
 }
 
 static void
-reg_save_str(HKEY key, const char *name, char *value)
-{ RegSetValueEx(key, name, 0, REG_SZ, (LPBYTE)value, strlen(value)+1);
+reg_save_str(HKEY key, const TCHAR *name, TCHAR *value)
+{ if ( RegSetValueEx(key, name, 0, REG_SZ,
+		     (LPBYTE)value, (_tcslen(value)+1)*sizeof(TCHAR)) != ERROR_SUCCESS )
+    DEBUG(MessageBox(NULL, _T("Failed to save string setting"), _T("Error"), MB_OK));
 }
 
 
@@ -598,23 +620,23 @@ rlc_save_options(RlcData b)
   if ( !(key = rlc_option_key(&attr, TRUE)) )
     return;
 
-  reg_save_int(key, "SaveLines",  b->height);
+  reg_save_int(key, _T("SaveLines"),  b->height);
   if ( b->modified_options & OPT_SIZE )
-  { reg_save_int(key, "Width",    b->width);
-    reg_save_int(key, "Height",   b->window_size);
+  { reg_save_int(key, _T("Width"),    b->width);
+    reg_save_int(key, _T("Height"),   b->window_size);
   }
   if ( b->modified_options & OPT_POSITION )
-  { reg_save_int(key, "X",	  b->win_x);
-    reg_save_int(key, "Y",	  b->win_y);
+  { reg_save_int(key, _T("X"),	  b->win_x);
+    reg_save_int(key, _T("Y"),	  b->win_y);
   }
 
   rlc_save_font_options(b->hfont, &attr);
   if ( attr.face_name[0] )
-  { reg_save_str(key, "FaceName",    attr.face_name);
-    reg_save_int(key, "FontFamily",  attr.font_family);
-    reg_save_int(key, "FontSize",    attr.font_size);
-    reg_save_int(key, "FontWeight",  attr.font_weight);
-    reg_save_int(key, "FontCharSet", attr.font_char_set);
+  { reg_save_str(key, _T("FaceName"),    attr.face_name);
+    reg_save_int(key, _T("FontFamily"),  attr.font_family);
+    reg_save_int(key, _T("FontSize"),    attr.font_size);
+    reg_save_int(key, _T("FontWeight"),  attr.font_weight);
+    reg_save_int(key, _T("FontCharSet"), attr.font_char_set);
   }
 
   RegCloseKey(key);
@@ -622,7 +644,7 @@ rlc_save_options(RlcData b)
 
 
 static void
-reg_get_int(HKEY key, const char *name, int mn, int def, int mx, int *value)
+reg_get_int(HKEY key, const TCHAR *name, int mn, int def, int mx, int *value)
 { DWORD type;
   BYTE  data[8];
   DWORD len = sizeof(data);
@@ -653,9 +675,9 @@ reg_get_int(HKEY key, const char *name, int mn, int def, int mx, int *value)
 
 
 static void
-reg_get_str(HKEY key, const char *name, char *value, int length)
+reg_get_str(HKEY key, const TCHAR *name, TCHAR *value, int length)
 { DWORD type;
-  BYTE  data[MAXREGSTRLEN];
+  BYTE  data[MAXREGSTRLEN*sizeof(TCHAR)];
   DWORD len = sizeof(data);
 
   if ( *value )
@@ -664,8 +686,8 @@ reg_get_str(HKEY key, const char *name, char *value, int length)
   if ( RegQueryValueEx(key, name, NULL, &type, data, &len) == ERROR_SUCCESS )
   { switch(type)
     { case REG_SZ:
-      { char *val = data;
-	strncpy(value, val, length-1);
+      { TCHAR *val = (TCHAR*)data;
+	_tcsncpy(value, val, length-1);
 	value[length-1] = '\0';
       }
     }
@@ -674,7 +696,7 @@ reg_get_str(HKEY key, const char *name, char *value, int length)
 
 
 HKEY
-reg_open_key(char **which, int create)
+reg_open_key(TCHAR **which, int create)
 { HKEY key = HKEY_CURRENT_USER;
   DWORD disp;
   LONG rval;
@@ -692,7 +714,7 @@ reg_open_key(char **which, int create)
 	return NULL;
     }
 
-    rval = RegCreateKeyEx(key, which[0], 0, "", 0,
+    rval = RegCreateKeyEx(key, which[0], 0, _T(""), 0,
 			  KEY_ALL_ACCESS, NULL, &tmp, &disp);
     RegCloseKey(key);
     if ( rval == ERROR_SUCCESS )
@@ -707,19 +729,19 @@ reg_open_key(char **which, int create)
 
 static HKEY
 rlc_option_key(rlc_console_attr *attr, int create)
-{ char Prog[256];
-  char *address[] = { "Software",
+{ TCHAR Prog[256];
+  TCHAR *address[] = { _T("Software"),
   		      RLC_VENDOR,
 		      Prog,
-		      "Console",
-		      (char *)attr->key,	/* possible secondary key */
+		      _T("Console"),
+		      (TCHAR *)attr->key,	/* possible secondary key */
 		      NULL
 		    };
-  const char *s;
-  char *q;
+  const TCHAR *s;
+  TCHAR *q;
 
   for(s=_rlc_program, q=Prog; *s; s++, q++) /* capitalise the key */
-  { *q = (s==_rlc_program ? toupper(*s) : tolower(*s));
+  { *q = (s==_rlc_program ? _totupper(*s) : _totlower(*s));
   }
   *q = EOS;
 
@@ -748,18 +770,19 @@ rlc_get_options(rlc_console_attr *attr)
   maxx = rect.right  - 40;
   maxy = rect.bottom - 40;
 
-  reg_get_int(key, "SaveLines",   200,  200, 100000, &attr->savelines);
-  reg_get_int(key, "Width",        20,	 80,    300, &attr->width);
-  reg_get_int(key, "Height",        5,	 24,    100, &attr->height);
-  reg_get_int(key, "X",		 minx, minx,   maxx, &attr->x);
-  reg_get_int(key, "Y",	         miny, miny,   maxy, &attr->y);
+  reg_get_int(key, _T("SaveLines"),   200,  200, 100000, &attr->savelines);
+  reg_get_int(key, _T("Width"),        20,	 80,    300, &attr->width);
+  reg_get_int(key, _T("Height"),        5,	 24,    100, &attr->height);
+  reg_get_int(key, _T("X"),		 minx, minx,   maxx, &attr->x);
+  reg_get_int(key, _T("Y"),	         miny, miny,   maxy, &attr->y);
 }
 
-  reg_get_str(key, "FaceName", attr->face_name, sizeof(attr->face_name));
-  reg_get_int(key, "FontFamily",    0,  0,  0, &attr->font_family);
-  reg_get_int(key, "FontSize",      0,  0,  0, &attr->font_size);
-  reg_get_int(key, "FontWeight",    0,  0,  0, &attr->font_weight);
-  reg_get_int(key, "FontCharSet",   0,  0,  0, &attr->font_char_set);
+  reg_get_str(key, _T("FaceName"), attr->face_name,
+	      sizeof(attr->face_name)/sizeof(TCHAR));
+  reg_get_int(key, _T("FontFamily"),    0,  0,  0, &attr->font_family);
+  reg_get_int(key, _T("FontSize"),      0,  0,  0, &attr->font_size);
+  reg_get_int(key, _T("FontWeight"),    0,  0,  0, &attr->font_weight);
+  reg_get_int(key, _T("FontCharSet"),   0,  0,  0, &attr->font_char_set);
 
   RegCloseKey(key);
 }
@@ -767,27 +790,25 @@ rlc_get_options(rlc_console_attr *attr)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Windows-'95 appears to quote names of files   (I guess because files may
-hold spaces).  rlc_breakargs()  will  pass  a   quoted  strings  as  one
-argument.  If it can't find the closing   quote, it will tread the quote
-as a normal character.
+Windows-'95 appears to quote  names  of   files  because  files may hold
+spaces. rlc_breakargs() will pass a quoted   strings as one argument. If
+it can't find the closing quote, it  will   tread  the quote as a normal
+character.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-rlc_breakargs(char *program, char *line, char **argv)
-{ int argc = 1;
-
-  argv[0] = program;
+rlc_breakargs(TCHAR *line, TCHAR **argv)
+{ int argc = 0;
 
   while(*line)
   { int q;
 
-    while(*line && isspace(*line))
+    while(*line && _istspace(*line))
       line++;
 
     if ( (q = *line) == '"' || q == '\'' )	/* quoted arguments */
-    { char *start = line+1;
-      char *end = start;
+    { TCHAR *start = line+1;
+      TCHAR *end = start;
 
       while( *end && *end != q )
 	end++;
@@ -801,7 +822,7 @@ rlc_breakargs(char *program, char *line, char **argv)
 
     if ( *line )
     { argv[argc++] = line;
-      while(*line && !isspace(*line))
+      while(*line && !_istspace(*line))
 	line++;
       if ( *line )
 	*line++ = '\0';
@@ -878,9 +899,9 @@ rlc_kill(RlcData b)
 			       &result) )
       { if ( b->window )
 	{ switch( MessageBox(b->window,
-			     "Main task is not responding."
-			     "Click \"OK\" to terminate it",
-			     "Error",
+			     _T("Main task is not responding.")
+			     _T("Click \"OK\" to terminate it"),
+			     _T("Error"),
 			     MB_OKCANCEL|MB_ICONEXCLAMATION|MB_APPLMODAL) )
 	  { case IDCANCEL:
 	      return FALSE;
@@ -988,7 +1009,7 @@ rlc_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
 
     case WM_COMMAND:
     { UINT  item  = (UINT) LOWORD(wParam);
-      const char *name;
+      const TCHAR *name;
 
       switch( item )
       { case IDM_PASTE:
@@ -1066,6 +1087,10 @@ rlc_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
     break2:
       break;
     }
+	case WM_UNICHAR:
+	  chr = wParam;
+	  typed_char(b, chr);
+	  return 0;
     case WM_SYSCHAR:	typed_char(b, ESC); /* Play escape-something */
     case WM_CHAR:	chr = wParam;
 
@@ -1258,17 +1283,17 @@ rlc_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
 
     case WM_RLC_WRITE:
     { int count = (int)wParam;
-      char *buf = (char *)lParam;
+      TCHAR *buf = (TCHAR *)lParam;
 
       if ( OQSIZE - b->output_queued > count )
-      { memcpy(&b->output_queue[b->output_queued], buf, count);
+      { _tcsncpy(&b->output_queue[b->output_queued], buf, count);
 	b->output_queued += count;
       } else
       { if ( b->output_queued > 0 )
 	  rlc_flush_output(b);
 
 	if ( count <= OQSIZE )
-	{ memcpy(b->output_queue, buf, count);
+	{ _tcsncpy(b->output_queue, buf, count);
 	  b->output_queued = count;
 	} else
 	  rlc_do_write(b, buf, count);
@@ -1333,7 +1358,7 @@ rlc_dispatch(RlcData b)
     rlc_flush_output(b);
     return;
   } else
-  { DEBUG(Dprintf("Thread %x got WM_RLC_CLOSEWIN\n",
+  { DEBUG(Dprintf(_T("Thread %x got WM_RLC_CLOSEWIN\n"),
 		  GetCurrentThreadId()));
     b->queue->flags |= RLC_EOF;
   }
@@ -1354,8 +1379,6 @@ rlc_yield()
 		 *	 CHARACTER TYPES	*
 		 *******************************/
 
-#define iswordchar(c) (_rlc_word_chars[(int)(c)])
-
 static void
 rlc_init_word_chars()
 { int i;
@@ -1377,7 +1400,7 @@ rlc_is_word_char(int chr)
 { if ( chr > 0 && chr < CHAR_MAX )
     return _rlc_word_chars[chr];
 
-  return FALSE;
+  return _istalnum((wint_t)chr);	/* only UNICODE version */
 }
 
 
@@ -1495,7 +1518,7 @@ rlc_translate_mouse(RlcData b, int x, int y, int *line, int *chr)
   } else if ( tl->size == 0 )
   { *chr = 0;
   } else
-  { char *s = tl->text;
+  { TCHAR *s = tl->text;
     HDC hdc = GetDC(b->window);
     int f = 0;
     int t = tl->size;
@@ -1572,12 +1595,12 @@ rlc_word_selection(RlcData b, int x, int y)
   if ( rlc_between(b, b->first, b->last, l) )
   { TextLine tl = &b->lines[l];
 
-    if ( c < tl->size && iswordchar(tl->text[c]) )
+    if ( c < tl->size && rlc_is_word_char(tl->text[c]) )
     { int f, t;
 
-      for(f=c; f>0 && iswordchar(tl->text[f-1]); f--)
+      for(f=c; f>0 && rlc_is_word_char(tl->text[f-1]); f--)
 	;
-      for(t=c; t<tl->size && iswordchar(tl->text[t]); t++)
+      for(t=c; t<tl->size && rlc_is_word_char(tl->text[t]); t++)
 	;
       rlc_set_selection(b, l, f, l, t);
     }
@@ -1597,8 +1620,8 @@ rlc_extend_selection(RlcData b, int x, int y)
     { if ( rlc_between(b, b->first, b->last, l) )
       { TextLine tl = &b->lines[l];
 
-	if ( c < tl->size && iswordchar(tl->text[c]) )
-	  for(; c > 0 && iswordchar(tl->text[c-1]); c--)
+	if ( c < tl->size && rlc_is_word_char(tl->text[c]) )
+	  for(; c > 0 && rlc_is_word_char(tl->text[c-1]); c--)
 	    ;
       }
     } else if ( b->sel_unit == SEL_LINE )
@@ -1609,8 +1632,8 @@ rlc_extend_selection(RlcData b, int x, int y)
     { if ( rlc_between(b, b->first, b->last, l) )
       { TextLine tl = &b->lines[l];
 
-	if ( c < tl->size && iswordchar(tl->text[c]) )
-	  for(; c < tl->size && iswordchar(tl->text[c]); c++)
+	if ( c < tl->size && rlc_is_word_char(tl->text[c]) )
+	  for(; c < tl->size && rlc_is_word_char(tl->text[c]); c++)
 	    ;
       }
     } else if ( b->sel_unit == SEL_LINE )
@@ -1620,15 +1643,15 @@ rlc_extend_selection(RlcData b, int x, int y)
 }
 
 
-static char *
+static TCHAR *
 rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
 { int bufsize = 256;
-  char *buf;
+  TCHAR *buf;
   int i = 0;
 
   if ( el < sl || el == sl && ec < sc )
     return NULL;			/* invalid region */
-  if ( !(buf = rlc_malloc(bufsize)) )
+  if ( !(buf = rlc_malloc(bufsize * sizeof(TCHAR))) )
     return NULL;			/* not enough memory */
 
   for( ; ; sc = 0, sl = NextLine(b, sl))
@@ -1642,7 +1665,7 @@ rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
       while(sc < e)
       { if ( i >= bufsize )
 	{ bufsize *= 2;
-	  if ( !(buf = rlc_realloc(buf, bufsize)) )
+	  if ( !(buf = rlc_realloc(buf, bufsize * sizeof(TCHAR))) )
 	    return NULL;		/* not enough memory */
 	}
 	buf[i++] = tl->text[sc++];
@@ -1657,7 +1680,7 @@ rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
     if ( tl && !tl->softreturn )
     { if ( i+1 >= bufsize )
       { bufsize *= 2;
-	if ( !(buf = rlc_realloc(buf, bufsize)) )
+	if ( !(buf = rlc_realloc(buf, bufsize * sizeof(TCHAR))) )
 	  return NULL;			/* not enough memory */
       }
       buf[i++] = '\r';			/* Bill ... */
@@ -1667,7 +1690,7 @@ rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
 }
 
 
-static char *
+static TCHAR *
 rlc_selection(RlcData b)
 { if ( SelEQ(b->sel_start_line, b->sel_start_char,
 	     b->sel_end_line,   b->sel_end_char) )
@@ -1681,16 +1704,16 @@ rlc_selection(RlcData b)
 
 static void
 rlc_copy(RlcData b)
-{ char *sel = rlc_selection(b);
+{ TCHAR *sel = rlc_selection(b);
 
   if ( sel && b->window )
-  { int size = strlen(sel);
-    HGLOBAL mem = GlobalAlloc(GMEM_MOVEABLE, size + 1);
-    char far *data;
+  { int size = _tcslen(sel);
+    HGLOBAL mem = GlobalAlloc(GMEM_MOVEABLE, (size + 1)*sizeof(TCHAR));
+    TCHAR far *data;
     int i;
 
     if ( !mem )
-    { MessageBox(NULL, "Not enough memory to paste", "Error", MB_OK);
+    { MessageBox(NULL, _T("Not enough memory to copy"), _T("Error"), MB_OK);
       return;
     }
     data = GlobalLock(mem);
@@ -1702,7 +1725,11 @@ rlc_copy(RlcData b)
     GlobalUnlock(mem);
     OpenClipboard(b->window);
     EmptyClipboard();
+#ifdef UNICODE
+    SetClipboardData(CF_UNICODETEXT, mem);
+#else
     SetClipboardData(CF_TEXT, mem);
+#endif
     CloseClipboard();
 
     rlc_free(sel);
@@ -1815,17 +1842,25 @@ rlc_redraw(RlcData b)
 
   for(; pl <= el; l = NextLine(b, l), pl++)
   { TextLine tl = &b->lines[l];
-    char text[MAXLINE];
+    TCHAR text[MAXLINE];
     int ty = b->ch * pl;
     int cx = b->cw;
 
     if ( !tl->text )
-    { tl->size = 0;
-      memset(text, ' ', b->width);
+    { int i;
+      TCHAR *t;
+
+      tl->size = 0;
+      for(i=0, t=text; i<b->width; i++)
+	*t++ = ' ';
     } else
-    { memcpy(text, tl->text, tl->size);
-      if ( b->width > tl->size )
-	memset(&text[tl->size], ' ', b->width - tl->size);
+    { int i;
+      TCHAR *t, *s;
+
+      for(i=0, t=text, s=tl->text; i<tl->size; i++)
+	*t++ = *s++;
+      for(; i<b->width; i++)
+	*t++ = ' ';
     }
 
     rect.top    = ty;
@@ -1955,7 +1990,7 @@ rlc_resize_pixel_units(RlcData b, int w, int h)
 { int nw = max(20, w/b->cw)-2;		/* 1 character space for margins */
   int nh = max(1, h/b->ch);
   
-  DEBUG(Dprintf("rlc_resize_pixel_units(%p, %d, %d) (%dx%d)\n",
+  DEBUG(Dprintf(_T("rlc_resize_pixel_units(%p, %d, %d) (%dx%d)\n"),
 		b, w, h, nw, nh));
 
   if ( b->width == nw && b->window_size == nh )
@@ -1999,7 +2034,7 @@ rlc_init_text_dimensions(RlcData b, HFONT font)
       lfont.lfWeight          = a->font_weight;
       lfont.lfPitchAndFamily  = a->font_family;
       lfont.lfCharSet	      = a->font_char_set;
-      strncpy(lfont.lfFaceName, a->face_name, 31);
+      _tcsncpy(lfont.lfFaceName, a->face_name, 31);
     
       if ( !(b->hfont = CreateFontIndirect(&lfont)) )
 	b->hfont = GetStockObject(ANSI_FIXED_FONT);
@@ -2032,7 +2067,7 @@ rlc_init_text_dimensions(RlcData b, HFONT font)
 
 
 static int
-text_width(RlcData b, HDC hdc, const char *text, int len)
+text_width(RlcData b, HDC hdc, const TCHAR *text, int len)
 { if ( b->fixedfont )
   { return len * b->cw;
   } else
@@ -2052,7 +2087,7 @@ rlc_save_font_options(HFONT font, rlc_console_attr *attr)
   { LOGFONT lf;
 
     if ( GetObject(font, sizeof(lf), &lf) )
-    { strncpy(attr->face_name, lf.lfFaceName, sizeof(attr->face_name)-1);
+    { memcpy(attr->face_name, lf.lfFaceName, sizeof(attr->face_name)-1);
 
       attr->font_family   = lf.lfPitchAndFamily;
       attr->font_size     = lf.lfHeight;
@@ -2183,7 +2218,8 @@ rlc_resize(RlcData b, int w, int h)
   if ( b->width == w && b->window_size == h )
     return;				/* no real change */
 
-  DEBUG(Dprintf("Resizing %dx%d --> %dx%d\n", b->width, b->window_size, w, h));
+  DEBUG(Dprintf(_T("Resizing %dx%d --> %dx%d\n"),
+		b->width, b->window_size, w, h));
 
   b->window_size = h;
   b->width = w;
@@ -2200,17 +2236,17 @@ rlc_resize(RlcData b, int w, int h)
 
 	rlc_shift_lines_down(b, i);
 	DEBUG(Dprint_lines(b, b->first, b->first));
-	DEBUG(Dprintf("b->first = %d, b->last = %d\n", b->first, b->last));
+	DEBUG(Dprintf(_T("b->first = %d, b->last = %d\n"), b->first, b->last));
 	pl = &b->lines[PrevLine(b, i)];	/* this is the moved line */
-	tl->text = rlc_malloc(pl->size - w);
-	memmove(tl->text, &pl->text[w], pl->size - w);
-	DEBUG(Dprintf("Copied %d chars from line %d to %d\n",
+	tl->text = rlc_malloc((pl->size - w)*sizeof(TCHAR));
+	memmove(tl->text, &pl->text[w], (pl->size - w)*sizeof(TCHAR));
+	DEBUG(Dprintf(_T("Copied %d chars from line %d to %d\n"),
 		      pl->size - w, pl - b->lines, i));
 	tl->size = pl->size - w;
 	tl->adjusted = TRUE;
 	tl->softreturn = FALSE;
 	pl->softreturn = TRUE;
-	pl->text = rlc_realloc(pl->text, w);
+	pl->text = rlc_realloc(pl->text, w * sizeof(TCHAR));
 	pl->size = w;
 	pl->adjusted = TRUE;
 	i = pl - b->lines;
@@ -2222,9 +2258,9 @@ rlc_resize(RlcData b, int w, int h)
 	if ( i == b->last )
 	  rlc_add_line(b);
 	nl = &b->lines[NextLine(b, i)];
-	nl->text = rlc_realloc(nl->text, nl->size + move);
-	memmove(&nl->text[move], nl->text, nl->size);
-	memmove(nl->text, &tl->text[w], move);
+	nl->text = rlc_realloc(nl->text, (nl->size + move)*sizeof(TCHAR));
+	memmove(&nl->text[move], nl->text, nl->size*sizeof(TCHAR));
+	memmove(nl->text, &tl->text[w], move*sizeof(TCHAR));
 	nl->size += move;
 	tl->size = w;
       }	
@@ -2235,9 +2271,9 @@ rlc_resize(RlcData b, int w, int h)
 	rlc_add_line(b);
       nl = &b->lines[NextLine(b, i)];
 
-      nl->text = rlc_realloc(nl->text, nl->size + tl->size);
-      memmove(&nl->text[tl->size], nl->text, nl->size);
-      memmove(nl->text, tl->text, tl->size);
+      nl->text = rlc_realloc(nl->text, (nl->size + tl->size)*sizeof(TCHAR));
+      memmove(&nl->text[tl->size], nl->text, nl->size*sizeof(TCHAR));
+      memmove(nl->text, tl->text, tl->size*sizeof(TCHAR));
       nl->size += tl->size;
       nl->adjusted = TRUE;
       rlc_shift_lines_up(b, i);
@@ -2290,7 +2326,9 @@ rlc_adjust_line(RlcData b, int line)
 { TextLine tl = &b->lines[line];
 
   if ( tl->text && !tl->adjusted )
-  { tl->text = rlc_realloc(tl->text, tl->size == 0 ? 4 : tl->size);
+  { tl->text = rlc_realloc(tl->text, tl->size == 0
+			   	? sizeof(TCHAR)
+				: tl->size * sizeof(TCHAR));
     tl->adjusted = TRUE;
   }
 }
@@ -2302,11 +2340,11 @@ rlc_unadjust_line(RlcData b, int line)
 
   if ( tl->text )
   { if ( tl->adjusted )
-    { tl->text = rlc_realloc(tl->text, b->width + 1);
+    { tl->text = rlc_realloc(tl->text, (b->width + 1)*sizeof(TCHAR));
       tl->adjusted = FALSE;
     }
   } else
-  { tl->text = rlc_malloc(b->width + 1);
+  { tl->text = rlc_malloc((b->width + 1)*sizeof(TCHAR));
     tl->adjusted = FALSE;
     tl->size = 0;
   }
@@ -2324,7 +2362,7 @@ rlc_open_line(RlcData b)
     b->first = NextLine(b, b->first);
   }
 
-  b->lines[i].text       = rlc_malloc(b->width + 1);
+  b->lines[i].text       = rlc_malloc((b->width + 1)*sizeof(TCHAR));
   b->lines[i].adjusted   = FALSE;
   b->lines[i].size       = 0;
   b->lines[i].softreturn = FALSE;
@@ -2528,7 +2566,7 @@ rlc_put(RlcData b, int chr)
 }
 
 #ifdef _DEBUG
-#define CMD(c) {cmd = #c; c;}
+#define CMD(c) {cmd = _T(#c); c;}
 #else
 #define CMD(c) {c;}
 #endif
@@ -2537,7 +2575,7 @@ static void
 rlc_putansi(RlcData b, int chr)
 {
 #ifdef _DEBUG
-  char *cmd;
+  TCHAR *cmd;
 #endif
 
   switch(b->cmdstat)
@@ -2579,7 +2617,7 @@ rlc_putansi(RlcData b, int chr)
       }
       break;
     case CMD_ANSI:			/* ESC [ */
-      if ( isdigit(chr) )
+      if ( _istdigit((wint_t)chr) )
       { if ( !b->argstat )
 	{ b->argv[b->argc] = (chr - '0');
 	  b->argstat = 1;		/* positive */
@@ -2655,7 +2693,21 @@ rlc_paste(RlcData b)
 
   if ( b->window )
   { OpenClipboard(b->window);
-    if ( (mem = GetClipboardData(CF_TEXT)) )
+    if ( (mem = GetClipboardData(CF_UNICODETEXT)) )
+    { wchar_t *data = GlobalLock(mem);
+      int i;
+      RlcQueue q = b->queue;
+  
+      if ( q )
+      { for(i=0; data[i]; i++)
+	{ rlc_add_queue(b, q, data[i]);
+	  if ( data[i] == '\r' && data[i+1] == '\n' )
+	    i++;
+	}
+      }
+  
+      GlobalUnlock(mem);
+    } else if ( (mem = GetClipboardData(CF_TEXT)) )
     { char far *data = GlobalLock(mem);
       int i;
       RlcQueue q = b->queue;
@@ -2688,7 +2740,7 @@ rlc_get_mark(rlc_console c, RlcMark m)
 
 
 void
-rlc_goto_mark(rlc_console c, RlcMark m, const char *data, int offset)
+rlc_goto_mark(rlc_console c, RlcMark m, const TCHAR *data, int offset)
 { RlcData b = rlc_get_data(c);
   
   b->caret_x = m->mark_x;
@@ -2739,10 +2791,10 @@ rlc_putchar(rlc_console c, int chr)
 }
 
 
-char *
+TCHAR *
 rlc_read_screen(rlc_console c, RlcMark f, RlcMark t)
 { RlcData b = rlc_get_data(c);
-  char *buf;
+  TCHAR *buf;
 
   buf = rlc_read_from_window(b, f->mark_y, f->mark_x, t->mark_y, t->mark_x);
 
@@ -2782,7 +2834,7 @@ window_loop(LPVOID arg)
   while(!b->closing)
   { switch( b->imode )
     { case IMODE_COOKED:
-      { char *line = read_line(b);
+      { TCHAR *line = read_line(b);
 
 	if ( line != RL_CANCELED_CHARP )
 	{ LQueued lq = rlc_malloc(sizeof(lqueued));
@@ -2821,10 +2873,10 @@ window_loop(LPVOID arg)
 
   if ( b->closing <= 2 )
   { MSG msg;
-    char *waiting = "\r\nWaiting for Prolog. "
-      		    "Close again to force termination ..";
+    TCHAR *waiting = _T("\r\nWaiting for Prolog. ")
+      		     _T("Close again to force termination ..");
       
-    rlc_write(b, waiting, strlen(waiting));
+    rlc_write(b, waiting, _tcslen(waiting));
 
     while ( b->closing <= 2 && rlc_get_message(&msg, NULL, 0, 0) )
     { TranslateMessage(&msg);
@@ -2969,7 +3021,7 @@ rlc_make_queue(int size)
     q->size = size;
     q->flags = 0;
 
-    if ( (q->buffer = rlc_malloc(sizeof(short) * size)) )
+    if ( (q->buffer = rlc_malloc(sizeof(TCHAR) * size)) )
       return q;
   }
 
@@ -2987,18 +3039,47 @@ rlc_free_queue(RlcQueue q)
 } 
 
 
+static int
+rlc_resize_queue(RlcQueue q, int size)
+{ TCHAR *newbuf;
+
+  if ( (newbuf = rlc_malloc(size*sizeof(TCHAR))) )
+  { TCHAR *o = newbuf;
+    int c;
+
+    while( (c=rlc_from_queue(q)) != -1 )
+      *o++ = c;
+
+    if ( q->buffer )
+      rlc_free(q->buffer);
+    q->buffer = newbuf;
+    q->first = 0;
+    q->last  = o-newbuf;
+    q->size  = size;
+
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+
 static int
 rlc_add_queue(RlcData b, RlcQueue q, int chr)
 { int empty = (q->first == q->last);
 
-  if ( QN(q, q->last) != q->first )
-  { q->buffer[q->last] = chr;
-    q->last = QN(q, q->last);
+  while(q->size < 50000)
+  { if ( QN(q, q->last) != q->first )
+    { q->buffer[q->last] = chr;
+      q->last = QN(q, q->last);
+      
+      if ( empty )
+	PostThreadMessage(b->application_thread_id, WM_RLC_INPUT, 0, 0);
 
-    if ( empty )
-      PostThreadMessage(b->application_thread_id, WM_RLC_INPUT, 0, 0);
+      return TRUE;
+    }
 
-    return TRUE;
+    rlc_resize_queue(q, q->size*2);
   }
 
   return FALSE;
@@ -3038,8 +3119,12 @@ rlc_from_queue(RlcQueue q)
 		 *	   BUFFERED I/O		*
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+When using UNICODE, count is in bytes!
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
-rlc_read(rlc_console c, char *buf, unsigned int count)
+rlc_read(rlc_console c, TCHAR *buf, unsigned int count)
 { RlcData d = rlc_get_data(c);
   int give;
   MSG msg;
@@ -3054,7 +3139,7 @@ rlc_read(rlc_console c, char *buf, unsigned int count)
     (*_rlc_update_hook)();
 
   d->promptbuf[d->promptlen] = EOS;
-  strcpy(d->prompt, d->promptbuf);
+  _tcscpy(d->prompt, d->promptbuf);
 
   if ( d->read_buffer.given >= d->read_buffer.length )
   { if ( d->read_buffer.line )
@@ -3085,7 +3170,7 @@ rlc_read(rlc_console c, char *buf, unsigned int count)
       rlc_free(lq);
     }
 
-    d->read_buffer.length = strlen(d->read_buffer.line);
+    d->read_buffer.length = _tcslen(d->read_buffer.line);
     d->read_buffer.given = 0;
   }
 
@@ -3094,7 +3179,7 @@ rlc_read(rlc_console c, char *buf, unsigned int count)
   else
     give = d->read_buffer.length - d->read_buffer.given;
 
-  memcpy(buf, d->read_buffer.line+d->read_buffer.given, give);
+  _tcsncpy(buf, d->read_buffer.line+d->read_buffer.given, give);
   d->read_buffer.given += give;
 
   return give;
@@ -3102,10 +3187,10 @@ rlc_read(rlc_console c, char *buf, unsigned int count)
 
 
 static void
-rlc_do_write(RlcData b, char *buf, int count)
+rlc_do_write(RlcData b, TCHAR *buf, int count)
 { if ( count > 0 )
   { int n = 0;
-    char *s = buf;
+    TCHAR *s = buf;
 
     while(n++ < count)
     { int chr = *s++;
@@ -3142,9 +3227,9 @@ rlc_flush_output(rlc_console c)
 
 
 int
-rlc_write(rlc_console c, char *buf, unsigned int count)
+rlc_write(rlc_console c, TCHAR *buf, unsigned int count)
 { DWORD result;
-  char *e, *s;
+  TCHAR *e, *s;
   RlcData b = rlc_get_data(c);
 
   if ( !b )
@@ -3234,20 +3319,20 @@ rlc_close(rlc_console c)
 }
 
 
-const char *
-rlc_prompt(rlc_console c, const char *new)
+const TCHAR *
+rlc_prompt(rlc_console c, const TCHAR *new)
 { RlcData b = rlc_get_data(c);
 
   if ( b )
   { if ( new )
-    { strncpy(b->prompt, new, MAXPROMPT);
+    { _tcsncpy(b->prompt, new, MAXPROMPT);
       b->prompt[MAXPROMPT-1] = EOS;
     }
 
     return b->prompt;
   }
   
-  return "";
+  return _T("");
 }
 
 
@@ -3266,20 +3351,20 @@ rlc_clearprompt(rlc_console c)
 		 *	    MISC STUFF		*
 		 *******************************/
 
-static char current_title[RLC_TITLE_MAX];
+static TCHAR current_title[RLC_TITLE_MAX];
 
 void
-rlc_title(rlc_console c, char *title, char *old, int size)
+rlc_title(rlc_console c, TCHAR *title, TCHAR *old, int size)
 { RlcData b = rlc_get_data(c);
 
   if ( old )
-    memmove(old, b->current_title, size);
+    memmove(old, b->current_title, size*sizeof(TCHAR));
 
   if ( title )
   { if ( b->window )
       SetWindowText(b->window, title);
 
-    memmove(b->current_title, title, RLC_TITLE_MAX);
+    memmove(b->current_title, title, RLC_TITLE_MAX*sizeof(TCHAR));
   }
 }
 
@@ -3469,7 +3554,7 @@ free_user_data(RlcData b)
 
 static void
 noMemory()
-{ MessageBox(NULL, "Not enough memory", "Console", MB_OK|MB_TASKMODAL);
+{ MessageBox(NULL, _T("Not enough memory"), _T("Console"), MB_OK|MB_TASKMODAL);
 
   ExitProcess(1);
 }
@@ -3533,12 +3618,12 @@ initHeapDebug(void)
 
 
 static void
-Dprintf(const char *fmt, ...)
-{ char buf[1024];
+Dprintf(const TCHAR *fmt, ...)
+{ TCHAR buf[1024];
   va_list args;
 
   va_start(args, fmt);
-  vsprintf(buf, fmt, args);
+  _vstprintf(buf, fmt, args);
   va_end(args);
 
   OutputDebugString(buf);
@@ -3546,14 +3631,14 @@ Dprintf(const char *fmt, ...)
 
 static void
 Dprint_lines(RlcData b, int from, int to)
-{ char buf[1024];
+{ TCHAR buf[1024];
 
   for( ; ; from = NextLine(b, from))
   { TextLine tl = &b->lines[from];
 
     memcpy(buf, tl->text, tl->size);
     buf[tl->size] = EOS;
-    Dprintf("%03d: (0x%08x) \"%s\"\n", from, tl->text, buf);
+    Dprintf(_T("%03d: (0x%08x) \"%s\"\n"), from, tl->text, buf);
 
     if ( from == to )
       break;
